# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import time
import traceback

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "utils_python": "PrimitiveLock",
    "bpy_inspect": "BlRna",
    "utils_ui": "BlUI",
    "utils_blender": "BlUtil",
    "utils_addon": "AddonManager",
})

from . import common
from . import decal_utils
from . import decal_gizmos
from . import palette
from . import palette_gizmo

from .common import Jobs, CommonUtils
from .decal_utils import DecalUtils, Decal
from .decal_gizmos import decal_gizmo_state, DecalTransformBase
from .palette import Palette
from .palette_gizmo import DecalPaletteGizmo

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

class Admin:
    depsgraph_update_lock = PrimitiveLock()
    
    active_scene = None
    
    def reset_addon():
        addon.runtime.initialized = False
        addon.runtime.frame_count = 0
        Jobs.clear()
        Palette.invalidate(reset_selection=True)
        DecalPaletteGizmo.reset()
    
    def check_active_scene():
        scene = bpy.context.scene
        if BlRna.is_valid(Admin.active_scene) and (Admin.active_scene == scene): return
        Admin.active_scene = scene
        Jobs.palette_needs_update = True
    
    def state_monitor():
        if not settings: return
        
        if not addon.runtime.initialized:
            Admin.on_load_post()
            addon.runtime.initialized = True
        
        settings.loading_options.sync_defaults()
        
        Admin.check_active_scene()
        
        if Jobs.background or Jobs.icon_loading or Jobs.palette_needs_update:
            time_end = time.perf_counter() + settings.background_job_duration
            
            while Jobs.background:
                job = Jobs.background.popleft()
                if job: job()
                if time.perf_counter() > time_end: break
            
            if Jobs.palette_needs_update:
                Jobs.palette_needs_update = False
                DecalPaletteGizmo.clear_icon_textures()
                Palette.invalidate()
            
            Jobs.load_icons(time_end)
            
            BlUI.tag_redraw()
            # We need to update depsgraph in order for the loaded previews to actually
            # get shown in the UI. Perhaps tag_redraw() only triggers a redraw of the
            # existing layout, whereas we need to actually re-layout.
            BlUtil.Depsgraph.trigger_update()
    
    # def on_object_update(obj):
    #     if not Decal.exists(obj): return
        
    #     if not BlUtil.Object.select_get(obj): return
    #     # ... ?
    
    # def on_material_update(material):
    #     if CommonUtils.is_template(material):
    #         preview = material.preview
    
    def process_depsgraph_changes():
        Palette.invalidate(reset_images=False)
        
        # # NOTE: depsgraph.updates are empty in depsgraph_update_pre
        # context = bpy.context
        # depsgraph = context.evaluated_depsgraph_get()
        
        # bpy_Object = bpy.types.Object
        # bpy_Material = bpy.types.Material
        
        # for update in depsgraph.updates:
        #     id_data = update.id
        #     if isinstance(id_data, bpy_Object):
        #         Admin.on_object_update(id_data.original)
        #     elif isinstance(id_data, bpy_Material):
        #         Admin.on_material_update(id_data.original)
    
    def process_render_jobs():
        if not Jobs.render: return
        
        while Jobs.render:
            job = Jobs.render.popleft()
            if not job: continue
            
            try:
                job()
            except Exception:
                print("Error in a render job:")
                print(traceback.format_exc().strip())
        
        #BlUI.tag_redraw()
    
    @addon.timer(persistent=True)
    def background_timer_update():
        addon.runtime.frame_count += 1
        Admin.state_monitor()
        return 0 # run each frame
    
    @addon.handler("depsgraph_update_pre", persistent=True)
    def on_depsgraph_update_pre(scene=None):
        #if not settings: return # just in case
        if Admin.depsgraph_update_lock: return
        
        with Admin.depsgraph_update_lock:
            # Previously, we prevented the edit mode here,
            # but with geometry nodes it's no longer useful
            pass
    
    @addon.handler("depsgraph_update_post", persistent=True)
    def on_depsgraph_update_post(scene=None):
        #if not settings: return # just in case
        if Admin.depsgraph_update_lock: return
        
        with Admin.depsgraph_update_lock:
            Admin.process_depsgraph_changes()
    
    @addon.handler("load_post", persistent=True)
    def on_load_post(scene=None):
        if not settings: return # just in case
        
        settings.files.refresh()
        Palette.invalidate(reset_selection=True)
        DecalPaletteGizmo.reset()
    
    @addon.handler("undo_post", persistent=True)
    @addon.handler("redo_post", persistent=True)
    def on_undo_redo(scene=None):
        Palette.invalidate()
        DecalPaletteGizmo.reset()
    
    @addon.draw_handler("SpaceView3D", 'POST_PIXEL')
    def on_draw_view_2d():
        Admin.process_render_jobs()
    
    @addon.draw_handler("SpaceView3D", 'POST_VIEW')
    def on_draw_view_3d():
        try:
            decal_gizmo_state.draw_volume_shape()
        except Exception:
            print("Error in on_draw_view_3d():")
            print(traceback.format_exc().strip())
