# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import os
import time
from collections import deque

import bpy

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "prop",
    "utils_ui": "BlUI",
    "utils_blender": "BpyPath",
    "utils_addon": "AddonManager",
})

addon = AddonManager()
settings = addon.settings
previews = addon.previews

# =========================================================================== #

class Jobs:
    background = deque()
    render = deque()
    icon_loading = deque()
    
    palette_needs_update = False
    
    @classmethod
    def clear(cls):
        cls.background.clear()
        cls.render.clear()
        cls.icon_loading.clear()
    
    @classmethod
    def load_icons(cls, time_end):
        if not settings.files_show_get("any"): return # pause when files are not shown
        
        while cls.icon_loading:
            job = cls.icon_loading.popleft()
            if job: job()
            if time.perf_counter() > time_end: break

class DelayedUndoRecorder:
    message = ""
    
    @classmethod
    def trigger(cls, message):
        cls.message = message

class CommonConfig:
    sidebar_category = "XDecal™"
    
    palette_zoom_step = 0.2

class CommonUtils:
    @classmethod
    def abspath(cls, src_path):
        if src_path.startswith("//"):
            abs_path = BpyPath.abspath(src_path)
        elif not os.path.isabs(src_path):
            abs_path = os.path.join(addon.path, src_path)
        else:
            abs_path = src_path
        
        return os.path.realpath(abs_path)
    
    @classmethod
    def split(cls, path):
        return BpyPath.split(path.replace(os.path.sep, "/"))
    
    @classmethod
    def find_linked(cls, lib_path, data_name, name):
        lib_path = cls.abspath(lib_path)
        
        for idblock in getattr(bpy.data, data_name):
            library = idblock.library
            if not library: continue
            if idblock.name != name: continue
            if lib_path != cls.abspath(library.filepath): continue
            return idblock
    
    @classmethod
    def gather_linked(cls, data_name):
        direct = {}
        overrides = {}
        
        for idblock in getattr(bpy.data, data_name):
            library = idblock.library
            if library:
                lib_path = cls.abspath(library.filepath)
                direct.setdefault(lib_path, []).append(idblock)
            
            override_library = idblock.override_library
            if override_library and override_library.reference:
                library = override_library.reference.library
                lib_path = cls.abspath(library.filepath)
                overrides.setdefault(lib_path, []).append(idblock)
        
        return direct, overrides
    
    @classmethod
    def load_node_group(cls, target, path, link, force=False):
        results = cls.load_resources("node_groups", [target], path, link, force)
        return next(iter(results.values()), None)
    
    @classmethod
    def load_resources(cls, data_name, items, path, link, force=False):
        datas = getattr(bpy.data, data_name)
        
        results = {}
        targets = {}
        
        for item in items:
            if isinstance(item, str):
                name, target = item, datas.get(item)
            else:
                name, target = item.name, item
            
            if (not target) or target.is_missing or force:
                targets[name] = target
            elif target:
                results[name] = target
        
        if not targets: return results
        
        abs_path = cls.abspath(path)
        if not (abs_path and os.path.isfile(abs_path)): return results
        
        with bpy.data.libraries.load(abs_path, link=link, relative=False) as (data_from, data_to):
            from_datas = getattr(data_from, data_name)
            names = [name for name in targets if name in from_datas]
            if not names: return results
            setattr(data_to, data_name, names)
        
        for item in getattr(data_to, data_name):
            lib_path = (item.library.filepath if item.library else "")
            results[item.name] = item
            if not link: continue
            target = targets.get(item.name)
            if not target: continue
            target.user_remap(item)
            datas.remove(target)
        
        return results
    
    ##################################################
    
    template_tag = "XDecal.Template"
    
    @classmethod
    def is_template(cls, idblock):
        if not idblock.asset_data: return False
        return cls.template_tag in idblock.asset_data.tags
    
    @classmethod
    def mark_as_template(cls, idblock):
        asset_data = idblock.asset_data
        
        if not asset_data:
            idblock.asset_mark()
            asset_data = idblock.asset_data
        
        asset_data.tags.new(cls.template_tag, skip_if_exists=True)
    
    @classmethod
    def unmark_as_template(cls, idblock, keep_as_asset=False):
        asset_data = idblock.asset_data
        if not asset_data: return
        if cls.template_tag not in asset_data.tags: return
        
        if keep_as_asset:
            while True:
                asset_tag = asset_data.tags.get(cls.template_tag)
                if not asset_tag: break
                asset_data.tags.remove(asset_tag)
        else:
            idblock.asset_clear()

class SimpleOpEditMixin:
    id = "" | prop()
    id_prev = "" | prop()
    
    def execute(self, context):
        presets = self.preset_manager
        
        value = self.id
        
        if getattr(self.preset_cls, "normalize_whitespace", None):
            value = " ".join(value.split())
        
        preset = self.preset_cls(value) if value else None
        
        if self.id_prev:
            index = presets.index(self.id_prev)
            presets.delete(self.id_prev)
            
            if value:
                presets.add(preset)
                if index is not None:
                    presets.move(preset, index)
        else:
            if not value: return {'CANCELLED'}
            presets.add(preset)
        
        BlUI.tag_redraw()
        return {'FINISHED'}
    
    def invoke(self, context, event):
        self.id_prev = self.id
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=200)
    
    def draw(self, context):
        self.layout.prop(self, "id", text="", translate=False)
