# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import time

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color
from bpy_extras.view3d_utils import region_2d_to_vector_3d, region_2d_to_origin_3d

import gpu

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "utils_math": "product, divide, orthogonal_XYZ, matrix_compose, matrix_LRS, dist_to_segment",
    "bounds": "Bounds",
    "bpy_inspect": "BlRna",
    "utils_gl": "cgl",
    "utils_view3d": "SmartView3D",
    "utils_blender": "Raycaster",
    "utils_addon": "AddonManager",
})

from . import icon_geometry
from . import common
from . import decal_utils

from .icon_geometry import gizmo_icons
from .common import Jobs
from .decal_utils import DecalUtils, Decal

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

class DecalGizmoState:
    last_time = 0.0
    anim_time = 0.0
    
    volume_shape = None
    show_volume_shape = False
    
    _target = None
    def _get(self):
        target = self._target
        if BlRna.is_valid(target): return target
        self._target = None
        return None
    def _set(self, value):
        if self._target == value: return
        self._target = value
        self.update_volume_shape()
    target = property(_get, _set)
    
    del _get
    del _set
    
    def update_volume_shape(self):
        target = self.target
        if not target: return
        decal = Decal(target)
        if not decal: return
        
        layers, count_x, count_y = decal.calc_volume_verts()
        
        verts = []
        def add_vert(ix, iy, iz):
            verts.append(layers[iz][ix+iy*count_x])
        def add_quad(ix, iy, iz, ax, ay, az, bx, by, bz):
            add_vert(ix, iy, iz)
            add_vert(ix+ax, iy+ay, iz+az)
            add_vert(ix+ax+bx, iy+ay+by, iz+az+bz)
            add_vert(ix, iy, iz)
            add_vert(ix+ax+bx, iy+ay+by, iz+az+bz)
            add_vert(ix+bx, iy+by, iz+bz)
        
        for iz in (0, 1):
            for iy in range(count_y-1):
                for ix in range(count_x-1):
                    add_quad(ix, iy, iz, 1, 0, 0, 0, 1, 0)
        
        for ix in range(count_x-1):
            iy = 0
            add_quad(ix, iy, 0, 1, 0, 0, 0, 0, 1)
            iy = count_y-1
            add_quad(ix, iy, 0, 1, 0, 0, 0, 0, 1)
        
        for iy in range(count_y-1):
            ix = 0
            add_quad(ix, iy, 0, 0, 1, 0, 0, 0, 1)
            ix = count_x-1
            add_quad(ix, iy, 0, 0, 1, 0, 0, 0, 1)
        
        self.volume_shape = GizmoShape('TRIS', verts)
    
    def draw_volume_shape(self):
        target = self.target
        if not target: return
        decal = Decal(target)
        if not decal: return
        
        if not decal.use_project: return
        if not self.show_volume_shape: return
        
        colors = settings.colors
        
        depth_test = cgl.depth_test
        
        color = colors["decal_volume"]
        cgl.depth_test = 'LESS_EQUAL'
        self.draw_custom_shape(self.volume_shape, color)
        
        color[3] *= 0.5
        cgl.depth_test = 'GREATER'
        self.draw_custom_shape(self.volume_shape, color)
        
        cgl.depth_test = depth_test
    
    def draw_custom_shape(self, shape, color=None, select_id=None):
        target = self.target
        if not target: return
        if not target.select_get(): return
        if not shape: return
        if select_id is not None: color = None
        shape.draw(target.matrix_world, color, select_id=select_id)
    
    exit_frame = 0
    blocked_count = 0

decal_gizmo_state = DecalGizmoState()

class DecalTransformBase:
    is_active = False
    
    gizmo_margin = 0.05
    
    default_mode = 'MOVE'
    
    # Note: there is no delta axis angle
    transform_attr_names = (
        "location",
        "rotation_mode",
        "rotation_axis_angle",
        "rotation_euler",
        "rotation_quaternion",
        "scale",
        "delta_location",
        "delta_rotation_euler",
        "delta_rotation_quaternion",
        "delta_scale",
    )
    
    def invoke(self, context, event):
        target = decal_gizmo_state.target
        if not target: return {'CANCELLED'}
        
        decal = Decal(target)
        if not decal: return {'CANCELLED'}
        
        self.invoke_ctrl = event.ctrl
        self.invoke_shift = event.shift
        self.invoke_alt = event.alt
        
        self.initial_frustum = Vector(decal.frustum)
        self.initial_curvature = Vector(decal.curvature)
        self.initial_displacement = decal.displacement
        self.initial_range_x = decal.range_x
        self.initial_range_y = decal.range_y
        self.initial_extra_scale = Vector(decal.extra_scale)
        self.initial_targets = decal.targets
        
        self.projection_options = {}
        
        if decal.use_project:
            projection_options_map = {
                "use_project": 'PROJECT',
                "use_clip": 'CLIP',
                "use_cull": 'CULL',
                "use_occlude": 'OCCLUDE',
                "use_intersect": 'INTERSECT',
            }
            keep_while_dragging = settings.placement_options.keep_while_dragging
            
            for attr_name, proj_option in projection_options_map.items():
                if proj_option in keep_while_dragging: continue
                self.projection_options[attr_name] = getattr(decal, attr_name)
                setattr(decal, attr_name, False)
        
        self.initial_is_dynamic = decal.is_dynamic
        if not self.initial_is_dynamic: decal.is_dynamic = True
        
        self.selected_targets = [obj for obj in context.selected_objects if Decal.can_use_as_target(obj)]
        
        self.mode = self.default_mode
        
        self.value = None
        
        target.select_set(True)
        
        self.is_snap_to_edge = False
        self.snap_edge_points = None
        
        self.init_parent_type = target.parent_type
        self.init_parent = target.parent
        self.matrix_parent_inverse = target.matrix_parent_inverse.copy()
        self.matrix_world = target.matrix_world.copy()
        for attr_name in self.transform_attr_names:
            value = getattr(target, attr_name)
            if not isinstance(value, str): value = tuple(value)
            setattr(self, attr_name, value)
        
        # Keep initial parent if no new parent was assigned
        self.new_parent = self.init_parent
        
        self.init_raycaster(context)
        
        coord = Vector((event.mouse_region_x, event.mouse_region_y))
        sv = SmartView3D(context)
        proj = sv.project(target.matrix_world.translation)
        if (proj is None) or any(not (abs(v) <= 1e6) for v in proj):
            self.origin_delta = Vector((0,0))
        else:
            self.origin_delta = proj - coord
        
        self.initial_mouse_pos = coord
        
        self._timer = None
        if settings.gizmo_anim_speed > 0:
            time_delta = 0.02 / min(1.0, settings.gizmo_anim_speed)
            self._timer = addon.event_timer_add(time_delta, context.window)
        
        # Clear parent to avoid dealing with skewed matrices (and to avoid relationship lines)
        self.set_parent(None)
        
        ray = sv.ray(coord)
        self.initial_intersection_XY = self.intersect_target_plane(ray)
        
        self.initial_rotation = self.rotation_get(target)
        self.initial_scale = self.scale_get(target)
        
        self.preferred_rotation = self.initial_rotation
        
        decal_gizmo_state.show_volume_shape = self.mode in ('DEPTH', 'HEIGHT')
        if decal_gizmo_state.show_volume_shape:
            decal_gizmo_state.update_volume_shape()
        
        self.is_active = True
        
        return self.modal(context, event)
    
    def exit(self, context, cancel):
        self.is_active = False
        
        if self._timer is not None:
            addon.remove(self._timer)
        
        # Apparently, there can be situations where these are None
        if context.area:
            context.area.header_text_set(None)
        if context.workspace:
            context.workspace.status_text_set(None)
        if context.window:
            context.window.cursor_modal_restore()
        
        decal_gizmo_state.last_time = 0.0
        
        decal_gizmo_state.show_volume_shape = False
        
        target = decal_gizmo_state.target
        
        # Stop drawing the edge on exit
        self.snap_edge_points = None
        
        # It seems that reparenting causes some EXPENSIVE depsgraph
        # recalculations when there are many objects; do it only on invoke/exit
        
        if cancel:
            if target:
                # Note: it seems that setting the transform_attr_names attributes
                # does not immediately update the internal state, so the decal
                # may project with an outdated matrix. However, explicitly
                # assigning the original matrix_world seems to fix that.
                
                target.parent_type = self.init_parent_type
                target.parent = self.init_parent
                target.matrix_parent_inverse = self.matrix_parent_inverse
                target.matrix_world = self.matrix_world
                for attr_name in self.transform_attr_names:
                    setattr(target, attr_name, getattr(self, attr_name))
                
                decal = Decal(target)
                
                decal.frustum = self.initial_frustum
                decal.curvature = self.initial_curvature
                decal.displacement = self.initial_displacement
                decal.extra_scale = self.initial_extra_scale
                decal.range_x = self.initial_range_x
                decal.range_y = self.initial_range_y
                decal.targets = self.initial_targets
                
                self.update_geometry()
                
                for attr_name, attr_value in self.projection_options.items():
                    setattr(decal, attr_name, attr_value)
                
                if not self.initial_is_dynamic: decal.is_dynamic = False
        else:
            self.set_parent(self.new_parent)
            
            if target:
                decal = Decal(target)
                
                for attr_name, attr_value in self.projection_options.items():
                    setattr(decal, attr_name, attr_value)
                
                if not self.initial_is_dynamic: decal.is_dynamic = False
            
            bpy.ops.ed.undo_push(message="Decal gizmo transform")
        
        if target: target.update_tag()
        
        self.raycaster = None
        self.new_parent = None
        
        self.value = None
    
    def update_geometry(self):
        target = decal_gizmo_state.target
        if not target: return
        
        target.update_tag()
        
        decal_gizmo_state.update_volume_shape()
    
    def project_target_origin(self, context):
        target = decal_gizmo_state.target
        if not target: return None
        sv = SmartView3D(context)
        pos = target.matrix_world.translation
        return sv.project(pos, coords='WINDOW')
    
    def intersect_target_plane(self, ray):
        target = decal_gizmo_state.target
        if not target: return None
        from mathutils.geometry import intersect_line_plane
        origin = Vector(target.matrix_world.translation)
        normal = self.get_axis(target.matrix_world, 2, normalize=True)
        p = intersect_line_plane(ray[0], ray[1], origin, normal)
        if p:
            ray_dir = ray[1] - ray[0]
            t = ray_dir.dot(p - ray[0]) / ray_dir.length_squared
            if (t >= 0) and (t <= 1): return Vector(p)
        return ray[1] - normal * (ray[1] - origin).dot(normal)
    
    def bound_box(self):
        target = decal_gizmo_state.target
        if not target: return None
        
        decal = Decal(target)
        if isinstance(decal.template, bpy.types.Material):
            extent_x = 0.5*abs(decal.range_x)
            extent_y = 0.5*abs(decal.range_y)
            extent_z = 0.0
            
            curv_x = decal.curv_min
            curv_y = decal.curv_maj
            angle = decal.angle
            
            curvature_threshold = 0.0001
            pi = math.pi
            pi_2 = pi/2.0
            pi_4 = pi/4.0
            
            if abs(curv_x) > curvature_threshold:
                r_x = 1.0 / abs(curv_x)
                phi_x = extent_x / r_x
                extent_z = (1.0 - math.cos(min(phi_x, pi))) * r_x
                extent_x = math.sin(min(phi_x, pi_2)) * r_x
            
            if abs(curv_y) > curvature_threshold:
                r_y = 1.0 / abs(curv_y)
                phi_y = extent_y / r_y
                if curv_x * curv_y < 0: r_y += extent_z
                extent_y = math.sin(min(phi_y, pi_2)) * r_y
            
            cos_a = abs(math.cos(angle))
            sin_a = abs(math.sin(angle))
            extent_x, extent_y = (
                extent_x*cos_a + extent_y*sin_a,
                extent_y*cos_a + extent_x*sin_a,
            )
            
            angle = abs(angle) % pi_2
            if angle > pi_4: angle = pi_2 - angle
            cos_a = abs(math.cos(angle))
            sin_a = abs(math.sin(angle))
            extent_x, extent_y = (
                extent_x*cos_a + extent_y*sin_a,
                extent_y*cos_a + extent_x*sin_a,
            )
        else:
            bbox = target.bound_box
            extent_x = max(abs(bbox[0][0]), abs(bbox[-2][0]))
            extent_y = max(abs(bbox[0][1]), abs(bbox[-2][1]))
        
        return Bounds.CenterSize(Vector(), Vector((2*extent_x, 2*extent_y, 0)))
    
    def set_parent(self, parent):
        target = decal_gizmo_state.target
        if not target: return
        
        if (target.parent == parent) and (target.parent_type == 'OBJECT'): return
        
        def calc_scale_signs(obj):
            signs = {
                "main": Vector(math.copysign(1, value) for value in obj.scale),
                "delta": Vector(math.copysign(1, value) for value in obj.delta_scale),
            }
            signs["combined"] = Vector(s * ds for s, ds in zip(signs["main"], signs["delta"]))
            signs["product"] = product(signs["combined"])
            signs["negative"] = signs["product"] < 0
            return signs
        
        signs = calc_scale_signs(target)
        
        ######################################################################
        
        matrix = target.matrix_world.copy()
        
        matrix_parent_inverse = (parent.matrix_world.inverted_safe() if parent else Matrix())
        
        if parent: target.parent_type = 'OBJECT'
        target.parent = parent
        target.matrix_parent_inverse = matrix_parent_inverse
        
        if not parent: matrix = matrix_LRS(*matrix.decompose())
        
        target.matrix_world = matrix
        
        ######################################################################
        
        matrix_basis = Matrix(target.matrix_basis)
        X = matrix_basis.col[0].to_3d()
        Y = matrix_basis.col[1].to_3d()
        Z = matrix_basis.col[2].to_3d()
        T = matrix_basis.col[3].to_3d()
        
        is_negative = X.cross(Y).dot(Z) < 0
        if is_negative != signs["negative"]:
            signs["combined"].x *= -1
        
        X *= signs["combined"].x
        Y *= signs["combined"].y
        Z *= signs["combined"].z
        
        L, R, S = matrix_compose(X, Y, Z, T).decompose()
        S = Vector(abs(value) * sign for value, sign in zip(S, signs["combined"]))
        
        self.location_set(target, L)
        self.rotation_set(target, R)
        self.scale_set(target, S)
    
    def init_raycaster(self, context):
        target = decal_gizmo_state.target
        if not target: return []
        
        scene = context.scene
        depsgraph = context.evaluated_depsgraph_get()
        
        exclude = DecalUtils.filter_decals(scene.objects, add_children=True)
        exclude.add(target)
        
        objs = context.visible_objects
        self.raycaster = Raycaster(objs=objs, depsgraph=depsgraph, exclude=exclude)
    
    @staticmethod
    def snap_to_edge(location, rotation, collider, collider_matrix, face_index):
        if face_index < 0: return None, rotation # just in case
        
        matrix = rotation.to_matrix()
        
        pos = collider_matrix.inverted_safe() @ location
        
        mesh = collider.data
        face = mesh.polygons[face_index]
        
        best_dist = math.inf
        best_edge = None
        for v0, v1 in face.edge_keys:
            v0 = mesh.vertices[v0]
            v1 = mesh.vertices[v1]
            dist = dist_to_segment(pos, v0.co, v1.co)
            if dist < best_dist:
                best_dist = dist
                best_edge = v0, v1
        
        axis_x = matrix.col[0].to_3d()
        axis_y = matrix.col[1].to_3d()
        axis_z = matrix.col[2].to_3d()
        
        v0 = collider_matrix @ best_edge[0].co
        v1 = collider_matrix @ best_edge[1].co
        ort = axis_z.cross(v1 - v0)
        
        dot_x = axis_x.dot(ort)
        if dot_x < 0: axis_x = -axis_x
        dot_y = axis_y.dot(ort)
        if dot_y < 0: axis_y = -axis_y
        axis = (axis_x if abs(dot_x) > abs(dot_y) else axis_y)
        
        return (v0, v1), (axis.rotation_difference(ort) @ rotation)
    
    @staticmethod
    def get_axis(matrix, i, normalize=True):
        axis = matrix.col[i].to_3d()
        if normalize: axis = axis.normalized()
        if axis.length_squared > 0: return axis
        A = matrix.col[(i+2) % 3].to_3d().normalized()
        B = matrix.col[(i+1) % 3].to_3d().normalized()
        return A.cross(B)
    
    # Modifying object transform via matrix is not stable in case of negative scale
    
    @staticmethod
    def location_get(obj):
        return obj.location + obj.delta_location
    
    @staticmethod
    def location_set(obj, value):
        obj.location = Vector(value) - obj.delta_location
    
    @staticmethod
    def rotation_get(obj):
        if obj.rotation_mode == 'QUATERNION':
            dr = obj.delta_rotation_quaternion
            r = Quaternion(obj.rotation_quaternion)
        elif obj.rotation_mode == 'AXIS_ANGLE':
            dr = Quaternion()
            aa = obj.rotation_axis_angle
            r = Quaternion(aa[1:], aa[0])
        else:
            dr = obj.delta_rotation_euler.to_quaternion()
            r = obj.rotation_euler.to_quaternion()
        return dr @ r
    
    @staticmethod
    def rotation_set(obj, value):
        if obj.rotation_mode == 'QUATERNION':
            dr = obj.delta_rotation_quaternion
            value = dr.inverted() @ value
            obj.rotation_quaternion = value
        elif obj.rotation_mode == 'AXIS_ANGLE':
            value = value.to_axis_angle()
            obj.rotation_axis_angle = (value[1], *value[0])
        else:
            dr = obj.delta_rotation_euler.to_quaternion()
            value = dr.inverted() @ value
            obj.rotation_euler = value.to_euler(obj.rotation_mode)
    
    @staticmethod
    def scale_get(obj):
        s, ds = obj.scale, obj.delta_scale
        return Vector((s.x*ds.x, s.y*ds.y, s.z*ds.z))
    
    @staticmethod
    def scale_set(obj, value):
        value = Vector(divide(s, ds) for s, ds in zip(value, obj.delta_scale))
        obj.scale = Vector((s if math.isfinite(s) else math.copysign(1, s)) for s in value)
    
    def modal(self, context, event):
        target = decal_gizmo_state.target
        if not target: return {'CANCELLED'}
        
        sv = SmartView3D(context)
        if not sv: return {'CANCELLED'}
        
        # Update animation on any event, but do actual calculations only on user input
        curr_time = time.perf_counter()
        diff_time = curr_time - (decal_gizmo_state.last_time or curr_time)
        decal_gizmo_state.last_time = curr_time
        decal_gizmo_state.anim_time += diff_time * settings.gizmo_anim_speed
        
        if 'TIMER' in event.type:
            context.region.tag_redraw()
            return {'RUNNING_MODAL'}
        
        scene = context.scene
        region = context.region
        rv3d = context.region_data
        view_layer = context.view_layer
        
        placement_options = settings.placement_options
        
        decal = Decal(target)
        
        coord = Vector((event.mouse_region_x, event.mouse_region_y))
        mouse_delta = Vector((event.mouse_x - event.mouse_prev_x, event.mouse_y - event.mouse_prev_y))
        wheel_delta = int(event.type == 'WHEELUPMOUSE') - int(event.type == 'WHEELDOWNMOUSE')
        
        is_click = (event.value == 'PRESS') or (event.value == 'DOUBLE_CLICK')
        
        is_mousemove = ('MOUSEMOVE' in event.type)
        is_invoke = (event.type in ('LEFTMOUSE', 'RIGHTMOUSE')) and is_click
        
        value_step = 0.1
        value_delta = mouse_delta * 0.01
        if event.shift:
            value_delta *= 0.1
            value_step *= 0.1
        
        if is_click:
            if event.type == 'F':
                decal.extra_scale *= Vector((-1, 1, 1))
            elif event.type == 'D':
                decal.extra_scale *= Vector((1, -1, 1))
            elif event.type == 'TAB':
                settings.cycle_decal_alignment_mode(event.shift)
        
        if event.type in ('MIDDLEMOUSE', 'A'):
            if event.value == 'PRESS':
                self.is_snap_to_edge = True
            elif event.value == 'RELEASE':
                self.is_snap_to_edge = False
        
        is_snap_to_edge = self.is_snap_to_edge
        
        decal_alignment_mode = settings.decal_alignment_mode
        
        if wheel_delta != 0:
            if event.alt:
                displacement_step = placement_options.displacement_step
                displacement_step_abs = abs(displacement_step)
                if displacement_step_abs > 1e-6:
                    displacement = decal.displacement + wheel_delta * displacement_step
                    decal.displacement = max(round(displacement / displacement_step_abs), 1) * displacement_step_abs
                    self.update_geometry()
            else:
                angle = wheel_delta * (placement_options.rotation_step_small if event.shift else placement_options.rotation_step)
                rotation = self.rotation_get(target)
                z_axis = rotation @ Vector((0, 0, 1))
                rotation.rotate(Quaternion(z_axis, math.radians(angle)))
                self.rotation_set(target, rotation)
                self.preferred_rotation = rotation
        
        if self.mode == 'MOVE':
            self.snap_edge_points = None
            
            if is_mousemove or is_invoke or is_snap_to_edge:
                ray = sv.ray(coord + self.origin_delta)
                
                success, location, normal, face_index, obj, collider, collider_matrix = self.raycaster.ray_cast(ray[0], ray[1])
                
                if placement_options.auto_parent: self.new_parent = obj
                
                if obj and (obj in self.initial_targets):
                    decal.targets = self.initial_targets + self.selected_targets
                else:
                    decal.targets = [obj] + self.selected_targets
                
                if success:
                    if decal_alignment_mode == 'UPRIGHT':
                        rotated_x = (self.preferred_rotation @ Vector((1, 0, 0)))
                        rotated_z = (self.preferred_rotation @ Vector((0, 0, 1)))
                        
                        if rotated_z.to_2d().magnitude > 1e-6:
                            ort_x = rotated_z.cross(Vector((0, 0, 1))).normalized()
                        else:
                            ort_x = Vector((1, 0, 0))
                        ort_y = rotated_z.cross(ort_x)
                        
                        angle_roll = math.atan2(ort_y.dot(rotated_x), ort_x.dot(rotated_x))
                        angle_pitch = math.asin(normal.z) - math.pi/2
                        angle_yaw = math.atan2(-normal.x, normal.y)
                        
                        rotation_roll = Quaternion((0, 0, 1), angle_roll)
                        rotation_pitch = Quaternion((1, 0, 0), angle_pitch)
                        rotation_yaw = Quaternion((0, 0, 1), angle_yaw)
                        rotation = rotation_yaw @ rotation_pitch @ rotation_roll
                    else:
                        use_preferred = (decal_alignment_mode == 'INITIAL')
                        rotation = (self.preferred_rotation if use_preferred else self.rotation_get(target))
                        z_axis = rotation @ Vector((0, 0, 1))
                        rotation = z_axis.rotation_difference(normal) @ rotation
                    
                    if decal_alignment_mode == 'EDGE':
                        is_snap_to_edge = not is_snap_to_edge
                    
                    if is_snap_to_edge:
                        result = self.snap_to_edge(location, rotation, collider, collider_matrix, face_index)
                        self.snap_edge_points, rotation = result
                        self.preferred_rotation = rotation
                    
                    self.location_set(target, location)
                    self.rotation_set(target, rotation)
                else:
                    forward = sv.forward
                    pos = self.location_get(target)
                    t0 = forward.dot(ray[0])
                    t1 = forward.dot(ray[1])
                    t = forward.dot(pos)
                    t = (t - t0) / (t1 - t0)
                    pos = ray[0].lerp(ray[1], t)
                    self.location_set(target, pos)
        elif self.mode == 'ROTATE':
            rotation_step = None
            if event.shift:
                #rotation_step = placement_options.rotation_step_small
                rotation_step = 45.0 # for now, use what moth3r asked
            elif event.ctrl:
                rotation_step = placement_options.rotation_step
            elif event.alt:
                #rotation_step = 45.0
                if placement_options.rotation_snap_by_default:
                    rotation_step = None
                else:
                    rotation_step = placement_options.rotation_step_small # for now, use what moth3r asked
            else:
                if placement_options.rotation_snap_by_default:
                    rotation_step = placement_options.rotation_step_small
                else:
                    rotation_step = None
            
            ray = sv.ray(coord)
            origin = self.location_get(target)
            normal = self.initial_rotation @ Vector((0, 0, 1))
            
            direction_src = (self.initial_intersection_XY - origin).normalized()
            direction_dst = (self.intersect_target_plane(ray) - origin).normalized()
            direction_src -= normal * direction_src.dot(normal)
            direction_dst -= normal * direction_dst.dot(normal)
            
            rotation_delta = direction_src.rotation_difference(direction_dst)
            axis, angle = rotation_delta.to_axis_angle()
            if rotation_step: angle = math.radians(round(math.degrees(angle) / rotation_step) * rotation_step)
            axis = (normal if axis.dot(normal) >= 0 else -normal)
            rotation_delta = Quaternion(axis, angle)
            
            rotation = rotation_delta @ self.initial_rotation
            self.rotation_set(target, rotation)
            self.preferred_rotation = rotation
        elif self.mode == 'SCALE':
            scale_step = None
            if event.shift:
                scale_step = 0.01
            elif event.ctrl:
                scale_step = 0.1
            
            ray = sv.ray(coord)
            origin = self.location_get(target)
            
            delta_src = (self.initial_intersection_XY - origin)
            delta_dst = (self.intersect_target_plane(ray) - origin)
            
            scale = delta_dst.magnitude / delta_src.magnitude
            if delta_src.dot(delta_dst) < 0: scale = -scale
            if scale_step: scale = round(scale / scale_step) * scale_step
            
            if self.invoke_alt:
                if (mouse_delta.x != 0) or (mouse_delta.y != 0):
                    if isinstance(decal.template, bpy.types.Material):
                        matrix = target.matrix_world.inverted_safe().to_3x3()
                        delta_src = matrix @ delta_src
                        delta_dst = matrix @ delta_dst
                        scale_x = delta_dst.x / delta_src.x
                        scale_y = delta_dst.y / delta_src.y
                        if scale_step: scale_x = round(scale_x / scale_step) * scale_step
                        if scale_step: scale_y = round(scale_y / scale_step) * scale_step
                        decal.range_x = self.initial_range_x * scale_x
                        decal.range_y = self.initial_range_y * scale_y
                    else:
                        decal.extra_scale = self.initial_extra_scale * scale
                    self.update_geometry()
            else:
                scale = self.initial_scale * scale
                self.scale_set(target, scale)
        elif self.mode == 'CURVATURE_X':
            if value_delta.x != 0:
                if self.value is None: self.value = decal.curv_min
                self.value = self.value + value_delta.x
                value = self.value
                if event.ctrl: value = round(value / value_step) * value_step
                decal.curv_min = value
                self.update_geometry()
        elif self.mode == 'CURVATURE_Y':
            if value_delta.x != 0:
                if self.value is None: self.value = decal.curv_maj
                self.value = self.value + value_delta.x
                value = self.value
                if event.ctrl: value = round(value / value_step) * value_step
                decal.curv_maj = value
                self.update_geometry()
        elif self.mode == 'CURVATURE':
            if value_delta.length_squared != 0:
                if self.value is None: self.value = Vector(decal.curvature)
                self.value.x += value_delta.x
                self.value.y += value_delta.y
                value = Vector(self.value)
                if event.ctrl:
                    value.x = round(value.x / value_step) * value_step
                    value.y = round(value.y / value_step) * value_step
                value.x = value.x
                value.y = value.y
                decal.curvature = Vector(value)
                self.update_geometry()
        elif self.mode in ('DEPTH', 'HEIGHT'):
            if value_delta.length_squared != 0:
                origin = self.location_get(target)
                axis = self.rotation_get(target) @ Vector((0, 0, sv.distance * 0.01))
                proj_start = sv.project(origin)
                proj_end = sv.project(origin + axis)
                if (proj_start is None) or (proj_end is None):
                    origin = sv.focus
                    proj_start = sv.project(origin)
                    proj_end = sv.project(origin + axis)
                screen_axis = (proj_end - proj_start).normalized()
                if screen_axis.length_squared == 0: screen_axis = Vector((1, 0))
                
                attr_name = ("depth_min" if self.mode == 'DEPTH' else "depth_max")
                if self.value is None: self.value = getattr(decal, attr_name)
                self.value = self.value + value_delta.dot(screen_axis)
                value = self.value
                if event.ctrl: value = round(value / value_step) * value_step
                setattr(decal, attr_name, value)
                self.update_geometry()
        
        decal_gizmo_state.show_volume_shape = self.mode in ('DEPTH', 'HEIGHT')
        
        if is_snap_to_edge:
            decal_alignment_name = "edge"
        elif decal_alignment_mode == 'UPRIGHT':
            decal_alignment_name = "upright"
        elif decal_alignment_mode == 'INITIAL':
            decal_alignment_name = "initial"
        else:
            decal_alignment_name = "current"
        
        context.area.header_text_set(self.get_header_text(decal_alignment_name))
        context.workspace.status_text_set(self.get_status_text())
        
        return {'RUNNING_MODAL'}
    
    def get_header_text(self, decal_alignment_name):
        target = decal_gizmo_state.target
        if not target: ""
        
        decal = Decal(target)
        
        mode_name = ""
        if self.mode == 'MOVE':
            mode_name = f"Move decal (align: {decal_alignment_name})"
        elif self.mode == 'ROTATE':
            mode_name = "Rotate decal"
        elif self.mode == 'SCALE':
            mode_name = "Scale decal"
        elif self.mode == 'CURVATURE_X':
            mode_name = f"X curvature = {decal.curvature.x:.3f}"
        elif self.mode == 'CURVATURE_Y':
            mode_name = f"Y curvature = {decal.curvature.y:.3f}"
        elif self.mode == 'CURVATURE':
            mode_name = f"Curvature: X={decal.curvature.x:.3f}, Y={decal.curvature.y:.3f}"
        elif self.mode == 'DEPTH':
            mode_name = f"Depth below = {decal.depth_min:.3f}"
        elif self.mode == 'HEIGHT':
            mode_name = f"Depth above = {decal.depth_max:.3f}"
        
        return f"{mode_name}; clearance = {decal.displacement:.4f}"
    
    def get_status_text(self):
        result = "F: flip x, D: flip y, Alt+Wheel: offset"
        
        if self.mode == 'MOVE':
            result = f"{result}, Tab: switch alignment, A or MMB: snap to edge, Wheel: rotate (+shift: small step)"
        elif self.mode == 'ROTATE':
            result = f"{result}, Shift: 45 degrees, Ctrl: step, Alt: small step"
        elif self.mode == 'SCALE':
            result = f"{result}, Ctrl: 0.1 step, Shift: 0.01 step"
        else:
            result = f"{result}, Ctrl: 0.1 step, Shift: 0.01 step"
        
        return result

class GizmoShape:
    def __init__(self, type, verts):
        if verts:
            self.dims = len(verts[0])
            if self.dims not in (2, 3): raise ValueError("Expected 2D or 3D vertex")
        else:
            self.dims = 3 # 2D or 3D doesn't really matter here
        
        if self.dims == 2:
            verts = [(v[0], v[1], 0.0) for v in verts]
            self.dims = 3
        
        self.type = type
        self.verts = verts
        
        self.shader = gpu.shader.from_builtin('UNIFORM_COLOR')
        
        from gpu.types import GPUBatch, GPUVertBuf, GPUVertFormat
        self.fmt = GPUVertFormat()
        self.pos_id = self.fmt.attr_add(id="pos", comp_type='F32', len=self.dims, fetch_mode='FLOAT')
        self.vbo = GPUVertBuf(len=len(self.verts), format=self.fmt)
        self.vbo.attr_fill(id=self.pos_id, data=self.verts)
        self.batch = GPUBatch(type=self.type, buf=self.vbo)
    
    def draw(self, matrix, color, select_id=None):
        if select_id is not None:
            # BLENDER BUG: it seems that selection will glitch if we try
            # to draw a beforehand-created shape. However, if we create
            # the shape right here, it seems fine.
            from gpu.types import GPUBatch
            batch = GPUBatch(type=self.type, buf=self.vbo)
        else:
            batch = self.batch
        
        self.shader.bind()
        
        # To truly solve the issue with antialiased seams,
        # we need an equivalent of GL_POLYGON_SMOOTH.
        # But always blending makes it less noticeable.
        
        if select_id is not None:
            gpu.select.load_id(select_id)
            use_blend = False
        else:
            self.shader.uniform_float("color", color)
            use_blend = True
        
        if use_blend:
            cgl.blend = 'ALPHA'
        
        with gpu.matrix.push_pop():
            gpu.matrix.multiply_matrix(matrix)
            batch.draw(self.shader)
        
        if use_blend:
            cgl.blend = 'NONE'

class DecalGizmoBase:
    # This attribute isn't defined in the base Gizmo class, so we
    # have to specify it here instead of in the @addon.Gizmo decorator
    #bl_target_properties = ()
    
    base = None
    
    ignore_modal = False
    
    drag_only = False # overridable
    
    initial_mouse_pos = Vector((0, 0))
    
    cursor = None
    
    def setup(self):
        self.base = DecalTransformBase()
    
    def invoke(self, context, event):
        self.ignore_modal = False
        
        self.initial_mouse_pos = Vector((event.mouse_x, event.mouse_y))
        
        if event.type == 'RIGHTMOUSE':
            if event.value != 'PRESS': return {'PASS_THROUGH'}
            
            bpy.ops.wm.call_panel(name="EMPTY_PT_xdecal_context_view3d")
            
            if self.cursor: context.window.cursor_modal_set(self.cursor)
            self.ignore_modal = True
            return {'RUNNING_MODAL'}
        elif event.type == 'LEFTMOUSE':
            if self.drag_only:
                if event.value != 'CLICK_DRAG': return {'PASS_THROUGH'}
                
                if decal_gizmo_state.blocked_count > 0:
                    decal_gizmo_state.blocked_count -= 1
                    return {'PASS_THROUGH'}
                
                frame_delta = addon.runtime.frame_count - decal_gizmo_state.exit_frame
                
                if frame_delta <= 1:
                    if self.cursor: context.window.cursor_modal_set(self.cursor)
                    self.ignore_modal = True
                    return {'RUNNING_MODAL'}
            else:
                if event.value != 'PRESS': return {'PASS_THROUGH'}
            
            return self.base.invoke(context, event)
        else:
            return {'PASS_THROUGH'}
    
    def exit(self, context, cancel):
        context.window.cursor_modal_restore()
        
        if self.ignore_modal:
            if self.use_grab_cursor:
                self.delayed_cursor_warp(self.initial_mouse_pos, context)
            return
        
        decal_gizmo_state.exit_frame = addon.runtime.frame_count
        
        self.base.exit(context, cancel)
        
        # If gizmo.use_grab_cursor is True, Blender automatically
        # resets mouse to its initial position after exit()
        if self.use_grab_cursor:
            proj = self.base.project_target_origin(context)
            if proj: self.delayed_cursor_warp(proj - self.base.origin_delta, context)
    
    def modal(self, context, event, tweak):
        if self.ignore_modal: return {'CANCELLED'}
        return self.base.modal(context, event)
    
    def delayed_cursor_warp(self, pos, context):
        x, y = pos
        window = context.window
        def job():
            window.cursor_warp(int(x), int(y))
        Jobs.background.appendleft(job) # high-priority

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_position")
class DecalPositionGizmo(DecalGizmoBase):
    drag_only = True
    
    def draw(self, context):
        colors = settings.colors
        
        if self.is_modal:
            color = colors["decal_active"]
        elif self.is_highlight:
            color = colors["decal_highlight"]
        else:
            color = colors["decal_main"]
        
        self.draw_shape(self.base, selection=False, color=color)
        
        self.draw_alignment_indicator()
    
    def draw_select(self, context, select_id):
        self.draw_shape(self.base, selection=True, select_id=select_id)
    
    def draw_alignment_indicator(self):
        if not self.base: return
        
        target = decal_gizmo_state.target
        if not target: return False
        
        matrix = target.matrix_world.to_3x3()
        bbox_size = self.base.bound_box().get("size", convert=Vector)
        size_x = (matrix @ Vector((bbox_size[0], 0, 0))).magnitude
        size_y = (matrix @ Vector((0, bbox_size[1], 0))).magnitude
        size_z = (matrix @ Vector((0, 0, bbox_size[2]))).magnitude
        size_max = max(size_x, size_y, size_z)
        
        decal_alignment_mode = settings.decal_alignment_mode
        
        snap_edge_points = getattr(self.base, "snap_edge_points", None)
        
        if snap_edge_points:
            p0, p1 = snap_edge_points
            midpoint = (p0 + p1) * 0.5
            delta = p1 - p0
            scale = size_max / delta.magnitude
            delta *= 0.5 * (scale if math.isfinite(scale) else 1)
            p0 = midpoint - delta
            p1 = midpoint + delta
        elif self.base.is_active:
            if decal_alignment_mode == 'UPRIGHT':
                p0 = target.matrix_world.translation
                p1 = p0 + Vector((0, 0, size_max * 0.5))
            elif decal_alignment_mode == 'INITIAL':
                p0 = self.base.matrix_world.translation
                p1 = target.matrix_world.translation
            else:
                return
        else:
            return
        
        colors = settings.colors
        
        color = colors["decal_edge"]
        color_inv = (1.0-color[0], 1.0-color[1], 1.0-color[2], color[3])
        
        stipple_count = 3
        stipple_start = 0.25 + decal_gizmo_state.anim_time
        stipple_delta = 0.5
        stipple = [*self.generate_stipple(p0, p1, stipple_count, stipple_start, stipple_delta)]
        
        shader = gpu.shader.from_builtin('UNIFORM_COLOR')
        shader.bind()
        with cgl(depth_test='ALWAYS', blend='ALPHA', depth_mask=False):
            shader.uniform_float("color", color_inv)
            batch = cgl.batch(shader, 'LINES', pos=[p0, p1])
            batch.draw(shader)
            
            shader.uniform_float("color", color)
            batch = cgl.batch(shader, 'LINES', pos=stipple)
            batch.draw(shader)
    
    # Also used by the palette gizmo
    @classmethod
    def draw_shape(cls, base, selection, color=None, select_id=None):
        decal_gizmo_state.draw_custom_shape(cls.make_shape(base, selection), color=color, select_id=select_id)
    
    @classmethod
    def generate_stipple(cls, p0, p1, count, f0=0.25, df=0.5):
        p0 = Vector(p0)
        p1 = Vector(p1)
        
        f0 = f0 % 1
        df = min(max(df, 0), 1)
        
        norm = 1.0 / count
        
        if f0+df > 1:
            count += 1
            f0 -= 1
        
        f0 *= norm
        df *= norm
        f1 = f0 + df
        
        for i in range(count):
            t = i * norm
            yield p0.lerp(p1, max(t + f0, 0))
            yield p0.lerp(p1, min(t + f1, 1))
    
    @classmethod
    def make_shape(cls, base, selection):
        target = decal_gizmo_state.target
        if not target: return GizmoShape('TRIS', [])
        
        center, half = base.bound_box().get("center", "extents", convert=Vector)
        
        if not selection: half += Vector((1, 1, 0)) * (max(half.x, half.y) * base.gizmo_margin)
        
        vert00 = (center.x-half.x, center.y-half.y, 0)
        vert01 = (center.x-half.x, center.y+half.y, 0)
        vert10 = (center.x+half.x, center.y-half.y, 0)
        vert11 = (center.x+half.x, center.y+half.y, 0)
        
        if selection:
            return GizmoShape('TRIS', (
                vert00, vert01, vert11,
                vert00, vert11, vert10,
            ))
        
        stipple_count = 10
        stipple_start = 0.25 + decal_gizmo_state.anim_time
        stipple_delta = 0.5
        
        return GizmoShape('LINES', [
            *cls.generate_stipple(vert00, vert01, stipple_count, stipple_start, stipple_delta),
            *cls.generate_stipple(vert01, vert11, stipple_count, stipple_start, stipple_delta),
            *cls.generate_stipple(vert11, vert10, stipple_count, stipple_start, stipple_delta),
            *cls.generate_stipple(vert10, vert00, stipple_count, stipple_start, stipple_delta),
        ])

class DecalGizmoIconBase(DecalGizmoBase):
    def active_gizmo_positions(self):
        return self.gizmo_positions
    
    def draw(self, context):
        if hasattr(self, "poll") and not self.poll(): return
        
        self.draw_select(context, None)
    
    def draw_select(self, context, select_id):
        target = decal_gizmo_state.target
        if not target: return
        
        decal = Decal(target)
        if not decal: return
        
        if not target.select_get(): return
        
        if hasattr(self, "poll") and not self.poll(): return
        
        colors = settings.colors
        
        if select_id is not None:
            color = None
        elif self.is_modal:
            color = colors["decal_active"]
        elif self.is_highlight:
            color = colors["decal_highlight"]
        else:
            color = colors["decal_main"]
        
        dS = target.delta_scale
        dS = (math.copysign(1, dS[0]), math.copysign(1, dS[1]), math.copysign(1, dS[2]))
        
        matrix = Matrix(target.matrix_world)
        
        matrix.col[0] *= dS[0]
        matrix.col[1] *= dS[1]
        matrix.col[2] *= dS[2]
        
        T = matrix.translation
        
        center, half = self.base.bound_box().get("center", "extents", convert=Vector)
        center.z = 0
        half.z = 0
        center.x *= dS[0]
        center.y *= dS[1]
        half += Vector((1, 1, 0)) * (max(half.x, half.y) * self.base.gizmo_margin)
        
        X, Y, Z = orthogonal_XYZ(matrix.col[0].to_3d(), matrix.col[1].to_3d(), None)
        X = X.normalized()
        Y = Y.normalized()
        Z = Z.normalized()
        
        shape = GizmoShape('TRIS', gizmo_icons[self.gizmo_icon])
        
        frustumX = (matrix @ Vector((half.x, 0, 0))) - T
        frustumY = (matrix @ Vector((0, half.y, 0))) - T
        frustumZ = -Z * decal.depth_min
        
        shape_size = (frustumX.magnitude + frustumY.magnitude) * 0.5 * settings.gizmo_scale
        
        dX = X * shape_size
        dY = Y * shape_size
        dZ = Z * shape_size
        
        for gizmo_pos in self.active_gizmo_positions():
            decal_signs = gizmo_pos["decal"]
            icon_signs = gizmo_pos["icon"]
            axes_signs = gizmo_pos["axes"]
            angle = gizmo_pos.get("angle")
            delta_signs = gizmo_pos.get("ds")
            ofs = matrix @ center
            ofs += frustumX*decal_signs[0] + frustumY*decal_signs[1] + frustumZ*decal_signs[2]
            ofs += dX*icon_signs[0] + dY*icon_signs[1] + dZ*icon_signs[2]
            axes = dX*axes_signs[0], dY*axes_signs[1], dZ*axes_signs[2]
            if delta_signs:
                axes = [(axis*ds*ds2 if ds2 else axis) for axis, ds, ds2 in zip(axes, dS, delta_signs)]
            matrix_draw = matrix_compose(axes[0], axes[1], axes[2], ofs)
            if angle: matrix_draw = matrix_draw @ Matrix.Rotation(math.radians(angle), 4, 'Z')
            shape.draw(matrix_draw, color, select_id=select_id)

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_rotation")
class DecalRotationGizmo(DecalGizmoIconBase):
    gizmo_icon = "rotate"
    gizmo_positions = (
        dict(decal=(-1, 1, 0), icon=(-1, 1, 0), axes=(1, 1, 1)),
        dict(decal=(1, -1, 0), icon=(1, -1, 0), axes=(-1, -1, 1)),
    )

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_scale")
class DecalScaleGizmo(DecalGizmoIconBase):
    gizmo_icon = "scale"
    gizmo_positions = (
        dict(decal=(-1, -1, 0), icon=(-1, -1, 0), axes=(1, -1, 1)),
        dict(decal=(1, 1, 0), icon=(1, 1, 0), axes=(-1, 1, 1)),
    )

class DecalFlipGizmoBase(DecalGizmoIconBase):
    def invoke(self, context, event):
        if (event.type != 'LEFTMOUSE') or (event.value != 'PRESS'):
            return {'PASS_THROUGH'}
        
        target = decal_gizmo_state.target
        if not target: return {'CANCELLED'}
        
        decal = Decal(target)
        if not decal: return {'CANCELLED'}
        
        decal.extra_scale = Vector(v*s for v, s in zip(decal.extra_scale, self.flip_scale))
        decal.update()
        
        bpy.ops.ed.undo_push(message="Decal Flip")
        
        # FINISHED does not consume the event, but RUNNING_MODAL does
        return {'RUNNING_MODAL'}
    
    def exit(self, context, cancel):
        context.window.cursor_modal_restore()
        return # do nothing (override the base behavior)
    
    def modal(self, context, event, tweak):
        return {'FINISHED'} # exit immediately

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_flip_x")
class DecalFlipXGizmo(DecalFlipGizmoBase):
    gizmo_icon = "flip"
    gizmo_positions = (
        dict(decal=(1, 0, 0), icon=(2, 0, 0), axes=(1, 1, 1), ds=(1, 1, 1), angle=90.0),
    )
    flip_scale = (-1, 1, 1)

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_flip_y")
class DecalFlipYGizmo(DecalFlipGizmoBase):
    gizmo_icon = "flip"
    gizmo_positions = (
        dict(decal=(0, -1, 0), icon=(0, -2, 0), axes=(1, 1, 1), ds=(1, 1, 1)),
    )
    flip_scale = (1, -1, 1)

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_curvature_x")
class DecalCurvatureXGizmo(DecalGizmoIconBase):
    gizmo_icon = "curvature"
    gizmo_positions = (
        dict(decal=(-1, 0, 0), icon=(-2, 0, 0), axes=(1, 1, 1), angle=90.0),
    )

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_curvature_y")
class DecalCurvatureYGizmo(DecalGizmoIconBase):
    gizmo_icon = "curvature"
    gizmo_positions = (
        dict(decal=(0, 1, 0), icon=(0, 2, 0), axes=(1, 1, 1)),
    )

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_curvature")
class DecalCurvatureGizmo(DecalGizmoIconBase):
    gizmo_icon = "curvature"
    gizmo_positions = (
        dict(decal=(-1, 0, 0), icon=(-2, 0, 0), axes=(1, 1, 1), angle=90.0),
    )

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_depth")
class DecalDepthGizmo(DecalGizmoIconBase):
    gizmo_icon = "depth"
    gizmo_positions = (
        # dict(decal=(0, 1, 0), icon=(3, 5, 0), axes=(1, 1, -1)),
        dict(decal=(0, 1, 0), icon=(3, 2, 0), axes=(1, 1, -1)),
    )
    
    def poll(self):
        target = decal_gizmo_state.target
        if not target: return False
        decal = Decal(target)
        if not decal: return False
        return decal.use_project

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_height")
class DecalHeightGizmo(DecalGizmoIconBase):
    gizmo_icon = "depth"
    gizmo_positions = (
        # dict(decal=(0, 1, 0), icon=(-3, 5, 0), axes=(1, 1, 1)),
        dict(decal=(0, 1, 0), icon=(-3, 2, 0), axes=(1, 1, 1)),
    )
    
    def poll(self):
        target = decal_gizmo_state.target
        if not target: return False
        decal = Decal(target)
        if not decal: return False
        return decal.use_project

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_project")
class DecalProjectGizmo(DecalGizmoIconBase):
    gizmo_icon = "resolution" # for now
    gizmo_positions = (
        # dict(decal=(0, 1, 0), icon=(0, 5, 0), axes=(1, 1, 1)),
        dict(decal=(0, 1, 0), icon=(0, 2, 0), axes=(1, 1, 1)),
    )
    
    def active_gizmo_positions(self):
        target = decal_gizmo_state.target
        decal = Decal(target)
        if decal and decal.use_embed:
            return (
                # dict(decal=(0, 1, 0), icon=(0, 5, 0), axes=(1, 1, 1), angle=45.0),
                dict(decal=(0, 1, 0), icon=(0, 2, 0), axes=(1, 1, 1), angle=45.0),
            )
        return self.gizmo_positions
    
    def invoke(self, context, event):
        if (event.type != 'LEFTMOUSE') or (event.value != 'PRESS'):
            return {'PASS_THROUGH'}
        
        target = decal_gizmo_state.target
        if not target: return {'CANCELLED'}
        
        decal = Decal(target)
        
        if event.shift:
            decal.embed(not decal.use_embed)
            bpy.ops.ed.undo_push(message="Decal Embed "+("ON" if decal.use_project else "OFF"))
        elif event.ctrl:
            decal.use_project = True
            decal.targets = decal.targets + list(context.selected_objects)
            decal.auto_depth()
            decal.update()
            bpy.ops.ed.undo_push(message="Decal Project Auto-Depth")
        else:
            decal.use_project = not decal.use_project
            decal.update()
            bpy.ops.ed.undo_push(message="Decal Project "+("ON" if decal.use_project else "OFF"))
        
        # FINISHED does not consume the event, but RUNNING_MODAL does
        return {'RUNNING_MODAL'}
    
    def exit(self, context, cancel):
        context.window.cursor_modal_restore()
        return # do nothing (override the base behavior)
    
    def modal(self, context, event, tweak):
        return {'FINISHED'} # exit immediately

@addon.GizmoGroup(idname = "OBJECT_GGT_xdecal", label = f"{addon.name} Widget",
    space_type = 'VIEW_3D', region_type = 'WINDOW', options = {'3D', 'PERSISTENT', 'SCALE'})
class DecalGizmoGroup:
    @classmethod
    def poll(cls, context):
        obj = context.object
        return (obj and obj.type == 'MESH' and obj.mode == 'OBJECT' and Decal.exists(obj))
    
    @classmethod
    def setup_keymap(cls, keyconfig):
        # By default, gizmos only have a left mouse press/tweak defined in their keymap
        km = keyconfig.keymaps.find("xdecal:gizmo", space_type='EMPTY', region_type='WINDOW')
        if km: keyconfig.keymaps.remove(km)
        km = keyconfig.keymaps.new("xdecal:gizmo", space_type='EMPTY', region_type='WINDOW')
        # LEFTMOUSE is also necessary if gizmo.use_event_handle_all == True
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFTMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'MIDDLEMOUSE', 'PRESS', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'RIGHTMOUSE', 'PRESS', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFTMOUSE', 'CLICK_DRAG', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFTMOUSE', 'DOUBLE_CLICK', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELUPMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELDOWNMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELINMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELOUTMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'MOUSEMOVE', 'ANY', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'INBETWEEN_MOUSEMOVE', 'ANY', any=True)
        return km
    
    @addon.on_unregister
    def clear_keymap():
        wm = bpy.context.window_manager
        for kc in wm.keyconfigs:
            km = kc.keymaps.find("xdecal:gizmo", space_type='EMPTY', region_type='WINDOW')
            if km: kc.keymaps.remove(km)
    
    def setup(self, context):
        # NOTE: if gizmo.use_grab_cursor = True, then, when
        # the mouse moves fast during the gizmo dragging,
        # mouse coordinates may jump in the opposite direction.
        # Could this be related to delayed_cursor_warp()?
        
        priority = -1000
        
        def add(gizmo_cls, mode=None, bias=0, grab=False, cursor=None):
            gizmo = self.gizmos.new(gizmo_cls.bl_idname)
            gizmo.scale_basis = 1.0
            gizmo.use_draw_modal = True
            gizmo.select_bias = priority + bias
            gizmo.use_grab_cursor = grab
            gizmo.cursor = cursor
            if mode: gizmo.base.default_mode = mode
        
        # NOTE: initialize the position gizmo last, so that
        # volume shape is drawn "under" the other gizmo icons.
        add(DecalRotationGizmo, mode='ROTATE', bias=1)
        add(DecalScaleGizmo, mode='SCALE', bias=1)
        add(DecalFlipXGizmo, bias=1)
        add(DecalFlipYGizmo, bias=1)
        # add(DecalCurvatureXGizmo, mode='CURVATURE_X', grab=True, bias=1)
        # add(DecalCurvatureYGizmo, mode='CURVATURE_Y', grab=True, bias=1)
        add(DecalCurvatureGizmo, mode='CURVATURE', grab=True, bias=1)
        add(DecalDepthGizmo, mode='DEPTH', grab=True, bias=1)
        add(DecalHeightGizmo, mode='HEIGHT', grab=True, bias=1)
        add(DecalProjectGizmo, bias=1)
        add(DecalPositionGizmo, mode='MOVE', cursor='NONE')
        
        self.refresh(context)
    
    def refresh(self, context):
        decal_gizmo_state.target = context.object
