# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import os

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "prop, IDTypes, IDProp",
    "utils_ui": "BlUI, NestedLayout",
    "utils_blender": "BlUtil, BpyPath",
    "utils_addon": "AddonManager",
})

from . import common
from . import material_utils
from . import decal_utils
from . import palette
from . import template_material_menu

from .common import Jobs, CommonUtils
from .material_utils import MaterialLibrary, TextureSynonym, MaterialMixer, MaterialUtils, NodeInfo
from .decal_utils import DecalUtils, Decal
from .palette import Palette
from .template_material_menu import TemplateMaterialMenu

addon = AddonManager()
settings = addon.settings
previews = addon.previews

# =========================================================================== #

def get_active_selected_object(context):
    obj = context.object
    return (obj if obj and obj.select_get() else None)

def get_selected_or_all_decals(context):
    if len(context.selected_objects) == 0:
        return context.scene.objects
    
    child_map = BlUtil.Object.map_children()[0]
    objs = set()
    for obj in context.selected_objects:
        objs.add(obj)
        BlUtil.Object.all_children(obj, result=objs, child_map=child_map)
    
    return objs

def unselect_decals(context):
    for obj in list(context.selected_objects):
        if not Decal.exists(obj): continue
        obj.select_set(False, view_layer=context.view_layer)

@addon.Operator(idname="xdecal.repair_node_group_refs", label="Repair", options={'INTERNAL'},
    description=f"Fix/update references to the {addon.name} node groups\n"+
    "(After this, you need to save and reload the file manually)")
class RepairAddonLibraryReferencesOperator:
    def execute(self, context):
        prefix = "XDecal."
        
        current_lib_path = CommonUtils.abspath(MaterialLibrary.default)
        
        reload_targets = []
        for node_group in bpy.data.node_groups:
            if not node_group.library: continue
            if not node_group.name.startswith(prefix): continue
            lib_path = CommonUtils.abspath(node_group.library.filepath)
            if lib_path != current_lib_path:
                reload_targets.append(node_group)
        
        CommonUtils.load_resources("node_groups", reload_targets, current_lib_path, link=True, force=True)
        
        bpy.ops.ed.undo_push(message=f"Repair {addon.name} references")
        
        BlUI.tag_redraw()
        
        return {'FINISHED'}

@addon.Operator(idname="xdecal.select_related_files", label="Select related", description="Select related PBR textures", options={'INTERNAL'})
class SelectRelatedFilesOperator:
    @classmethod
    def poll(cls, context):
        return settings.files_show_get("any") and any(item.selected for item in settings.files.items)
    
    def execute(self, context):
        settings.files.select_related()
        context.region.tag_redraw()
        return {'FINISHED'}

@addon.Operator(idname="xdecal.replace_decals_with_files", label="Replace decals", options={'INTERNAL'})
class ReplaceDecalsWithFilesOperator:
    "Replace the selected decals with the chosen decal"
    
    @classmethod
    def poll(cls, context):
        return settings.files_show_get("any") and any(item.selected for item in settings.files.items)
    
    def execute(self, context):
        selected_names = settings.files.selected_names
        if not selected_names: return {'CANCELLED'}
        settings.files.replace(selected_names[0])
        context.region.tag_redraw()
        return {'FINISHED'}

@addon.Operator(idname="xdecal.replace_images_with_files", label="Replace images", options={'INTERNAL'})
class ReplaceDecalsWithFilesOperator:
    "Make copies of the selected decals' materials, and replace images in them with the chosen file"
    
    @classmethod
    def poll(cls, context):
        return settings.files_show_get("any") and any(item.selected for item in settings.files.items)
    
    def execute(self, context):
        selected_names = settings.files.selected_names
        if not selected_names: return {'CANCELLED'}
        settings.files.replace_images(selected_names)
        context.region.tag_redraw()
        return {'FINISHED'}

@addon.Operator(idname="xdecal.convert", label="Convert", description="Convert meshes/empties to decals", options={'INTERNAL'})
class ConvertToDecalsOperator:
    push_undo: False | prop()
    
    def execute(self, context):
        # Don't create decals if the scene is linked from a library
        if context.scene.library is not None: return {'CANCELLED'}
        
        depsgraph = context.evaluated_depsgraph_get()
        
        convertible_types = ('MESH', 'EMPTY')
        
        has_empties = False
        candidates = set()
        for obj in context.selected_objects:
            if obj.type not in convertible_types: continue
            if Decal.exists(obj): continue
            if obj.type == 'EMPTY':
                if obj.empty_display_type != 'IMAGE': continue
                if not obj.data: continue
                has_empties = True
            candidates.add(obj)
        
        if not candidates: return {'CANCELLED'}
        
        if has_empties:
            operator = (self if getattr(self, "is_invoked", None) else None)
            material = MaterialLibrary.try_load_material(operator=operator)
        else:
            material = None
        
        view_layer = context.view_layer
        scene = context.scene
        
        eval_objs = {}
        for obj in scene.objects:
            if obj.type not in Decal.geo_nodes_usable_types: continue
            if obj in candidates: continue
            if Decal.exists(obj): continue
            eval_objs[obj] = obj.evaluated_get(depsgraph)
        
        new_objs = []
        for obj in candidates:
            new_obj = DecalUtils.convert_to_decal(obj, depsgraph, eval_objs, material)
            if new_obj is None: continue
            new_objs.append((obj, new_obj))
        
        # Remap objects afterwards to avoid extra modifications to depsgraph
        with BlUtil.Data.OrphanCleanup(exclude=["objects"]):
            for obj, new_obj in new_objs:
                obj.user_remap(new_obj)
                bpy.data.objects.remove(obj)
        
        if not new_objs: return {'CANCELLED'}
        
        DecalUtils.render_preview(Decal(new_obj).template for obj, new_obj in new_objs)
        
        view_layer.update()
        
        if self.push_undo: bpy.ops.ed.undo_push(message="Convert to Decals")
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        self.is_invoked = True
        self.push_undo = True
        return self.execute(context)

@addon.Operator(idname="xdecal.load", label="Open", description="Select a folder / load decals", options={'INTERNAL'})
class LoadDecalsOperator:
    directory: "" | prop(maxlen=1024, subtype='FILE_PATH', options={'HIDDEN', 'SKIP_SAVE'})
    files: [bpy.types.OperatorFileListElement] | prop(options={'HIDDEN', 'SKIP_SAVE'})
    
    filter_folder: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    filter_image: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    filter_movie: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    filter_blender: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    
    go_to_directory: False | prop()
    
    push_undo: False | prop()
    
    def execute(self, context):
        if self.go_to_directory: settings.files.go(self.directory)
        
        # Allow navigating folders, but don't create decals if the scene is linked from a library
        if context.scene.library is not None: return {'CANCELLED'}
        
        paths = [os.path.join(self.directory, item.name) for item in self.files if item.name]
        if not paths: return {'CANCELLED'}
        
        view_layer = context.view_layer
        scene = context.scene
        
        active_obj = get_active_selected_object(context)
        
        kwargs = dict(paths=paths, context=context)
        objs = DecalUtils.load_libraries_as_decals(**kwargs)
        
        operator = (self if getattr(self, "is_invoked", None) else None)
        material = MaterialLibrary.try_load_material(operator=operator)
        
        if material:
            kwargs = settings.loading_options.kwargs(paths=paths, material=material, context=context)
            objs.extend(DecalUtils.load_images_as_decals(**kwargs))
        
        if not objs: return {'CANCELLED'}
        
        unselect_decals(context)
        
        if objs:
            DecalUtils.copy_properties(active_obj, objs)
            DecalUtils.arrange(objs, settings.get_default_matrix(context))
            BlUtil.Object.select_activate(objs, mode='SELECT', active='FIRST', view_layer=view_layer)
        
        DecalUtils.render_preview(Decal(obj).template for obj in objs)
        
        view_layer.update()
        
        if self.push_undo: bpy.ops.ed.undo_push(message="Load Decals")
        
        if settings.use_setup_snapping:
            DecalUtils.setup_snap_settings(scene.tool_settings)
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        directory = BpyPath.abspath(settings.files.directory)
        if not os.path.isdir(directory): directory = os.path.dirname(directory)
        if os.path.isdir(directory): self.directory = directory
        
        self.is_invoked = True
        self.go_to_directory = True
        self.push_undo = True
        
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
    def draw(self, context):
        settings.loading_options.draw(self.layout)

@addon.Operator(idname="xdecal.add_objects_to_palette", label="Add objects to palette")
class AddObjectsToPaletteOperator:
    "Add selected objects to decal palette (if already in palette, update the preview)"
    
    @classmethod
    def poll(cls, context):
        if context.space_data.type != 'VIEW_3D': return False
        
        if context.mode != 'OBJECT':
            cls.poll_message_set("Works only in Object mode")
            return False
        
        return True
    
    def execute(self, context):
        if not context.selected_objects: return {'CANCELLED'}
        
        preview_queue = []
        
        for obj in tuple(context.selected_objects):
            if obj.library: continue
            if Decal.exists(obj): continue
            CommonUtils.mark_as_template(obj)
            preview_queue.append(obj)
        
        DecalUtils.render_preview(preview_queue)
        
        BlUI.tag_redraw()
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        result = self.execute(context)
        if 'FINISHED' in result: bpy.ops.ed.undo_push(message="Add to Decal Palette")
        return result

@addon.Operator(idname="xdecal.add_materials_to_palette", label="Add materials to palette")
class AddMaterialsToPaletteOperator:
    "Add selected objects' materials to decal palette"
    
    def execute(self, context):
        if not context.selected_objects: return {'CANCELLED'}
        
        preview_queue = []
        
        for obj in tuple(context.selected_objects):
            if Decal.exists(obj): continue
            
            for ms in obj.material_slots:
                material = ms.material
                if not material: continue
                if material.library: continue
                CommonUtils.mark_as_template(material)
                preview_queue.append(material)
                material.preview_render_type = 'FLAT'
        
        DecalUtils.render_preview(preview_queue)
        
        BlUI.tag_redraw()
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        result = self.execute(context)
        if 'FINISHED' in result: bpy.ops.ed.undo_push(message="Add to Decal Palette")
        return result

class TemplateOperatorBase:
    type: "" | prop()
    name: "" | prop()
    palette_type: "ui" | prop()
    
    push_undo: False | prop()
    
    applicability = 'ALL'
    
    def setup(self, context, event):
        return True
    
    def preprocess(self, context):
        return True
    
    def process(self, context, template):
        return True
    
    def postprocess(self, context):
        return True
    
    def process_templates(self, context, templates):
        if not templates: return False
        
        if self.preprocess(context) is False: return False
        
        if hasattr(self, "process_many"):
            result = not (self.process_many(context, templates) is False)
        else:
            result = False
            for template in templates:
                result |= not (self.process(context, template) is False)
        
        if not result: return False
        
        if self.postprocess(context) is False: return False
        
        return True
    
    def process_single(self, context):
        if self.type == "Material":
            template = bpy.data.materials.get(self.name)
            if not template: return False
        elif self.type == "Object":
            template = bpy.data.objects.get(self.name)
            if not template: return False
        else:
            return False
        
        return self.process_templates(context, [template])
    
    def process_selection(self, context):
        selected_names = set(settings.decals.selected_names)
        if (not selected_names) or (self.name and (self.name not in selected_names)):
            selected_names = {self.name}
        
        metadata = BlUI.get_ui_metadata(context)
        # default to 3d, since we can't always set context there
        palette_type = metadata.get("palette_type", "3d")
        
        items = Palette.filter_items(f"decals-{palette_type}", context)
        templates = [item.template for item in items if item.name in selected_names]
        
        return self.process_templates(context, templates)
    
    def process_all(self, context):
        filter_implicit = settings.decals.filter_implicit_get(self.palette_type)
        filter_explicit = settings.decals.filter_explicit_get(self.palette_type)
        filter_scene = settings.decals.filter_scene_get(self.palette_type)
        
        templates = DecalUtils.find_all_templates(filter_implicit, filter_explicit)
        
        if filter_scene:
            scene_templates = DecalUtils.find_scene_templates(context.scene)
            templates = [template for template in templates if template in scene_templates]
        
        return self.process_templates(context, templates)
    
    def process_assets(self, context):
        templates = self.gather_asset_templates(context)
        return self.process_templates(context, templates)
    
    def gather_asset_templates(self, context):
        id_types = ('MATERIAL', 'OBJECT')
        id_type_infos = IDTypes.map("id_type")
        
        template_tag = CommonUtils.template_tag
        
        linked_idblocks = {}
        
        for id_type in id_types:
            data_name = id_type_infos[id_type].data_name
            
            direct, overrides = CommonUtils.gather_linked(data_name)
            
            for lib_path, idblocks in direct.items():
                for idblock in idblocks:
                    key = (lib_path, id_type, idblock.name)
                    linked_idblocks[key] = idblock
            
            for lib_path, idblocks in overrides.items():
                for idblock in idblocks:
                    ref_idblock = idblock.override_library.reference
                    key = (lib_path, id_type, ref_idblock.name)
                    linked_idblocks[key] = idblock
        
        templates = []
        
        external_assets = {}
        
        import_prefs_map = {}
        
        for asset_info in BlUtil.Assets.get_selection(context=context):
            file_path = asset_info["file_path"]
            metadata = asset_info["metadata"]
            local_id = asset_info["local_id"]
            id_type = asset_info["id_type"]
            name = asset_info["name"]
            
            if id_type not in id_types: continue
            
            # Perhaps requiring the tag here is not strictly necessary
            # (tag is mainly for filtering what to show in the palette)
            #if template_tag in metadata.tags: continue
            
            if local_id:
                templates.append(local_id)
                continue
            
            file_path = CommonUtils.abspath(file_path)
            key = (file_path, id_type, name)
            local_id = linked_idblocks.get(key)
            
            import_prefs = import_prefs_map.get(file_path)
            if not import_prefs:
                import_prefs = BlUtil.Assets.get_import_prefs(file_path, context=context)
                import_prefs_map[file_path] = import_prefs
            
            if local_id:
                import_method = import_prefs["method"]
                should_reuse = (import_method == 'APPEND_REUSE')
                
                if (not should_reuse) or local_id.override_library:
                    templates.append(local_id)
                    continue
            
            data_name = id_type_infos[id_type].data_name
            file_collections = external_assets.setdefault(file_path, {})
            data_collections = file_collections.setdefault(data_name, set())
            data_collections.add(name)
        
        for file_path, file_collections in external_assets.items():
            import_prefs = import_prefs_map[file_path]
            
            # It seems that Blender currently does not provide an API
            # to append-with-reuse, only to link with liboverrides
            
            import_method = import_prefs["method"]
            should_reuse = (import_method == 'APPEND_REUSE')
            load_kwargs = dict(
                relative = import_prefs["relative"],
                link = (import_method == 'LINK') or should_reuse,
                create_liboverrides = should_reuse,
                reuse_liboverrides = should_reuse,
                create_liboverrides_runtime = should_reuse,
            )
            
            with bpy.data.libraries.load(file_path, **load_kwargs) as (data_from, data_to):
                for data_name, data_collections in file_collections.items():
                    setattr(data_to, data_name, list(data_collections))
            
            for data_name, data_collections in file_collections.items():
                templates.extend(getattr(data_to, data_name))
        
        return templates
    
    def process_main(self, context):
        metadata = BlUI.get_ui_metadata(context)
        palette_type = metadata.get("palette_type", self.palette_type)
        
        if palette_type == "asset":
            return self.process_assets(context)
        elif self.type and self.name:
            if self.applicability == 'ACTIVE':
                return self.process_single(context)
            else:
                return self.process_selection(context)
        elif self.applicability == 'ALL':
            return self.process_all(context)
        else:
            return False
    
    def execute(self, context):
        result = self.process_main(context)
        if result is False: return {'CANCELLED'}
        
        if self.push_undo:
            undo_label = getattr(self, "undo_label", self.bl_label)
            if undo_label: bpy.ops.ed.undo_push(message=undo_label)
        
        context.view_layer.update()
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        if (self.applicability == 'ALL') and event.shift:
            self.type = ""
            self.name = ""
        
        if self.setup(context, event) is False: return {'CANCELLED'}
        
        self.push_undo = True
        result = self.execute(context)
        
        return result

@addon.Operator(idname="xdecal.instantiate", label="Instantiate", options={'INTERNAL'})
class InstantiateDecalOperator(TemplateOperatorBase):
    "Add decal(s) based on the selected template(s)"
    
    undo_label = "Add Decal"
    
    applicability = 'SELECTION'
    
    def preprocess(self, context):
        # Don't create decals if the scene is linked from a library
        if context.scene.library is not None: return False
        
        self.active_obj = get_active_selected_object(context)
        
        unselect_decals(context)
    
    def process(self, context, template):
        view_layer = context.view_layer
        
        obj = Decal.instantiate(template, context=context)
        obj.select_set(True, view_layer=context.view_layer)
    
    def postprocess(self, context):
        # We need to update the depsgraph in order to evaluate
        # the modifiers and get the correct object dimensions
        context.evaluated_depsgraph_get()
        
        scene = context.scene
        view_layer = context.view_layer
        
        objs = DecalUtils.filter_decals(context.selected_objects, result=[])
        
        DecalUtils.copy_properties(self.active_obj, objs)
        DecalUtils.arrange(objs, settings.get_default_matrix(context))
        BlUtil.Object.select_activate(objs, mode='SELECT', active='FIRST', view_layer=view_layer)

@addon.Operator(idname="xdecal.mat_props_copy_palette", label="Copy parameters", options={'INTERNAL'})
class PaletteMaterialPropertiesCopyOperator(TemplateOperatorBase):
    "Copy material control parameters"
    
    undo_label = "" # don't push undo
    
    applicability = 'ACTIVE'
    
    def process(self, context, template):
        if not isinstance(template, bpy.types.Material): return
        MaterialUtils.clipboard = NodeInfo.mat_props_copy(template)

@addon.Operator(idname="xdecal.mat_props_paste_palette", label="Paste parameters", options={'INTERNAL'})
class PaletteMaterialPropertiesPasteOperator(TemplateOperatorBase):
    "Paste material control parameters"
    
    undo_label = "Paste material parameters"
    
    applicability = 'SELECTION'
    
    def process(self, context, template):
        if not isinstance(template, bpy.types.Material): return
        NodeInfo.mat_props_paste(template, MaterialUtils.clipboard)

@addon.Operator(idname="xdecal.select_instances", label="Select instances", options={'INTERNAL'})
class SelectInstancesOperator(TemplateOperatorBase):
    "Select decals corresponding to the selected template(s) (Ctrl+Click to also zoom to the selection)"
    
    undo_label = "Select Decals"
    
    zoom: False | prop()
    
    def setup(self, context, event):
        self.zoom = event.ctrl
    
    def preprocess(self, context):
        BlUtil.Object.select_all('DESELECT', context=context)
    
    def process(self, context, template):
        scene = context.scene
        view_layer = context.view_layer
        objs = DecalUtils.filter_instances(scene.collection.all_objects, template)
        BlUtil.Object.select_activate(objs, 'SELECT', view_layer=view_layer)
    
    def postprocess(self, context):
        if self.zoom: bpy.ops.view3d.view_selected('INVOKE_DEFAULT')

@addon.Operator(idname="xdecal.select_template", label="Select template", options={'INTERNAL'})
class SelectTemplateOperator(TemplateOperatorBase):
    "Select the template object(s)"
    
    undo_label = "Select Decal Template"
    
    def preprocess(self, context):
        BlUtil.Object.select_all('DESELECT', context=context)
    
    def process(self, context, template):
        if not isinstance(template, bpy.types.Object): return False
        view_layer = context.view_layer
        BlUtil.Object.select_activate(template, 'SELECT', view_layer=view_layer)

@addon.Operator(idname="xdecal.update_icon", label="Update icon", options={'INTERNAL'})
class UpdateTemplateIconOperator(TemplateOperatorBase):
    "Update the selected templates' icons\nWARNING: works only in Object mode"
    
    undo_label = "Update Decal Icon"
    
    @classmethod
    def poll(cls, context):
        if context.space_data.type != 'VIEW_3D': return False
        
        if context.mode != 'OBJECT':
            cls.poll_message_set("Works only in Object mode")
            return False
        
        return True
    
    def process_many(self, context, templates):
        DecalUtils.render_preview(templates)
        BlUI.tag_redraw()

@addon.Operator(idname="xdecal.update_instances", label="Update instances", options={'INTERNAL'})
class UpdateTemplateInstancesOperator(TemplateOperatorBase):
    "Update instances of the selected template(s)"
    
    undo_label = "Update Decals"
    
    def process(self, context, template):
        collection = context.scene.collection
        objs = DecalUtils.filter_instances(collection.all_objects, template)
        return DecalUtils.update(objs)

@addon.Operator(idname="xdecal.replace_decals", label="Replace decals", options={'INTERNAL'})
class ReplaceDecalsOperator(TemplateOperatorBase):
    "Assign this template to the selected decals"
    
    undo_label = "Replace Decals"
    
    applicability = 'ACTIVE'
    
    def process(self, context, template):
        decals = []
        for obj in context.selected_objects:
            if obj.library: continue
            decal = Decal(obj)
            if decal: decals.append(decal)
        
        if not decals: return False
        
        for decal in decals:
            decal.template = template
            decal.update()
        
        return True

class ReplaceMaterialsBase:
    undo_label = "Replace Materials"
    
    prototype_name: "" | prop()
    
    prototype = None
    
    def preprocess(self, context):
        self.prototype = bpy.data.materials.get(self.prototype_name)
        if self.prototype: return True
        
        operator = (self if getattr(self, "is_invoked", None) else None)
        self.prototype = MaterialLibrary.try_load_material(operator=operator)
        return bool(self.prototype)
    
    def invoke(self, context, event):
        self.is_invoked = True
        self.push_undo = True
        
        wm = context.window_manager
        width = min(settings.popup_dialog_width, context.window.width)
        return wm.invoke_props_dialog(self, width=int(width))
    
    def draw(self, context):
        layout = self.layout
        
        prototype = bpy.data.materials.get(self.prototype_name)
        is_generic = bool(prototype)
        
        row = layout.row()
        row.label(icon=('RADIOBUT_ON' if is_generic else 'RADIOBUT_OFF'))
        row.prop_search(self, "prototype_name", bpy.data, "materials", text="")
        
        row = layout.row()
        row.label(icon=('RADIOBUT_ON' if not is_generic else 'RADIOBUT_OFF'))
        row.enabled = not is_generic
        TemplateMaterialMenu.draw_main(row)
        
        if hasattr(self, "replace_everywhere"):
            layout.prop(self, "replace_everywhere")

@addon.Operator(idname="xdecal.replace_materials", label="Replace materials", options={'INTERNAL'})
class ReplaceMaterialsOperator(ReplaceMaterialsBase, TemplateOperatorBase):
    "Create template(s) based on a different material + copy over images and main properties"
    
    applicability = 'SELECTION'
    
    def process(self, context, template):
        objs = DecalUtils.filter_instances(bpy.data.objects, template)
        DecalUtils.replace_material(objs, self.prototype)
        return True

@addon.Operator(idname="xdecal.replace_materials_in_objects", label="Replace materials", options={'INTERNAL'})
class ReplaceMaterialsInObjectsOperator(ReplaceMaterialsBase):
    "Replace selected decals' materials + copy over images and main properties"
    
    replace_everywhere: False | prop("Replace everywhere", "Affect all decals which use the same materials, instead of just the selected ones")
    
    push_undo: False | prop()
    
    def execute(self, context):
        if not self.preprocess(context): return {'CANCELLED'}
        
        objs = context.selected_objects
        
        if self.replace_everywhere:
            templates_sel = DecalUtils.filter_templates(objs, types='MATERIAL', editable='OBJECT')
            templates_all = DecalUtils.filter_templates(bpy.data.objects, types='MATERIAL', editable='OBJECT')
            objs = [obj for template in templates_sel.keys() for obj in templates_all[template]]
        
        DecalUtils.replace_material(objs, self.prototype)
        
        if self.push_undo:
            undo_label = getattr(self, "undo_label", self.bl_label)
            bpy.ops.ed.undo_push(message=undo_label)
        
        context.view_layer.update()
        
        return {'FINISHED'}

class EnableDisableBase(TemplateOperatorBase):
    mode: 'VIEW_LAYER' | prop(items=[
        ('SELECT', "Select", "Select"),
        ('VIEW_LAYER', "View Layer", "View Layer"),
        ('VIEWPORT', "Viewport", "Viewport"),
        ('RENDER', "Render", "Render"),
    ])
    
    param_names = {
        'SELECT': "selection",
        'VIEW_LAYER': "scene layer visibility",
        'VIEWPORT': "viewport visibility",
        'RENDER': "render visibility",
    }
    
    @classmethod
    def description(cls, context, properties):
        prefix = "Enable" if cls.value else "Disable"
        return f"{prefix} decals' {cls.param_names[properties.mode]}"
    
    def process(self, context, template):
        collection = context.scene.collection
        # Changing hide_viewport / hide_select / hide_render will trigger a rebuild
        # of Collection caches, breaking any current iteration over all_objects
        objs = list(DecalUtils.filter_instances(collection.all_objects, template))
        
        if not objs: return False
        
        hide = not self.value
        
        if self.mode == 'SELECT':
            for obj in objs: obj.hide_select = hide
        elif self.mode == 'VIEWPORT':
            for obj in objs: obj.hide_viewport = hide
        elif self.mode == 'RENDER':
            for obj in objs: obj.hide_render = hide
        elif self.mode == 'VIEW_LAYER':
            view_layer = context.view_layer
            for obj in objs:
                try:
                    obj.hide_set(hide, view_layer=view_layer)
                except RuntimeError:
                    pass
        
        return True

@addon.Operator(idname="xdecal.enable_decals", label="Enable decals", options={'INTERNAL'})
class EnableDecalsOperator(EnableDisableBase):
    "Enable decals (in the current scene)"
    undo_label: "Enable Decals" | prop(options={'HIDDEN', 'SKIP_SAVE'})
    value = True

@addon.Operator(idname="xdecal.disable_decals", label="Disable decals", options={'INTERNAL'})
class DisableDecalsOperator(EnableDisableBase):
    "Disable decals (in the current scene)"
    undo_label: "Disable Decals" | prop(options={'HIDDEN', 'SKIP_SAVE'})
    value = False

@addon.Operator(idname="xdecal.remove_from_palette", label="Remove from palette", options={'INTERNAL'})
class RemoveDecalsOperator(TemplateOperatorBase):
    "Remove the selected template(s) from the palette (Click: remove the decal tag, Ctrl+Click: clear the asset mark)"
    
    undo_label = "Remove Decal Template from Palette"
    
    keep_as_asset: False | prop()
    
    def setup(self, context, event):
        self.keep_as_asset = event.ctrl
    
    def process(self, context, template):
        if (not template) or template.library: return False
        CommonUtils.unmark_as_template(template, self.keep_as_asset)
        Palette.invalidate()
        BlUI.tag_redraw()
        return True

@addon.Operator(idname="xdecal.restore_to_palette", label="Restore to palette", options={'INTERNAL'})
class RestoreDecalsOperator(TemplateOperatorBase):
    "Add the selected implicit templates back to the palette (mark as explicit templates)"
    
    undo_label = "Restore Decal Template to Palette"
    
    def process(self, context, template):
        if (not template) or template.library: return False
        CommonUtils.mark_as_template(template)
        Palette.invalidate()
        BlUI.tag_redraw()
        return True

@addon.Operator(idname="xdecal.delete_decals", label="Delete instances", options={'INTERNAL'})
class DeleteDecalsOperator(TemplateOperatorBase):
    "Delete instances of the selected template(s)"
    
    undo_label = "Delete Decals"
    
    def process(self, context, template):
        objs = DecalUtils.filter_instances(bpy.data.objects, template)
        if not objs: return False
        
        with BlUtil.Data.OrphanCleanup(exclude=["objects"]):
            for obj in objs:
                bpy.data.objects.remove(obj)
        
        return True

@addon.Operator(idname="xdecal.delete_template", label="Delete completely", options={'INTERNAL'})
class DeleteTemplateOperator(TemplateOperatorBase):
    "Delete the selected templates and all of their instances"
    
    undo_label = "Delete Decal Template"
    
    def process(self, context, template):
        objs = DecalUtils.filter_instances(bpy.data.objects, template)
        
        with BlUtil.Data.OrphanCleanup(exclude=["objects"]):
            for obj in objs:
                bpy.data.objects.remove(obj)
            
            if isinstance(template, bpy.types.Object):
                bpy.data.objects.remove(template)
            elif isinstance(template, bpy.types.Material):
                bpy.data.materials.remove(template)
        
        return True

@addon.Operator(idname="xdecal.update_decals", label="Update decal(s)", options={'INTERNAL'})
class UpdateDecalOperator:
    "Update decals in selected objects (or in scene, if no objects are selected)"
    
    def execute(self, context):
        result = DecalUtils.update(get_selected_or_all_decals(context))
        context.view_layer.update()
        return ({'FINISHED'} if result else {'CANCELLED'})
    
    def invoke(self, context, event):
        result = self.execute(context)
        if 'FINISHED' in result: bpy.ops.ed.undo_push(message="Update Decal(s)")
        return result

@addon.Operator(idname="xdecal.sync_collections", label="Sync collections", options={'INTERNAL'})
class SyncCollectionsOperator:
    "Synchronize decal collections in selected objects (or in scene, if no objects are selected)"
    
    def execute(self, context):
        result = DecalUtils.update(get_selected_or_all_decals(context), geometry=False, collections=True)
        context.view_layer.update()
        return ({'FINISHED'} if result else {'CANCELLED'})
    
    def invoke(self, context, event):
        result = self.execute(context)
        if 'FINISHED' in result: bpy.ops.ed.undo_push(message="Sync Decal Collections")
        return result

@addon.Operator(idname="xdecal.select_targets", label="Select target(s)", options={'INTERNAL'})
class SelectTargetsOperator:
    "Select projection target(s)\nClick: set, Shift+Click: add, Ctrl+Click: remove"
    
    mode: 'SET' | prop(items=['SET', 'ADD', 'REMOVE'])
    
    def execute(self, context):
        targets = set()
        decals = set()
        non_decals = set()
        for obj in context.selected_objects:
            decal = Decal(obj)
            if decal:
                decals.add(obj)
                targets.update(decal.targets)
            else:
                non_decals.add(obj)
        
        view_layer = context.view_layer
        select_get = BlUtil.Object.select_get
        select_set = BlUtil.Object.select_set
        
        if self.mode == 'ADD':
            objs_diff = targets - non_decals
            for obj in objs_diff:
                select_set(obj, True, view_layer=view_layer)
        elif self.mode == 'REMOVE':
            objs_diff = non_decals & targets
            for obj in objs_diff:
                select_set(obj, False, view_layer=view_layer)
        else:
            objs_diff = targets.symmetric_difference(non_decals)
            for obj in objs_diff:
                is_selected = select_get(obj, view_layer=view_layer)
                select_set(obj, not is_selected, view_layer=view_layer)
        
        context.view_layer.update()
        BlUI.tag_redraw()
        
        return ({'FINISHED'} if objs_diff else {'CANCELLED'})
    
    def invoke(self, context, event):
        self.mode = ('ADD' if event.shift else ('REMOVE' if event.ctrl else 'SET'))
        result = self.execute(context)
        if 'FINISHED' in result: bpy.ops.ed.undo_push(message="Select Target(s)")
        return result

@addon.Operator(idname="xdecal.assign_targets", label="Assign target(s)", options={'INTERNAL'})
class SelectTargetsOperator:
    "Update projection targets with selected object(s)\nClick: set, Shift+Click: add, Ctrl+Click: remove"
    
    mode: 'SET' | prop(items=['SET', 'ADD', 'REMOVE'])
    
    def execute(self, context):
        decals = []
        candidates = []
        for obj in context.selected_objects:
            decal = Decal(obj)
            if decal:
                decals.append(decal)
            else:
                candidates.append(obj)
        
        if not decals: return {'CANCELLED'}
        
        for decal in decals:
            if self.mode == 'ADD':
                targets = decal.targets
                for obj in candidates:
                    if obj not in targets: targets.append(obj)
            elif self.mode == 'REMOVE':
                targets = decal.targets
                for obj in candidates:
                    if obj in targets: targets.remove(obj)
            else:
                targets = candidates
            
            decal.targets = targets
            decal.update()
        
        context.view_layer.update()
        BlUI.tag_redraw()
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        self.mode = ('ADD' if event.shift else ('REMOVE' if event.ctrl else 'SET'))
        result = self.execute(context)
        if 'FINISHED' in result: bpy.ops.ed.undo_push(message="Assign Target(s)")
        return result

@addon.ui_draw("ASSETBROWSER_MT_asset", 'APPEND')
def asset_browser_draw_in_asset_menu(self, context):
    layout = self.layout
    layout.separator()
    
    col = layout.column()
    BlUI.set_ui_metadata(col, {"palette_type": "asset"})
    
    col.operator("xdecal.instantiate", text="Add as decal(s)", icon='IMAGE_PLANE')

@addon.Operator(idname="xdecal.node_settings", label="Node Settings", options={'INTERNAL'})
class NodeSettingsOperator:
    "Configure the settings and UI controls for the active node"
    
    group: "" | prop("Group", "The group of controls to which this node belongs")
    id: "" | prop("ID", "Unique identifier within the group (used to disambiguate nodes when copy-pasting)")
    mix_alpha: True | prop("Mix by Alpha", "Turn this off if the node's effect is already dependent on Alpha")
    mix_socket: "" | prop("Mix-socket", "The input socket to which the base material's value will be provided upon embedding")
    toggles: set() | prop("Toggles", "Which kinds of controls to show", items=[
        ('B', "Buttons", "Show node buttons"), # OPTIONS
        ('I', "Inputs", "Show node inputs"), # ANCHOR_LEFT
        ('O', "Outputs", "Show node outputs"), # ANCHOR_RIGHT
    ])
    
    socket_infos: [{
        "toggle": False | prop("Socket", "Toggle whether this socket will show in the controls UI"),
        "label": "" | prop("Label", "Custom name for this socket in the controls UI\n"+
            "(If empty, socket's original name will be used)"),
    }] | prop()
    
    label: "" | prop("Label", "Node label (for image nodes, used as the name of a PBR channel)")
    
    mix_socket_candidates: [""] | prop()
    
    def get_is_decal(self):
        return MaterialUtils.node_ignore_mark not in self.label
    def set_is_decal(self, value):
        if value == (MaterialUtils.node_ignore_mark not in self.label): return
        if value:
            self.label = self.label.replace(MaterialUtils.node_ignore_mark, "")
        else:
            self.label = MaterialUtils.node_ignore_mark + self.label
    is_decal: True | prop("Is decal", "Does this image node contain a decal image / PBR channel",
        get="get_is_decal", set="set_is_decal")
    
    separators = {
        "input": "<",
        "output": ">",
    }
    
    @classmethod
    def poll(cls, context):
        snode = context.space_data
        if snode is None: return False
        
        # Node controls make sense only for material-level nodes, not node groups
        tree = snode.edit_tree
        if tree is None: return False
        
        if not tree.is_embedded_data:
            cls.poll_message_set("Node controls inside of node groups are not supported")
            return False
        
        if context.active_node not in context.selected_nodes:
            cls.poll_message_set("An active node needs to be selected")
            return False
        
        return True
    
    def execute(self, context):
        node = context.active_node
        if not node: return {'CANCELLED'}
        
        sep_output = self.separators["output"]
        
        inputs = []
        outputs = []
        for socket_info in self.socket_infos:
            is_output = socket_info.name.endswith(sep_output)
            sockets = (outputs if is_output else inputs)
            socket_name = socket_info.name[:-1]
            socket_label = socket_info.label
            sockets.append(dict(name=socket_name, label=socket_label, toggle=socket_info.toggle))
        
        info = dict(
            group = self.group,
            id = self.id,
            alpha = not self.mix_alpha,
            mix_socket = self.mix_socket,
            toggles = "".join(sorted(self.toggles)).upper(),
            inputs = inputs,
            outputs = outputs,
        )
        
        NodeInfo.set(node, info)
        node.label = self.label
        
        bpy.ops.ed.undo_push(message="Node Settings")
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        node = context.active_node
        if not node: return {'CANCELLED'}
        
        self.is_image = (node.bl_idname == "ShaderNodeTexImage")
        
        self.label = node.label
        
        self.mix_candidate_counts = {}
        
        self.socket_infos.clear()
        
        self.mix_socket_candidates.clear()
        self.mix_socket_candidates.add() # empty item
        
        for socket in node.inputs:
            self.add_socket(socket)
        
        for socket in node.outputs:
            self.add_socket(socket)
        
        mix_candidate_indices = {}
        for item in self.mix_socket_candidates:
            if self.mix_candidate_counts.get(item.name, 0) <= 1: continue
            count = mix_candidate_indices.get(item.name, 0) + 1
            mix_candidate_indices[item.name] = count
            item.name += f"{NodeInfo.mix_info_separator}{count}"
        
        info = NodeInfo.get(node)
        if info:
            self.group = info["group"]
            self.id = info["id"]
            self.mix_alpha = not info["alpha"]
            self.mix_socket = info["mix_socket"]
            self.toggles = set(info["toggles"])
            self.map_sockets(info["inputs"], self.separators["input"])
            self.map_sockets(info["outputs"], self.separators["output"])
        else:
            self.group = node.name
            self.id = ""
            self.mix_alpha = True
            self.mix_socket = ""
            self.toggles = set()
        
        self.group_ids = self.gather_group_ids(context.space_data.edit_tree, node)
        
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=260)
    
    def gather_group_ids(self, tree, exclude_node):
        group_ids = {}
        for node in tree.nodes:
            if node == exclude_node: continue
            info = NodeInfo.get(node)
            if not info: continue
            group_ids.setdefault(info["group"], set()).add(info["id"])
        return group_ids
    
    def add_socket(self, socket):
        if (not socket.enabled) or socket.is_unavailable: return
        if socket.is_linked != socket.is_output: return
        socket_info = self.socket_infos.add()
        socket_info.name = socket.name + self.separators["output" if socket.is_output else "input"]
        if not socket.is_output:
            mix_candidate = self.mix_socket_candidates.add()
            mix_candidate.name = socket.name
            self.mix_candidate_counts[socket.name] = self.mix_candidate_counts.get(socket.name, 0) + 1
    
    def find_socket(self, name, count=1):
        for socket_info in self.socket_infos:
            if socket_info.name != name: continue
            count -= 1
            if count == 0: return socket_info
        return None
    
    def map_sockets(self, sockets, suffix):
        name_counts = {}
        for socket_info in sockets:
            if not hasattr(socket_info, "get"): continue
            name = socket_info.get("name", "")
            label = socket_info.get("label", "")
            toggle = socket_info.get("toggle", False)
            count = name_counts.get(name, 0) + 1
            name_counts[name] = count
            socket_info = self.find_socket(name+suffix, count)
            if not socket_info: continue
            socket_info.label = label
            socket_info.toggle = toggle
    
    def draw(self, context):
        layout = NestedLayout(self.layout)
        
        ids = self.group_ids.get(self.group)
        is_conflicting = (ids is not None) and (self.id in ids)
        
        with layout.row(align=True):
            with layout.column(align=True)(alignment='LEFT'):
                layout.label(text="Mix:")
                layout.label(text="Label:")
                layout.label(text="Group:")
                layout.label(text="ID:")
            with layout.column(align=True):
                with layout.row():
                    layout.prop_search(self, "mix_socket", self,"mix_socket_candidates", text="")
                    with layout.row()(alignment='RIGHT'):
                        layout.prop(self, "mix_alpha", text="By Alpha")
                with layout.row(align=True):
                    layout.prop(self, "label", text="")
                    if self.is_image:
                        layout.prop(self, "is_decal", text="", icon='IMAGE_PLANE')
                layout.prop(self, "group", text="")
                with layout.row(align=True)(alert=is_conflicting):
                    layout.prop(self, "id", text="")
        
        with layout.row(align=True):
            layout.prop_enum_filtered(self, "toggles")
        
        sep_output = self.separators["output"]
        invert_inputs = 'I' in self.toggles
        invert_outputs = 'O' in self.toggles
        
        was_output = None
        
        with layout.column(align=True):
            for socket_info in self.socket_infos:
                is_output = socket_info.name.endswith(sep_output)
                if (was_output is False) and is_output: layout.separator()
                was_output = is_output
                
                invert = (invert_outputs if is_output else invert_inputs)
                text = socket_info.name[:-1]
                icon = ('ANCHOR_RIGHT' if is_output else 'ANCHOR_LEFT')
                with layout.row(align=True):
                    layout.prop(socket_info, "toggle", text=text, icon=icon, toggle=True, invert_checkbox=invert)
                    with layout.row(align=True)(enabled=(socket_info.toggle != invert)):
                        layout.prop(socket_info, "label", text="")

class AttributeNodeOperatorBase:
    priorities = ["bool", "int", "int[]", "float", "rgb", "rgba", "float[]"]
    type_icons = {
        "bool": 'CHECKBOX_HLT',
        "int": 'SNAP_INCREMENT',
        "int[]": 'SNAP_VERTEX',
        "float": 'SNAP_MIDPOINT',
        "float[]": 'OPTIONS',
        "rgb": 'IMAGE_RGB',
        "rgba": 'IMAGE_RGB_ALPHA',
    }
    
    @classmethod
    def filter_nodes(cls, nodes, attribute_types=('OBJECT', 'INSTANCER')):
        if not nodes: return
        for node in nodes:
            if not node: continue
            if node.bl_idname != "ShaderNodeAttribute": continue
            if attribute_types and (node.attribute_type not in attribute_types): continue
            yield node
    
    @classmethod
    def filter_node_tree_users(cls, objs, node_tree):
        users = []
        for obj in objs:
            if not obj: continue
            for ms in obj.material_slots:
                mat = ms.material
                if not (mat and mat.node_tree): continue
                # some_tree.contains_tree(some_tree) also returns True
                if not mat.node_tree.contains_tree(node_tree): continue
                users.append(obj)
        return users
    
    @classmethod
    def get_default(cls, node):
        outputs = node.outputs
        # Color's default is 4D, even if shader nodes ignore the 4th channel
        return outputs["Color"].default_value
    
    @classmethod
    def set_default(cls, node, value):
        if hasattr(value, "__len__"):
            size = len(value)
            x = (value[0] if size > 0 else 0.0)
            y = (value[1] if size > 1 else 0.0)
            z = (value[2] if size > 2 else 0.0)
            a = (value[3] if size > 3 else 1.0)
        else:
            value = float(value)
            x, y, z, a = value, value, value, 1.0
        
        outputs = node.outputs
        outputs["Color"].default_value = (x, y, z, a)
        outputs["Vector"].default_value = (x, y, z)
        outputs["Fac"].default_value = (x + y + z) / 3.0
        outputs["Alpha"].default_value = a

@addon.Operator(idname="xdecal.mat_obj_attr_select", label="Attribute Select", options={'INTERNAL'})
class MaterialObjectAttributeSelectOperator(AttributeNodeOperatorBase):
    "Assign a name of an existing custom property for the selected Attribute node(s)"
    
    name: "" | prop()
    
    @classmethod
    def poll(cls, context):
        if not any(cls.filter_nodes(context.selected_nodes)):
            cls.poll_message_set("Attribute node(s) of Object or Instancer type need to be selected")
            return False
        
        return True
    
    def execute(self, context):
        for node in self.filter_nodes(context.selected_nodes):
            node.attribute_name = self.name
        
        bpy.ops.ed.undo_push(message="Select Attribute")
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        selected_nodes = context.selected_nodes
        if not selected_nodes: return {'CANCELLED'}
        
        priorities = self.priorities
        type_icons = self.type_icons
        
        node_tree = selected_nodes[0].id_data
        objs = self.filter_node_tree_users(bpy.data.objects, node_tree)
        
        entries = {}
        
        for obj in objs:
            for key, value in obj.items():
                type_name, size = IDProp.get_type_info(value, True)
                type_limit = MaterialMixer.shader_attr_size_limits.get(type_name)
                if (type_limit is None) or (size > type_limit): continue
                
                metadata = IDProp.get_metadata(obj, key)
                subtype = metadata.get("subtype", "")
                
                if 'COLOR' in subtype:
                    type_name = ("rgb" if size < 4 else "rgba")
                elif type_name == "float":
                    type_name = ("float" if size == 0 else "float[]")
                elif type_name == "int":
                    type_name = ("int" if size == 0 else "int[]")
                elif type_name == "bool":
                    type_name = "bool"
                
                priority = priorities.index(type_name)
                entries[key] = max(entries.get(key, 0), priority)
        
        entries = sorted(entries.items())
        
        op_idname = self.bl_idname
        
        def draw(self, context):
            layout = self.layout
            
            for name, priority in entries:
                icon = type_icons[priorities[priority]]
                op = layout.operator(op_idname, text=name, icon=icon)
                op.name = name
        
        context.window_manager.popup_menu(draw, title="Choose attribute:", icon='OBJECT_DATA')
        
        return {'CANCELLED'}

@addon.Operator(idname="xdecal.mat_obj_attr_set_default", label="Attribute Defaults", options={'INTERNAL'})
class MaterialObjectAttributeSetDefaultOperator(AttributeNodeOperatorBase):
    "Configure the default value for the selected Attribute node(s)"
    
    value_x: 0.0 | prop("X", "Default value")
    value_y: 0.0 | prop("Y", "Default value")
    value_z: 0.0 | prop("Z", "Default value")
    value_a: 1.0 | prop("α", "Default value")
    
    use_linear: True | prop("Linear", "Edit as linear color")
    
    def color_get(self):
        return (self.value_x, self.value_y, self.value_z, self.value_a)
    def color_set(self, value):
        self.value_x, self.value_y, self.value_z, self.value_a = value
    color_linear: (0.0, 0.0, 0.0, 1.0) | prop("Color", "", subtype='COLOR', min=0, max=1, get=color_get, set=color_set)
    color_gamma: (0.0, 0.0, 0.0, 1.0) | prop("Color", "", subtype='COLOR_GAMMA', min=0, max=1, get=color_get, set=color_set)
    
    @classmethod
    def poll(cls, context):
        if not any(cls.filter_nodes(context.selected_nodes)):
            cls.poll_message_set("Attribute node(s) of Object or Instancer type need to be selected")
            return False
        
        return True
    
    def execute(self, context):
        value = tuple(self.color_linear)
        
        # The default values are ignored by Blender, but at least
        # we can use them to add custom properties on decal creation
        for node in self.filter_nodes(context.selected_nodes):
            self.set_default(node, value)
        
        bpy.ops.ed.undo_push(message="Set Attribute Defaults")
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        selected_nodes = context.selected_nodes
        if not selected_nodes: return {'CANCELLED'}
        
        value = self.get_default(selected_nodes[0])
        self.value_x = value[0]
        self.value_y = value[1]
        self.value_z = value[2]
        self.value_a = value[3]
        
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=160)
    
    def draw(self, context):
        layout = NestedLayout(self.layout)
        
        with layout.row():
            if self.use_linear:
                layout.prop(self, "color_linear", text="")
            else:
                layout.prop(self, "color_gamma", text="")
            layout.prop(self, "use_linear", text="Linear")
        
        with layout.column(align=True):
            layout.prop(self, "value_x", text="X")
            layout.prop(self, "value_y", text="Y")
            layout.prop(self, "value_z", text="Z")
            layout.prop(self, "value_a", text="α")

@addon.Operator(idname="xdecal.mat_obj_attr_setup", label="Attribute Setup", options={'INTERNAL'})
class MaterialObjectAttributeSetupOperator(AttributeNodeOperatorBase):
    "Configure the objects' custom property value(s) for the active Attribute node"
    
    initialize_default: False | prop()
    
    name: "" | prop("Name", "The name of the attribute / custom property")
    
    scope: 'SELECTED' | prop("Scope", "", items=[
        ('ACTIVE', "Active object", "Apply to the active object"),
        ('SELECTED', "Selected objects", "Apply to the selected objects"),
        ('SCENE', "Scene objects", "Apply to objects in the current scene"),
        ('FILE', "All objects", "Apply to all objects in the current file"),
    ])
    
    mode: 'INITIALIZE' | prop("Mode", "", items=[
        ('UPDATE', "Update existing", "Update values / settings, for objects where property exists"),
        ('INITIALIZE', "Initialize new", "Add a new property, only if it did not exist"),
        ('OVERRIDE', "Override", "Add a property if necessary, and override any previous values / settings"),
    ])
    
    value_type: 'FLOAT' | prop("Data type", "", items=[
        ('FLOAT', "Float", ""),
        ('INT', "Integer", ""),
        ('BOOL', "Boolean", ""),
    ])
    
    value_size: 4 | prop("Dimension", "", min=0, max=4)
    
    dummy_subtype: 'NONE' | prop("Subtype", "", items=[('NONE', "Plain Data", "")])
    float_scalar_subtype: 'NONE' | prop("Subtype", "", items=IDProp.float_subtypes)
    float_vector_subtype: 'NONE' | prop("Subtype", "", items=IDProp.float_array_subtypes)
    
    value_min: -math.inf | prop("Min", "")
    value_max: math.inf | prop("Max", "")
    value_step: 0.1 | prop("Step", "", min=0.0)
    value_precision: 3 | prop("Precision", "", min=0, max=8)
    
    bool0: True | prop("Value", "")
    int0: 1 | prop("X", "")
    int1: 1 | prop("Y", "")
    int2: 1 | prop("Z", "")
    int3: 1 | prop("α", "")
    float0: 1.0 | prop("X", "", precision=3)
    float1: 1.0 | prop("Y", "", precision=3)
    float2: 1.0 | prop("Z", "", precision=3)
    float3: 1.0 | prop("α", "", precision=3)
    
    def color_get(self):
        return (self.float0, self.float1, self.float2, self.float3)
    def color_set(self, value):
        self.float0, self.float1, self.float2, self.float3 = value
    color_linear: (0.0, 0.0, 0.0, 1.0) | prop("Color", "", subtype='COLOR', min=0, max=1, get=color_get, set=color_set)
    color_gamma: (0.0, 0.0, 0.0, 1.0) | prop("Color", "", subtype='COLOR_GAMMA', min=0, max=1, get=color_get, set=color_set)
    
    @classmethod
    def poll(cls, context):
        if context.active_node not in context.selected_nodes:
            cls.poll_message_set("An Attribute node needs to be active & selected")
            return False
        
        if not any(cls.filter_nodes([context.active_node], None)):
            cls.poll_message_set("An Attribute node needs to be active & selected")
            return False
        
        return True
    
    def execute(self, context):
        node = context.active_node
        if not node: return {'CANCELLED'}
        
        if not self.name: return {'CANCELLED'}
        
        if node.attribute_name != self.name:
            node.attribute_name = self.name
        
        if node.attribute_type not in ('OBJECT', 'INSTANCER'):
            node.attribute_type = 'INSTANCER'
        
        if self.value_type == 'FLOAT':
            if self.value_size == 0:
                value = self.float0
            else:
                value = [getattr(self, f"float{i}") for i in range(self.value_size)]
        elif self.value_type == 'INT':
            if self.value_size == 0:
                value = self.int0
            else:
                value = [getattr(self, f"int{i}") for i in range(self.value_size)]
        elif self.value_type == 'BOOL':
            value = self.bool0
        
        if self.initialize_default:
            self.set_default(node, value)
        
        metadata = {
            "min": self.value_min,
            "max": self.value_max,
            "step": self.value_step,
            "precision": self.value_precision,
        }
        
        if self.value_type == 'FLOAT':
            if self.value_size == 0:
                metadata["subtype"] = self.float_scalar_subtype
            else:
                metadata["subtype"] = self.float_vector_subtype
        
        if self.scope == 'ACTIVE':
            objs = [context.active_object]
        elif self.scope == 'SELECTED':
            objs = context.selected_objects
        elif self.scope == 'SCENE':
            objs = context.scene.objects
        elif self.scope == 'FILE':
            objs = bpy.data.objects
        
        node_tree = node.id_data
        objs = self.filter_node_tree_users(objs, node_tree)
        
        can_create = (self.mode != 'UPDATE')
        can_modify = (self.mode != 'INITIALIZE')
        
        key = self.name
        data = {"value": value, "metadata": metadata}
        
        for obj in objs:
            if key in obj:
                if not can_modify: continue
            else:
                if not can_create: continue
            
            IDProp.set_data(obj, key, data)
            obj.update_tag()
        
        context.view_layer.update()
        BlUI.tag_redraw()
        
        bpy.ops.ed.undo_push(message="Setup Attribute")
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        node = context.active_node
        if not node: return {'CANCELLED'}
        
        self.name = node.attribute_name
        self.initialize_default = not self.name
        
        self.value_size = 4
        self.float_vector_subtype = 'NONE'
        
        attr_type = MaterialMixer.get_attribute_node_type(node)
        if attr_type == 'FLOAT':
            self.value_size = 0
        elif attr_type != 'NONE':
            self.value_size = len(attr_type)
            if attr_type.startswith('RGB'):
                self.float_vector_subtype = 'COLOR'
        
        self.color_linear = node.outputs["Color"].default_value
        
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=300)
    
    def draw(self, context):
        layout = NestedLayout(self.layout)
        
        with layout.row()(alert=(not self.name)):
            layout.prop(self, "name")
        
        layout.prop(self, "scope")
        
        layout.prop(self, "mode")
        
        layout.prop(self, "value_type")
        
        if self.value_type == 'FLOAT':
            if self.value_size == 0:
                layout.prop(self, "float_scalar_subtype")
            else:
                layout.prop(self, "float_vector_subtype")
        else:
            with layout.row()(enabled=False):
                layout.prop(self, "dummy_subtype")
        
        with layout.column()(active=(self.value_type != 'BOOL')):
            with layout.row(align=True):
                with layout.column(align=True):
                    layout.prop(self, "value_min")
                    layout.prop(self, "value_max")
                with layout.column(align=True):
                    layout.prop(self, "value_step")
                    layout.prop(self, "value_precision")
            
            layout.prop(self, "value_size")
        
        with layout.column(align=True):
            if self.value_type == 'FLOAT':
                if self.value_size == 0:
                    layout.prop(self, "float0", text="Value")
                elif 'COLOR' in self.float_vector_subtype:
                    if self.float_vector_subtype == 'COLOR':
                        layout.prop(self, "color_linear")
                    else:
                        layout.prop(self, "color_gamma")
                else:
                    for i in range(max(self.value_size, 1)):
                        layout.prop(self, f"float{i}")
            elif self.value_type == 'INT':
                if self.value_size == 0:
                    layout.prop(self, "int0", text="Value")
                else:
                    for i in range(max(self.value_size, 1)):
                        layout.prop(self, f"int{i}")
            elif self.value_type == 'BOOL':
                layout.prop(self, "bool0", text="Value")

@addon.Menu(idname="VIEW3D_MT_xdecal_node_context_menu", label=addon.name)
class DecalNodeContextMenu:
    def draw(self, context):
        layout = self.layout
        layout.operator("xdecal.node_settings", text="Node Settings")
        layout.operator("xdecal.mat_obj_attr_select", text="Attribute: Select")
        layout.operator("xdecal.mat_obj_attr_set_default", text="Attribute: Defaults")
        layout.operator("xdecal.mat_obj_attr_setup", text="Attribute: Value")

@addon.ui_draw("NODE_MT_context_menu", 'APPEND')
def node_context_menu_draw(self, context):
    snode = context.space_data
    if snode is None: return
    if snode.tree_type != "ShaderNodeTree": return
    
    self.layout.separator()
    self.layout.menu("VIEW3D_MT_xdecal_node_context_menu", text=addon.name, icon='IMAGE_PLANE')
