# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import os
import math

import bpy
from bpy.app.translations import pgettext_iface as iface_
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import gpu

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "BlRna, prop, IDProp",
    "utils_ui": "NestedLayout, BlUI",
    "utils_addon": "AddonManager",
    "utils_blender": "BlUtil, UndoBlock, BpyPath",
})

from . import common

from .common import Jobs, DelayedUndoRecorder, CommonUtils, SimpleOpEditMixin, CommonConfig
from .decal_utils import DecalUtils, Decal
from .material_utils import MaterialMixer, NodeInfo, MaterialUtils

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

@addon.PropertyGroup
class DecalWrapper:
    @classmethod
    def obtain(cls):
        return DecalWrapperHolder.obtain("decal_ui_props")
    
    @classmethod
    def refresh(cls, context=None):
        if not context: context = bpy.context
        
        active_obj = context.object
        
        attributes = {}
        
        def add_value(attr_name, value, weight):
            values = attributes.setdefault(attr_name, {})
            values[value] = values.get(value, 0) + weight
        
        def add(decal, attr_name, fallback):
            value = getattr(decal, attr_name, fallback)
            decal_obj = (decal.obj if decal else None)
            weight = (math.inf if decal_obj == active_obj else 1)
            add_value(attr_name, value, weight)
        
        def add_all(decal):
            # If we do this, the user won't be able to change the template type
            # template = (decal.template if decal else None)
            # template_type = ('OBJECT' if isinstance(template, bpy.types.Object) else 'MATERIAL')
            # add_value("template_type", template_type)
            
            add(decal, "template", None)
            add(decal, "is_dynamic", False)
            add(decal, "mask_name", "")
            add(decal, "mask_min", 0.0)
            add(decal, "mask_max", 1.0)
            add(decal, "use_world_scale", False)
            add(decal, "aspect", 0.0)
            add(decal, "subdivs", 0.0)
            add(decal, "curv_min", 0.0)
            add(decal, "curv_maj", 0.0)
            add(decal, "angle", 0.0)
            add(decal, "range_x", 0.0)
            add(decal, "range_y", 0.0)
            add(decal, "depth_min", 0.0)
            add(decal, "depth_max", 0.0)
            add(decal, "displacement", 0.0)
            add(decal, "use_project", False)
            add(decal, "use_clip", False)
            add(decal, "use_cull", False)
            add(decal, "use_occlude", False)
            add(decal, "use_intersect", False)
            add(decal, "extra_location_x", 0.0)
            add(decal, "extra_location_y", 0.0)
            add(decal, "extra_location_z", 0.0)
            add(decal, "extra_rotation_x", 0.0)
            add(decal, "extra_rotation_y", 0.0)
            add(decal, "extra_rotation_z", 0.0)
            add(decal, "extra_scale_x", 0.0)
            add(decal, "extra_scale_y", 0.0)
            add(decal, "extra_scale_z", 0.0)
            add(decal, "use_uv_map", False)
            add(decal, "use_embed", False)
            add(decal, "extrude_layers", 0)
            add(decal, "extrude_amount", 0.0)
        
        for obj in context.selected_objects:
            decal = Decal(obj)
            if decal: add_all(decal)
        
        # Do this before add_all(None)
        cls.has_decals = bool(attributes)
        
        if not attributes: add_all(None)
        
        cls.lock_updates = True
        
        wrapper = cls.obtain()
        
        cls.initial_values = {}
        cls.multi_values = {}
        
        for attr_name, values in attributes.items():
            is_multi_value = (len(values) > 1)
            best_value = None
            best_count = 0
            if (attr_name != "template") or not is_multi_value:
                for value, count in values.items():
                    if count > best_count:
                        best_value, best_count = value, count
            
            cls.initial_values[attr_name] = best_value
            cls.multi_values[attr_name] = is_multi_value
            
            # Avoid relayouting if nothing has changed
            if getattr(wrapper, attr_name) != best_value:
                setattr(wrapper, attr_name, best_value)
        
        cls.lock_updates = False
    
    has_decals = False
    initial_values = {}
    multi_values = {}
    
    lock_updates = False
    
    rebake_trigger_attrs = {"template", "mask_name", "use_world_scale", "use_project",
        "use_clip", "use_cull", "use_occlude", "use_intersect", "use_uv_map", "use_embed"}
    
    @classmethod
    def update_value(cls, self, context, attr_name, should_rebake=False):
        if cls.lock_updates: return
        
        old_value = cls.initial_values.get(attr_name)
        value = getattr(self, attr_name)
        
        is_embed = (attr_name == "use_embed")
        should_rebake |= (attr_name in cls.rebake_trigger_attrs)
        
        for obj in context.selected_objects:
            decal = Decal(obj)
            if not decal: continue
            
            setattr(decal, attr_name, value)
            
            obj.update_tag()
            
            if is_embed and decal.use_embed:
                decal.embed()
            elif should_rebake:
                decal.update()
        
        DelayedUndoRecorder.trigger(message="Configure Decal(s)")
    
    # Geometry, materials, images
    tab_items = []
    tab_index: 0 | prop() # a bpy property so that Blender would update the UI on changes
    
    # TODO: relative mode? (might need to recalculate values if switching between relative/absolute)
    
    # Note: this reference counts among the datablock users, but we can't do anything about it
    template: bpy.types.ID | prop("Template", "The material or object on which this decal is based",
        poll=(lambda self, idblock: Decal.can_use_as_template(idblock)),
        update=(lambda self, context: DecalWrapper.update_value(self, context, "template")))
    template_type: 'MATERIAL' | prop("Template type", "Which datablock type to select a template from", items=[
        ('MATERIAL', "Material", "", 'MATERIAL', 0),
        ('OBJECT', "Object", "", 'OBJECT_DATA', 1),
    ])
    
    is_dynamic: True | prop("Dynamic", "Whether the decal is dynamic or baked",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "is_dynamic")))
    
    mask_name: "" | prop("Mask Name", "Vertex group to which the projection should be limited",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "mask_name")))
    mask_min: 0.0 | prop("Mask Min", "Minimal threshold of the mask weights", step=10, min=0, max=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "mask_min")))
    mask_max: 1.0 | prop("Mask Max", "Maximal threshold of the mask weights", step=10, min=0, max=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "mask_max")))
    use_world_scale: False | prop("World Scale", "Use template object's world scale",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_world_scale")))
    aspect: 0.0 | prop("Aspect ratio", "Image aspect ratio", step=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "aspect")))
    subdivs: 0.0 | prop("Subdivisions", "Subdivision factor (shrinkwrap if > 0)", step=100, min=-8, max=8,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "subdivs")))
    curv_min: 0.0 | prop("Minor curvature", "Minor curvature (along the X-axis)", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "curv_min")))
    curv_maj: 0.0 | prop("Major curvature", "Major curvature (along the Y-axis)", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "curv_maj")))
    angle: 0.0 | prop("Angle", "Rotation of curvature axes around decal's Z axis", subtype='ANGLE', step=100,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "angle")))
    range_x: 0.0 | prop("Range X", "Decal size along X (minor) axis", step=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "range_x")))
    range_y: 0.0 | prop("Range Y", "Decal size along Y (major) axis", step=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "range_y")))
    depth_min: 0.0 | prop("Depth below", "Projection depth below the decal", precision=3, step=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "depth_min")))
    depth_max: 0.0 | prop("Depth above", "Projection depth above the decal", precision=3, step=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "depth_max")))
    displacement: 0.0 | prop("Clearance", "Offset from the underlying surface", precision=3, step=1,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "displacement")))
    use_project: False | prop("Project", "Project the decal on the target object(s)",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_project")))
    use_clip: False | prop("Clip", "Remove polygons outside of the decal's bounding box",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_clip")))
    use_cull: False | prop("Cull", "Remove backfacing polygons",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_cull")))
    use_occlude: False | prop("Occlude", "Remove occluded polygons",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_occlude")))
    use_intersect: False | prop("Intersect", "Use boolean intersections (exact result, but slow)",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_intersect")))
    use_uv_map: False | prop("Add UV Map", "Calculate vertex UVs",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_uv_map")))
    use_embed: False | prop("Embed", "Mix decal material with the materials of target surfaces",
        update=(lambda self, context: DecalWrapper.update_value(self, context, "use_embed")))
    
    extrude_layers: 0 | prop("Layers", "Number of extrusion layers", min=0, max=1000,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extrude_layers")))
    extrude_amount: 0.0 | prop("Extrude", "Extrusion distance", min=0.0,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extrude_amount")))
    
    extra_location_x: 0.0 | prop("X", "Location adjustment", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_location_x")))
    extra_location_y: 0.0 | prop("Y", "Location adjustment", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_location_y")))
    extra_location_z: 0.0 | prop("Z", "Location adjustment", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_location_z")))
    extra_location: Vector() | prop("Location", "Location adjustment", step=10,
        get=(lambda self: self.get_vector("extra_location")),
        set=(lambda self, value: self.set_vector("extra_location", value)))
    
    extra_rotation_x: 0.0 | prop("X", "Rotation adjustment", subtype='ANGLE', step=100,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_rotation_x")))
    extra_rotation_y: 0.0 | prop("Y", "Rotation adjustment", subtype='ANGLE', step=100,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_rotation_y")))
    extra_rotation_z: 0.0 | prop("Z", "Rotation adjustment", subtype='ANGLE', step=100,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_rotation_z")))
    extra_rotation: Vector() | prop("Rotation", "Rotation adjustment", step=10,
        get=(lambda self: self.get_vector("extra_rotation")),
        set=(lambda self, value: self.set_vector("extra_rotation", value)))
    
    extra_scale_x: 0.0 | prop("X", "Scale adjustment", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_scale_x")))
    extra_scale_y: 0.0 | prop("Y", "Scale adjustment", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_scale_y")))
    extra_scale_z: 0.0 | prop("Z", "Scale adjustment", step=10,
        update=(lambda self, context: DecalWrapper.update_value(self, context, "extra_scale_z")))
    extra_scale: Vector() | prop("Scale", "Scale adjustment", step=10,
        get=(lambda self: self.get_vector("extra_scale")),
        set=(lambda self, value: self.set_vector("extra_scale", value)))
    
    def get_vector(self, name):
        x = getattr(self, name+"_x")
        y = getattr(self, name+"_y")
        z = getattr(self, name+"_z")
        return Vector((x, y, z))
    
    def set_vector(self, name, value):
        setattr(self, name+"_x", value[0])
        setattr(self, name+"_y", value[1])
        setattr(self, name+"_z", value[2])

@addon.PropertyGroup
class ShaderObjAttrEntry:
    active_material = None
    
    attr_type: "" | prop()
    is_linear: True | prop()
    lock_counter: 0 | prop()
    
    def is_matching(self, obj, material):
        decal = Decal(obj)
        if decal: return (decal.template == material)
        
        for ms in obj.material_slots:
            if ms.material == material: return True
        
        return False
    
    def update(self, context, index=-1):
        if self.lock_counter > 0: return
        
        active_material = ShaderObjAttrEntry.active_material
        
        if self.attr_type == 'FLOAT':
            value = self.factor
        else:
            value = [getattr(self, f"float{i}") for i in range(len(self.attr_type))]
        
        for obj in context.selected_objects:
            if not self.is_matching(obj, active_material): continue
            
            old_value = None
            clamped = value
            
            if self.name in obj:
                old_value = obj[self.name]
                
                value_type, value_size = IDProp.get_type_info(old_value)
                size_limit = MaterialMixer.shader_attr_size_limits.get(value_type)
                if size_limit is None:
                    old_value = None
                elif value_size > size_limit:
                    old_value = (list(old_value)[:size_limit] if size_limit else old_value[0])
                elif hasattr(old_value, "__len__"):
                    old_value = list(old_value)
                
                metadata = IDProp.get_metadata(obj, self.name)
                min_value = metadata.get("min")
                max_value = metadata.get("max")
                if isinstance(clamped, float):
                    if min_value is not None:
                        clamped = max(float(min_value), clamped)
                    if max_value is not None:
                        clamped = min(float(max_value), clamped)
                else:
                    for i in range(len(clamped)):
                        if min_value is not None:
                            clamped[i] = max(float(min_value), clamped[i])
                        if max_value is not None:
                            clamped[i] = min(float(max_value), clamped[i])
            
            if old_value is None:
                metadata = {}
                if self.attr_type.startswith('RGB'):
                    metadata["subtype"] = ('COLOR' if self.is_linear else 'COLOR_GAMMA')
                data = {"value": clamped, "metadata": metadata}
                
                IDProp.set_data(obj, self.name, data)
            elif index >= 0:
                if not hasattr(old_value, "__len__"):
                    old_value = [old_value] * 3
                
                size = len(old_value)
                new_size = max(size, index+1)
                new_value = [0.0] * new_size
                for i in range(new_size):
                    if i == index:
                        new_value[i] = clamped[i]
                    elif i < size:
                        new_value[i] = float(old_value[i])
                    elif i == 3:
                        new_value[i] = 1.0
                
                obj[self.name] = new_value
            elif not hasattr(old_value, "__len__"):
                obj[self.name] = clamped
            else:
                vec_value = (clamped if hasattr(clamped, "__len__") else [clamped] * 3)
                
                size = len(vec_value)
                new_size = max(size, len(old_value))
                new_value = [0.0] * new_size
                for i in range(new_size):
                    new_value[i] = (vec_value[i] if i < size else float(old_value[i]))
                
                obj[self.name] = new_value
            
            obj.update_tag()
        
        DelayedUndoRecorder.trigger(message="Set material attributes")
    
    float0: 1.0 | prop("X", "", precision=3, step=1, update=(lambda self, context: self.update(context, 0)))
    float1: 1.0 | prop("Y", "", precision=3, step=1, update=(lambda self, context: self.update(context, 1)))
    float2: 1.0 | prop("Z", "", precision=3, step=1, update=(lambda self, context: self.update(context, 2)))
    float3: 1.0 | prop("α", "", precision=3, step=1, update=(lambda self, context: self.update(context, 3)))
    
    factor: 1.0 | prop("Value", "", precision=3, step=1, update=(lambda self, context: self.update(context)))
    
    def color_get(self):
        return (self.float0, self.float1, self.float2, self.float3)
    def color_set(self, value):
        self.lock_counter += 1
        self.float0, self.float1, self.float2, self.float3 = value
        self.lock_counter -= 1
        self.update(bpy.context)
    
    color_linear: (0.0, 0.0, 0.0, 1.0) | prop("Color", "", min=0.0, max=1.0, subtype='COLOR',
        get=color_get, set=color_set)
    color_gamma: (0.0, 0.0, 0.0, 1.0) | prop("Color", "", min=0.0, max=1.0, subtype='COLOR_GAMMA',
        get=color_get, set=color_set)

@addon.PropertyGroup
class ShaderObjAttrSet:
    entries: [ShaderObjAttrEntry] | prop()
    
    def gather_values(self, objs):
        for attr_entry in self.entries:
            value_sum = [0.0, 0.0, 0.0, 0.0]
            value_counts = [0, 0, 0, 0]
            
            gamma_count = 0
            total_count = 0
            
            for obj in objs:
                if attr_entry.name not in obj: continue
                
                value = obj[attr_entry.name]
                
                value_type, value_size = IDProp.get_type_info(value)
                size_limit = MaterialMixer.shader_attr_size_limits.get(value_type)
                if (size_limit is None) or (value_size > size_limit): continue
                
                if value_size == 0: value = [value] * 3
                
                for i in range(len(value)):
                    value_sum[i] += float(value[i])
                    value_counts[i] += 1
                
                metadata = IDProp.get_metadata(obj, attr_entry.name)
                is_gamma = (metadata.get("subtype") == 'COLOR_GAMMA')
                gamma_count += int(is_gamma)
                total_count += 1
            
            # If alpha is not specified, Blender assumes the value 1.0
            if value_counts[3] == 0: value_sum[3] = 1.0
            
            vector_value = tuple(v/max(n, 1) for v, n in zip(value_sum, value_counts))
            factor_value = sum(vector_value[:3]) / 3
            is_linear = (gamma_count * 2 <= total_count)
            
            attr_entry.lock_counter += 1
            attr_entry.color_linear = vector_value
            attr_entry.factor = factor_value
            attr_entry.is_linear = is_linear
            attr_entry.lock_counter -= 1

@addon.External.Include
class DecalWrapperHolder:
    decal_ui_props: DecalWrapper | prop()
    shader_obj_attrs: [ShaderObjAttrSet] | prop()
    
    @classmethod
    def obtain(cls, name):
        return getattr(addon.external, name)

@addon.Operator(idname="xdecal.popup_tab_select", label="Select", options={'INTERNAL'})
class SelectTargetsOperator:
    "Select what to edit"
    
    index: 0 | prop()
    
    def execute(self, context):
        wrapper = DecalWrapper.obtain()
        wrapper.tab_index = self.index
        # context.region.tag_redraw()
        return {'FINISHED'}

@addon.Operator(idname="xdecal.popup_propagate_value", label="Propagate", options={'INTERNAL'})
class PropagateDecalPropOperator:
    "Propagate this value to all selected decals"
    
    attr_name: "" | prop()
    
    def execute(self, context):
        wrapper = DecalWrapper.obtain()
        if not hasattr(wrapper, self.attr_name): return {'CANCELLED'}
        if self.attr_name == "mask_name":
            # mask_min and mask_max don't have dedicated icons
            DecalWrapper.update_value(wrapper, context, "mask_min")
            DecalWrapper.update_value(wrapper, context, "mask_max")
        DecalWrapper.update_value(wrapper, context, self.attr_name, should_rebake=True)
        # context.region.tag_redraw()
        return {'FINISHED'}

@addon.Menu(idname="VIEW3D_MT_xdecal_popup_tab_selector", label="Select")
class PopupTabSelectorMenu:
    "Which settings to show"
    
    def draw(self, context):
        layout = self.layout
        for i, tab_item in enumerate(DecalWrapper.tab_items):
            op = layout.operator("xdecal.popup_tab_select", **tab_item)
            op.index = i

@addon.Menu(idname="VIEW3D_MT_xdecal_popup_extra_actions", label="Extras")
class PopupExtraActionsMenu:
    "Extra settings and operations"
    
    def draw(self, context):
        layout = self.layout
        
        text = layout.enum_item_name(addon.internal, "decal_alignment_mode", settings.decal_alignment_mode)
        layout.prop_menu_enum(addon.internal, "decal_alignment_mode", text=text, icon='SNAP_NORMAL')
        
        layout.operator("xdecal.repair_node_group_refs", text="Repair", icon='RECOVER_LAST')
        layout.operator("xdecal.convert", text="Convert", icon='IMAGE_PLANE')
        layout.operator("xdecal.update_decals", text="Update", icon='FILE_REFRESH')
        layout.operator("xdecal.sync_collections", text="Sync collections", icon='OUTLINER_COLLECTION')
        layout.operator("xdecal.replace_materials_in_objects", text="Replace materials", icon='AUTOMERGE_ON')

@addon.Preset("folded_groups.preset", popup='PANEL', title="Folded groups", options={'ADD', 'RESET', 'INLINE'})
class InvertedGroupFoldState:
    OpEditMixin = SimpleOpEditMixin
    normalize_whitespace = False
    type_name = "material control group name"

class DecalUI:
    collapse_prop_name = "xdecal.mat_props_collapse"
    
    material_current = None
    
    @classmethod
    def mat_props_copy(cls):
        MaterialUtils.clipboard = NodeInfo.mat_props_copy(cls.material_current)
    
    @classmethod
    def mat_props_paste(cls):
        NodeInfo.mat_props_paste(cls.material_current, MaterialUtils.clipboard)
    
    @classmethod
    def get_collapsed(cls, screen, key):
        collapse_dict = IDProp.normalize(screen.get(cls.collapse_prop_name))
        return bool(collapse_dict.get(key) if isinstance(collapse_dict, dict) else None)
    
    @classmethod
    def set_collapsed(cls, screen, key, value):
        collapse_dict = IDProp.normalize(screen.get(cls.collapse_prop_name))
        if not isinstance(collapse_dict, dict): collapse_dict = {}
        collapse_dict[key] = bool(value)
        screen[cls.collapse_prop_name] = collapse_dict
    
    @classmethod
    def gather_materials(cls, context):
        materials = set()
        
        for obj in context.selected_objects:
            decal = Decal(obj)
            
            if decal:
                template = decal.template
                if isinstance(template, bpy.types.Material):
                    materials.add(template)
            else:
                for ms in obj.material_slots:
                    if ms.material: materials.add(ms.material)
        
        return materials
    
    @classmethod
    def gather_tabs(cls, context):
        tabs = []
        
        def draw_func(layout, context):
            DecalWrapper.refresh(context=context)
            cls.draw_decal_wrapper_controls(layout)
        
        tabs.append((0, "Geometry", 'MESH_DATA', draw_func))
        
        materials = cls.gather_materials(context)
        
        shader_obj_attrs = DecalWrapperHolder.obtain("shader_obj_attrs")
        shader_obj_attrs.clear()
        
        image_keys = set()
        material_controls = []
        
        for material in materials:
            image_controls, node_controls = cls.gather_material_controls(material)
            
            for key, draw_func in image_controls:
                if key in image_keys: continue
                image_keys.add(key)
                
                source_image, image = key
                tabs.append((1, source_image.name, 'IMAGE_DATA', draw_func))
            
            if node_controls:
                # Make sure it's captured in a closure
                def make_draw_func(node_controls, material):
                    def draw_func(layout, context):
                        cls.material_current = material
                        
                        row = layout.row()
                        row.operator("xdecal.mat_props_copy", text="Copy", icon='COPYDOWN')
                        row.operator("xdecal.mat_props_paste", text="Paste", icon='PASTEDOWN')
                        
                        for key, draw_func in node_controls:
                            draw_func(layout, context)
                    
                    return draw_func
                
                tabs.append((2, material.name, 'MATERIAL', make_draw_func(node_controls, material)))
        
        tabs.sort()
        
        return tabs
    
    @classmethod
    def draw_all_controls(cls, layout, context):
        tabs = cls.gather_tabs(context)
        
        DecalWrapper.tab_items = []
        draw_funcs = []
        
        for _, text, icon, draw_func in tabs:
            tab_item = {"text": text}
            tab_item["icon" if isinstance(icon, str) else "icon_value"] = icon
            DecalWrapper.tab_items.append(tab_item)
            draw_funcs.append(draw_func)
        
        wrapper = DecalWrapper.obtain()
        tab_index = wrapper.tab_index
        if tab_index >= len(tabs): tab_index = 0
        
        cls.draw_enable_disable_controls(layout)
        
        tab_item = DecalWrapper.tab_items[tab_index]
        row = layout.row(align=True)
        row.menu("VIEW3D_MT_xdecal_popup_extra_actions", text="", icon='THREE_DOTS')
        row.menu("VIEW3D_MT_xdecal_popup_tab_selector", **tab_item)
        
        draw_func = draw_funcs[tab_index]
        draw_func(layout, context)
    
    @classmethod
    def draw_enable_disable_controls(cls, layout):
        layout = NestedLayout(layout)
        
        def draw_op(prefix, mode, text, icon):
            idname = f"xdecal.{prefix}_decals"
            op = layout.operator(idname, text=" ", icon=icon)
            op.name = ""
            op.type = ""
            op.mode = mode
            op.undo_label = "Decals: "+text
        
        with layout.row():
            with layout.row(align=True):
                draw_op("enable", 'SELECT', "Enable Selection", 'RESTRICT_SELECT_OFF')
                draw_op("disable", 'SELECT', "Disable Selection", 'RESTRICT_SELECT_ON')
            with layout.row(align=True):
                draw_op("enable", 'VIEW_LAYER', "Show in View Layer", 'HIDE_OFF')
                draw_op("disable", 'VIEW_LAYER', "Hide in View Layer", 'HIDE_ON')
            with layout.row(align=True):
                draw_op("enable", 'VIEWPORT', "Enable in Viewport", 'RESTRICT_VIEW_OFF')
                draw_op("disable", 'VIEWPORT', "Disable in Viewport", 'RESTRICT_VIEW_ON')
            with layout.row(align=True):
                draw_op("enable", 'RENDER', "Enable in Rendering", 'RESTRICT_RENDER_OFF')
                draw_op("disable", 'RENDER', "Disable in Rendering", 'RESTRICT_RENDER_ON')
    
    @classmethod
    def draw_decal_wrapper_controls(cls, layout):
        layout = NestedLayout(layout)
        
        wrapper = DecalWrapper.obtain()
        
        def draw_md_prop(key, text=None, icon='NONE', **kwargs):
            if text is None: text = key
            
            if bpy.app.version < (4, 1, 0):
                kwargs.pop("placeholder", None)
            
            if (icon != 'NONE') and (not isinstance(getattr(wrapper, key), bool)):
                with layout.row(align=True):
                    op = layout.operator("xdecal.popup_propagate_value",
                        text="", icon=icon, emboss=bool(wrapper.multi_values.get(key)))
                    op.attr_name = key
                    layout.prop(wrapper, key, text=text, **kwargs)
            else:
                kwargs["toggle"] = True
                layout.prop(wrapper, key, text=text, icon=icon, **kwargs)
        
        with layout.row()(operator_context='INVOKE_DEFAULT'):
            prop_text = ("Dynamic" if wrapper.is_dynamic else "Baked")
            prop_icon = ('PLAY' if wrapper.is_dynamic else 'PAUSE')
            with layout.row()(enabled=wrapper.has_decals):
                layout.prop(wrapper, "is_dynamic", text=prop_text, icon=prop_icon)
            layout.operator("xdecal.update_decals", text="Update", icon='FILE_REFRESH')
            
        with layout.column()(operator_context='INVOKE_DEFAULT', enabled=wrapper.has_decals):
            with layout.column(align=True):
                with layout.row(align=True):
                    layout.custom_any_ID(wrapper, "template", "template_type", text="")
            
            with layout.column(align=True):
                with layout.row(align=True):
                    draw_md_prop("use_uv_map", text="UV map", icon='UV_DATA')
                    draw_md_prop("use_world_scale", text="Use Scale", icon='WORLD')
            
            with layout.row(align=True):
                layout.operator("xdecal.select_targets", text="Select", icon='RESTRICT_SELECT_OFF')
                layout.operator("xdecal.assign_targets", text="Assign", icon='CHECKMARK')
            
            with layout.column(align=True):
                with layout.row(align=True):
                    with layout.column(align=True):
                        draw_md_prop("use_project", text="Project", icon='MOD_SHRINKWRAP')
                        draw_md_prop("use_clip", text="Clip", icon='SHADING_BBOX')
                        draw_md_prop("use_cull", text="Cull", icon='NORMALS_FACE')
                    with layout.column(align=True):
                        draw_md_prop("use_embed", text="Embed", icon='EYEDROPPER')
                        draw_md_prop("use_occlude", text="Occlude", icon='LIGHT_AREA')
                        draw_md_prop("use_intersect", text="Intersect", icon='MOD_BOOLEAN')
            
            with layout.column(align=True):
                draw_md_prop("mask_name", text="", icon='MOD_MASK', placeholder="Mask")
                with layout.row(align=True):
                    draw_md_prop("mask_min", text="Min")
                    draw_md_prop("mask_max", text="Max")
                
            with layout.column(align=True):
                with layout.row(align=False):
                    with layout.column(align=True):
                        draw_md_prop("subdivs", text="", icon='MESH_GRID')
                        draw_md_prop("aspect", text="", icon='IMAGE_PLANE')
                    with layout.column(align=True):
                        draw_md_prop("range_x", text="X", icon='OBJECT_HIDDEN')
                        draw_md_prop("range_y", text="Y", icon='OBJECT_HIDDEN')
            
            with layout.column(align=True):
                with layout.row(align=False):
                    with layout.column(align=True):
                        draw_md_prop("curv_min", text="", icon='MESH_CYLINDER')
                        draw_md_prop("curv_maj", text="", icon='MESH_TORUS')
                        draw_md_prop("angle", text="", icon='DRIVER_ROTATIONAL_DIFFERENCE')
                    with layout.column(align=True):
                        draw_md_prop("depth_max", text="", icon='SORT_DESC')
                        draw_md_prop("depth_min", text="", icon='SORT_ASC')
                        draw_md_prop("displacement", text="", icon='MOD_DISPLACE')
            
            with layout.column(align=True):
                with layout.row(align=False):
                    with layout.column(align=True):
                        draw_md_prop("extrude_layers", text="", icon='MOD_WAVE')
                    with layout.column(align=True):
                        draw_md_prop("extrude_amount", text="", icon='RNDCURVE')
            
            with layout.column(align=True):
                draw_md_prop("extra_location", text="", icon='OBJECT_ORIGIN')
                draw_md_prop("extra_rotation", text="", icon='ORIENTATION_GIMBAL')
                draw_md_prop("extra_scale", text="", icon='MOD_LENGTH')
    
    @classmethod
    def draw_material_controls(cls, layout, material, context):
        image_controls, node_controls = cls.gather_material_controls(material)
        
        cls.material_current = material
        
        row = layout.row()
        row.operator("xdecal.mat_props_copy", text="Copy", icon='COPYDOWN')
        row.operator("xdecal.mat_props_paste", text="Paste", icon='PASTEDOWN')
        
        for key, draw_func in image_controls:
            draw_func(layout, context)
        
        for key, draw_func in node_controls:
            draw_func(layout, context)
    
    @classmethod
    def gather_material_controls(cls, material):
        image_controls = []
        node_controls = {}
        grouped_node_controls = []
        
        if material and material.node_tree:
            nodes = sorted(material.node_tree.nodes, key=(lambda node: node.name))
            
            for node in nodes:
                if node.bl_idname == "ShaderNodeTexImage":
                    cls.gather_image_controls(node, image_controls)
                
                cls.gather_node_controls(node, node_controls)
            
            cls.gather_attributes_controls(material, node_controls)
        
        for label in sorted(node_controls.keys()):
            draw_func = cls.group_node_controls(label, node_controls[label])
            grouped_node_controls.append((label, draw_func))
        
        return image_controls, grouped_node_controls
    
    @classmethod
    def gather_attributes_controls(cls, material, node_controls):
        prop_infos = MaterialMixer.gather_custom_props(material)
        if not prop_infos: return
        
        shader_obj_attrs = DecalWrapperHolder.obtain("shader_obj_attrs")
        
        shader_obj_attr_set = shader_obj_attrs.add()
        attr_entries = shader_obj_attr_set.entries
        
        for attr_name, attr_type in prop_infos.items():
            if attr_type == 'NONE': continue
            
            attr_entry = attr_entries.add()
            attr_entry.name = attr_name
            attr_entry.attr_type = attr_type
            cls.gather_attribute_controls(attr_entry, node_controls)
        
        if len(attr_entries) == 0: return
        
        def draw_func(layout, context, count):
            ShaderObjAttrEntry.active_material = material
            shader_obj_attr_set.gather_values(context.selected_objects)
        
        cls.add_node_draw_func(node_controls, cls.obj_attrs_label, draw_func, controls_count=0)
    
    obj_attrs_label = "<Per-object attributes>"
    
    @classmethod
    def gather_attribute_controls(cls, attr_entry, node_controls):
        attr_name = attr_entry.name
        attr_type = attr_entry.attr_type
        
        def draw_func(layout, context, count):
            if attr_type == 'FLOAT':
                layout.prop(attr_entry, "factor", text=attr_name)
            elif attr_type in ('RGB', 'RGBA'):
                prop_name = ("color_linear" if attr_entry.is_linear else "color_gamma")
                row = layout.row()
                row.prop(attr_entry, prop_name, text=attr_name)
            else:
                col = layout.column(align=True)
                col.prop(attr_entry, "float0", text=f"{attr_name} X")
                col.prop(attr_entry, "float1", text=f"{attr_name} Y")
                col.prop(attr_entry, "float2", text=f"{attr_name} Z")
                if len(attr_type) == 4:
                    col.prop(attr_entry, "float3", text=f"{attr_name} α")
        
        cls.add_node_draw_func(node_controls, cls.obj_attrs_label, draw_func, controls_count=2)
    
    @classmethod
    def gather_image_controls(cls, node, image_controls):
        image = node.image
        if not image: return
        
        source_image = image.xdecal.source_image
        if not source_image: return
        
        if source_image.xdecal.normal_map == image:
            def draw_func(layout, context):
                cls.draw_normalmap_controls(layout, source_image)
            image_controls.append(((source_image, image), draw_func))
    
    @classmethod
    def draw_group_collapser(cls, layout, label):
        value = cls.get_collapsed(bpy.context.screen, label)
        
        invert = (label in InvertedGroupFoldState.presets)
        effective_value = value != invert
        
        row = layout.row()
        row.alignment = 'LEFT'
        row.emboss = 'NONE'
        
        icon = ('DISCLOSURE_TRI_RIGHT' if effective_value else 'DISCLOSURE_TRI_DOWN')
        op = row.operator("xdecal.mat_prop_collapse", text=label+":", icon=icon)
        op.key = label
        op.value = not value
        
        return effective_value
    
    @classmethod
    def group_node_controls(cls, label, node_controls_info):
        draw_funcs = node_controls_info["draw"]
        count = node_controls_info["count"]
        
        def draw_func(layout, context):
            box = layout.box()
            col = box.column(align=False)
            
            if (count > 1) and label:
                if cls.draw_group_collapser(col, label): return
            
            for draw_func in draw_funcs:
                draw_func(col, context, count)
        
        return draw_func
    
    @classmethod
    def gather_node_controls(cls, node, node_controls):
        info = NodeInfo.get(node)
        if not info: return
        
        label = info["group"]
        show_buttons = 'B' in info["toggles"]
        
        input_ids = NodeInfo.get_socket_ids(node.inputs, info)
        output_ids = NodeInfo.get_socket_ids(node.outputs, info)
        
        # Note: if total controls_count > 1 will use an explicit box label, controls_count == 1 will not.
        if (node.bl_idname == "ShaderNodeTexImage") and not show_buttons:
            controls_count = 2
            
            def draw_func(layout, context, count):
                cls.draw_image_controls(layout, node)
        else:
            draw_buttons = show_buttons and hasattr(node, "draw_buttons")
            draw_buttons_ext = show_buttons and hasattr(node, "draw_buttons_ext")
            show_buttons &= draw_buttons or draw_buttons_ext
            controls_count = len(input_ids) + len(output_ids) + int(show_buttons)*2
            if controls_count == 0: return
            
            def draw_func(layout, context, count):
                cls.draw_node_controls(layout, node, label, show_buttons, input_ids, output_ids, count, context)
        
        cls.add_node_draw_func(node_controls, label, draw_func, controls_count)
    
    @classmethod
    def add_node_draw_func(cls, node_controls, label, draw_func, controls_count):
        # Note: controls_count is allowed to be 0 in special cases
        
        node_controls_info = node_controls.get(label)
        if not node_controls_info:
            node_controls_info = {"draw":[], "count":0}
            node_controls[label] = node_controls_info
        
        node_controls_info["draw"].append(draw_func)
        node_controls_info["count"] += controls_count
    
    @classmethod
    def draw_normalmap_controls(cls, layout, source_image):
        box = layout.box()
        col = box.column(align=True)
        col.label(text="Generated normal map:")
        row = col.row(align=True)
        row.prop(source_image.xdecal, "normal_map_blur")
        row.prop(source_image.xdecal, "normal_map_blur_type", text="", icon_only=True)
        col.prop(source_image.xdecal, "normal_map_scale")
    
    @classmethod
    def draw_image_controls(cls, layout, node):
        col = layout.column(align=True)
        col.context_pointer_set("node", node)
        
        # We can't use built-in image.new / image.open, since they
        # don't assign the image to the node, for some reason
        op_open = "xdecal.shader_image_open"
        op_unlink = "xdecal.shader_image_unlink"
        col.template_ID(node, "image", new="", open=op_open, unlink=op_unlink)
        # col.template_ID_preview(node, "image", new="", open=op_open, unlink=op_unlink)
    
    @classmethod
    def draw_node_controls(cls, layout, node, label, show_buttons, input_ids, output_ids, count, context):
        col = layout.column(align=True)
        col.context_pointer_set("node", node)
        
        if show_buttons or (count > 1):
            if show_buttons and hasattr(node, "draw_buttons_ext"):
                node.draw_buttons_ext(context, col)
            elif show_buttons and hasattr(node, "draw_buttons"):
                node.draw_buttons(context, col)
            
            cls.draw_node_sockets(col, node, node.inputs, False, None, input_ids, context)
            cls.draw_node_sockets(col, node, node.outputs, True, None, output_ids, context)
        else:
            row = col.row(align=True)
            cls.draw_node_sockets(row, node, node.inputs, False, label, input_ids, context)
            cls.draw_node_sockets(row, node, node.outputs, True, label, output_ids, context)
    
    @classmethod
    def draw_node_sockets(cls, layout, node, sockets, draw_default, label, socket_ids, context):
        if not socket_ids: return
        for socket in sockets:
            socket_label = socket_ids.get(socket.identifier)
            if not socket_label: continue
            socket_label = iface_(label or socket_label, socket.bl_rna.translation_context)
            
            if socket.type == 'VECTOR':
                col = layout.column(align=True)
                col.row(align=True).label(text=socket_label+":")
                col.row(align=True).prop(socket, "default_value", text="")
            elif draw_default:
                layout.row(align=True).prop(socket, "default_value", text=socket_label)
            else:
                socket.draw(context, layout.row(align=True), node, socket_label)

@addon.Operator(idname="xdecal.shader_image_open", label="Open", options={'INTERNAL'})
class OpenShaderImageOperator:
    "Open an image"
    from pathlib import Path
    filepath: "" | prop("File Path", "File path used for exporting", maxlen=1024, subtype='FILE_PATH')
    filter_image: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    filter_movie: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    filter_folder: True | prop(options={'HIDDEN', 'SKIP_SAVE'})
    
    alpha_mode_rna = bpy.types.Image.bl_rna.properties["alpha_mode"]
    alpha_mode: alpha_mode_rna.default | prop(alpha_mode_rna.name, alpha_mode_rna.description,
        items=[(item.identifier, item.name, item.description) for item in alpha_mode_rna.enum_items])
    relative: True | prop("Relative Paths", "Use relative file paths")
    check_existing: False | prop("Use Existing", "Use existing data-block if this file is already loaded")
    
    # Adapted from "Import Images As Planes" addon
    def postprocess(self, image):
        image.alpha_mode = self.alpha_mode
        
        if self.relative:
            try:
                image.filepath = BpyPath.relpath(image.filepath)
            except ValueError:
                pass
    
    def execute(self, context):
        node = getattr(self, "node", None) or getattr(context, "node", None)
        if node is None: return {'CANCELLED'}
        
        try:
            image = bpy.data.images.load(self.filepath, check_existing=self.check_existing)
        except Exception:
            return {'CANCELLED'}
        
        if not image: return {'CANCELLED'}
        
        self.postprocess(image)
        
        node.image = image
        
        bpy.ops.ed.undo_push(message="Load image(s)")
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        node = getattr(context, "node", None)
        if node is None: return {'CANCELLED'}
        
        self.node = node
        
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
    def draw(self, context):
        layout = self.layout
        layout.prop(self, "relative")
        layout.prop(self, "check_existing")
        layout.prop(self, "alpha_mode")

@addon.Operator(idname="xdecal.shader_image_unlink", label="Unlink", options={'INTERNAL'})
class UnlinkShaderImageOperator:
    "Unlink the image"
    
    def execute(self, context):
        node = getattr(context, "node", None)
        if node is None: return {'CANCELLED'}
        
        with UndoBlock("Add image"):
            node.image = None
        
        return {'FINISHED'}

@addon.Operator(idname="xdecal.mat_prop_collapse", label="(Un)collapse", options={'INTERNAL'})
class MaterialPropertyCollapseOperator:
    "(Un)collapse"
    
    key: "" | prop()
    value: False | prop()
    
    def execute(self, context):
        DecalUI.set_collapsed(context.screen, self.key, self.value)
        return {'FINISHED'}

@addon.Operator(idname="xdecal.mat_props_copy", label="Copy material parameters", options={'INTERNAL'})
class MaterialPropertiesCopyOperator:
    "Copy material control parameters"
    
    def execute(self, context):
        DecalUI.mat_props_copy()
        return {'FINISHED'}

@addon.Operator(idname="xdecal.mat_props_paste", label="Paste material parameters", options={'INTERNAL'})
class MaterialPropertiesCopyOperator:
    "Paste material control parameters to all selected decals (Ctrl+Click: paste to the current material)"
    
    to_selected_objects: True | prop()
    
    def execute(self, context):
        if self.to_selected_objects:
            objs = context.selected_objects
            types = ['MATERIAL']
            templates = DecalUtils.filter_templates(objs, types=types, editable=types)
            for template in templates:
                NodeInfo.mat_props_paste(template, MaterialUtils.clipboard)
        else:
            DecalUI.mat_props_paste()
        return {'FINISHED'}
    
    def invoke(self, context, event):
        self.to_selected_objects = not event.ctrl
        self.execute(context)
        bpy.ops.ed.undo_push(message="Paste material parameters")
        return {'FINISHED'}

def trigger_redraw(redraw=True, update=False):
    # Note: depsgraph update is not allowed in certain contexts,
    # so we have to be careful about using it
    if redraw: BlUI.tag_redraw()
    if update: BlUtil.Depsgraph.trigger_update()

@addon.Panel(idname="NODE_EDITOR_PT_xdecal", label="Decal Extras",
    space_type='NODE_EDITOR', region_type='UI', category=CommonConfig.sidebar_category)
class DecalsPanelNodeExtras:
    @classmethod
    def poll(cls, context):
        space = context.space_data
        return (space.type != 'NODE_EDITOR') or (space.shader_type == 'OBJECT')
    
    def draw(self, context):
        obj = context.object
        if obj: DecalUI.draw_material_controls(self.layout, obj.active_material, context=context)

@addon.Panel(idname="EMPTY_PT_xdecal_context_view3d", label="Decal settings", space_type='CONSOLE', region_type='WINDOW')
class DecalsPanelContextView3D:
    # @classmethod
    # def poll(cls, context):
    #     return bool(context.object)
    
    def draw(self, context):
        DecalUI.draw_all_controls(self.layout, context=context)

@addon.ui_draw("VIEW3D_MT_object_context_menu", 'PREPEND')
def object_context_menu_draw(self, context):
    obj = context.object
    # if not obj: return
    # if not obj.xdecal.template: return
    self.layout.popover("EMPTY_PT_xdecal_context_view3d", text=addon.name, icon='IMAGE_PLANE')
    self.layout.separator()

@addon.Preset("favorite_textures.preset", popup='PANEL', title="Favorite textures", options={'ADD', 'RESET', 'INLINE'})
class FavoriteTexture:
    def serialize(self):
        return {"path": self.path}
    
    def update(self, path=None, context=None, **kwargs):
        self.path = path
    
    class OpEditMixin:
        id: "" | prop("Name", "Category / name of the favorite image")
        path: "" | prop("Path", "Path to the image")
        
        def get_image_node(self, context):
            node = getattr(context, "node", None)
            return (node if isinstance(node, bpy.types.ShaderNodeTexImage) else None)
        
        def execute(self, context):
            if not (self.id and self.path): return {'CANCELLED'}
            
            FavoriteTexture.presets.add(FavoriteTexture(self.id, context=context, path=self.path))
            
            old_id = getattr(self, "old_id", None)
            if old_id and (old_id != self.id):
                FavoriteTexture.presets.delete(old_id)
            
            BlUI.tag_redraw()
            
            return {'FINISHED'}
        
        def invoke(self, context, event):
            if not self.id:
                node = self.get_image_node(context)
                image = (node.image if node else None)
                if image:
                    self.id = image.name
                    self.path = CommonUtils.abspath(image.filepath)
            
            wm = context.window_manager
            return wm.invoke_props_dialog(self, width=300)
        
        def draw(self, context):
            self.layout.alert = not self.id
            self.layout.prop(self, "id")
            row = self.layout.row()
            row.alert = not os.path.isfile(self.path)
            row.prop(self, "path")

@addon.Operator(idname="xdecal.favorite_texture_click", label="Favorite", options={'INTERNAL'})
class FavoriteTextureClickOperator(FavoriteTexture.OpEditMixin):
    "Click: apply, Shift+Click: edit, Ctrl+Click: discard"
    
    def load_image(self, preset):
        from bpy_extras.image_utils import load_image
        
        filepath = CommonUtils.abspath(preset.path)
        if not os.path.isfile(filepath): return None
        
        image = load_image(filepath, check_existing=True, force_reload=False)
        if not image: return None
        
        if max(image.size) <= 0: # e.g. corrupted file
            if image.users == 0: bpy.data.images.remove(image)
            return None
        
        return image
    
    def invoke(self, context, event):
        presets = FavoriteTexture.presets
        preset = presets.get(self.id)
        if not preset: return {'CANCELLED'}
        
        if event.shift:
            self.old_id = self.id
            self.path = preset.path
            wm = context.window_manager
            return wm.invoke_props_dialog(self, width=300)
        
        if event.ctrl:
            presets.delete(preset)
        else:
            node = self.get_image_node(context)
            if node:
                image = self.load_image(preset)
                if image:
                    node.image = image
                    bpy.ops.ed.undo_push(message="Apply favorite texture")
        
        BlUI.tag_redraw()
        
        return {'FINISHED'}

@addon.Menu(idname="SCENE_MT_xdecal_favorite_textures", label="Favorite textures", description="")
class FavoriteTexturesMenu:
    def draw(self, context):
        self.draw_body(self.layout, context)
    
    @classmethod
    def draw_body(cls, layout, context):
        metadata = BlUI.get_ui_metadata(context)
        category_prefix = metadata.get("category_prefix") or ""
        prefix_len = len(category_prefix)
        
        folders = []
        entries = []
        presets = FavoriteTexture.presets
        for preset in presets:
            if not preset.id.startswith(category_prefix): continue
            parts = preset.id[prefix_len:].split("/", maxsplit=1)
            if len(parts) > 1:
                if parts[0] not in folders:
                    folders.append(parts[0])
            else:
                entries.append((parts[0], preset))
        
        folders.sort()
        for folder in folders:
            BlUI.set_ui_metadata(layout, {"category_prefix": f"{category_prefix}{folder}/"})
            layout.menu(cls.bl_idname, text=folder, icon='FILE_FOLDER')
        BlUI.set_ui_metadata(layout, {"category_prefix": ""})
        
        entries.sort()
        for name, preset in entries:
            exists = os.path.isfile(preset.path)
            icon = ('IMAGE_DATA' if exists else 'LIBRARY_DATA_BROKEN')
            op = layout.operator("xdecal.favorite_texture_click", text=name, icon=icon)
            op.id = preset.id

@addon.context_menu('APPEND')
def decal_popup_context_menu_draw(self, context):
    layout = NestedLayout(self.layout)
    
    # Works both for the decal popup and for the shader editor
    node = getattr(context, "node", None)
    
    if isinstance(node, bpy.types.ShaderNodeTexImage):
        layout.separator()
        
        op = layout.operator(FavoriteTexture.op_add, text="<Add to favorites>", icon='ADD')
        op.id = ""
        op.path = ""
        
        FavoriteTexturesMenu.draw_body(layout, context)
