# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import os
import sys
import math
import time
import random

import numpy as np

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color
from mathutils.geometry import tessellate_polygon, closest_point_on_tri
from gpu_extras.presets import draw_texture_2d

import gpu

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "utils_math": "orthogonal_XYZ, matrix_compose, projection_matrix",
    "bpy_inspect": "BlRna, BpyData, prop, bpy_file_filters, IDTypes",
    "utils_gl": "cgl",
    "utils_ui": "BlUI",
    "utils_view3d": "SmartView3D",
    "utils_blender": "BlUtil, Raycaster",
    "utils_addon": "AddonManager",
})

from . import common
from . import material_utils

from .common import Jobs, CommonUtils
from .material_utils import MaterialLibrary, MaterialUtils, MaterialLoader, MaterialMixer, NodeInfo

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

class Decal:
    geo_nodes_library_path = MaterialLibrary.default
    
    geo_node_name_main = "XDecal.Geo"
    geo_node_name_join = "XDecal.Geo.Join"
    geo_node_name_extrude = "XDecal.Geo.Extrude"
    mat_node_name_uvw = "XDecal.UVW"
    
    geo_attr_name_target = "XDecal.IsTarget"
    geo_attr_name_embed_uv = "XDecal.UVMap"
    geo_attr_name_mat_id = "XDecal.SrcMatIndex"
    
    default_uv_name = "UVMap"
    
    decal_collection_name = ".Decals"
    
    geo_nodes_usable_types = {'MESH', 'CURVE', 'SURFACE', 'META', 'FONT', 'CURVES', 'POINTCLOUD', 'VOLUME'}
    
    def __new__(cls, obj):
        if not obj: return None
        
        if isinstance(obj, cls): return obj
        
        md_main = cls._get_geo_main_modifier(obj)
        if not md_main: return None
        
        md_extrude = cls._get_geo_extrude_modifier(obj)
        
        self = object.__new__(cls)
        self.obj = obj
        self.md_main = md_main
        self.md_extrude = md_extrude
        self._target_key = None
        self.input_map = {}
        
        if md_extrude:
            tree_interface = BlUtil.NodeTree.get_interface(md_extrude.node_group)
            for item in tree_interface.items_tree:
                self.input_map["Extrude/"+item.name] = (md_extrude, item.identifier)
        
        tree_interface = BlUtil.NodeTree.get_interface(md_main.node_group)
        for item in tree_interface.items_tree:
            self.input_map["Main/"+item.name] = (md_main, item.identifier)
        
        return self
    
    def _get_target_key(self, node_group):
        if not self._target_key:
            tree_interface = BlUtil.NodeTree.get_interface(node_group)
            for item in tree_interface.items_tree:
                if item.name != "Object": continue
                self._target_key = item.identifier
        return self._target_key
    
    def __getitem__(self, key):
        md, idname = self.input_map[key]
        return md[idname]
    
    def __setitem__(self, key, value):
        md, idname = self.input_map[key]
        md[idname] = value
    
    @classmethod
    def _should_calculate_uv(cls, material):
        if material.node_tree:
            for node in material.node_tree.nodes:
                if node.bl_idname == "ShaderNodeGroup":
                    if not node.node_tree: continue
                    if node.node_tree.name != cls.mat_node_name_uvw: continue
                    return False
                elif node.bl_idname == "ShaderNodeAttribute":
                    if node.attribute_type != 'GEOMETRY': continue
                    if not node.attribute_name.startswith("XDecal."): continue
                    return False
        return True
    
    @classmethod
    def _init_custom_props_from_material(cls, obj, mat):
        if not mat.node_tree: return
        
        props_map = {}
        attribute_types = ('OBJECT', 'INSTANCER')
        def process(node):
            if node.bl_idname != "ShaderNodeAttribute": return
            if node.attribute_type not in attribute_types: return
            props_map[node.attribute_name] = tuple(node.outputs["Color"].default_value)
        
        node_trees = set()
        def traverse(node_tree):
            if node_tree in node_trees: return
            node_trees.add(node_tree)
            for node in node_tree.nodes:
                process(node)
                sub_tree = getattr(node, "node_tree", None)
                if sub_tree: traverse(sub_tree)
        
        traverse(mat.node_tree)
        
        for name, value in props_map.items():
            if name in obj: continue
            obj[name] = value
    
    def get_template(self):
        template = self["Main/Object"]
        if template: return template
        materials = self.obj.data.materials
        return (materials[0] if materials else None)
    
    def set_template(self, template, image_size='AUTO'):
        if not self.can_use_as_template(template): return
        
        materials = self.obj.data.materials
        
        if isinstance(template, bpy.types.Material):
            if not isinstance(self.get_template(), bpy.types.Material):
                self.obj.visible_shadow = False
            
            if materials:
                materials[0] = template
            else:
                materials.append(template)
            
            if isinstance(image_size, str):
                if image_size == 'AUTO':
                    image_size = MaterialUtils.estimate_image_size(template)
                elif image_size == 'KEEP':
                    image_size = None
            
            if image_size is not None:
                size_x = max(image_size[0], 1)
                size_y = max(image_size[1], 1)
                max_size = max(size_x, size_y)
                
                scale_x = self.range_x / min(self.aspect, 1.0)
                scale_y = self.range_y * max(self.aspect, 1.0)
                
                self.range_x = size_x/max_size * scale_x
                self.range_y = size_y/max_size * scale_y
                self.aspect = size_x/size_y
            
            self.use_uv_map = self._should_calculate_uv(template)
            
            self["Main/Object"] = None
            
            self._init_custom_props_from_material(self.obj, template)
        elif isinstance(template, bpy.types.Object):
            if not isinstance(self.get_template(), bpy.types.Object):
                self.obj.visible_shadow = True
            
            if materials:
                materials[0] = None
            
            self.range_x = 1.0
            self.range_y = 1.0
            self.aspect = 1.0
            
            self.use_uv_map = False # just for consistency
            
            self["Main/Object"] = template
    
    template = property(get_template, set_template)
    
    def get_targets(self, sort_parent=False):
        targets = []
        
        for md in self.obj.modifiers:
            if (md.type != 'NODES') or (not md.node_group): continue
            if md.node_group.name != self.geo_node_name_join: continue
            target_key = self._get_target_key(md.node_group)
            if not target_key: continue
            target = md[target_key]
            if target: targets.append(target)
        
        if sort_parent:
            try:
                i = targets.index(self.obj.parent)
                targets[0], targets[i] = targets[i], targets[0]
            except Exception:
                pass
        
        return targets
    
    def set_targets(self, targets):
        geo_group_join = self._get_decal_geo_group(self.geo_node_name_join)
        if not geo_group_join: return
        
        target_key = self._get_target_key(geo_group_join)
        if not target_key: return
        
        unique_targets = []
        for target in targets:
            if (not target) or (target in unique_targets): continue
            if not BlRna.is_valid(target): continue
            if Decal.exists(target): continue
            unique_targets.append(target)
        
        modifiers = self.obj.modifiers
        
        join_modifiers = []
        for md in modifiers:
            if (md.type != 'NODES') or (not md.node_group): continue
            if md.node_group.name != self.geo_node_name_join: continue
            join_modifiers.append(md)
        
        targets_count = len(unique_targets)
        md_count = max(targets_count, 1)
        
        while len(join_modifiers) < md_count:
            md = modifiers.new("AddTarget", 'NODES')
            md.node_group = geo_group_join
            modifiers.move(len(modifiers)-1, 0)
            join_modifiers.append(md)
        
        while len(join_modifiers) > md_count:
            modifiers.remove(join_modifiers.pop())
        
        for i, md in enumerate(join_modifiers):
            target = (unique_targets[i] if i < targets_count else None)
            md[target_key] = target
        
        if settings.placement_options.sync_collections:
            self.sync_collections()
    
    targets = property(get_targets, set_targets)
    
    def get_dynamic(self):
        return self.md_main.show_viewport
    
    def set_dynamic(self, value):
        template = self.template
        is_material = isinstance(template, bpy.types.Material)
        
        MaterialMixer.clear_custom_props(self.obj)
        
        depsgraph = None
        armature_data = None
        
        if value:
            mesh = self.make_empty_mesh()
            
            mat = (template if is_material else None)
            if mat: mesh.materials.append(mat)
        else:
            # Do this before evaluating the depsgraph
            armature_data = self.setup_armature_data()
            
            depsgraph = bpy.context.evaluated_depsgraph_get()
            
            obj_eval = self.obj.evaluated_get(depsgraph)
            
            mesh = bpy.data.meshes.new_from_object(obj_eval)
            
            target_attr = mesh.attributes.get(self.geo_attr_name_target)
            if target_attr: mesh.attributes.remove(target_attr)
            
            if is_material:
                use_project = self.use_project
                use_embed = self.use_embed and (self.geo_attr_name_mat_id in mesh.attributes)
                targets = (self.get_targets(sort_parent=True) if use_embed and use_project else ())
                self.activate_default_uv_layer(mesh, targets)
                
                if not use_project:
                    while len(mesh.materials) > 1:
                        mesh.materials.pop()
                elif use_embed:
                    mixed_materials = {}
                    uv_name = self.geo_attr_name_embed_uv
                    for i, mat in enumerate(mesh.materials):
                        if mat == template: continue
                        mat_mix = mixed_materials.get(mat)
                        if not mat_mix:
                            mat_mix = MaterialUtils.mix(mat, template, uv_overlay=uv_name)
                            mixed_materials[mat] = mat_mix
                        mesh.materials[i] = mat_mix
                    
                    mat_ids_attr = mesh.attributes[self.geo_attr_name_mat_id]
                    for item, face in zip(mat_ids_attr.data, mesh.polygons):
                        face.material_index = item.value
                    
                    prop_infos = MaterialMixer.gather_custom_props(mixed_materials.keys())
                    MaterialMixer.copy_custom_props(self.targets, self.obj, prop_infos)
        
        old_mesh = self.obj.data
        
        self.obj.data = mesh
        
        if old_mesh.users == 0:
            mesh.name = old_mesh.name
            
            if is_material:
                while len(old_mesh.materials) > 1:
                    mat = old_mesh.materials.pop()
                    if mat and (mat.users == 0):
                        bpy.data.materials.remove(mat)
            
            bpy.data.meshes.remove(old_mesh)
        
        # Use show_expanded to store whether the modifier was enabled
        for md in self.obj.modifiers:
            if not value:
                md.show_expanded = md.show_viewport
                md.show_viewport = False
                md.show_render = False
            else:
                md.show_viewport = md.show_expanded
                md.show_render = md.show_expanded
        
        # Do this after storing/modifying the enabled/disabled state
        self.clear_armature_data()
        if armature_data: self.copy_armature_data(armature_data, depsgraph)
        
        self.obj.update_tag()
    
    @staticmethod
    def find_armature_modifier(obj):
        is_last = True
        for md in reversed(obj.modifiers):
            if md.type != 'ARMATURE':
                is_last &= not md.show_viewport
                continue
            obj = md.object
            if (not obj) or (obj.type != 'ARMATURE'): continue
            return md, is_last
        return None, None
    
    def clear_armature_data(self):
        for md in list(self.obj.modifiers):
            if md.type != 'ARMATURE': continue
            self.obj.modifiers.remove(md)
    
    def setup_armature_data(self):
        if not self.use_project: return None
        
        obj = self.obj
        parent = obj.parent
        if not parent: return None
        
        md_src, is_last = self.find_armature_modifier(parent)
        if not md_src: return None
        
        armature_data = {
            "source": md_src,
            "is_last": is_last,
            "use_uv_map": self.use_uv_map,
            "use_intersect": self.use_intersect,
        }
        
        self.use_uv_map = True
        self.use_intersect = True
        self.obj.update_tag()
        
        return armature_data
    
    def copy_armature_data(self, armature_data, depsgraph):
        # The mesh is baked, so now we can restore the original settings
        self.use_uv_map = armature_data["use_uv_map"]
        self.use_intersect = armature_data["use_intersect"]
        
        obj = self.obj
        parent = obj.parent
        scene = bpy.context.scene
        md_src = armature_data["source"]
        is_last = armature_data["is_last"]
        
        # We need to transfer data anyway, because any extra vertices
        # (e.g. from boolean intersections) and object-based decals
        # don't have any bone weights associated with them.
        
        obj.vertex_groups.clear()
        for name in sorted(g.name for g in parent.vertex_groups):
            vg = obj.vertex_groups.new(name=name)
        
        md_transfer = obj.modifiers.new("DataTransfer", 'DATA_TRANSFER')
        md_transfer.object = parent
        md_transfer.use_vert_data = True
        md_transfer.data_types_verts = {'VGROUP_WEIGHTS'}
        md_transfer.vert_mapping = 'POLYINTERP_NEAREST'
        obj.modifiers.move(len(obj.modifiers)-1, 0)
        
        # Applying the transferred weights is important for correct behavior
        with bpy.context.temp_override(object=obj, active_object=obj):
            bpy.ops.object.modifier_apply(modifier=md_transfer.name)
        
        if is_last and (parent.type == 'MESH') and md_src.show_viewport:
            self._calc_undeformed_positions(md_src, depsgraph)
        
        md_dst = obj.modifiers.new("Armature", 'ARMATURE')
        md_dst.show_viewport = md_src.show_viewport
        md_dst.show_render = md_src.show_render
        md_dst.invert_vertex_group = md_src.invert_vertex_group
        md_dst.use_bone_envelopes = md_src.use_bone_envelopes
        md_dst.use_deform_preserve_volume = md_src.use_deform_preserve_volume
        md_dst.use_multi_modifier = md_src.use_multi_modifier
        md_dst.use_vertex_groups = md_src.use_vertex_groups
        md_dst.vertex_group = md_src.vertex_group
        md_dst.object = md_src.object
    
    def _calc_undeformed_positions(self, md_src, depsgraph):
        obj = self.obj
        mesh = obj.data
        matrix = obj.matrix_world
        matrix_inv = matrix.inverted_safe()
        
        parent = obj.parent
        
        parent_post = parent.evaluated_get(depsgraph)
        mesh_post = parent_post.data
        verts_post = mesh_post.vertices
        faces_post = mesh_post.polygons
        verts_post_len = len(verts_post)
        faces_post_len = len(faces_post)
        matrix_post = parent_post.matrix_world.inverted_safe() @ matrix
        
        tesselations = {}
        
        def get_tri_matrix(v0, v1, v2, invert):
            axis_x = v2 - v0
            axis_y = v1 - v0
            axis_z = axis_y.cross(axis_x)
            axis_z.normalize()
            m = Matrix((axis_x, axis_y, axis_z))
            m.transpose()
            if invert: m.invert_safe()
            return m
        
        def get_tri_info(pos, index):
            tesselation = tesselations.get(index)
            if tesselation is None:
                indices = list(faces_post[index].vertices)
                verts = [verts_post[vi].co for vi in indices]
                tris = tessellate_polygon([verts])
                tesselations[index] = (verts, tris, indices)
            else:
                verts, tris, indices = tesselation
            
            best_dist = math.inf
            best_tri = None
            for tri in tris:
                i0, i1, i2 = tri
                v0 = verts[i0]
                v1 = verts[i1]
                v2 = verts[i2]
                p = closest_point_on_tri(pos, v0, v1, v2)
                dist = (p - pos).length_squared
                if dist < best_dist:
                    best_dist = dist
                    best_tri = tri
            
            i0, i1, i2 = best_tri
            v0 = verts[i0]
            v1 = verts[i1]
            v2 = verts[i2]
            p = get_tri_matrix(v0, v1, v2, True) @ (pos - v0)
            return (indices[i0], indices[i1], indices[i2], p)
        
        local_positions = [None] * len(mesh.vertices)
        
        for vi, v in enumerate(mesh.vertices):
            pos = matrix_post @ v.co
            success, _, _, index = parent_post.closest_point_on_mesh(pos)
            if not success: continue
            local_positions[vi] = get_tri_info(pos, index)
        
        md_src.show_viewport = False
        parent.update_tag()
        
        depsgraph = bpy.context.evaluated_depsgraph_get()
        parent_pre = parent.evaluated_get(depsgraph)
        mesh_pre = parent_pre.data
        verts_pre = mesh_pre.vertices
        faces_pre = mesh_pre.polygons
        verts_pre_len = len(verts_pre)
        faces_pre_len = len(faces_pre)
        matrix_pre = matrix_inv @ parent.matrix_world
        
        md_src.show_viewport = True
        parent.update_tag()
        
        if len(verts_pre) != verts_post_len: return
        if len(faces_pre) != faces_post_len: return
        
        for v, info in zip(mesh.vertices, local_positions):
            if not info: continue
            i0, i1, i2, p = info
            v0 = verts_pre[i0].co
            v1 = verts_pre[i1].co
            v2 = verts_pre[i2].co
            v.co = matrix_pre @ ((get_tri_matrix(v0, v1, v2, False) @ p) + v0)
    
    is_dynamic = property(get_dynamic, set_dynamic)
    
    def _md_property(name, default=None, getter=None, setter=None):
        def _get(self):
            try:
                value = self[name]
            except KeyError:
                return default
            return (getter(value) if getter else value)
        
        def _set(self, value):
            if setter: value = setter(value)
            try:
                self[name] = value
            except KeyError:
                return
        
        return property(_get, _set)
    
    mask_name = _md_property("Main/Mask Name")
    mask_min = _md_property("Main/Mask Min")
    mask_max = _md_property("Main/Mask Max")
    use_world_scale = _md_property("Main/World Scale")
    aspect = _md_property("Main/Aspect")
    subdivs = _md_property("Main/Subdivs")
    curv_min = _md_property("Main/CurvMinor")
    curv_maj = _md_property("Main/CurvMajor")
    angle = _md_property("Main/Angle")
    range_x = _md_property("Main/RangeX")
    range_y = _md_property("Main/RangeY")
    depth_min = _md_property("Main/DepthMin")
    depth_max = _md_property("Main/DepthMax")
    displacement = _md_property("Main/Clearance")
    use_project = _md_property("Main/Project")
    use_clip = _md_property("Main/Clip")
    use_cull = _md_property("Main/Cull")
    use_occlude = _md_property("Main/Occlude")
    use_intersect = _md_property("Main/Intersect")
    extra_location = _md_property("Main/Location", getter=Vector)
    extra_rotation = _md_property("Main/Rotation", getter=Euler)
    extra_scale = _md_property("Main/Scale", getter=Vector)
    use_uv_map = _md_property("Main/UV map")
    use_embed = _md_property("Main/Embed")
    
    extrude_layers = _md_property("Extrude/Layers", 0)
    extrude_amount = _md_property("Extrude/Distance", 0.0)
    
    def _property_vector_component(name, i):
        def _get(self):
            return getattr(self, name)[i]
        def _set(self, value):
            vector = getattr(self, name)
            vector[i] = value
            setattr(self, name, vector)
        return property(_get, _set)
    
    extra_location_x = _property_vector_component("extra_location", 0)
    extra_location_y = _property_vector_component("extra_location", 1)
    extra_location_z = _property_vector_component("extra_location", 2)
    
    extra_rotation_x = _property_vector_component("extra_rotation", 0)
    extra_rotation_y = _property_vector_component("extra_rotation", 1)
    extra_rotation_z = _property_vector_component("extra_rotation", 2)
    
    extra_scale_x = _property_vector_component("extra_scale", 0)
    extra_scale_y = _property_vector_component("extra_scale", 1)
    extra_scale_z = _property_vector_component("extra_scale", 2)
    
    def _get(self):
        return Vector((self.range_x, self.range_y, self.depth_min, self.depth_max))
    def _set(self, value):
        self.range_x = value[0]
        self.range_y = value[1]
        self.depth_min = value[2]
        self.depth_max = value[3]
    frustum = property(_get, _set)
    
    def _get(self):
        return Vector((self.curv_min, self.curv_maj))
    def _set(self, value):
        self.curv_min = value[0]
        self.curv_maj = value[1]
    curvature = property(_get, _set)
    
    def update(self):
        if not self.is_dynamic:
            self.is_dynamic = True
            self.is_dynamic = False
        self.obj.update_tag()
    
    def sync_collections(self):
        decals_name = self.decal_collection_name
        targets = set(self.targets)
        active_collection = bpy.context.collection
        
        def sync(coll, use=False):
            use |= (coll.name_full == decals_name) or not targets.isdisjoint(coll.objects)
            
            try:
                if use:
                    coll.objects.link(self.obj)
                else:
                    coll.objects.unlink(self.obj)
            except Exception:
                pass # already linked/unlinked
        
        for coll in bpy.data.collections:
            sync(coll)
        
        sync(active_collection, not targets)
    
    def embed(self, value=True):
        self.use_embed = value
        
        if not self.use_embed:
            self.update()
            return
        
        if not self.use_project:
            self.is_dynamic = True
            self.use_project = True
        
        if self.is_dynamic:
            self.is_dynamic = False
        else:
            self.update()
    
    def auto_depth(self):
        depth_min, depth_max = self.estimate_depth()
        displacement = self.displacement
        self.depth_min = depth_min + min(displacement, 0)
        self.depth_max = depth_max + max(displacement, 0)
    
    def calc_geometry_params(self):
        params = {}
        
        range_x = self.range_x
        range_y = self.range_y
        angle = self.angle
        
        depth_min = self.depth_min
        depth_max = self.depth_max
        
        cos_a = abs(math.cos(angle))
        sin_a = abs(math.sin(angle))
        range_x, range_y = (
            range_x*cos_a + range_y*sin_a,
            range_y*cos_a + range_x*sin_a,
        )
        
        threshold = 0.0001
        curv_x = self.curv_min
        curv_y = self.curv_maj
        
        if (abs(curv_x) > threshold) or (abs(curv_y) > threshold):
            rxs = (1 if curv_x < 0 else -1)
            rys = (1 if curv_y < 0 else -1)
            rxa = 1.0 / max(abs(curv_x), threshold)
            rya = 1.0 / max(abs(curv_y), threshold)
            
            rx = rxs * rxa
            ry = rys * rya
            
            ry -= rx
            rys = (-1 if ry < 0 else 1)
            rya = max(abs(ry), threshold)
            ry = rys * rya
            
            is_lemon = (rxs*rys < 0) and (rxa > rya)
            
            pi = math.pi
            pi2 = pi*2
            
            rw = rx + ry
            rwa = max(abs(rw), 0.000001)
            
            limit = 2*math.acos(min(rya, rxa) / rxa)
            if not is_lemon: limit = pi2 - limit
            limit = min(limit, pi2)
            
            sector_x = min(max(range_x / rxa, 0.0), limit)
            sector_y = min(max(range_y / rwa, 0.0), pi2)
            
            range_x = sector_x * rxa
            range_y = sector_y * rwa
            
            params["torus"] = True
            params["lemon"] = is_lemon
            params["radius_min"] = rx
            params["radius_maj"] = ry
            params["radius_sum"] = rw
            params["sector_min"] = sector_x
            params["sector_maj"] = sector_y
        
        params["angle"] = angle
        params["range_min"] = range_x
        params["range_maj"] = range_y
        params["depth_min"] = depth_min
        params["depth_max"] = depth_max
        
        return params
    
    def transform_points(self, points, depths, use_rotate=True, params=None):
        if params is None: params = self.calc_geometry_params()
        
        points = np.asarray(points)
        
        count = points.shape[0]
        
        depth_min = -math.inf
        depth_max = math.inf
        
        def rotate(x, y, sin_a, cos_a):
            return (x*cos_a - y*sin_a, x*sin_a + y*cos_a)
        
        def rotate_around(x, y, sin_a, cos_a, cx, cy):
            x, y = rotate(x - cx, y - cy, sin_a, cos_a)
            x += cx
            y += cy
            return (x, y)
        
        angle = params["angle"]
        
        is_torus = bool(params.get("torus"))
        
        if is_torus:
            rmin = params["radius_min"]
            rsum = params["radius_sum"]
            
            is_lemon = bool(params.get("lemon"))
            clamp_axis_positive = (is_lemon == (rmin > 0.0))
            
            if rmin < 0: depth_min = rmin
            if rmin > 0: depth_max = rmin
            
            phi_min = points[:, 0] / abs(rmin)
            phi_maj = points[:, 1] / abs(rsum)
            
            x, y, z = np.zeros(count), np.zeros(count), np.zeros(count)
            nx, ny, nz = np.zeros(count), np.zeros(count), np.ones(count)
            
            phi_min_sin = np.sin(phi_min)
            phi_min_cos = np.cos(phi_min)
            nx, nz = rotate(nx, nz, phi_min_sin, phi_min_cos)
            x, z = rotate_around(x, z, phi_min_sin, phi_min_cos, 0, rmin)
            
            phi_maj_sin = np.sin(phi_maj)
            phi_maj_cos = np.cos(phi_maj)
            ny, nz = rotate(ny, nz, phi_maj_sin, phi_maj_cos)
            y, z = rotate_around(y, z, phi_maj_sin, phi_maj_cos, 0, rsum)
        else:
            x = points[:, 0]
            y = points[:, 1]
            z = np.zeros(count)
            nx, ny, nz = np.zeros(count), np.zeros(count), np.ones(count)
        
        results = []
        
        for depth in depths:
            depth = np.clip(depth, depth_min, depth_max)
            
            x1 = x + nx*depth
            y1 = y + ny*depth
            z1 = z + nz*depth
            
            if is_torus:
                cz = z - rsum
                cz1 = z1 - rsum
                
                yz0_mag = np.sqrt(y*y + cz*cz)
                yz1_mag = np.sqrt(y1*y1 + cz1*cz1)
                
                m_mask = (yz0_mag <= 0.000001)
                y[m_mask] = 0.0
                cz[m_mask] = 0.0
                
                if clamp_axis_positive:
                    d_mask = (depth < 0.0)
                else:
                    d_mask = (depth > 0.0)
                
                r_mask = ((y*y1 + cz*cz1) >= 0.0)
                
                mask = (m_mask & d_mask) | ((~m_mask) & r_mask)
                
                t1 = yz0_mag / np.maximum(yz0_mag + yz1_mag, sys.float_info.min)
                t1[mask] = 1.0
                t0 = 1.0 - t1
                x1 = x*t0 + x1*t1
                y1 = y*t0 + y1*t1
                z1 = z*t0 + z1*t1
            
            if use_rotate:
                angle_sin = np.sin(angle)
                angle_cos = np.cos(angle)
                x1, y1 = rotate(x1, y1, angle_sin, angle_cos)
            
            results.append(list(zip(x1, y1, z1)))
        
        return results
    
    def calc_volume_verts(self, resolution=25, use_rotate=True, depth_min=None, depth_max=None):
        params = self.calc_geometry_params()
        range_x = params["range_min"]
        range_y = params["range_maj"]
        sector_x = params.get("sector_min", 0.0)
        sector_y = params.get("sector_maj", 0.0)
        
        if depth_min is None: depth_min = params["depth_min"]
        if depth_max is None: depth_max = params["depth_max"]
        
        detail_power = 0.35
        detail_x = (sector_x/(math.pi*2)) ** detail_power
        detail_y = (sector_y/(math.pi*2)) ** detail_power
        
        half_resolution = resolution // 2
        
        points = []
        depths = [depth_min, depth_max]
        count_x = max(math.ceil(half_resolution * detail_x), 1)*2 + 1
        count_y = max(math.ceil(half_resolution * detail_y), 1)*2 + 1
        for iy in range(count_y):
            for ix in range(count_x):
                u = ix/(count_x-1)
                v = iy/(count_y-1)
                points.append(((u-0.5)*range_x, (v-0.5)*range_y, 0))
        
        layers = self.transform_points(points, depths, use_rotate=use_rotate, params=params)
        
        return layers, count_x, count_y
    
    def estimate_depth(self, resolution=25):
        epsilon = 0.001
        layers, count_x, count_y = self.calc_volume_verts(
            depth_min=-epsilon, depth_max=epsilon, resolution=resolution)
        
        positions = []
        normals = []
        layer_min, layer_max = layers
        for i in range(len(layer_min)):
            p_min = Vector(layer_min[i])
            p_max = Vector(layer_max[i])
            positions.append(p_min.lerp(p_max, 0.5))
            normals.append((p_max - p_min).normalized())
        
        obj = self.obj
        raycaster = Raycaster(objs=self.targets, matrix=obj.matrix_world)
        
        depth_min = -0.001
        depth_max = 0.001
        
        for p, n in zip(positions, normals):
            result = raycaster.closest_point(p)
            if not result.success: continue # or actually break/return?
            
            delta = result.location - p
            dist = delta.magnitude * 2
            
            is_above = (n.dot(delta) > 0)
            if not is_above: n = -n
            
            result = raycaster.ray_cast(p, n * dist)
            if result.success:
                delta = result.location - p
                dist = max(dist, delta.magnitude * 1.5)
            
            if is_above:
                depth_max = max(depth_max, dist)
            else:
                depth_min = min(depth_min, -dist)
        
        return depth_min, depth_max
    
    @classmethod
    def _get_decal_geo_group(cls, group_name):
        return CommonUtils.load_node_group(group_name, cls.geo_nodes_library_path, link=True)
    
    @classmethod
    def _get_geo_main_modifier(cls, obj):
        if obj.type != 'MESH': return None
        for md in reversed(obj.modifiers):
            if md.type != 'NODES': continue
            node_group = md.node_group
            if (not node_group) or node_group.is_missing: continue
            if node_group.name == cls.geo_node_name_main: return md
        return None
    
    @classmethod
    def _get_geo_extrude_modifier(cls, obj):
        if obj.type != 'MESH': return None
        for md in reversed(obj.modifiers):
            if (md.type != 'NODES') or (not md.node_group): continue
            if md.node_group.name == cls.geo_node_name_extrude: return md
        return None
    
    @classmethod
    def exists(cls, obj):
        return bool(cls._get_geo_main_modifier(obj))
    
    @classmethod
    def can_use_as_template(cls, template):
        if isinstance(template, bpy.types.Material): return True
        if not isinstance(template, bpy.types.Object): return False
        if Decal.exists(template): return False
        return template.type in cls.geo_nodes_usable_types
    
    @classmethod
    def can_use_as_target(cls, obj):
        return (obj.type in cls.geo_nodes_usable_types) and not Decal.exists(obj)
    
    @classmethod
    def activate_default_uv_layer(cls, mesh, targets=()):
        def find_active_uvmap():
            for target in targets:
                if target.type != 'MESH': continue
                for uv_layer in target.data.uv_layers:
                    if uv_layer.active_render: return uv_layer.name
            return cls.default_uv_name
        active_uvmap = find_active_uvmap()
        
        best_uv_layer = None
        for uv_layer in mesh.uv_layers:
            if uv_layer.name == active_uvmap:
                best_uv_layer = uv_layer
                break
            elif uv_layer.name != cls.geo_attr_name_embed_uv:
                best_uv_layer = uv_layer
        
        if best_uv_layer is None: return
        
        mesh.uv_layers.active = best_uv_layer
        best_uv_layer.active = True
        best_uv_layer.active_clone = True
        best_uv_layer.active_render = True
    
    @classmethod
    def make_empty_mesh(cls, name="Mesh"):
        mesh = bpy.data.meshes.new(name)
        
        # need at least some geometry to propagate materials
        mesh.vertices.add(1)
        
        # Seems like geometry nodes can't create uv layers, only write to existing ones
        mesh.uv_layers.new(name=cls.default_uv_name)
        mesh.uv_layers.new(name=cls.geo_attr_name_embed_uv)
        
        cls.activate_default_uv_layer(mesh)
        
        return mesh
    
    @classmethod
    def instantiate(cls, template, *, context=None, collection=None, image_size='AUTO'):
        if not context: context = bpy.context
        
        geo_group_main = cls._get_decal_geo_group(cls.geo_node_name_main)
        if not geo_group_main: return None
        
        geo_group_join = cls._get_decal_geo_group(cls.geo_node_name_join)
        if not geo_group_join: return None
        
        geo_group_extrude = cls._get_decal_geo_group(cls.geo_node_name_extrude)
        
        if collection is None: collection = context.collection
        
        decal_collection = bpy.data.collections.get(cls.decal_collection_name)
        if decal_collection is None:
            decal_collection = bpy.data.collections.new(cls.decal_collection_name)
        decal_collection.use_fake_user = True
        
        mesh_name = template.name
        if not mesh_name.startswith("."): mesh_name = "."+mesh_name
        
        mesh = cls.make_empty_mesh(mesh_name)
        
        obj = bpy.data.objects.new(template.name, mesh)
        collection.objects.link(obj)
        decal_collection.objects.link(obj)
        
        # just to keep referenced
        md_join = obj.modifiers.new("AddTarget", 'NODES')
        md_join.node_group = geo_group_join
        
        md_main = obj.modifiers.new("Decal", 'NODES')
        md_main.node_group = geo_group_main
        
        # With default parameters, the extrude geo group
        # does nothing, so we can just add it by default
        # Note: add AFTER the main geo modifier
        if geo_group_extrude:
            md_extrude = obj.modifiers.new("Extrude", 'NODES')
            md_extrude.node_group = geo_group_extrude
        
        decal = Decal(obj)
        decal.set_template(template, image_size=image_size)
        decal.displacement = settings.placement_options.displacement_step
        decal.use_intersect = settings.default_intersect
        
        if not settings.default_dynamic: decal.is_dynamic = False
        
        return obj

class DecalUtils:
    @classmethod
    def load_images_as_decals(cls, paths, material, categories, context=None, **kwargs):
        if context is None: context = bpy.context
        
        material_infos = MaterialUtils.load_images_to_materials(paths, material, categories, **kwargs)
        
        objs = []
        
        for material, image_size in material_infos:
            obj = Decal.instantiate(material, image_size=image_size, context=context)
            objs.append(obj)
        
        #context.view_layer.update()
        
        return objs
    
    @classmethod
    def load_libraries_as_decals(cls, paths, **kwargs):
        objs = []
        
        allowed_exts = set(bpy_file_filters["blender"])
        
        def remove_data(datas, data, force):
            if not force:
                users = data.users - int(data.use_fake_user)
                if users > 1: return
            datas.remove(data)
        
        def add_library(path):
            relative = path.startswith("//")
            
            with bpy.data.libraries.load(path, link=True, relative=relative) as (data_from, data_to):
                data_to.objects = data_from.objects
                data_to.materials = data_from.materials
            
            for obj in data_to.objects:
                if not CommonUtils.is_template(obj):
                    remove_data(bpy.data.objects, obj, True)
                    continue
                
                obj = Decal.instantiate(obj, **kwargs)
                objs.append(obj)
            
            for material in data_to.materials:
                if not CommonUtils.is_template(material):
                    remove_data(bpy.data.materials, material, False)
                    continue
                
                obj = Decal.instantiate(material, **kwargs)
                objs.append(obj)
        
        for path in paths:
            ext = os.path.splitext(path)[1].lower()
            if ext not in allowed_exts: continue
            
            try:
                add_library(path)
            except OSError as exc:
                # In some cases Blender may fail to open the file
                # Print so user may have some idea why loading failed
                print(f"Failed to load decal library {path}")
        
        #context.view_layer.update()
        
        return objs
    
    _inheritable_geometry_props = [
        "mask_name",
        "mask_min",
        "mask_max",
        "use_world_scale",
        "subdivs",
        "curv_min",
        "curv_maj",
        "angle",
        "depth_min",
        "depth_max",
        "displacement",
        "use_project",
        "use_clip",
        "use_cull",
        "use_occlude",
        "use_intersect",
        "extra_location",
        "extra_rotation",
        "extra_scale",
        "use_uv_map",
        "use_embed",
        "extrude_layers",
        "extrude_amount",
        "targets",
    ]
    
    @classmethod
    def copy_properties(cls, source, targets, options=None):
        if options is None:
            options = (settings.inherit_options if settings.use_inherit else None)
        
        if not options: return
        
        source = Decal(source)
        if not source: return
        
        use_scale = 'SCALE' in options
        use_geometry = 'GEOMETRY' in options
        use_material = 'MATERIAL' in options
        use_clipboard = 'CLIPBOARD' in options
        
        scale_x = source.range_x / min(source.aspect, 1.0)
        scale_y = source.range_y * max(source.aspect, 1.0)
        
        source_clipboard = None
        if use_material and isinstance(source.template, bpy.types.Material):
            source_clipboard = NodeInfo.mat_props_copy(source.template)
        
        clipboard = (MaterialUtils.clipboard if use_clipboard else None)
        
        material_map = {}
        
        for target in targets:
            target = Decal(target)
            if not target: continue
            
            if use_scale:
                target.obj.scale = Vector(source.obj.scale)
            
            if use_geometry:
                for prop_name in cls._inheritable_geometry_props:
                    setattr(target, prop_name, getattr(source, prop_name))
                
                target.range_x = scale_x * min(target.aspect, 1.0)
                target.range_y = scale_y / max(target.aspect, 1.0)
                
                target.is_dynamic = source.is_dynamic
            
            if isinstance(target.template, bpy.types.Material):
                material_map.setdefault(target.template, []).append(target)
        
        need_preview_update = []
        
        for template, template_decals in material_map.items():
            if (template.users - int(template.use_fake_user)) > 1:
                template = template.copy()
                need_preview_update.append(template)
                for decal in template_decals:
                    decal.template = template
            
            NodeInfo.mat_props_paste(template, source_clipboard)
            NodeInfo.mat_props_paste(template, clipboard)
        
        cls.render_preview(need_preview_update)
    
    @classmethod
    def arrange(cls, objs, matrix, spacing=1.0, use_template_scale=True):
        max_size = Vector()
        for obj in objs:
            size = obj.dimensions
            max_size.x = max(max_size.x, size.x)
            max_size.y = max(max_size.y, size.y)
        spacing = max_size * spacing
        
        side_count = math.ceil(math.sqrt(len(objs)))
        offset = (side_count-1) * 0.5
        
        for i, obj in enumerate(objs):
            y, x = divmod(i, side_count)
            pos = Vector((x-offset, y-offset, 0)) * spacing
            scale_matrix = Matrix.Diagonal((obj.scale * obj.delta_scale).to_4d())
            obj.matrix_world = matrix @ Matrix.Translation(pos) @ scale_matrix
    
    @classmethod
    def setup_snap_settings(cls, tool_settings, pivot='BOUNDING_BOX_CENTER'):
        # CURSOR, BOUNDING_BOX_CENTER, MEDIAN_POINT, ACTIVE_ELEMENT, INDIVIDUAL_ORIGINS
        snap_target_map = {'BOUNDING_BOX_CENTER': 'CENTER', 'MEDIAN_POINT': 'MEDIAN', 'ACTIVE_ELEMENT': 'ACTIVE'}
        
        #tool_settings.use_snap = True
        tool_settings.transform_pivot_point = pivot
        tool_settings.use_snap_align_rotation = True
        tool_settings.snap_target = snap_target_map.get(pivot, 'MEDIAN')
        
        if hasattr(tool_settings, "use_snap_project"): # Blender 3.6
            tool_settings.snap_elements = {'FACE'}
            tool_settings.use_snap_project = (pivot == 'INDIVIDUAL_ORIGINS')
        else: # Blender 4.0 and above
            tool_settings.snap_elements = ({'FACE'} if pivot != 'INDIVIDUAL_ORIGINS' else {'FACE_PROJECT'})
    
    @classmethod
    def find_all_decals(cls):
        # We ignore objects not contained in scenes because in most cases they
        # simply were deleted, and we don't want them to affect the behavior
        objs = set()
        for scene in bpy.data.scenes:
            for obj in scene.objects:
                if obj in objs: continue
                objs.add(obj)
                decal = Decal(obj)
                if decal: yield decal
    
    @classmethod
    def find_all_templates(cls, implicit=True, explicit=True):
        templates = {}
        
        overridden = set()
        
        def add(template, increment):
            if not template: return
            templates[template] = templates.get(template, 0) + increment
            
            override_library = template.override_library
            if not override_library: return
            
            ref_idblock = override_library.reference
            if ref_idblock: overridden.add(ref_idblock)
        
        if explicit:
            for obj in bpy.data.objects:
                if CommonUtils.is_template(obj): add(obj, 0)
            
            for mat in bpy.data.materials:
                if CommonUtils.is_template(mat): add(mat, 0)
        
        if implicit:
            for decal in cls.find_all_decals():
                add(decal.template, 1)
        
        # If both raw linked blocks and their overrides are present,
        # ignore the raw linked blocks if they have 0 actual users
        for idblock in overridden:
            count = templates.get(idblock)
            if count is None: continue
            if count == 0: templates.pop(idblock)
        
        return templates
    
    @classmethod
    def find_scene_templates(cls, scene=None):
        if scene is None: scene = bpy.context.scene
        scene_templates = set()
        for obj in scene.objects:
            decal = Decal(obj)
            if not decal: continue
            template = decal.template
            if template: scene_templates.add(template)
        return scene_templates
    
    @classmethod
    def count_template_users(cls):
        template_counts = {}
        for decal in cls.find_all_decals():
            template = decal.template
            template_counts[template] = template_counts.get(template, 0) + 1
        return template_counts
    
    @classmethod
    def filter_templates(cls, objs, types=None, editable=()):
        if types is None: types = ('OBJECT', 'MATERIAL')
        
        type_map = IDTypes.map("id_type")
        if isinstance(types, str):
            types = (type_map[types].type,)
        else:
            types = tuple(type_map[id_type].type for id_type in types)
        
        check_editable_object = 'OBJECT' in editable
        check_editable_template = 'TEMPLATE' in editable
        
        template_objs = {}
        for obj in objs:
            if check_editable_object and obj.library: continue
            
            decal = Decal(obj)
            if not decal: continue
            
            template = decal.template
            if not (template and isinstance(template, types)): continue
            if check_editable_template and template.library: continue
            
            template_objs.setdefault(template, []).append(obj)
        
        return template_objs
    
    @classmethod
    def filter_decals(cls, objs, add_children=False, result=None, child_map=None):
        if result is None: result = set()
        
        add = (result.append if isinstance(result, list) else result.add)
        
        if add_children and (child_map is None):
            child_map, parent_map = BlUtil.Object.map_children(objs)
        
        for obj in objs:
            if not Decal.exists(obj): continue
            add(obj)
            if add_children: BlUtil.Object.all_children(obj, result, child_map)
        
        return result
    
    @classmethod
    def filter_instances(cls, objs, template):
        for obj in objs:
            decal = Decal(obj)
            if decal and (decal.template == template): yield obj
    
    @classmethod
    def update(cls, objs, geometry=True, collections=False):
        if not objs: return False
        
        if isinstance(objs, bpy.types.Object): objs = [objs]
        
        was_updated = False
        
        for obj in objs:
            decal = Decal(obj)
            if not decal: continue
            if geometry: decal.update()
            if collections: decal.sync_collections()
            was_updated = True
        
        return was_updated
    
    @classmethod
    def replace_images(cls, objs, paths):
        kwargs = settings.loading_options.kwargs(paths=paths)
        
        template_counts = cls.count_template_users()
        
        template_objs = cls.filter_templates(objs, types='MATERIAL', editable='OBJECT')
        
        templates = []
        for template, objs in template_objs.items():
            if template.library or (template_counts[template] > len(objs)):
                material = template.copy()
                CommonUtils.mark_as_template(material)
            else:
                material = template
            
            material_loader = MaterialLoader(material=material, **kwargs)
            material_loader.load_override()
            material, image_size = material_loader.material_infos[0]
            
            for obj in objs:
                decal = Decal(obj)
                decal.set_template(material, image_size=image_size)
                decal.update()
            
            templates.append(material)
        
        cls.render_preview(templates)
    
    @classmethod
    def replace_material(cls, objs, prototype):
        if not isinstance(prototype, bpy.types.Material): return
        
        template_counts = cls.count_template_users()
        
        template_objs = cls.filter_templates(objs, types='MATERIAL', editable='OBJECT')
        
        def get_base_name(name):
            index = name.rfind("(")
            return (name[:index] if index >= 0 else name).strip()
        
        def get_proto_name(name):
            index = name.rfind("(")
            if index < 0: return name.strip()
            name = name[index+1:]
            index = name.find(")")
            if index < 0: return name.strip()
            name = name[:index]
            return name.strip()
        
        prototype_name = get_proto_name(prototype.name)
        
        templates = []
        for template, objs in template_objs.items():
            material = prototype.copy()
            CommonUtils.mark_as_template(material)
            
            clipboard = NodeInfo.mat_props_copy(template)
            NodeInfo.mat_props_paste(material, clipboard, all_images=True)
            
            material.name = f"{get_base_name(template.name)} ({prototype_name})"
            
            # It's probably better not to delete the original, since
            # the user might still need to copy some stuff manually
            # if template_counts[template] <= len(objs):
            #     template.user_remap(material)
            #     bpy.data.materials.remove(template)
            #     continue
            
            for obj in objs:
                decal = Decal(obj)
                decal.set_template(material, image_size='KEEP')
                decal.update()
            
            templates.append(material)
        
        cls.render_preview(templates)
    
    @classmethod
    def convert_to_decal(cls, obj, depsgraph=None, eval_objs=None, prototype=None, context=None):
        info = None
        if obj.type == 'MESH':
            info = cls.mesh_to_decal_info(obj, depsgraph, eval_objs)
        elif obj.type == 'EMPTY':
            info = cls.empty_to_decal_info(obj, prototype)
        
        if not info: return None
        
        template = info["material"]
        
        new_obj = Decal.instantiate(template, context=context)
        new_obj.matrix_world = info["matrix"]
        
        decal = Decal(new_obj)
        
        if info["frustum"]:
            decal.frustum = info["frustum"]
            decal.depth_min = min(decal.depth_min, -0.01)
            decal.depth_max = max(decal.depth_max, 0.01)
            
            aspect = decal.aspect
            scale = new_obj.scale
            if aspect > 1:
                decal.range_y = decal.range_x / aspect
                scale.y *= aspect
            else:
                decal.range_x = decal.range_y * aspect
                scale.x /= aspect
            # If scale is not uniform, things like parallax won't work correctly
            scale = (scale.x + scale.y) * 0.5
            new_obj.scale = Vector((scale, scale, scale))
            
            decal.use_intersect = True
        
        decal.targets = info["targets"]
        
        decal.update()
        
        return decal.obj
    
    @classmethod
    def empty_to_decal_info(cls, obj, prototype):
        if not prototype: return None
        
        if obj.type != 'EMPTY': return None
        if obj.empty_display_type != 'IMAGE': return None
        
        image = obj.data
        if not image: return None
        
        if obj.use_empty_image_alpha:
            alpha = obj.color[3]
        else:
            alpha = 1.0
        
        size = obj.empty_display_size
        offset_x = obj.empty_image_offset[0] + 0.5
        offset_y = obj.empty_image_offset[1] + 0.5
        
        matrix = Matrix.Translation((offset_x, offset_y, 0))
        matrix = obj.matrix_world @ Matrix.Scale(size, 4) @ matrix
        
        material = prototype.copy()
        CommonUtils.mark_as_template(material)
        
        clipboard = NodeInfo.image_to_clipboard(image, alpha, image_user=obj.image_user)
        NodeInfo.mat_props_paste(material, clipboard, all_images=True)
        
        material_name = f"{image.name} ({prototype.name})"
        old_material = bpy.data.materials.get(material_name)
        
        material.name = material_name
        if old_material:
            old_material.user_remap(material)
            bpy.data.materials.remove(old_material)
        
        result = dict(
            material = material,
            matrix = matrix,
            frustum = None,
            targets = [],
        )
        
        return result
    
    @classmethod
    def mesh_to_decal_info(cls, obj, depsgraph=None, eval_objs=None):
        if obj.type != 'MESH': return None
        
        material = next((ms.material for ms in obj.material_slots if ms.material), None)
        if not material: return None
        
        if depsgraph is None: depsgraph = bpy.context.evaluated_depsgraph_get()
        
        obj_eval = obj.evaluated_get(depsgraph)
        src_matrix = obj_eval.matrix_world
        
        targets = cls._find_shrinkwrap_targets(obj) or []
        
        mesh = obj_eval.to_mesh(preserve_all_data_layers=True, depsgraph=depsgraph)
        verts_eval = np.array([src_matrix @ v.co for v in mesh.vertices])
        # If an object has Shrinkwrap, we assume that the local mesh
        # is a grid and therefore much more likely to have linear UVs
        if targets:
            verts = np.array([src_matrix @ v.co for v in obj.data.vertices])
            uvs, uvs_use = cls._get_vertex_uvs(obj.data)
        else:
            verts = verts_eval
            uvs, uvs_use = cls._get_vertex_uvs(mesh)
        obj_eval.to_mesh_clear()
        
        if (uvs is None) or (uvs.shape[0] < 3): return None
        
        verts_eval = verts_eval[uvs_use]
        verts = verts[uvs_use]
        uvs = uvs[uvs_use]
        
        verts_mean = np.mean(verts, axis=0)
        uvs_mean = np.mean(uvs, axis=0)
        
        verts_centered = verts - verts_mean
        uvs_centered = uvs - uvs_mean
        
        dot = np.dot
        # It's easier to find the gradient directions first, then calculate the scale
        dU = np.zeros(3)
        dV = np.zeros(3)
        for vert, uv in zip(verts_centered, uvs_centered):
            vert = vert / dot(vert, vert)
            dU += vert * uv[0]
            dV += vert * uv[1]
        
        dU /= np.linalg.norm(dU)
        dV /= np.linalg.norm(dV)
        # Linear regression without intercept
        u_xy, u_xx = 0.0, 0.0
        v_xy, v_xx = 0.0, 0.0
        for vert, uv in zip(verts_centered, uvs_centered):
            pU = dot(vert, dU)
            u_xy += pU * uv[0]
            u_xx += pU * pU
            pV = dot(vert, dV)
            v_xy += pV * uv[1]
            v_xx += pV * pV
        sU = u_xy / u_xx
        sV = v_xy / v_xx
        X = dU / sU
        Y = dV / sV
        
        matrix = Matrix()
        matrix.col[0][:3] = X
        matrix.col[1][:3] = Y
        matrix.col[2][:3] = Vector(np.cross(X, Y)).normalized()
        du = (uvs_mean[0] % 1.0) - 0.5
        dv = (uvs_mean[1] % 1.0) - 0.5
        matrix.translation = verts_mean + X*du + Y*dv
        
        matrix_inv = matrix.inverted_safe()
        
        verts_local = np.array([matrix_inv @ Vector(v) for v in verts])
        bmin, bmax = verts_local.min(axis=0), verts_local.max(axis=0)
        range_x = 2*max(abs(bmin[0]), abs(bmax[0]))
        range_y = 2*max(abs(bmin[1]), abs(bmax[1]))
        
        if targets:
            # A local mesh is likely flat, so we can't estimate depth from it
            verts_local = np.array([matrix_inv @ Vector(v) for v in verts_eval])
            bmin, bmax = verts_local.min(axis=0), verts_local.max(axis=0)
        
        depth_min, depth_max = bmin[2], bmax[2]
        
        if (not targets) and ((depth_max - depth_min) > max(range_x, range_y) * 0.001):
            pos = Vector(verts_eval[0])
            
            best_dist = math.inf
            best_target = None
            for src_obj, eval_obj in eval_objs.items():
                if src_obj == obj: continue
                eval_matrix = eval_obj.matrix_world
                matrix_local = eval_matrix.inverted_safe() @ matrix
                origin = matrix_local.translation
                size_x = matrix_local.col[0].to_3d().magnitude * range_x
                size_y = matrix_local.col[1].to_3d().magnitude * range_y
                size_z = matrix_local.col[2].to_3d().magnitude * (depth_max - depth_min)
                distance = Vector((size_x, size_y, size_z)).magnitude
                result = eval_obj.closest_point_on_mesh(origin, distance=distance, depsgraph=depsgraph)
                if not result[0]: continue
                dist = ((eval_matrix @ result[1]) - pos).magnitude
                if dist < best_dist:
                    best_target, best_dist = src_obj, dist
            targets = ([best_target] if best_target else [])
        
        result = dict(
            material = material,
            matrix = matrix,
            frustum = (range_x, range_y, depth_min, depth_max),
            targets = targets,
        )
        
        return result
    
    @classmethod
    def _find_shrinkwrap_targets(cls, obj):
        for md in reversed(obj.modifiers):
            if not md.show_viewport: continue
            if md.type != 'SHRINKWRAP': continue
            targets = {md.target, md.auxiliary_target}
            targets.discard(None)
            if targets: return targets
    
    @classmethod
    def _get_vertex_uvs(cls, mesh):
        uv_layer = next(iter(mesh.uv_layers), None)
        if not uv_layer: return None, None
        
        uv_ranges = [[0, 0, 0] for i in range(len(mesh.vertices))]
        for loop, uv_item in zip(mesh.loops, uv_layer.uv):
            u, v = uv_item.vector
            uv_range = uv_ranges[loop.vertex_index]
            uv_range[0] += u
            uv_range[1] += v
            uv_range[2] += 1
        
        uvs_use = [(item[-1] > 0) for item in uv_ranges]
        uvs = np.array(uv_ranges)
        uvs = uvs[:, :2] / np.maximum(uvs[:, 2:], 1)
        
        return uvs, uvs_use
    
    @classmethod
    def render_preview(cls, templates):
        if isinstance(templates, bpy.types.ID): templates = [templates]
        
        main_size = settings.preview_render_size
        
        # It seems that GPUOffscreen.draw_view3d() already works without the need for
        # bpy.ops.wm.redraw_timer(), so long as a matching SpaceView3D and a region
        # of 'WINDOW' type are supplied as arguments. Since we no longer care about
        # redraw_timer crashing if it's ivoked more than once per frame, we can
        # potentially render as many previews per frame as we need.
        with CustomPreviewRenderer(main_size=main_size) as preview_renderer:
            for template in templates:
                if (not template) or template.library: continue
                preview_renderer.render(template)
        
        Jobs.palette_needs_update = True

class CustomPreviewRenderer:
    # Note: we can't simply disable overlays entirely, because
    # wireframe objects currently only render in overlay mode
    view3d_config = {
        "shading": {"type": 'MATERIAL'},
        "overlay": {p.identifier: (p.identifier == "show_overlays")
            for p in bpy.types.View3DOverlay.bl_rna.properties
            if (p.type == 'BOOLEAN') and (not p.is_readonly)
        },
    }
    
    object_config = {
        "show_name": False,
        "show_axis": False,
        "show_wire": False,
        "show_texture_space": False,
        "show_shadows": False,
        "show_in_front": False,
        "show_bounds": False,
        "display_type": 'TEXTURED',
    }
    
    def __init__(self, context=None, main_size=128, icon_size=32, margin=0.01):
        self.context = context or bpy.context
        
        self.main_size = main_size
        self.icon_size = icon_size
        self.margin = margin
        
        self.view3d, self.region = self.find_3d_viewport(self.context)
        
        self.scene = self.context.scene
        self.view_layer = None
        
        self.decal = None
    
    def __enter__(self):
        self.view3d_config_prev = self.reconfigure(self.view3d, self.view3d_config)
        
        self.view_layer = self.scene.view_layers.new("PreviewRenderingLayer")
        
        self.objs_in_scene = set(BlRna.iter_valid(self.view_layer.objects))
        
        for obj in self.objs_in_scene:
            obj.hide_set(True, view_layer=self.view_layer)
        
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.reconfigure(self.view3d, self.view3d_config_prev)
        
        self.scene.view_layers.remove(self.view_layer)
        
        if self.decal is not None:
            decal_mesh = self.decal.obj.data
            bpy.data.objects.remove(self.decal.obj)
            bpy.data.meshes.remove(decal_mesh)
    
    def render(self, target):
        if not BlRna.is_valid(target): return
        
        obj = self.get_object(target)
        if not obj: return
        
        object_config_prev = self.reconfigure(obj, self.object_config)
        
        depsgraph = self.context.evaluated_depsgraph_get()
        
        if obj in self.objs_in_scene:
            obj.hide_set(False, view_layer=self.view_layer)
        else:
            self.scene.collection.objects.link(obj)
        
        self.view_layer.update()
        
        if isinstance(target, bpy.types.Material):
            view_proj_info = self.calc_view_proj_material(target)
        else:
            view_proj_info = self.calc_view_proj_object(depsgraph, obj, self.margin)
        
        if view_proj_info: self.render_preview(target, **view_proj_info)
        
        self.reconfigure(obj, object_config_prev)
        
        if obj in self.objs_in_scene:
            obj.hide_set(True, view_layer=self.view_layer)
        else:
            self.scene.collection.objects.unlink(obj)
    
    def get_object(self, target):
        if isinstance(target, bpy.types.Object): return target
        
        if not isinstance(target, bpy.types.Material): return None
        
        if not self.decal:
            obj = Decal.instantiate(target)
            self.decal = Decal(obj)
            
            # A trick to hide the decal instead of linking
            # (since it's already linked, Blender would raise an error)
            self.objs_in_scene.add(obj)
        else:
            obj = self.decal.obj
            
            # Clear any previous custom attributes
            for key in list(obj.keys()):
                obj.pop(key)
            
            self.decal.set_template(target)
        
        self.decal.range_x = 1.0
        self.decal.range_y = 1.0
        self.decal.aspect = 1.0
        
        return obj
    
    def render_preview(self, target, scale, view_matrix, proj_matrix):
        main_offscreen = self.make_offscreen(self.main_size * scale.x, self.main_size * scale.y)
        
        icon_offscreen = self.make_offscreen(self.icon_size * scale.x, self.icon_size * scale.y)
        
        main_offscreen.draw_view3d(self.scene, self.view_layer, self.view3d, self.region,
            view_matrix, proj_matrix, do_color_management=True, draw_background=False)
        cgl.depth_test = 'NONE'
        cgl.blend = 'NONE'
        
        with main_offscreen.bind():
            main_buffer = self.read_color(main_offscreen)
        
        with icon_offscreen.bind():
            self.blit_texture(main_offscreen.texture_color)
            icon_buffer = self.read_color(icon_offscreen)
        
        preview = target.preview_ensure()
        preview.image_size = (main_offscreen.width, main_offscreen.height)
        preview.image_pixels[:] = main_buffer
        preview.icon_size = (icon_offscreen.width, icon_offscreen.height)
        preview.icon_pixels[:] = icon_buffer
        
        main_offscreen.free()
        icon_offscreen.free()
    
    @classmethod
    def find_3d_viewport(cls, context):
        space_data = getattr(context, "space_data", None)
        region = getattr(context, "region", None)
        
        if not (space_data or region): return None, None
        
        if not isinstance(space_data, bpy.types.SpaceView3D):
            area_info = BlUI.find_ui_area(area_type='VIEW_3D', region_type='WINDOW')
            return ((area_info["space_data"], area_info["region"]) if area_info else (None, None))
        
        if region.type != 'WINDOW':
            window, area, region = BlUI.ui_hierarchy(context.region)
            region = next(region for region in area.regions if region.type == 'WINDOW')
        
        return space_data, region
    
    @classmethod
    def reconfigure(cls, target, config):
        prev_config = {}
        for key, value in config.items():
            if not hasattr(target, key): continue
            prev_value = getattr(target, key)
            if isinstance(value, dict):
                prev_value = cls.reconfigure(prev_value, value)
            else:
                setattr(target, key, value)
            prev_config[key] = prev_value
        return prev_config
    
    @classmethod
    def calc_view_proj_material(cls, material):
        # Invert Z here too, so that normals (and lighting) would match the projection
        view_matrix = Matrix.Diagonal((1, 1, -1, 1))
        
        # Invert Z to render from decal's front-side (due to Blender's camera conventions)
        proj_matrix = cls.calc_proj_matrix((-0.5, -0.5, 0.5), (0.5, 0.5, -0.5))
        
        xy_size = Vector(MaterialUtils.estimate_image_size(material))
        scale = xy_size / max(xy_size.x, xy_size.y)
        
        return dict(scale=scale, view_matrix=view_matrix, proj_matrix=proj_matrix)
    
    @classmethod
    def calc_view_proj_object(cls, depsgraph, obj, margin):
        obj_eval = obj.evaluated_get(depsgraph)
        
        matrix_world_inv = obj_eval.matrix_world.inverted_safe()
        
        bbox = cls.calc_bbox_info(depsgraph, obj_eval, matrix_world_inv)
        if not bbox: return None
        
        mesh = obj_eval.to_mesh(preserve_all_data_layers=False, depsgraph=depsgraph)
        view_dir = cls.calc_view_direction(mesh, bbox["size"])
        obj_eval.to_mesh_clear()
        
        view_matrix = matrix_compose(*orthogonal_XYZ(None, None, view_dir)).to_4x4()
        view_matrix = matrix_world_inv @ view_matrix.inverted_safe()
        
        bbox = cls.calc_bbox_info(depsgraph, obj_eval, view_matrix)
        if not bbox: return None
        
        proj_matrix = cls.calc_proj_matrix(bbox["min"], bbox["max"], bbox["size"] * margin)
        
        xy_size = Vector((bbox["size"].x, bbox["size"].y)) * (1 + 2 * margin)
        scale = xy_size / max(xy_size.x, xy_size.y)
        
        return dict(scale=scale, view_matrix=view_matrix, proj_matrix=proj_matrix)
    
    @classmethod
    def calc_bbox_info(cls, depsgraph, obj, matrix, min_ratio=0.001):
        bbox = BlUtil.Depsgraph.bounding_box(depsgraph, obj, matrix, origins='NONE', use_bbox=True)
        if not bbox.is_defined: return None
        
        bbox_min, bbox_max, size = bbox.get("min", "max", "size", convert=Vector)
        size_max = max(size)
        if size_max == 0: return None
        
        size = Vector([max(v, size_max*min_ratio) for v in size])
        if min(size) == 0: return None # essentially zero size
        
        return dict(min=bbox_min, max=bbox_max, size=size)
    
    @classmethod
    def calc_view_direction(cls, mesh, size):
        ax = Vector((size.x, 0, 0))
        ay = Vector((0, size.y, 0))
        az = Vector((0, 0, size.z))
        view_dir_abs = (ax-az).cross(ay-az).normalized()
        view_dir_abs.x = abs(view_dir_abs.x)
        view_dir_abs.y = abs(view_dir_abs.y)
        view_dir_abs.z = abs(view_dir_abs.z)
        
        # Decals are always viewed "from above"
        view_dirs = [view_dir_abs * Vector((sx, sy, 1)) for sy in (-1, 1) for sx in (-1, 1)]
        
        correlations = [0.0] * len(view_dirs)
        correlations_abs = [0.0] * len(view_dirs)
        
        for polygon in mesh.polygons:
            scaled_normal = polygon.normal * polygon.area
            for i, view_dir in enumerate(view_dirs):
                correlation = view_dir.dot(scaled_normal)
                correlations[i] += correlation
                correlations_abs[i] += abs(correlation)
        
        best_view_dir = None
        best_correlation_abs = 0.0
        best_correlation = -math.inf
        for view_dir, correlation, correlation_abs in zip(view_dirs, correlations, correlations_abs):
            if correlation_abs < best_correlation_abs: continue
            if correlation_abs == best_correlation_abs:
                if correlation < best_correlation: continue
            best_view_dir = view_dir
            best_correlation_abs = correlation_abs
            best_correlation = correlation
        
        return best_view_dir
    
    @classmethod
    def calc_proj_matrix(cls, bmin, bmax, margins=None):
        bmin, bmax = Vector(bmin), Vector(bmax)
        if margins is not None: bmin, bmax = (bmin - margins, bmax + margins)
        return projection_matrix(bmin.x, bmax.x, bmin.y, bmax.y, -bmax.z, -bmin.z, perspective=False)
    
    @classmethod
    def blit_texture(cls, texture):
        with gpu.matrix.push_pop_projection():
            gpu.matrix.load_projection_matrix(Matrix())
            with gpu.matrix.push_pop():
                gpu.matrix.load_matrix(Matrix())
                draw_texture_2d(texture, (-1, -1), 2, 2)
    
    @classmethod
    def make_offscreen(cls, size_x, size_y):
        return gpu.types.GPUOffScreen(int(size_x), int(size_y))
    
    @classmethod
    def read_color(cls, offscreen):
        data = cgl.read_color((0, 0), (offscreen.width, offscreen.height), format='UBYTE')
        return np.asarray(data, dtype='B').view('i')
