# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import os
import math

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import gpu
from gpu_extras.batch import batch_for_shader

# =========================================================================== #

# Note: Blender's icons reference is at release/datafiles/blender_icons.svg
# https://developer.blender.org/diffusion/B/browse/master/release/datafiles

palette_icons_images = {}

def load_icon_image(name):
    pcoll = bpy.utils.previews.new()
    
    icons_dir = os.path.join(os.path.dirname(__file__), "icons")
    pcoll.load(name, os.path.join(icons_dir, f"{name}.png"), 'IMAGE')
    
    preview = pcoll[name]
    size = tuple(preview.image_size) # make sure it's a copy
    
    # If image does not exist, a preview is still creates, but its size is zero
    buffer = None
    if size[0] and size[1]:
        buffer = gpu.types.Buffer('INT', size[0]*size[1], preview.image_pixels)
    
    bpy.utils.previews.remove(pcoll)
    
    if buffer is None: return None
    return size, buffer

def get_palette_icon_image(name):
    image = palette_icons_images.get(name, None)
    
    if not image:
        image = load_icon_image(name)
        if image: palette_icons_images[name] = image
    
    return image

gizmo_icons = dict(
    rotate = (
        (+0.500, +0.850, +0.000), (+0.950, +0.700, +0.000), (+0.500, +1.000, +0.000),
        (+0.279, +0.841, +0.000), (+0.950, +0.700, +0.000), (+0.500, +0.850, +0.000),
        (+0.079, +0.814, +0.000), (+0.950, +0.700, +0.000), (+0.279, +0.841, +0.000),
        (-0.101, +0.769, +0.000), (+0.950, +0.700, +0.000), (+0.079, +0.814, +0.000),
        (-0.261, +0.705, +0.000), (+0.950, +0.700, +0.000), (-0.101, +0.769, +0.000),
        (-0.401, +0.623, +0.000), (+0.950, +0.700, +0.000), (-0.261, +0.705, +0.000),
        (-0.401, +0.623, +0.000), (+0.500, +0.550, +0.000), (+0.950, +0.700, +0.000),
        (+0.500, +0.550, +0.000), (+0.500, +0.400, +0.000), (+0.950, +0.700, +0.000),
        (-0.522, +0.522, +0.000), (+0.500, +0.550, +0.000), (-0.401, +0.623, +0.000),
        (-0.522, +0.522, +0.000), (+0.331, +0.540, +0.000), (+0.500, +0.550, +0.000),
        (-0.522, +0.522, +0.000), (+0.177, +0.517, +0.000), (+0.331, +0.540, +0.000),
        (-0.623, +0.401, +0.000), (+0.177, +0.517, +0.000), (-0.522, +0.522, +0.000),
        (-0.623, +0.401, +0.000), (+0.039, +0.481, +0.000), (+0.177, +0.517, +0.000),
        (-0.623, +0.401, +0.000), (-0.085, +0.432, +0.000), (+0.039, +0.481, +0.000),
        (-0.623, +0.401, +0.000), (-0.194, +0.368, +0.000), (-0.085, +0.432, +0.000),
        (-0.705, +0.261, +0.000), (-0.194, +0.368, +0.000), (-0.623, +0.401, +0.000),
        (-0.705, +0.261, +0.000), (-0.287, +0.290, +0.000), (-0.194, +0.368, +0.000),
        (-0.705, +0.261, +0.000), (-0.367, +0.197, +0.000), (-0.287, +0.290, +0.000),
        (-0.769, +0.101, +0.000), (-0.367, +0.197, +0.000), (-0.705, +0.261, +0.000),
        (-0.769, +0.101, +0.000), (-0.431, +0.090, +0.000), (-0.367, +0.197, +0.000),
        (-0.814, -0.079, +0.000), (-0.431, +0.090, +0.000), (-0.769, +0.101, +0.000),
        (-0.814, -0.079, +0.000), (-0.482, -0.034, +0.000), (-0.431, +0.090, +0.000),
        (-0.814, -0.079, +0.000), (-0.519, -0.173, +0.000), (-0.482, -0.034, +0.000),
        (-0.841, -0.279, +0.000), (-0.519, -0.173, +0.000), (-0.814, -0.079, +0.000),
        (-0.841, -0.279, +0.000), (-0.541, -0.328, +0.000), (-0.519, -0.173, +0.000),
        (-0.850, -0.500, +0.000), (-0.541, -0.328, +0.000), (-0.841, -0.279, +0.000),
        (-0.850, -0.500, +0.000), (-0.550, -0.500, +0.000), (-0.541, -0.328, +0.000),
        (-0.700, -0.950, +0.000), (-0.850, -0.500, +0.000), (-1.000, -0.500, +0.000),
        (-0.700, -0.950, +0.000), (-0.550, -0.500, +0.000), (-0.850, -0.500, +0.000),
        (-0.700, -0.950, +0.000), (-0.400, -0.500, +0.000), (-0.550, -0.500, +0.000),
    ),
    scale = (
        (-1.000, -1.000, +0.000), (-0.700, +0.700, +0.000), (-1.000, +1.000, +0.000),
        (-0.700, +0.700, +0.000), (+1.000, +1.000, +0.000), (-1.000, +1.000, +0.000),
        (-0.700, +0.700, +0.000), (+1.000, +0.700, +0.000), (+1.000, +1.000, +0.000),
        (-1.000, -1.000, +0.000), (-0.700, -1.000, +0.000), (-0.700, +0.700, +0.000),
    ),
    flip = (
        (+0.000, -0.500, +0.000), (+0.600, +0.400, +0.000), (-0.600, +0.400, +0.000),
    ),
    depth = (
        (+0.000, +0.700, +0.035), (+0.000, +0.000, +1.200), (+0.268, +0.647, +0.035),
        (+0.268, +0.647, +0.035), (+0.000, +0.000, +1.200), (+0.495, +0.495, +0.035),
        (+0.495, +0.495, +0.035), (+0.000, +0.000, +1.200), (+0.647, +0.268, +0.035),
        (+0.647, +0.268, +0.035), (+0.000, +0.000, +1.200), (+0.700, -0.000, +0.035),
        (+0.700, -0.000, +0.035), (+0.000, +0.000, +1.200), (+0.647, -0.268, +0.035),
        (+0.647, -0.268, +0.035), (+0.000, +0.000, +1.200), (+0.495, -0.495, +0.035),
        (+0.495, -0.495, +0.035), (+0.000, +0.000, +1.200), (+0.268, -0.647, +0.035),
        (+0.268, -0.647, +0.035), (+0.000, +0.000, +1.200), (+0.000, -0.700, +0.035),
        (+0.000, -0.700, +0.035), (+0.000, +0.000, +1.200), (-0.268, -0.647, +0.035),
        (-0.268, -0.647, +0.035), (+0.000, +0.000, +1.200), (-0.495, -0.495, +0.035),
        (-0.495, -0.495, +0.035), (+0.000, +0.000, +1.200), (-0.647, -0.268, +0.035),
        (-0.647, -0.268, +0.035), (+0.000, +0.000, +1.200), (-0.700, +0.000, +0.035),
        (-0.700, +0.000, +0.035), (+0.000, +0.000, +1.200), (-0.647, +0.268, +0.035),
        (-0.647, +0.268, +0.035), (+0.000, +0.000, +1.200), (-0.495, +0.495, +0.035),
        (-0.495, +0.495, +0.035), (+0.000, +0.000, +1.200), (-0.268, +0.647, +0.035),
        (-0.268, +0.647, +0.035), (+0.000, +0.000, +1.200), (+0.000, +0.700, +0.035),
        (-0.268, +0.647, +0.035), (+0.000, +0.700, +0.035), (+0.268, +0.647, +0.035),
        (+0.268, +0.647, +0.035), (+0.495, +0.495, +0.035), (+0.647, +0.268, +0.035),
        (+0.647, +0.268, +0.035), (+0.700, -0.000, +0.035), (+0.647, -0.268, +0.035),
        (+0.647, -0.268, +0.035), (+0.495, -0.495, +0.035), (+0.268, -0.647, +0.035),
        (+0.268, -0.647, +0.035), (+0.000, -0.700, +0.035), (-0.268, -0.647, +0.035),
        (-0.268, -0.647, +0.035), (-0.495, -0.495, +0.035), (-0.647, -0.268, +0.035),
        (-0.647, -0.268, +0.035), (-0.700, +0.000, +0.035), (-0.647, +0.268, +0.035),
        (-0.647, +0.268, +0.035), (-0.495, +0.495, +0.035), (-0.268, +0.647, +0.035),
        (-0.268, +0.647, +0.035), (+0.268, +0.647, +0.035), (+0.647, +0.268, +0.035),
        (+0.647, +0.268, +0.035), (+0.647, -0.268, +0.035), (+0.268, -0.647, +0.035),
        (+0.268, -0.647, +0.035), (-0.268, -0.647, +0.035), (-0.647, -0.268, +0.035),
        (-0.647, -0.268, +0.035), (-0.647, +0.268, +0.035), (-0.268, +0.647, +0.035),
        (-0.268, +0.647, +0.035), (+0.647, +0.268, +0.035), (+0.268, -0.647, +0.035),
        (+0.268, -0.647, +0.035), (-0.647, -0.268, +0.035), (-0.268, +0.647, +0.035),
    ),
    resolution = (
        (-0.500, +1.000, +0.000), (-0.300, +0.500, +0.000), (-0.300, +1.000, +0.000),
        (+0.300, +0.500, +0.000), (+0.500, +1.000, +0.000), (+0.300, +1.000, +0.000),
        (-0.500, +0.500, +0.000), (-0.300, +0.500, +0.000), (-0.500, +1.000, +0.000),
        (+0.300, +0.500, +0.000), (+0.500, +0.500, +0.000), (+0.500, +1.000, +0.000),
        (-1.000, +0.300, +0.000), (-0.500, +0.500, +0.000), (-1.000, +0.500, +0.000),
        (-1.000, +0.300, +0.000), (-0.300, +0.500, +0.000), (-0.500, +0.500, +0.000),
        (-1.000, +0.300, +0.000), (+0.300, +0.500, +0.000), (-0.300, +0.500, +0.000),
        (-1.000, +0.300, +0.000), (+0.500, +0.500, +0.000), (+0.300, +0.500, +0.000),
        (-1.000, +0.300, +0.000), (+1.000, +0.500, +0.000), (+0.500, +0.500, +0.000),
        (-1.000, +0.300, +0.000), (+1.000, +0.300, +0.000), (+1.000, +0.500, +0.000),
        (-0.500, +0.300, +0.000), (+1.000, +0.300, +0.000), (-1.000, +0.300, +0.000),
        (-0.500, -0.300, +0.000), (-0.300, +0.300, +0.000), (-0.500, +0.300, +0.000),
        (-0.300, +0.300, +0.000), (+1.000, +0.300, +0.000), (-0.500, +0.300, +0.000),
        (-0.500, -0.300, +0.000), (-0.300, -0.300, +0.000), (-0.300, +0.300, +0.000),
        (+0.300, +0.300, +0.000), (+1.000, +0.300, +0.000), (-0.300, +0.300, +0.000),
        (+0.300, -0.300, +0.000), (+0.500, +0.300, +0.000), (+0.300, +0.300, +0.000),
        (+0.500, +0.300, +0.000), (+1.000, +0.300, +0.000), (+0.300, +0.300, +0.000),
        (+0.300, -0.300, +0.000), (+0.500, -0.300, +0.000), (+0.500, +0.300, +0.000),
        (-1.000, -0.500, +0.000), (-0.500, -0.300, +0.000), (-1.000, -0.300, +0.000),
        (-1.000, -0.500, +0.000), (-0.300, -0.300, +0.000), (-0.500, -0.300, +0.000),
        (-1.000, -0.500, +0.000), (+0.300, -0.300, +0.000), (-0.300, -0.300, +0.000),
        (-1.000, -0.500, +0.000), (+0.500, -0.300, +0.000), (+0.300, -0.300, +0.000),
        (-1.000, -0.500, +0.000), (+1.000, -0.300, +0.000), (+0.500, -0.300, +0.000),
        (-1.000, -0.500, +0.000), (+1.000, -0.500, +0.000), (+1.000, -0.300, +0.000),
        (-0.500, -0.500, +0.000), (+1.000, -0.500, +0.000), (-1.000, -0.500, +0.000),
        (-0.500, -1.000, +0.000), (-0.300, -0.500, +0.000), (-0.500, -0.500, +0.000),
        (-0.300, -0.500, +0.000), (+1.000, -0.500, +0.000), (-0.500, -0.500, +0.000),
        (-0.500, -1.000, +0.000), (-0.300, -1.000, +0.000), (-0.300, -0.500, +0.000),
        (+0.300, -0.500, +0.000), (+1.000, -0.500, +0.000), (-0.300, -0.500, +0.000),
        (+0.300, -1.000, +0.000), (+0.500, -0.500, +0.000), (+0.300, -0.500, +0.000),
        (+0.500, -0.500, +0.000), (+1.000, -0.500, +0.000), (+0.300, -0.500, +0.000),
        (+0.300, -1.000, +0.000), (+0.500, -1.000, +0.000), (+0.500, -0.500, +0.000),
    ),
    curvature = (
        (-1.000, -0.800, +0.000), (+1.000, -1.000, +0.000), (+1.000, -0.800, +0.000),
        (-1.000, -1.000, +0.000), (+1.000, -1.000, +0.000), (-1.000, -0.800, +0.000),
        (-0.136, +0.991, +0.000), (+0.136, +0.991, +0.000), (+0.000, +1.000, +0.000),
        (-0.136, +0.991, +0.000), (+0.266, +0.964, +0.000), (+0.136, +0.991, +0.000),
        (-0.266, +0.964, +0.000), (+0.266, +0.964, +0.000), (-0.136, +0.991, +0.000),
        (-0.266, +0.964, +0.000), (+0.389, +0.921, +0.000), (+0.266, +0.964, +0.000),
        (-0.389, +0.921, +0.000), (+0.389, +0.921, +0.000), (-0.266, +0.964, +0.000),
        (-0.389, +0.921, +0.000), (+0.505, +0.863, +0.000), (+0.389, +0.921, +0.000),
        (-0.505, +0.863, +0.000), (+0.505, +0.863, +0.000), (-0.389, +0.921, +0.000),
        (-0.505, +0.863, +0.000), (+0.611, +0.792, +0.000), (+0.505, +0.863, +0.000),
        (-0.611, +0.792, +0.000), (+0.000, +0.800, +0.000), (-0.505, +0.863, +0.000),
        (+0.000, +0.800, +0.000), (+0.611, +0.792, +0.000), (-0.505, +0.863, +0.000),
        (-0.611, +0.792, +0.000), (-0.109, +0.793, +0.000), (+0.000, +0.800, +0.000),
        (+0.109, +0.793, +0.000), (+0.611, +0.792, +0.000), (+0.000, +0.800, +0.000),
        (-0.611, +0.792, +0.000), (-0.213, +0.771, +0.000), (-0.109, +0.793, +0.000),
        (+0.213, +0.771, +0.000), (+0.611, +0.792, +0.000), (+0.109, +0.793, +0.000),
        (+0.213, +0.771, +0.000), (+0.707, +0.707, +0.000), (+0.611, +0.792, +0.000),
        (-0.707, +0.707, +0.000), (-0.213, +0.771, +0.000), (-0.611, +0.792, +0.000),
        (+0.311, +0.737, +0.000), (+0.707, +0.707, +0.000), (+0.213, +0.771, +0.000),
        (-0.707, +0.707, +0.000), (-0.311, +0.737, +0.000), (-0.213, +0.771, +0.000),
        (+0.404, +0.691, +0.000), (+0.707, +0.707, +0.000), (+0.311, +0.737, +0.000),
        (-0.707, +0.707, +0.000), (-0.404, +0.691, +0.000), (-0.311, +0.737, +0.000),
        (+0.404, +0.691, +0.000), (+0.792, +0.611, +0.000), (+0.707, +0.707, +0.000),
        (-0.792, +0.611, +0.000), (-0.404, +0.691, +0.000), (-0.707, +0.707, +0.000),
        (+0.489, +0.633, +0.000), (+0.792, +0.611, +0.000), (+0.404, +0.691, +0.000),
        (-0.792, +0.611, +0.000), (-0.489, +0.633, +0.000), (-0.404, +0.691, +0.000),
        (+0.566, +0.566, +0.000), (+0.792, +0.611, +0.000), (+0.489, +0.633, +0.000),
        (-0.792, +0.611, +0.000), (-0.566, +0.566, +0.000), (-0.489, +0.633, +0.000),
        (+0.566, +0.566, +0.000), (+0.863, +0.505, +0.000), (+0.792, +0.611, +0.000),
        (-0.863, +0.505, +0.000), (-0.566, +0.566, +0.000), (-0.792, +0.611, +0.000),
        (+0.633, +0.489, +0.000), (+0.863, +0.505, +0.000), (+0.566, +0.566, +0.000),
        (-0.863, +0.505, +0.000), (-0.633, +0.489, +0.000), (-0.566, +0.566, +0.000),
        (+0.633, +0.489, +0.000), (+0.921, +0.389, +0.000), (+0.863, +0.505, +0.000),
        (-0.921, +0.389, +0.000), (-0.633, +0.489, +0.000), (-0.863, +0.505, +0.000),
        (+0.691, +0.404, +0.000), (+0.921, +0.389, +0.000), (+0.633, +0.489, +0.000),
        (-0.921, +0.389, +0.000), (-0.691, +0.404, +0.000), (-0.633, +0.489, +0.000),
        (+0.737, +0.311, +0.000), (+0.921, +0.389, +0.000), (+0.691, +0.404, +0.000),
        (-0.921, +0.389, +0.000), (-0.737, +0.311, +0.000), (-0.691, +0.404, +0.000),
        (+0.737, +0.311, +0.000), (+0.964, +0.266, +0.000), (+0.921, +0.389, +0.000),
        (-0.964, +0.266, +0.000), (-0.737, +0.311, +0.000), (-0.921, +0.389, +0.000),
        (+0.771, +0.213, +0.000), (+0.964, +0.266, +0.000), (+0.737, +0.311, +0.000),
        (-0.964, +0.266, +0.000), (-0.771, +0.213, +0.000), (-0.737, +0.311, +0.000),
        (+0.771, +0.213, +0.000), (+0.991, +0.136, +0.000), (+0.964, +0.266, +0.000),
        (-0.991, +0.136, +0.000), (-0.771, +0.213, +0.000), (-0.964, +0.266, +0.000),
        (-0.991, +0.136, +0.000), (-0.793, +0.109, +0.000), (-0.771, +0.213, +0.000),
        (+0.793, +0.109, +0.000), (+0.991, +0.136, +0.000), (+0.771, +0.213, +0.000),
        (+0.793, +0.109, +0.000), (+1.000, +0.000, +0.000), (+0.991, +0.136, +0.000),
        (-1.000, +0.000, +0.000), (-0.793, +0.109, +0.000), (-0.991, +0.136, +0.000),
        (-1.000, +0.000, +0.000), (-0.800, +0.000, +0.000), (-0.793, +0.109, +0.000),
        (+0.800, +0.000, +0.000), (+1.000, +0.000, +0.000), (+0.793, +0.109, +0.000),
        (-0.995, -0.117, +0.000), (-0.800, +0.000, +0.000), (-1.000, +0.000, +0.000),
        (-0.995, -0.117, +0.000), (-0.797, -0.094, +0.000), (-0.800, +0.000, +0.000),
        (+0.797, -0.094, +0.000), (+1.000, +0.000, +0.000), (+0.800, +0.000, +0.000),
        (+0.797, -0.094, +0.000), (+0.995, -0.117, +0.000), (+1.000, +0.000, +0.000),
        (+0.788, -0.177, +0.000), (+0.995, -0.117, +0.000), (+0.797, -0.094, +0.000),
        (-0.995, -0.117, +0.000), (-0.788, -0.177, +0.000), (-0.797, -0.094, +0.000),
        (-0.982, -0.219, +0.000), (-0.788, -0.177, +0.000), (-0.995, -0.117, +0.000),
        (+0.788, -0.177, +0.000), (+0.982, -0.219, +0.000), (+0.995, -0.117, +0.000),
        (+0.772, -0.250, +0.000), (+0.982, -0.219, +0.000), (+0.788, -0.177, +0.000),
        (-0.982, -0.219, +0.000), (-0.772, -0.250, +0.000), (-0.788, -0.177, +0.000),
        (-0.962, -0.306, +0.000), (-0.772, -0.250, +0.000), (-0.982, -0.219, +0.000),
        (+0.772, -0.250, +0.000), (+0.962, -0.306, +0.000), (+0.982, -0.219, +0.000),
        (+0.751, -0.315, +0.000), (+0.962, -0.306, +0.000), (+0.772, -0.250, +0.000),
        (-0.962, -0.306, +0.000), (-0.751, -0.315, +0.000), (-0.772, -0.250, +0.000),
        (-0.936, -0.381, +0.000), (-0.751, -0.315, +0.000), (-0.962, -0.306, +0.000),
        (+0.751, -0.315, +0.000), (+0.936, -0.381, +0.000), (+0.962, -0.306, +0.000),
        (+0.725, -0.373, +0.000), (+0.936, -0.381, +0.000), (+0.751, -0.315, +0.000),
        (-0.936, -0.381, +0.000), (-0.725, -0.373, +0.000), (-0.751, -0.315, +0.000),
        (+0.692, -0.425, +0.000), (+0.936, -0.381, +0.000), (+0.725, -0.373, +0.000),
        (-0.936, -0.381, +0.000), (-0.692, -0.425, +0.000), (-0.725, -0.373, +0.000),
        (-0.906, -0.446, +0.000), (-0.692, -0.425, +0.000), (-0.936, -0.381, +0.000),
        (+0.692, -0.425, +0.000), (+0.906, -0.446, +0.000), (+0.936, -0.381, +0.000),
        (+0.655, -0.473, +0.000), (+0.906, -0.446, +0.000), (+0.692, -0.425, +0.000),
        (-0.906, -0.446, +0.000), (-0.655, -0.473, +0.000), (-0.692, -0.425, +0.000),
        (-0.871, -0.500, +0.000), (-0.655, -0.473, +0.000), (-0.906, -0.446, +0.000),
        (+0.655, -0.473, +0.000), (+0.871, -0.500, +0.000), (+0.906, -0.446, +0.000),
        (+0.612, -0.519, +0.000), (+0.871, -0.500, +0.000), (+0.655, -0.473, +0.000),
        (-0.871, -0.500, +0.000), (-0.612, -0.519, +0.000), (-0.655, -0.473, +0.000),
        (-0.835, -0.546, +0.000), (-0.612, -0.519, +0.000), (-0.871, -0.500, +0.000),
        (+0.612, -0.519, +0.000), (+0.835, -0.546, +0.000), (+0.871, -0.500, +0.000),
        (+0.564, -0.562, +0.000), (+0.835, -0.546, +0.000), (+0.612, -0.519, +0.000),
        (-0.835, -0.546, +0.000), (-0.564, -0.562, +0.000), (-0.612, -0.519, +0.000),
        (-0.797, -0.585, +0.000), (-0.564, -0.562, +0.000), (-0.835, -0.546, +0.000),
        (+0.564, -0.562, +0.000), (+0.797, -0.585, +0.000), (+0.835, -0.546, +0.000),
        (+0.511, -0.606, +0.000), (+0.797, -0.585, +0.000), (+0.564, -0.562, +0.000),
        (-0.797, -0.585, +0.000), (-0.511, -0.606, +0.000), (-0.564, -0.562, +0.000),
        (-0.760, -0.619, +0.000), (-0.511, -0.606, +0.000), (-0.797, -0.585, +0.000),
        (+0.511, -0.606, +0.000), (+0.760, -0.619, +0.000), (+0.797, -0.585, +0.000),
        (+0.453, -0.652, +0.000), (+0.760, -0.619, +0.000), (+0.511, -0.606, +0.000),
        (-0.760, -0.619, +0.000), (-0.453, -0.652, +0.000), (-0.511, -0.606, +0.000),
        (-0.723, -0.648, +0.000), (-0.453, -0.652, +0.000), (-0.760, -0.619, +0.000),
        (+0.453, -0.652, +0.000), (+0.723, -0.648, +0.000), (+0.760, -0.619, +0.000),
        (-0.690, -0.675, +0.000), (-0.453, -0.652, +0.000), (-0.723, -0.648, +0.000),
        (+0.453, -0.652, +0.000), (+0.690, -0.675, +0.000), (+0.723, -0.648, +0.000),
        (+0.390, -0.700, +0.000), (+0.690, -0.675, +0.000), (+0.453, -0.652, +0.000),
        (-0.690, -0.675, +0.000), (-0.390, -0.700, +0.000), (-0.453, -0.652, +0.000),
        (-0.660, -0.700, +0.000), (-0.390, -0.700, +0.000), (-0.690, -0.675, +0.000),
        (+0.390, -0.700, +0.000), (+0.660, -0.700, +0.000), (+0.690, -0.675, +0.000),
    ),
    deform = (
        (-0.538, -0.540, +0.000), (-0.362, +0.529, +0.000), (-0.538, +0.688, +0.000),
        (-0.362, +0.529, +0.000), (-0.061, +0.688, +0.000), (-0.538, +0.688, +0.000),
        (-0.362, +0.529, +0.000), (+0.047, +0.681, +0.000), (-0.061, +0.688, +0.000),
        (-0.362, +0.529, +0.000), (+0.145, +0.662, +0.000), (+0.047, +0.681, +0.000),
        (-0.362, +0.529, +0.000), (+0.232, +0.632, +0.000), (+0.145, +0.662, +0.000),
        (-0.362, +0.529, +0.000), (+0.310, +0.593, +0.000), (+0.232, +0.632, +0.000),
        (-0.362, +0.529, +0.000), (+0.378, +0.545, +0.000), (+0.310, +0.593, +0.000),
        (-0.362, +0.529, +0.000), (-0.090, +0.529, +0.000), (+0.378, +0.545, +0.000),
        (-0.090, +0.529, +0.000), (-0.008, +0.525, +0.000), (+0.378, +0.545, +0.000),
        (-0.008, +0.525, +0.000), (+0.436, +0.489, +0.000), (+0.378, +0.545, +0.000),
        (-0.538, -0.540, +0.000), (-0.362, -0.382, +0.000), (-0.362, +0.529, +0.000),
        (+0.066, +0.513, +0.000), (+0.436, +0.489, +0.000), (-0.008, +0.525, +0.000),
        (+0.133, +0.493, +0.000), (+0.436, +0.489, +0.000), (+0.066, +0.513, +0.000),
        (+0.193, +0.466, +0.000), (+0.436, +0.489, +0.000), (+0.133, +0.493, +0.000),
        (+0.193, +0.466, +0.000), (+0.484, +0.428, +0.000), (+0.436, +0.489, +0.000),
        (+0.244, +0.433, +0.000), (+0.484, +0.428, +0.000), (+0.193, +0.466, +0.000),
        (+0.289, +0.394, +0.000), (+0.484, +0.428, +0.000), (+0.244, +0.433, +0.000),
        (+0.289, +0.394, +0.000), (+0.524, +0.362, +0.000), (+0.484, +0.428, +0.000),
        (+0.327, +0.350, +0.000), (+0.524, +0.362, +0.000), (+0.289, +0.394, +0.000),
        (+0.327, +0.350, +0.000), (+0.554, +0.292, +0.000), (+0.524, +0.362, +0.000),
        (+0.357, +0.301, +0.000), (+0.554, +0.292, +0.000), (+0.327, +0.350, +0.000),
        (+0.381, +0.249, +0.000), (+0.554, +0.292, +0.000), (+0.357, +0.301, +0.000),
        (+0.381, +0.249, +0.000), (+0.575, +0.220, +0.000), (+0.554, +0.292, +0.000),
        (+0.397, +0.193, +0.000), (+0.575, +0.220, +0.000), (+0.381, +0.249, +0.000),
        (+0.397, +0.193, +0.000), (+0.588, +0.147, +0.000), (+0.575, +0.220, +0.000),
        (+0.407, +0.135, +0.000), (+0.588, +0.147, +0.000), (+0.397, +0.193, +0.000),
        (+0.407, +0.135, +0.000), (+0.592, +0.074, +0.000), (+0.588, +0.147, +0.000),
        (+0.410, +0.074, +0.000), (+0.592, +0.074, +0.000), (+0.407, +0.135, +0.000),
        (+0.408, +0.033, +0.000), (+0.592, +0.074, +0.000), (+0.410, +0.074, +0.000),
        (+0.408, +0.033, +0.000), (+0.587, -0.004, +0.000), (+0.592, +0.074, +0.000),
        (+0.402, -0.012, +0.000), (+0.587, -0.004, +0.000), (+0.408, +0.033, +0.000),
        (+0.402, -0.012, +0.000), (+0.572, -0.081, +0.000), (+0.587, -0.004, +0.000),
        (+0.390, -0.061, +0.000), (+0.572, -0.081, +0.000), (+0.402, -0.012, +0.000),
        (+0.372, -0.111, +0.000), (+0.572, -0.081, +0.000), (+0.390, -0.061, +0.000),
        (+0.372, -0.111, +0.000), (+0.547, -0.155, +0.000), (+0.572, -0.081, +0.000),
        (+0.347, -0.162, +0.000), (+0.547, -0.155, +0.000), (+0.372, -0.111, +0.000),
        (+0.347, -0.162, +0.000), (+0.513, -0.225, +0.000), (+0.547, -0.155, +0.000),
        (+0.315, -0.211, +0.000), (+0.513, -0.225, +0.000), (+0.347, -0.162, +0.000),
        (+0.274, -0.256, +0.000), (+0.513, -0.225, +0.000), (+0.315, -0.211, +0.000),
        (+0.274, -0.256, +0.000), (+0.470, -0.291, +0.000), (+0.513, -0.225, +0.000),
        (+0.224, -0.297, +0.000), (+0.470, -0.291, +0.000), (+0.274, -0.256, +0.000),
        (+0.224, -0.297, +0.000), (+0.418, -0.351, +0.000), (+0.470, -0.291, +0.000),
        (+0.164, -0.332, +0.000), (+0.418, -0.351, +0.000), (+0.224, -0.297, +0.000),
        (+0.094, -0.358, +0.000), (+0.418, -0.351, +0.000), (+0.164, -0.332, +0.000),
        (+0.094, -0.358, +0.000), (+0.358, -0.405, +0.000), (+0.418, -0.351, +0.000),
        (+0.013, -0.376, +0.000), (+0.358, -0.405, +0.000), (+0.094, -0.358, +0.000),
        (-0.081, -0.382, +0.000), (+0.358, -0.405, +0.000), (+0.013, -0.376, +0.000),
        (-0.538, -0.540, +0.000), (-0.081, -0.382, +0.000), (-0.362, -0.382, +0.000),
        (-0.538, -0.540, +0.000), (+0.358, -0.405, +0.000), (-0.081, -0.382, +0.000),
        (-0.538, -0.540, +0.000), (+0.290, -0.451, +0.000), (+0.358, -0.405, +0.000),
        (-0.538, -0.540, +0.000), (+0.213, -0.488, +0.000), (+0.290, -0.451, +0.000),
        (-0.538, -0.540, +0.000), (+0.129, -0.516, +0.000), (+0.213, -0.488, +0.000),
        (-0.538, -0.540, +0.000), (+0.038, -0.534, +0.000), (+0.129, -0.516, +0.000),
        (-0.538, -0.540, +0.000), (-0.059, -0.540, +0.000), (+0.038, -0.534, +0.000),
    ),
    intersect = (
        (-0.088, -0.540, +0.000), (+0.088, +0.688, +0.000), (-0.088, +0.688, +0.000),
        (-0.088, -0.540, +0.000), (+0.088, -0.540, +0.000), (+0.088, +0.688, +0.000),
    ),
    unwrap = (
        (-0.527, -0.076, +0.000), (-0.351, +0.688, +0.000), (-0.527, +0.688, +0.000),
        (-0.527, -0.076, +0.000), (-0.351, -0.038, +0.000), (-0.351, +0.688, +0.000),
        (+0.353, -0.027, +0.000), (+0.529, +0.688, +0.000), (+0.353, +0.688, +0.000),
        (+0.353, -0.027, +0.000), (+0.529, -0.058, +0.000), (+0.529, +0.688, +0.000),
        (+0.352, -0.055, +0.000), (+0.529, -0.058, +0.000), (+0.353, -0.027, +0.000),
        (-0.527, -0.076, +0.000), (-0.351, -0.071, +0.000), (-0.351, -0.038, +0.000),
        (+0.351, -0.082, +0.000), (+0.529, -0.058, +0.000), (+0.352, -0.055, +0.000),
        (+0.351, -0.082, +0.000), (+0.525, -0.131, +0.000), (+0.529, -0.058, +0.000),
        (-0.527, -0.076, +0.000), (-0.349, -0.101, +0.000), (-0.351, -0.071, +0.000),
        (-0.522, -0.155, +0.000), (-0.349, -0.101, +0.000), (-0.527, -0.076, +0.000),
        (+0.350, -0.109, +0.000), (+0.525, -0.131, +0.000), (+0.351, -0.082, +0.000),
        (-0.522, -0.155, +0.000), (-0.347, -0.128, +0.000), (-0.349, -0.101, +0.000),
        (+0.347, -0.134, +0.000), (+0.525, -0.131, +0.000), (+0.350, -0.109, +0.000),
        (-0.522, -0.155, +0.000), (-0.344, -0.152, +0.000), (-0.347, -0.128, +0.000),
        (+0.347, -0.134, +0.000), (+0.514, -0.199, +0.000), (+0.525, -0.131, +0.000),
        (+0.343, -0.158, +0.000), (+0.514, -0.199, +0.000), (+0.347, -0.134, +0.000),
        (-0.522, -0.155, +0.000), (-0.340, -0.175, +0.000), (-0.344, -0.152, +0.000),
        (-0.508, -0.227, +0.000), (-0.340, -0.175, +0.000), (-0.522, -0.155, +0.000),
        (+0.338, -0.181, +0.000), (+0.514, -0.199, +0.000), (+0.343, -0.158, +0.000),
        (-0.508, -0.227, +0.000), (-0.335, -0.195, +0.000), (-0.340, -0.175, +0.000),
        (+0.331, -0.203, +0.000), (+0.514, -0.199, +0.000), (+0.338, -0.181, +0.000),
        (-0.508, -0.227, +0.000), (-0.329, -0.213, +0.000), (-0.335, -0.195, +0.000),
        (+0.331, -0.203, +0.000), (+0.495, -0.262, +0.000), (+0.514, -0.199, +0.000),
        (+0.323, -0.223, +0.000), (+0.495, -0.262, +0.000), (+0.331, -0.203, +0.000),
        (-0.508, -0.227, +0.000), (-0.321, -0.230, +0.000), (-0.329, -0.213, +0.000),
        (+0.314, -0.243, +0.000), (+0.495, -0.262, +0.000), (+0.323, -0.223, +0.000),
        (-0.484, -0.292, +0.000), (-0.321, -0.230, +0.000), (-0.508, -0.227, +0.000),
        (-0.484, -0.292, +0.000), (-0.312, -0.246, +0.000), (-0.321, -0.230, +0.000),
        (+0.303, -0.261, +0.000), (+0.495, -0.262, +0.000), (+0.314, -0.243, +0.000),
        (-0.484, -0.292, +0.000), (-0.302, -0.261, +0.000), (-0.312, -0.246, +0.000),
        (-0.484, -0.292, +0.000), (-0.290, -0.276, +0.000), (-0.302, -0.261, +0.000),
        (+0.290, -0.279, +0.000), (+0.495, -0.262, +0.000), (+0.303, -0.261, +0.000),
        (+0.290, -0.279, +0.000), (+0.469, -0.320, +0.000), (+0.495, -0.262, +0.000),
        (-0.484, -0.292, +0.000), (-0.277, -0.290, +0.000), (-0.290, -0.276, +0.000),
        (+0.275, -0.295, +0.000), (+0.469, -0.320, +0.000), (+0.290, -0.279, +0.000),
        (-0.484, -0.292, +0.000), (-0.259, -0.307, +0.000), (-0.277, -0.290, +0.000),
        (-0.452, -0.349, +0.000), (-0.259, -0.307, +0.000), (-0.484, -0.292, +0.000),
        (+0.261, -0.309, +0.000), (+0.469, -0.320, +0.000), (+0.275, -0.295, +0.000),
        (-0.452, -0.349, +0.000), (-0.241, -0.322, +0.000), (-0.259, -0.307, +0.000),
        (+0.244, -0.322, +0.000), (+0.469, -0.320, +0.000), (+0.261, -0.309, +0.000),
        (+0.244, -0.322, +0.000), (+0.435, -0.373, +0.000), (+0.469, -0.320, +0.000),
        (-0.452, -0.349, +0.000), (-0.221, -0.336, +0.000), (-0.241, -0.322, +0.000),
        (+0.226, -0.334, +0.000), (+0.435, -0.373, +0.000), (+0.244, -0.322, +0.000),
        (+0.206, -0.346, +0.000), (+0.435, -0.373, +0.000), (+0.226, -0.334, +0.000),
        (-0.452, -0.349, +0.000), (-0.201, -0.348, +0.000), (-0.221, -0.336, +0.000),
        (+0.184, -0.356, +0.000), (+0.435, -0.373, +0.000), (+0.206, -0.346, +0.000),
        (-0.452, -0.349, +0.000), (-0.180, -0.359, +0.000), (-0.201, -0.348, +0.000),
        (-0.414, -0.399, +0.000), (-0.180, -0.359, +0.000), (-0.452, -0.349, +0.000),
        (+0.161, -0.366, +0.000), (+0.435, -0.373, +0.000), (+0.184, -0.356, +0.000),
        (-0.414, -0.399, +0.000), (-0.158, -0.368, +0.000), (-0.180, -0.359, +0.000),
        (+0.136, -0.374, +0.000), (+0.435, -0.373, +0.000), (+0.161, -0.366, +0.000),
        (-0.414, -0.399, +0.000), (-0.136, -0.376, +0.000), (-0.158, -0.368, +0.000),
        (+0.136, -0.374, +0.000), (+0.394, -0.420, +0.000), (+0.435, -0.373, +0.000),
        (+0.110, -0.381, +0.000), (+0.394, -0.420, +0.000), (+0.136, -0.374, +0.000),
        (-0.414, -0.399, +0.000), (-0.113, -0.383, +0.000), (-0.136, -0.376, +0.000),
        (+0.082, -0.387, +0.000), (+0.394, -0.420, +0.000), (+0.110, -0.381, +0.000),
        (-0.414, -0.399, +0.000), (-0.089, -0.388, +0.000), (-0.113, -0.383, +0.000),
        (+0.052, -0.391, +0.000), (+0.394, -0.420, +0.000), (+0.082, -0.387, +0.000),
        (-0.414, -0.399, +0.000), (-0.064, -0.391, +0.000), (-0.089, -0.388, +0.000),
        (+0.020, -0.393, +0.000), (+0.394, -0.420, +0.000), (+0.052, -0.391, +0.000),
        (-0.414, -0.399, +0.000), (-0.039, -0.393, +0.000), (-0.064, -0.391, +0.000),
        (-0.013, -0.394, +0.000), (+0.394, -0.420, +0.000), (+0.020, -0.393, +0.000),
        (-0.414, -0.399, +0.000), (-0.013, -0.394, +0.000), (-0.039, -0.393, +0.000),
        (-0.414, -0.399, +0.000), (+0.394, -0.420, +0.000), (-0.013, -0.394, +0.000),
        (-0.369, -0.441, +0.000), (+0.394, -0.420, +0.000), (-0.414, -0.399, +0.000),
        (-0.369, -0.441, +0.000), (+0.344, -0.460, +0.000), (+0.394, -0.420, +0.000),
        (-0.319, -0.477, +0.000), (+0.344, -0.460, +0.000), (-0.369, -0.441, +0.000),
        (-0.319, -0.477, +0.000), (+0.288, -0.494, +0.000), (+0.344, -0.460, +0.000),
        (-0.264, -0.506, +0.000), (+0.288, -0.494, +0.000), (-0.319, -0.477, +0.000),
        (-0.264, -0.506, +0.000), (+0.223, -0.521, +0.000), (+0.288, -0.494, +0.000),
        (-0.205, -0.528, +0.000), (+0.223, -0.521, +0.000), (-0.264, -0.506, +0.000),
        (-0.205, -0.528, +0.000), (+0.151, -0.540, +0.000), (+0.223, -0.521, +0.000),
        (-0.144, -0.544, +0.000), (+0.151, -0.540, +0.000), (-0.205, -0.528, +0.000),
        (-0.144, -0.544, +0.000), (+0.071, -0.552, +0.000), (+0.151, -0.540, +0.000),
        (-0.081, -0.553, +0.000), (+0.071, -0.552, +0.000), (-0.144, -0.544, +0.000),
        (-0.081, -0.553, +0.000), (-0.016, -0.556, +0.000), (+0.071, -0.552, +0.000),
    ),
    embed = (
        (-0.315, -0.540, +0.000), (-0.139, +0.529, +0.000), (-0.315, +0.688, +0.000),
        (-0.139, +0.529, +0.000), (+0.382, +0.688, +0.000), (-0.315, +0.688, +0.000),
        (-0.139, +0.529, +0.000), (+0.382, +0.529, +0.000), (+0.382, +0.688, +0.000),
        (-0.315, -0.540, +0.000), (-0.139, +0.158, +0.000), (-0.139, +0.529, +0.000),
        (-0.315, -0.540, +0.000), (-0.139, +0.000, +0.000), (-0.139, +0.158, +0.000),
        (-0.139, +0.000, +0.000), (+0.364, +0.158, +0.000), (-0.139, +0.158, +0.000),
        (-0.139, +0.000, +0.000), (+0.364, +0.000, +0.000), (+0.364, +0.158, +0.000),
        (-0.315, -0.540, +0.000), (-0.139, -0.382, +0.000), (-0.139, +0.000, +0.000),
        (-0.315, -0.540, +0.000), (+0.398, -0.382, +0.000), (-0.139, -0.382, +0.000),
        (-0.315, -0.540, +0.000), (+0.398, -0.540, +0.000), (+0.398, -0.382, +0.000),
    ),
)
