# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "prop",
    "utils_ui": "NestedLayout",
    "utils_blender": "BlUtil",
    "utils_addon": "AddonManager",
})

from . import material_utils
from . import template_material_menu

from .material_utils import MaterialLibraryLocation, TextureSynonym
from .template_material_menu import TemplateMaterialMenu

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

@addon.PropertyGroup
class DecalLoadingOptions:
    draw_category: 'MATERIAL' | prop("Import options category", "Which category of decal loading options to show", items=[
        ('FILE', "File", "File path and image reuse options"),
        ('FRAMES', "Frames", "Options for movies and image sequences"),
        ('DECAL', "Decal", "Decal size and alpha mode options"),
        ('MATERIAL', "Material", "Material options"),
    ])
    
    path_mode: 'PREFERENCES' | prop("Convert paths", "Convert paths to absolute or relative style on import", items=[
        ('PREFERENCES', "Use Preferences", "Use the path mode defined in the addon's preferences"),
        ('DONT_CHANGE', "Don't change", "Don't change file paths"),
        ('ABSOLUTE', "Absolute", "Make paths absolute"),
        ('RELATIVE', "Relative", "Make paths relative"),
    ])
    
    reuse_mode: 'REUSE' | prop("Reuse images", "Use already loaded images", items=[
        ('NONE', "No", "New image datablocks will be created even if the same files have already been loaded"),
        ('REUSE', "Yes", "Use already loaded images when possible"),
        ('RELOAD', "Reload", "Use already loaded images + force reload them"),
    ])
    
    use_auto_refresh: True | prop("Auto refresh", "Always refresh image on frame changes")
    
    frame_start: 1 | prop("Start frame", "Starting frame for movies and image sequences")
    
    sequences_mode: 'NONE' | prop("Image sequences", "Automatically detect image sequences", items=[
        ('NONE', "None", "Don't detect image sequences"),
        ('NO_GAPS', "Without gaps", "Gaps in frame numbers will result in separate sequences"),
        ('WITH_GAPS', "With gaps", "Gaps in frame numbers won't result in spearate sequences"),
    ])
    
    alpha_mode: 'DONT_CHANGE' | prop("Alpha mode", "How to interpret the alpha channel", items=[
        ('DONT_CHANGE', "Don't change", "Don't change alpha mode"), # used in BlUtil.Image.load
        ('STRAIGHT', "Straight", "Store RGB and alpha channels separately, with alpha acting as a mask"),
        ('PREMUL', "Premultiplied", "Store RGB channels with alpha multiplied in"),
        ('CHANNEL_PACKED', "Channel Packed", "Different images are packed in the RGB and alpha channels"),
        ('NONE', "Ignore alpha", "Ignore alpha channel from the file and make the image fully opaque"),
    ])
    
    template_material_path: "" | prop()
    
    overwrite_material: False | prop("Overwrite material", "Overwrite existing material (based on material name)")
    
    use_multi_texture: True | prop("Multi-texture", "Try to auto-detect which images correspond to which shader nodes (based on node labels)")
    
    @property
    def real_path_mode(self):
        return (self.path_mode if self.path_mode != 'PREFERENCES' else settings.path_mode)
    
    @property
    def categories(self):
        return [preset.id for preset in TextureSynonym.presets]
    
    def kwargs(self, **kwargs):
        kwargs.update(categories=self.categories,
            use_auto_refresh=self.use_auto_refresh, frame_start=self.frame_start,
            use_multi_texture=self.use_multi_texture,
            overwrite_material=self.overwrite_material,
            path_mode=self.real_path_mode, reuse=self.reuse_mode,
            sequences=self.sequences_mode, alpha_mode=self.alpha_mode)
        return kwargs
    
    def sync_defaults(self):
        if not self.template_material_path:
            self.template_material_path = settings.template_material_path
    
    def draw(self, layout, mode=None):
        layout = NestedLayout(layout)
        
        if (not mode) or (mode == 'FILE'):
            with layout.box():
                with layout.row():
                    with layout.row()(alignment='LEFT'):
                        layout.label(text="Paths:")
                    layout.prop(self, "path_mode", text="")
                with layout.row():
                    with layout.row()(alignment='LEFT'):
                        layout.label(text="Reuse:")
                    layout.prop(self, "reuse_mode", text="")
        
        if (not mode) or (mode == 'FRAMES'):
            with layout.box():
                with layout.row():
                    layout.prop(self, "use_auto_refresh", toggle=True)
                    layout.prop(self, "frame_start")
                with layout.row():
                    with layout.row()(alignment='LEFT'):
                        layout.label(text="Image sequences:")
                    layout.prop(self, "sequences_mode", text="")
        
        if (not mode) or (mode == 'DECAL'):
            with layout.box():
                with layout.row():
                    with layout.row(align=True):
                        icon = ('PLAY' if settings.default_dynamic else 'PAUSE')
                        layout.prop(*settings("default_dynamic"), text="", icon=icon)
                        layout.prop(*settings("default_intersect"), text="", icon='MOD_BOOLEAN')
                    with layout.row(align=True):
                        layout.prop(*settings("default_scale"))
                        layout.prop(*settings("default_rotation"), text="", icon_only=True)
                with layout.row():
                    with layout.row()(alignment='LEFT'):
                        layout.label(text="Alpha:")
                    layout.prop(self, "alpha_mode", text="")
        
        if (not mode) or (mode == 'MATERIAL'):
            with layout.box():
                with layout.row():
                    with layout.row()(alignment='LEFT'):
                        layout.label(text="Material:")
                    with layout.row(align=True):
                        TemplateMaterialMenu.draw_main(layout)
                        MaterialLibraryLocation.draw_popup(layout, icon='PREFERENCES')
                with layout.row():
                    layout.prop(self, "overwrite_material", text="Overwrite", toggle=True)
                    with layout.row(align=True):
                        layout.prop(self, "use_multi_texture", text="Multi Tex", toggle=True)
                        TextureSynonym.draw_popup(layout, icon='PREFERENCES')
