# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import os
from collections import deque

import numpy as np

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bounds": "Bounds",
    "bpy_inspect": "BlRna, prop, IDProp, IDTypes",
    "utils_ui": "BlUI",
    "utils_blender": "BlUtil, BpyPath, TextureFilenameParser",
    "utils_addon": "AddonManager",
})

from . import common
from .common import Jobs, CommonUtils, SimpleOpEditMixin

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

class MaterialLibrary:
    default = "default.blend"
    
    availability = {}
    
    @classmethod
    def read(cls):
        processed_paths = set()
        material_infos = []
        
        for preset in MaterialLibraryLocation.presets:
            src_path = preset.id
            abs_path = CommonUtils.abspath(src_path)
            
            if abs_path in processed_paths: continue
            processed_paths.add(abs_path)
            
            if not os.path.isfile(abs_path): continue
            
            relative = src_path.startswith("//")
            
            with bpy.data.libraries.load(abs_path, link=True, relative=relative) as (data_from, data_to):
                for material_name in data_from.materials:
                    material_path = os.path.join(src_path, material_name).replace(os.path.sep, "/")
                    material_infos.append((material_path, abs_path, material_name))
        
        return material_infos
    
    @classmethod
    def try_load_material(cls, library_path=None, material_name=None, operator=None):
        if library_path is None: library_path = settings.loading_options.template_material_path
        
        try:
            material = cls.load_material(library_path, material_name)
        except Exception as exc:
            if operator is not None:
                # ERROR_INVALID_CONTEXT displays a popup
                operator.report({'ERROR_INVALID_CONTEXT'}, str(exc))
            else:
                print(exc)
            
            material = None
        
        cls.availability[library_path] = bool(material)
        
        return material
    
    @classmethod
    def load_material(cls, library_path, material_name=None):
        if material_name is None: library_path, material_name = CommonUtils.split(library_path)
        
        abs_path = CommonUtils.abspath(library_path)
        
        if not (abs_path and material_name):
            raise ValueError("No template material is specified")
        
        # Avoid the "ID ... is already linked" warning
        material = CommonUtils.find_linked(abs_path, "materials", material_name)
        if material: return material
        
        if not os.path.isfile(abs_path):
            raise FileNotFoundError(f"Can't find template material: file '{library_path}' does not exist")
        
        relative = library_path.startswith("//")
        
        with bpy.data.libraries.load(abs_path, link=True, relative=relative) as (data_from, data_to):
            if material_name not in data_from.materials:
                raise RuntimeError(f"Can't find template material {material_name} in file '{library_path}'")
            
            data_to.materials = [material_name]
        
        material = data_to.materials[0]
        
        if not material:
            raise RuntimeError(f"Could not load template material {material_name} in file '{library_path}'")
        
        if not (material.use_nodes and material.node_tree):
            raise RuntimeError(f"Template material {material_name} in file '{library_path}' does not use nodes")
        
        for node in material.node_tree.nodes:
            if node.bl_idname == "ShaderNodeTexImage": break
        else:
            raise RuntimeError(f"Template material {material_name} in file '{library_path}' has no Image Texture nodes")
        
        return material

@addon.Preset("library_locations.preset", sorting='NONE', popup='PANEL', title="Library locations", options={'ADD', 'RESET', 'INLINE'})
class MaterialLibraryLocation:
    OpEditMixin = SimpleOpEditMixin
    normalize_whitespace = False

@addon.Preset("texture_synonyms.preset", sorting='NONE', popup='PANEL', title="Texture synonyms", options={'ADD', 'RESET', 'INLINE'})
class TextureSynonym:
    OpEditMixin = SimpleOpEditMixin
    normalize_whitespace = True

class MaterialUtils:
    node_ignore_mark = "#"
    
    clipboard = None
    
    @classmethod
    def estimate_image_size(cls, material):
        image_size = (1, 1)
        
        if material and material.node_tree:
            for node in material.node_tree.nodes:
                if node.bl_idname != "ShaderNodeTexImage": continue
                
                image = node.image
                if not image: continue
                
                label = node.label.strip()
                if cls.node_ignore_mark in label: continue
                
                if max(image.size) > max(image_size):
                    image_size = tuple(image.size)
        
        return image_size
    
    @classmethod
    def load_images_to_materials(cls, paths, material, categories, use_multi_texture=True,
                                use_auto_refresh=True, frame_start=1, overwrite_material=True, **kwargs):
        material_loader = MaterialLoader(paths, material, categories, use_multi_texture,
            use_auto_refresh, frame_start, overwrite_material, **kwargs)
        material_loader.load_new()
        return material_loader.material_infos
    
    @classmethod
    def mix(cls, base, overlay, uv_overlay=""):
        if base == overlay: return overlay
        if not overlay: return base
        
        key_base = "Base Material"
        key_overlay = "Overlay Material"
        
        # Gather old materials before attempting to mix
        old_materials = []
        for mat in bpy.data.materials:
            if mat.library: continue
            if (mat.get(key_base) == base) and (mat.get(key_overlay) == overlay):
                old_materials.append(mat)
        
        result = MaterialMixer.mix(base, overlay, uv_overlay)
        if not result: return overlay
        
        result[key_base] = base
        result[key_overlay] = overlay
        
        for mat in old_materials:
            mat.user_remap(result)
            bpy.data.materials.remove(mat)
        
        return result

class MaterialLoader:
    height_to_normal = "height->normal"
    non_color = "non-color"
    
    material_infos = None
    
    def __init__(self, paths, material, categories, use_multi_texture=True,
                 use_auto_refresh=True, frame_start=1, overwrite_material=True, **kwargs):
        self.paths = paths
        self.material = material
        self.categories = categories
        self.use_multi_texture = use_multi_texture
        self.use_auto_refresh = use_auto_refresh
        self.frame_start = frame_start
        self.overwrite_material = overwrite_material
        self.kwargs = kwargs
        self.normal_maps = {}
        self.material_infos = []
        
        self.image_nodes, self.has_labeled_nodes = self.find_image_nodes(material)
        
        if self.image_nodes:
            kwargs["parser"] = self.init_tex_parser()
            self.find_old_materials()
    
    def load_new(self):
        if not self.image_nodes: return
        
        for shared_key, image_infos in BlUtil.Image.load(self.paths, **self.kwargs).items():
            new_material = self.material.copy()
            new_material.preview_render_type = 'FLAT'
            CommonUtils.mark_as_template(new_material)
            
            image_path, image_size = self.setup_nodes(new_material, image_infos)
            if not (image_path and max(image_size)): continue
            
            self.rename_material(new_material, shared_key, image_path)
            self.material_infos.append((new_material, image_size))
    
    def load_override(self):
        if not self.image_nodes: return
        
        for shared_key, image_infos in BlUtil.Image.load(self.paths, **self.kwargs).items():
            image_path, image_size = self.setup_nodes(self.material, image_infos)
            if not (image_path and max(image_size)): continue
            
            self.rename_material(self.material, shared_key, image_path)
            self.material_infos.append((self.material, image_size))
            
            break # stop after the first successful override
    
    def find_image_nodes(self, material):
        if not (material and material.use_nodes and material.node_tree): return None, False
        
        image_nodes = {}
        has_labeled_nodes = False
        
        for node in material.node_tree.nodes:
            if node.bl_idname != "ShaderNodeTexImage": continue
            
            label = node.label.strip()
            if MaterialUtils.node_ignore_mark in label: continue
            
            colorspace_settings = getattr(node.image, "colorspace_settings", None)
            is_data = getattr(colorspace_settings, "is_data", None)
            colorspace = getattr(colorspace_settings, "name", None)
            
            has_labeled_nodes |= bool(label)
            image_nodes[node.name] = dict(label=label, is_data=is_data,
                colorspace=colorspace, category_ids=[None], color_hints=[True],
                skip=False, image_info=None, color_hint=None)
        
        return image_nodes, has_labeled_nodes
    
    def init_tex_parser(self):
        has_labeled_nodes = self.has_labeled_nodes
        use_multi_texture = self.use_multi_texture
        
        if not (self.categories and has_labeled_nodes): return None
        
        parser = TextureFilenameParser(self.categories)
        parser.filter = set()
        
        def get_color_hint(category_id):
            # Can't pre-compute this information in general case,
            # since parser.find() may potentially add new categories
            return ("color" in parser.categories[category_id].lower())
        
        for node_name, node_info in self.image_nodes.items():
            category_ids = parser.find(node_info["label"])
            parser.filter.update(category_ids)
            
            if use_multi_texture:
                node_info["category_ids"] = category_ids
                node_info["color_hints"] = [get_color_hint(category_id) for category_id in category_ids]
            else:
                node_info["color_hints"] = [any(get_color_hint(category_id) for category_id in category_ids)]
            
            node_info["skip"] = (not use_multi_texture) and has_labeled_nodes and (not node_info["color_hints"][0])
        
        return (parser if use_multi_texture else None)
    
    def setup_nodes(self, new_material, image_infos):
        image_path = None
        image_size = (0, 0)
        
        for node_name, node_info in self.image_nodes.items():
            if node_info["skip"]: continue
            
            node = new_material.node_tree.nodes[node_name]
            
            image_info = self.get_image_info(image_infos, node_info)
            
            if image_info:
                image, index_start, frame_count = image_info
                
                image_path = image.filepath
                if max(image.size) > max(image_size):
                    image_size = tuple(image.size)
                
                self.setup_node_image(node, image, node_info)
                self.setup_node_frames(node, index_start, frame_count)
            
            node.mute = not node.image
        
        return image_path, image_size
    
    def get_image_info(self, image_infos, node_info):
        for category_id, color_hint in zip(node_info["category_ids"], node_info["color_hints"]):
            image_info = image_infos.get(category_id)
            if image_info is None: continue
            node_info["image_info"] = image_info
            node_info["color_hint"] = color_hint
            return image_info
        return None
    
    def setup_node_image(self, node, image, node_info):
        node_name_low = node.name.lower()
        
        if self.height_to_normal in node_name_low:
            normal_map = self.normal_maps.get(image)
            if not normal_map:
                normal_map = image.xdecal.make_normalmap()
                self.normal_maps[image] = normal_map
            image = normal_map
        else:
            colorspace = node_info["colorspace"]
            if not colorspace:
                non_color = (not node_info["color_hint"]) or (self.non_color in node_name_low)
                colorspace = ("Non-Color" if non_color else "sRGB")
            image.colorspace_settings.name = colorspace
        
        node.image = image
        node.show_texture = True
        #node.extension = 'EXTEND' # hmm... maybe use 'CLIP' for alpha?
        node.extension = 'REPEAT' # hmm... maybe use 'CLIP' for alpha?
    
    def setup_node_frames(self, node, index_start, frame_count):
        image_user = node.image_user
        image_user.use_auto_refresh = self.use_auto_refresh
        image_user.frame_start = self.frame_start
        image_user.frame_offset = index_start - self.frame_start
        image_user.frame_duration = frame_count
    
    def find_old_materials(self):
        if not self.overwrite_material: return
        self.old_materials = {old_material.name_full: old_material for old_material in bpy.data.materials}
    
    def replace_old_material(self, name, new_material):
        if not self.overwrite_material: return
        old_material = self.old_materials.get(name)
        if not old_material: return
        if old_material == self.material: return # just in case
        old_material.user_remap(new_material)
        bpy.data.materials.remove(old_material)
    
    def make_name(self, shared_key, image_path):
        filename = BpyPath.abspath(image_path)
        filename = os.path.basename(filename)
        filename = os.path.splitext(filename)[0]
        shared_key = BpyPath.abspath(shared_key)
        shared_key = os.path.basename(shared_key)
        filename = filename[:len(shared_key)].strip().rstrip("_-[{(< \t\r\n")
        filename = BpyPath.display_name_from_filepath(filename)
        filename = f"{filename} ({self.material.name})"
        return filename
    
    def rename_material(self, new_material, shared_key, image_path):
        if not (shared_key and image_path): return
        name = self.make_name(shared_key, image_path)
        self.replace_old_material(name, new_material)
        new_material.name = name

class MaterialMixer:
    class FrameInfo:
        size = property(lambda self: self.max - self.min)
        center = property(lambda self: (self.max + self.min)*0.5)
        
        def __init__(self, frame, bounds):
            self.frame = frame
            self.min, self.max = bounds.get("min", "max", convert=Vector, nan=0.0)
        
        # Note: when a frame node is fitted to its contents, its location
        # might not correspond to the lower-left corner in the editor UI
        def move(self, delta):
            delta = Vector(delta)
            self.frame.location += delta
            self.min += delta
            self.max += delta
        
        def align(self, pos, pivot=(0,0)):
            delta = Vector(pos) - self.min
            size = self.size
            delta -= Vector((pivot[0]*size.x, pivot[1]*size.y))
            self.move(delta)
    
    @classmethod
    def map_links(cls, tree):
        links_fwd = {}
        links_bwd = {}
        for link in tree.links:
            links_fwd.setdefault(link.from_socket, []).append(link.to_socket)
            links_bwd.setdefault(link.to_socket, []).append(link.from_socket)
        return links_fwd, links_bwd
    
    @classmethod
    def copy_nodes(cls, src_mat, dst_mat, parent=None):
        old_nodes = set(dst_mat.node_tree.nodes)
        BlUtil.Material.copy_nodes(src_mat, dst_mat)
        new_nodes = [node for node in dst_mat.node_tree.nodes if node not in old_nodes]
        for node in new_nodes:
            if not node.parent: node.parent = parent
        return new_nodes
    
    @classmethod
    def frame(cls, tree, label, nodes, force=False):
        if not nodes: return None
        
        # Node dimensions may have weird values until they are shown in UI
        node_dimensions = (100.0, 100.0) # use some reasonable estimate
        
        # Calculate bounds before adding a new node
        bounds = Bounds(Bounds.MinSize(node.location, node_dimensions) for node in nodes)
        
        frame = tree.nodes.new("NodeFrame")
        frame.label = label
        
        for node in nodes:
            if (not node.parent) or force: node.parent = frame
        
        return cls.FrameInfo(frame, bounds)
    
    @classmethod
    def import_node_group(cls, group_name):
        return CommonUtils.load_node_group(group_name, MaterialLibrary.default, link=False)
    
    output_nodes = ("ShaderNodeOutputMaterial", "ShaderNodeOutputAOV")
    
    uv_map_nodes = ("ShaderNodeUVMap", "ShaderNodeNormalMap", "ShaderNodeTangent")
    
    # Other data types cannot be meaningfully mixed
    socket_types = {'VALUE', 'INT', 'BOOLEAN', 'VECTOR', 'ROTATION', 'RGBA'}
    vector_types = ('VECTOR', 'ROTATION', 'RGBA')
    
    @classmethod
    def is_output_node(cls, node):
        return node.bl_idname in cls.output_nodes
    
    @classmethod
    def is_uv_map_node(cls, node):
        return node.bl_idname in cls.uv_map_nodes
    
    @classmethod
    def is_shader_node(cls, node):
        return any(socket.type == 'SHADER' for socket in node.outputs)
    
    @classmethod
    def is_texture_node(cls, node):
        return node.bl_idname.startswith("ShaderNodeTex")
    
    @classmethod
    def find_connected_nodes(cls, nodes, links_map, is_forward, node_filter, socket_filter):
        nodes = list(filter(node_filter, nodes))
        
        i = 0
        while i < len(nodes):
            node = nodes[i]
            i += 1
            
            for socket0 in (node.outputs if is_forward else node.inputs):
                if socket_filter and (not socket_filter(socket0)): continue
                
                for socket1 in links_map.get(socket0, ()):
                    if socket1 and (socket1.node not in nodes):
                        nodes.append(socket1.node)
        
        return nodes
    
    @classmethod
    def find_endpoint_nodes(cls, nodes, links_bwd):
        socket_filter = (lambda socket: socket.type == 'SHADER')
        return cls.find_connected_nodes(nodes, links_bwd, is_forward=False,
            node_filter=cls.is_output_node, socket_filter=socket_filter)
    
    @classmethod
    def reachable_from_textures(cls, nodes, links_fwd):
        node_filter = (lambda node: cls.is_texture_node(node) and (not node.mute))
        return cls.find_connected_nodes(nodes, links_fwd, is_forward=True,
            node_filter=node_filter, socket_filter=None)
    
    @classmethod
    def ensure_nodes(cls, material):
        should_reset = not material.use_nodes
        
        # Auto creates Principled BSDF and Output if nodes weren't initialized
        material.use_nodes = True
        
        tree = material.node_tree
        
        if not should_reset:
            should_reset = not any(cls.is_output_node(node) for node in tree.nodes)
        
        if not should_reset: return
        
        tree.nodes.clear()
        shader_node = tree.nodes.new("ShaderNodeBsdfPrincipled")
        output_node = tree.nodes.new("ShaderNodeOutputMaterial")
        tree.links.new(output_node.inputs["Surface"], shader_node.outputs["BSDF"])
        shader_node.inputs["Base Color"].default_value = material.diffuse_color
        shader_node.inputs["Metallic"].default_value = material.metallic
        shader_node.inputs["Roughness"].default_value = material.roughness
        if "Specular" in shader_node.inputs: # before Blender 4
            shader_node.inputs["Specular"].default_value = material.specular_intensity
        elif "Specular IOR Level" in shader_node.inputs: # Blender 4+
            shader_node.inputs["Specular IOR Level"].default_value = material.specular_intensity
        output_node.location.x = shader_node.width + 20
    
    @classmethod
    def add_uv_map(cls, tree, uv_map, nodes):
        uv_sockets = []
        
        for node in nodes:
            if not cls.is_texture_node(node):
                if cls.is_uv_map_node(node):
                    # Technically, empty UV map uses the active UV map,
                    # but in basic cases we can assume it's "UVMap"
                    if (not node.uv_map) or (node.uv_map == "UVMap"):
                        node.uv_map = uv_map
                continue
            
            for socket in node.inputs:
                if socket.type != 'VECTOR': continue
                if not socket.is_linked: uv_sockets.append(socket)
                break
        
        uv_map_node = tree.nodes.new("ShaderNodeUVMap")
        uv_map_node.uv_map = uv_map
        from_socket = uv_map_node.outputs[0]
        for to_socket in uv_sockets:
            tree.links.new(to_socket, from_socket)
        
        return uv_map_node
    
    @classmethod
    def gather_node_groups(cls, materials):
        if isinstance(materials, bpy.types.Material):
            materials = [materials]
        
        node_groups = set()
        
        for material in materials:
            if not material: continue
            if not material.use_nodes: continue
            node_group = material.node_tree
            if not node_group: continue
            node_groups.add(node_group)
        
        queue = list(node_groups)
        
        while queue:
            node_group = queue.pop()
            
            for node in node_group.nodes:
                node_group = getattr(node, "node_tree", None)
                if not node_group: continue
                if node_group in node_groups: continue
                node_groups.add(node_group)
                queue.append(node_group)
        
        return node_groups
    
    @classmethod
    def get_attribute_node_type(cls, node):
        outputs = node.outputs
        
        is_color = outputs["Color"].is_linked and not outputs["Vector"].is_linked
        
        if outputs["Alpha"].is_linked:
            return ('RGBA' if is_color else 'XYZW')
        elif outputs["Color"].is_linked or outputs["Vector"].is_linked:
            return ('RGB' if is_color else 'XYZ')
        elif outputs["Fac"].is_linked:
            return 'FLOAT'
        else:
            return 'NONE'
    
    @classmethod
    def gather_custom_props(cls, materials):
        attr_types = ('OBJECT', 'INSTANCER')
        prop_infos = {}
        for node_group in cls.gather_node_groups(materials):
            for node in node_group.nodes:
                if node.bl_idname != "ShaderNodeAttribute": continue
                if node.attribute_type not in attr_types: continue
                prop_infos[node.attribute_name] = cls.get_attribute_node_type(node)
        return prop_infos
    
    # Blender accepts as shader attributes only int/float properties
    # (either scalar or up to 4 components)
    # Also bool scalars, but not bool vectors, for some reason
    shader_attr_size_limits = {"bool": 0, "int": 4, "float": 4}
    
    @classmethod
    def can_value_be_attribute(cls, value):
        value_type, value_size = IDProp.get_type_info(value)
        size_limit = cls.shader_attr_size_limits.get(value_type)
        return (size_limit is not None) and (value_size <= size_limit)
    
    @classmethod
    def copy_custom_props(cls, src_objs, dst_obj, prop_infos):
        custom_props = {}
        for src_obj in src_objs:
            for key, value in src_obj.items():
                if key not in prop_infos: continue
                if not cls.can_value_be_attribute(value): continue
                value_type, value_size = IDProp.get_type_info(value)
                entry = custom_props.get(key)
                if entry and (value_size < entry[-1]): continue
                custom_props[key] = (src_obj, value, value_size)
        
        for key, entry in custom_props.items():
            src_obj, value, value_size = entry
            dst_obj[key] = value
            
            prop_path = "[\"%s\"]" % bpy.utils.escape_identifier(key)
            cls.setup_property(dst_obj, prop_path, src_obj, prop_path, var_name="decal_embed_var")
    
    @classmethod
    def clear_custom_props(cls, obj):
        anim_data = obj.animation_data
        if not anim_data: return
        
        data_paths = set()
        for fcurve in anim_data.drivers:
            driver = fcurve.driver
            if not driver: continue
            if driver.expression != "decal_embed_var": continue
            data_paths.add(fcurve.data_path)
        
        for key in tuple(obj.keys()):
            prop_path = "[\"%s\"]" % bpy.utils.escape_identifier(key)
            if prop_path in data_paths: del obj[key]
    
    @classmethod
    def arrange_nodes(cls, frame_output, frame_base, frame_overlay, uv_overlay_node, mix_nodes):
        if frame_base:
            base_min, base_max = frame_base.min, frame_base.max
            base_top_y = base_max.y + 300
        else:
            base_min, base_max = Vector(), Vector()
            base_top_y = (frame_output.min.y if frame_output else 0)
        
        if frame_overlay:
            frame_overlay.align((base_min.x, base_top_y))
            
            if uv_overlay_node:
                uv_overlay_node.location = Vector((
                    frame_overlay.min.x - uv_overlay_node.width - 60,
                    frame_overlay.min.y + (frame_overlay.size.y - uv_overlay_node.height)*0.5,
                ))
            
            mix_left_x = max(base_max.x, frame_overlay.max.x) + 60
            
            if frame_output:
                frame_output.move((max(base_max.x, frame_overlay.max.x) + 500 - frame_output.min.x, 0))
        
        mix_y = base_min.y
        for node in reversed(mix_nodes):
            node.location = Vector((mix_left_x, mix_y))
            # Case-specific, because Blender has no way to estimate
            # the size of a newly created node before a UI update
            if node.bl_idname == "ShaderNodeNewGeometry":
                node_height = 180
                node.location.y += 20
                mix_y += 20
            else:
                node_height = 180
            mix_y += node_height + 20
    
    props_from_overlay = [
        "use_transparent_shadow",
        "cycles.use_bump_map_correction",
        "use_backface_culling",
        "use_backface_culling_shadow", # >= Blender 4.2 (added)
        "use_backface_culling_lightprobe_volume", # >= Blender 4.2 (added)
        "blend_method", # < Blender 4.2 (renamed)
        "surface_render_method", # >= Blender 4.2 (renamed)
        "shadow_method",
        "alpha_threshold",
        "show_transparent_back", # < Blender 4.2 (renamed)
        "use_transparency_overlap", # >= Blender 4.2 (renamed)
        "use_screen_refraction", # < Blender 4.2 (renamed)
        "use_raytrace_refraction", # >= Blender 4.2 (renamed)
        "refraction_depth",
        "use_sss_translucency", # < Blender 4.2 (deprecated)
        "pass_index",
    ]
    
    @classmethod
    def mix(cls, base, overlay, uv_overlay=""):
        """Combine materials' shader nodes"""
        if base == overlay: return None
        if not (overlay and overlay.use_nodes and overlay.node_tree): return None
        if len(overlay.node_tree.nodes) == 0: return None
        
        if base:
            result = base.copy()
            result.name = f".{base.name}+{overlay.name}"
        else:
            result = bpy.data.materials.new(f".<No-material>+{overlay.name}")
        
        for prop_path in cls.props_from_overlay:
            attr_names = prop_path.split(".")
            src_container = overlay
            dst_container = result
            for attr_name in attr_names[:-1]:
                if not hasattr(src_container, attr_name): continue
                src_container = getattr(src_container, attr_name)
                dst_container = getattr(dst_container, attr_name)
            attr_name = attr_names[-1]
            if not hasattr(src_container, attr_name): continue
            setattr(dst_container, attr_name, getattr(src_container, attr_name))
        
        cls.ensure_nodes(result)
        tree = result.node_tree
        
        base_nodes = list(tree.nodes) # make a snapshot
        overlay_nodes = cls.copy_nodes(overlay, result)
        
        mixer = MaterialMixer(tree, base, base_nodes, overlay, overlay_nodes)
        mixer.gather_channels()
        mixer.mix_channels()
        mixer.setup_drivers()
        mixer.cleanup()
        
        if uv_overlay: uv_overlay_node = cls.add_uv_map(tree, uv_overlay, mixer.over_logic_nodes)
        
        if mixer.has_alpha:
            if hasattr(result, "surface_render_method"): # >= Blender 4.2
                result.surface_render_method = 'BLENDED'
            elif hasattr(result, "blend_method"): # < Blender 4.2
                result.blend_method = 'BLEND'
        
        # shadow_method was not removed in Blender 4.2, but doesn't work anyway
        # (though it won't create problems because EEVEE now uses the Cycles'
        # bpy.types.Object property "visible_shadow")
        if hasattr(result, "shadow_method"):
            result.shadow_method = 'NONE'
        
        frame_output = cls.frame(tree, "Shader/Output", mixer.base_final_nodes, force=True)
        
        frame_base = None
        if mixer.base_logic_nodes:
            frame_base = cls.frame(tree, "Base", mixer.base_logic_nodes)
        
        frame_overlay = cls.frame(tree, "Overlay", mixer.over_logic_nodes)
        
        cls.arrange_nodes(frame_output, frame_base, frame_overlay, uv_overlay_node, mixer.new_nodes)
        
        return result
    
    @classmethod
    def make_node_map(cls, mat, nodes):
        if not (mat and mat.use_nodes and mat.node_tree): return {}
        node_map = {}
        for src, dst in zip(mat.node_tree.nodes, nodes):
            node_map[src] = dst
            node_map[dst] = src
        return node_map
    
    def __init__(self, tree, base_mat, base_nodes, over_mat, over_nodes):
        self.tree = tree
        self.base_mat = base_mat
        self.over_mat = over_mat
        
        self.base_nodes_map = self.make_node_map(base_mat, base_nodes)
        self.over_nodes_map = self.make_node_map(over_mat, over_nodes)
        
        self.links_fwd, self.links_bwd = self.map_links(tree)
        
        self.base_final_nodes = set(self.find_endpoint_nodes(base_nodes, self.links_bwd))
        self.over_final_nodes = set(self.find_endpoint_nodes(over_nodes, self.links_bwd))
        
        self.base_logic_nodes = set(base_nodes) - self.base_final_nodes
        self.over_logic_nodes = set(over_nodes) - self.over_final_nodes
        
        self.driven_props = {}
        self.gather_driven_props(self.base_final_nodes, self.base_nodes_map, base_mat, True)
        self.gather_driven_props(self.base_logic_nodes, self.base_nodes_map, base_mat)
        self.gather_driven_props(self.over_logic_nodes, self.over_nodes_map, over_mat)
        
        self.new_nodes = []
        self.common_nodes = {}
    
    def gather_driven_props(self, nodes, nodes_map, src_mat, use_fallback=False):
        if not src_mat: return
        
        if src_mat.use_nodes:
            for dst_node in nodes:
                src_node = nodes_map.get(dst_node)
                if not src_node: continue
                self.gather_node_props(src_node, dst_node, src_mat)
            return
        
        if not use_fallback: return
        
        for dst_node in nodes:
            for dst_socket in dst_node.inputs:
                if dst_socket.type not in self.socket_types: continue
                if dst_socket.is_linked: continue
                
                prop_name = "default_value"
                
                channel_name = self.get_channel_name(dst_socket)
                
                if channel_name == "Diffuse Color":
                    self.driven_props[(dst_socket, prop_name)] = (src_mat, "diffuse_color")
                elif channel_name == "Metallic":
                    self.driven_props[(dst_socket, prop_name)] = (src_mat, "metallic")
                elif channel_name == "Specular":
                    self.driven_props[(dst_socket, prop_name)] = (src_mat, "specular_intensity")
                elif channel_name == "Roughness":
                    self.driven_props[(dst_socket, prop_name)] = (src_mat, "roughness")
    
    def gather_channels(self):
        self.channel_overrides = {}
        
        def get_from_socket(to_socket):
            from_sockets = self.links_bwd.get(to_socket)
            return (from_sockets[0] if from_sockets else None)
        
        for node in self.over_final_nodes:
            for to_socket in node.inputs:
                if to_socket.type not in self.socket_types: continue
                
                channel_name = self.get_channel_name(to_socket)
                if channel_name in self.channel_overrides: continue
                
                from_socket = get_from_socket(to_socket)
                if not from_socket: continue
                
                self.channel_overrides[channel_name] = from_socket
        
        self.has_alpha = ("Alpha" in self.channel_overrides) or ("1-Alpha" in self.channel_overrides)
    
    def cleanup(self):
        self.links_fwd, self.links_bwd = self.map_links(self.tree)
        
        used_nodes = set(self.find_connected_nodes(self.base_final_nodes, self.links_bwd,
            is_forward=False, node_filter=self.is_output_node, socket_filter=None))
        
        for node in self.tree.nodes:
            if node in used_nodes: continue
            
            # This must be done BEFORE removing the node,
            # since after that it may no longer work
            self.base_final_nodes.discard(node)
            self.over_final_nodes.discard(node)
            self.base_logic_nodes.discard(node)
            self.over_logic_nodes.discard(node)
            
            self.tree.nodes.remove(node)
    
    def add_node(self, node_type):
        node = self.tree.nodes.new(node_type)
        self.new_nodes.append(node)
        return node
    
    def get_common_node(self, node_type):
        node = self.common_nodes.get(node_type)
        if not node:
            node = self.add_node(node_type)
            self.common_nodes[node_type] = node
        return node
    
    def add_link(self, to_socket, from_socket):
        self.tree.links.new(to_socket, from_socket)
    
    def add_group(self, group_name):
        group_node = self.add_node("ShaderNodeGroup")
        group_node.node_tree = self.import_node_group(group_name)
        return group_node
    
    @classmethod
    def is_normal(cls, socket):
        return (socket.type == 'VECTOR') and ("Normal" in socket.identifier.split())
    
    @classmethod
    def find_socket(cls, sockets, socket_type, name=None):
        # Note: sockets are indexed by name instead of
        # identifier, but names can be non-unique
        for socket in sockets:
            if socket.type != socket_type: continue
            if name and (socket.name != name): continue
            return socket
    
    def make_inverted_socket(self, src_socket):
        node = self.add_node("ShaderNodeMath")
        node.operation = 'SUBTRACT'
        node.inputs[0].default_value = 1
        self.add_link(node.inputs[1], src_socket)
        return node.outputs[0]
    
    def get_override_socket(self, channel_name):
        override_socket = self.channel_overrides.get(channel_name)
        
        if not override_socket:
            if (channel_name == "1-Alpha") and ("Alpha" in self.channel_overrides):
                override_socket = self.make_inverted_socket(self.channel_overrides["Alpha"])
                self.channel_overrides["1-Alpha"] = override_socket
            elif (channel_name == "Alpha") and ("1-Alpha" in self.channel_overrides):
                override_socket = self.make_inverted_socket(self.channel_overrides["1-Alpha"])
                self.channel_overrides["Alpha"] = override_socket
        
        return override_socket
    
    def get_base_logic_sockets(self, to_socket):
        from_sockets = self.links_bwd.get(to_socket)
        if from_sockets: return from_sockets
        
        if self.is_normal(to_socket):
            node = self.get_common_node("ShaderNodeNewGeometry")
            return [node.outputs["Normal"]]
        
        if hasattr(to_socket, "default_value"): return [to_socket]
        
        return None
    
    socket_to_mix_node = {
        'VALUE': 'FLOAT',
    }
    
    def add_mix_node(self, to_socket, channel_name, override_socket):
        alpha = self.get_override_socket("Alpha")
        
        mix_node = None
        if (not alpha) and self.is_normal(to_socket):
            mix_node = self.add_group("XDecal.MixWorldNormals")
            if mix_node:
                factor = mix_node.inputs[0]
                value0 = mix_node.inputs[1]
                value1 = mix_node.inputs[2]
                output = mix_node.outputs[0]
        
        if mix_node is None:
            if channel_name == "Alpha":
                mix_node = self.add_node("ShaderNodeMath")
                mix_node.operation = 'MULTIPLY'
                
                factor = None
                value0 = mix_node.inputs[0]
                value1 = mix_node.inputs[1]
                output = mix_node.outputs[0]
            else:
                mix_node = self.add_node("ShaderNodeMix")
                mix_node.data_type = self.socket_to_mix_node.get(to_socket.type, to_socket.type)
                
                if alpha:
                    mix_node.blend_type = 'MIX'
                    if NodeInfo.should_ignore_alpha(override_socket.node):
                        alpha = None
                else:
                    mix_node.blend_type = 'ADD'
                
                factor = self.find_socket(mix_node.inputs, 'VALUE', "Factor")
                value0 = self.find_socket(mix_node.inputs, to_socket.type, "A")
                value1 = self.find_socket(mix_node.inputs, to_socket.type, "B")
                output = self.find_socket(mix_node.outputs, to_socket.type)
        
        return dict(node=mix_node, alpha=alpha, factor=factor, value0=value0, value1=value1, output=output)
    
    def mix_channels(self):
        for node in self.base_final_nodes:
            for to_socket in node.inputs:
                if to_socket.type not in self.socket_types: continue
                
                channel_name = self.get_channel_name(to_socket)
                
                override_socket = self.get_override_socket(channel_name)
                if not override_socket: continue
                
                mix_socket = NodeInfo.get_mix_socket(override_socket.node)
                
                from_sockets = self.get_base_logic_sockets(to_socket)
                
                if from_sockets is None:
                    # Not connected in the base shader, and has no default value
                    # (this shouldn't really happen, but just in case)
                    self.add_link(to_socket, override_socket)
                    continue
                
                def connect_from_socket(target_socket, from_socket):
                    if from_socket.is_output:
                        self.add_link(target_socket, from_socket)
                    else:
                        try:
                            target_socket.default_value = from_socket.default_value
                        except Exception as exc:
                            print("Cannot assign default value (mismatching socket types):", exc)
                        
                        driver_target = self.driven_props.get((to_socket, "default_value"))
                        if driver_target:
                            self.driven_props[(target_socket, "default_value")] = driver_target
                
                for from_socket in from_sockets:
                    info = self.add_mix_node(to_socket, channel_name, override_socket)
                    
                    info["node"].name = "Mix "+channel_name
                    info["node"].label = channel_name
                    
                    if info["factor"]:
                        info["factor"].default_value = 1.0
                        if info["alpha"]: self.add_link(info["factor"], info["alpha"])
                    
                    connect_from_socket(info["value0"], from_socket)
                    
                    if mix_socket:
                        connect_from_socket(mix_socket, from_socket)
                    
                    self.add_link(info["value1"], override_socket)
                    
                    self.add_link(to_socket, info["output"])
    
    node_color_map = {
        "ShaderNodeBsdfTransparent": "Transparent Color",
        "ShaderNodeBsdfDiffuse": "Diffuse Color",
        "ShaderNodeBsdfRefraction": "Refraction Color",
        "ShaderNodeSubsurfaceScattering": "Subsurface Color",
        "ShaderNodeEmission": "Emission Color",
        "ShaderNodeBsdfTranslucent": "Translucent Color",
        "ShaderNodeBsdfGlass": "Glass Color",
        "ShaderNodeBsdfGlossy": "Glossy Color", # merged with ShaderNodeBsdfAnisotropic in Blender 4+
        "ShaderNodeBsdfAnisotropic": "Glossy Color",
    }
    
    def is_connected_to_transparent(self, socket):
        from_sockets = self.links_bwd.get(socket)
        if not from_sockets: return False
        return (from_sockets[0].node.bl_idname == "ShaderNodeBsdfTransparent")
    
    def get_channel_name(self, socket):
        node = socket.node
        channel_name = socket.name
        
        if node.bl_idname == "ShaderNodeOutputAOV":
            if hasattr(node, "aov_name"): # >= Blender 4.2
                channel_name = f"AOV/{node.aov_name}/{channel_name}"
            else: # < Blender 4.2
                channel_name = f"AOV/{node.name}/{channel_name}"
        elif node.bl_idname == "ShaderNodeMixShader":
            if self.is_connected_to_transparent(node.inputs[1]):
                channel_name = "Alpha"
            elif self.is_connected_to_transparent(node.inputs[2]):
                channel_name = "1-Alpha"
        elif socket.type == 'RGBA':
            channel_name_low = channel_name.lower()
            if channel_name_low in ("specular", "specular tint"):
                channel_name = "Specular Color"
            elif channel_name_low in ("emission", "emission color"):
                channel_name = "Emission Color"
            elif channel_name_low == "subsurface": # removed in Blender 4
                channel_name = "Subsurface Color"
            elif channel_name_low == "base color":
                channel_name = "Diffuse Color"
            elif channel_name_low == "color":
                channel_name = self.node_color_map.get(node.bl_idname, channel_name)
        
        return channel_name
    
    base_prop_count = len(bpy.types.ShaderNode.bl_rna.properties)
    
    rna_prop_types = {'FLOAT', 'INT', 'BOOLEAN', 'ENUM'}
    
    nodes_with_outputs = {
        "ShaderNodeValue": [0],
        "ShaderNodeRGB": [0],
        "ShaderNodeNormal": [0],
    }
    
    @classmethod
    def get_data_path(cls, container, prop_name):
        try:
            return "node_tree." + container.path_from_id(prop_name)
        except Exception as exc:
            print(f"FAIL to get property path: {container} {prop_name}\n", exc)
            return None
    
    def gather_node_props(self, src_node, dst_node, src_mat):
        if dst_node.bl_idname != src_node.bl_idname: return
        
        # E.g. Frame and Reroute are instances of NodeInternal
        if not isinstance(dst_node, bpy.types.ShaderNode): return
        
        # Note: trying to create drivers for node ui properties mostly works,
        # but in most cases it results in Blender printing warnings to console,
        # and I'm not sure how useful it is anyway.
        
        # node_rna_props = dst_node.bl_rna.properties
        # for i in range(self.base_prop_count, len(node_rna_props)):
        #     rna_prop = node_rna_props[i]
        #     if rna_prop.type not in self.rna_prop_types: continue
        #     prop_name = rna_prop.identifier
        #     data_path = self.get_data_path(src_node, prop_name)
        #     self.driven_props[(dst_node, prop_name)] = (src_mat, data_path)
        
        # Note: input/output sockets are indexable via name, not identifier
        # However, names might not be unique, so we can only rely on index
        
        for i, dst_socket in enumerate(dst_node.inputs):
            if dst_socket.type not in self.socket_types: continue
            if dst_socket.is_linked: continue
            src_socket = src_node.inputs[i]
            if src_socket.hide: continue
            if src_socket.hide_value: continue
            if src_socket.is_unavailable: continue
            prop_name = "default_value"
            data_path = self.get_data_path(src_socket, prop_name)
            self.driven_props[(dst_socket, prop_name)] = (src_mat, data_path)
        
        output_indices = self.nodes_with_outputs.get(dst_node.bl_idname)
        if output_indices:
            for i in output_indices:
                dst_socket = dst_node.outputs[i]
                if not dst_socket.is_linked: continue
                src_socket = src_node.outputs[i]
                prop_name = "default_value"
                data_path = self.get_data_path(src_socket, prop_name)
                self.driven_props[(dst_socket, prop_name)] = (src_mat, data_path)
    
    def setup_drivers(self):
        for dst_info, src_info in self.driven_props.items():
            material, data_path = src_info
            prop_container, prop_name = dst_info
            if not data_path: continue
            self.setup_property(prop_container, prop_name, material, data_path)
    
    @classmethod
    def setup_property(cls, prop_container, prop_name, idblock, data_path, var_name=None):
        try:
            fcurves = prop_container.driver_add(prop_name)
        except Exception as exc:
            print(f"FAIL to add a driver: {prop_container} {prop_name}\n", exc)
            return
        
        if isinstance(fcurves, bpy.types.FCurve):
            cls.setup_fcurve(fcurves, idblock, data_path, var_name)
        else:
            for i, fcurve in enumerate(fcurves):
                cls.setup_fcurve(fcurve, idblock, f"{data_path}[{i}]", var_name)
    
    @classmethod
    def setup_fcurve(cls, fcurve, idblock, data_path, var_name=None):
        id_type_info = IDTypes.map("type").get(type(idblock))
        
        if not var_name: var_name = "var"
        
        driver = fcurve.driver
        
        if len(driver.variables) > 0:
            driver_var = driver.variables[0]
        else:
            driver_var = driver.variables.new()
        
        driver_var.type = 'SINGLE_PROP'
        driver_var.name = var_name
        
        driver_target = driver_var.targets[0]
        driver_target.id_type = id_type_info.id_type
        driver_target.id = idblock
        driver_target.data_path = data_path
        
        driver.expression = var_name

class ImageUtils:
    @classmethod
    def get_pixels(cls, image):
        width, height = image.size
        # Creating a numpy array from image.pixels directly is quite slow.
        # It's better to use image.pixels[:] (which converts it to a
        # tuple on the Blender side), and tuple->ndarray is much faster.
        # In total, image.pixels[:] is ~3 times faster than image.pixels.
        return np.reshape(image.pixels[:], (height, width, image.channels))
    
    @classmethod
    def set_pixels(cls, image, pixels):
        # image.pixels[:] seems to be somewhat faster than image.pixels
        image.pixels[:] = np.reshape(pixels, (-1,))
    
    @classmethod
    def grayscale(cls, pixels, reuse=False):
        height, width, channels = pixels.shape
        
        data = pixels[:, :, 0]
        if not reuse: data = data.copy()
        
        if channels > 1:
            color_channels = min(channels, 3)
            for channel in range(1, color_channels):
                data += pixels[:, :, channel]
            data /= float(color_channels)
            if channels > 3: data *= pixels[:, :, 3]
        
        return data
    
    @classmethod
    def blur(cls, data, size, mode):
        if size <= 0: return data
        
        if mode == 'BOX':
            # Box blur using the Summed Area Table, based on
            # https://www.gamasutra.com/view/feature/3102/four_tricks_for_fast_blurring_in_.php
            size = round(size)
            if size <= 0: return data
            sat = np.pad(data, size, 'edge')
            sat.cumsum(axis=0, out=sat)
            sat.cumsum(axis=1, out=sat)
            data[:, :] = sat[size*2:, size*2:]
            data[:, :] += sat[:-size*2, :-size*2]
            data[:, :] -= sat[size*2:, :-size*2]
            data[:, :] -= sat[:-size*2, size*2:]
            data *= 1.0 / ((size*2+1) * (size*2+1))
        elif mode == 'GAUSSIAN':
            size *= 0.5 # to empirically have similar size as box blur
            kernel_size = math.ceil(size*3)
            kernel = np.arange(-kernel_size, kernel_size+1)
            kernel = np.exp((kernel**2) / (-2 * size**2)) / (2 * np.pi * size**2)
            kernel /= np.sum(kernel)
            data = np.pad(data, kernel_size, 'edge')
            data = np.apply_along_axis(np.convolve, 0, data, kernel, mode='valid')
            data = np.apply_along_axis(np.convolve, 1, data, kernel, mode='valid')
        
        return data
    
    @classmethod
    def normalmap(cls, data, channels=3, scale=255.0):
        height, width = data.shape
        
        normals = np.zeros((height, width, channels))
        normals[:, :, 2] = 1
        
        if width > 1:
            dx = data[:, 1:] - data[:, :-1]
            normals[:, :-1, 0] = dx
            normals[:, 1:, 0] += dx
            normals[:, :, 0] *= -0.5 * scale
        
        if height > 1:
            dy = data[1:, :] - data[:-1, :]
            normals[:-1, :, 1] = dy
            normals[1:, :, 1] += dy
            normals[:, :, 1] *= -0.5 * scale
        
        normals[:, :, :3] /= np.linalg.norm(normals[:, :, :3], axis=2, keepdims=True)
        
        normals[:, :, :3] += 1.0
        normals[:, :, :3] *= 0.5
        
        return normals
    
    @classmethod
    def _get_image_kwargs(cls, image):
        if not image: return {}
        width, height = image.size
        # Note: colorspace_settings.is_data is not always synced with "Non-Color"
        # (which is probably a bug, since assigning "Non-Color" makes is_data True)
        return dict(
            width = width,
            height = height,
            alpha = (image.channels > 3),
            float_buffer = image.is_float,
            stereo3d = image.is_stereo_3d,
            is_data = (image.colorspace_settings.name == "Non-Color"),
            tiled = (image.source == 'TILED'),
        )
    
    @classmethod
    def _must_be_regenerated(cls, image, kwargs):
        if (not image) or image.library: return True
        alpha = kwargs.get("alpha")
        float_buffer = kwargs.get("float_buffer")
        stereo3d = kwargs.get("stereo3d")
        kwargs = cls._get_image_kwargs(image)
        alpha = (alpha is not None) and (kwargs.get("alpha") != alpha)
        float_buffer = (float_buffer is not None) and (kwargs.get("float_buffer") != float_buffer)
        stereo3d = (stereo3d is not None) and (kwargs.get("stereo3d") != stereo3d)
        return alpha or float_buffer or stereo3d
    
    @classmethod
    def _make_output_image(cls, image, name, result=None, **kwargs):
        kwargs = {**cls._get_image_kwargs(image), **kwargs}
        
        tiles_supported = "tiles" in bpy.types.Image.bl_rna.properties
        
        if not cls._must_be_regenerated(result, kwargs):
            width = kwargs.get("width")
            height = kwargs.get("height")
            is_data = kwargs.get("is_data")
            tiled = kwargs.get("tiled")
            
            kwargs = cls._get_image_kwargs(result)
            if width is None: width = kwargs["width"]
            if height is None: height = kwargs["height"]
            if is_data is None: is_data = kwargs["is_data"]
            if tiled is None: tiled = kwargs["tiled"]
            
            if (kwargs["width"] != width) or (kwargs["height"] != height):
                result.scale(width, height)
            
            if is_data != kwargs["is_data"]:
                colorspace = image.colorspace_settings.name if image else "sRGB"
                if colorspace == "Non-Color": colorspace = "sRGB"
                result.colorspace_settings.name = ("Non-Color" if is_data else colorspace)
            
            if tiles_supported and (tiled != kwargs["tiled"]):
                image.source = 'TILED'
            
            return
        
        if not tiles_supported: kwargs.pop("tiled", None)
        
        new_result = bpy.data.images.new(name, **kwargs)
        
        if result and (not result.library):
            result.user_remap(new_result)
            bpy.data.images.remove(result)
        
        return new_result
    
    @classmethod
    def heightmap_to_normalmap(cls, image, result=None, scale=255.0, blur=0, blur_type='BOX'):
        result = cls._make_output_image(image, image.name+"->normal", result, alpha=False, is_data=True)
        pixels = cls.get_pixels(image)
        data = cls.grayscale(pixels, reuse=True)
        data = cls.blur(data, blur, blur_type)
        normals = cls.normalmap(data, result.channels, scale)
        cls.set_pixels(result, normals)
        result.pack()
        return result

class Serializers:
    default = ((lambda value: value), setattr)
    
    vector = (tuple, setattr)
    
    def color_ramp_elements_get(elements):
        return [(e.position, tuple(e.color), e.alpha) for e in elements]
    def color_ramp_elements_set(target, name, infos):
        elements = getattr(target, name)
        for i in range(len(elements)-1, len(infos)-1, -1):
            elements.remove(elements[i])
        for i, info in enumerate(infos):
            position, color, alpha = info
            e = (elements[i] if i < len(elements) else elements.new(position))
            e.position = position
            e.color = color
            e.alpha = alpha
    color_ramp_elements = (color_ramp_elements_get, color_ramp_elements_set)
    
    def curve_maps_get(curve_maps):
        return [[(p.handle_type, tuple(p.location), p.select) for p in curve_map.points]
            for curve_map in curve_maps]
    def curve_maps_set(target, name, infos):
        curve_maps = getattr(target, name)
        for curve_map, curve_data in zip(curve_maps, infos):
            points = curve_map.points
            for i in range(len(points)-1, len(curve_data)-1, -1):
                points.remove(points[i])
            for i, point_data in enumerate(curve_data):
                handle_type, location, select = point_data
                p = (points[i] if i < len(points) else points.new(location[0], location[1]))
                p.handle_type = handle_type
                p.location = location
                p.select = select
    curve_maps = (curve_maps_get, curve_maps_set)
    
    @classmethod
    def serialize(cls, target, attr_infos, data=None):
        if data is None: data = {}
        for attr_name, attr_info in attr_infos.items():
            attr_value = getattr(target, attr_name)
            if isinstance(attr_info, dict):
                data[attr_name] = cls.serialize(attr_value, attr_info)
            else:
                serializer, deserializer = attr_info
                data[attr_name] = serializer(attr_value)
        return data
    
    @classmethod
    def deserialize(cls, target, attr_infos, data):
        if not data: return
        for attr_name, attr_info in attr_infos.items():
            if attr_name not in data: continue
            data_value = data[attr_name]
            attr_value = getattr(target, attr_name)
            if isinstance(attr_info, dict):
                cls.deserialize(attr_value, attr_info, data_value)
            else:
                serializer, deserializer = attr_info
                deserializer(target, attr_name, data_value)

class NodeInfo:
    prop_key = "XDecal"
    
    mix_info_separator = "#"
    
    @classmethod
    def should_ignore_alpha(cls, node):
        info = cls.get(node)
        return info and info["alpha"]
    
    @classmethod
    def get_mix_socket(cls, node):
        info = cls.get(node)
        if not info: return None
        
        mix_socket_info = info.setdefault("mix_socket", "")
        if not mix_socket_info: return None
        
        parts = mix_socket_info.rsplit(cls.mix_info_separator, maxsplit=1)
        
        count = None
        if len(parts) > 1:
            try:
                count = int(parts[-1]) - 1
            except Exception:
                pass
        
        if count is None:
            name = mix_socket_info
            count = 0
        else:
            name = parts[0]
        
        for socket in node.inputs:
            if (not socket.enabled) or socket.is_unavailable: continue
            if socket.is_linked != socket.is_output: continue
            if socket.name != name: continue
            if count <= 0: return socket
            count -= 1
        
        return None
    
    @classmethod
    def get_mix_socket_info(cls, node, mix_socket):
        if mix_socket is None: return ""
        
        count = 0
        
        for socket in node.inputs:
            if (not socket.enabled) or socket.is_unavailable: continue
            if socket.is_linked != socket.is_output: continue
            if socket.name != mix_socket.name: continue
            if socket == mix_socket:
                if count == 0: return socket.name
                return f"{socket.name}{cls.mix_info_separator}{count+1}"
            count += 1
        
        return ""
    
    @classmethod
    def get(cls, node):
        info = node.get(cls.prop_key)
        if not isinstance(info, dict):
            if not hasattr(info, "to_dict"): return None
            info = info.to_dict()
        info.setdefault("group", "")
        info.setdefault("id", "")
        info.setdefault("alpha", False)
        info.setdefault("mix_socket", "")
        info["toggles"] = info.get("toggles", "").upper()
        info.setdefault("inputs", {})
        info.setdefault("outputs", {})
        return info
    
    @classmethod
    def set(cls, node, info):
        node[cls.prop_key] = info
    
    @classmethod
    def get_socket_ids(cls, sockets, info):
        toggles = info.get("toggles") or ""
        inputs = info.get("inputs") or {}
        outputs = info.get("outputs") or {}
        
        invert_inputs = 'I' in toggles
        invert_outputs = 'O' in toggles
        
        def get_label(socket, count=1):
            socket_name = socket.name
            invert = (invert_outputs if socket.is_output else invert_inputs)
            
            for socket_info in (outputs if socket.is_output else inputs):
                if not hasattr(socket_info, "get"): continue
                name = socket_info.get("name", "")
                if socket_name != name: continue
                label = socket_info.get("label", "")
                toggle = socket_info.get("toggle", False)
                count -= 1
                if count > 0: continue
                return ((label or socket_name) if toggle != invert else None)
            
            return (socket_name if invert else None)
        
        name_counts = {}
        socket_ids = {}
        
        for socket in sockets:
            if (not socket.enabled) or socket.is_unavailable: continue
            if socket.is_linked != socket.is_output: continue
            
            count = name_counts.get(socket.name, 0) + 1
            name_counts[socket.name] = count
            
            label = get_label(socket, count)
            if label is None: continue
            
            socket_ids[socket.identifier] = label
        
        return socket_ids
    
    # COPY/PASTE:
    
    curves_copy_attrs = {
        "black_level": Serializers.vector,
        "clip_max_x": Serializers.default,
        "clip_max_y": Serializers.default,
        "clip_min_x": Serializers.default,
        "clip_min_y": Serializers.default,
        "curves": Serializers.curve_maps,
        "extend": Serializers.default,
        "tone": Serializers.default,
        "use_clip": Serializers.default,
        "white_level": Serializers.vector,
    }
    
    ramp_copy_attrs = {
        "color_mode": Serializers.default,
        "elements": Serializers.color_ramp_elements,
        "hue_interpolation": Serializers.default,
        "interpolation": Serializers.default,
    }
    
    img_copy_attrs = {
        "color_mapping": {
            "blend_color": Serializers.vector,
            "blend_factor": Serializers.default,
            "blend_type": Serializers.default,
            "brightness": Serializers.default,
            "color_ramp": {
                "color_mode": Serializers.default,
                "elements": Serializers.color_ramp_elements,
                "hue_interpolation": Serializers.default,
                "interpolation": Serializers.default,
            },
            "contrast": Serializers.default,
            "saturation": Serializers.default,
            "use_color_ramp": Serializers.default,
        },
        "extension": Serializers.default,
        "image_user": {
            "frame_current": Serializers.default,
            "frame_duration": Serializers.default,
            "frame_offset": Serializers.default,
            "frame_start": Serializers.default,
            "tile": Serializers.default,
            "use_auto_refresh": Serializers.default,
            "use_cyclic": Serializers.default,
        },
        "interpolation": Serializers.default,
        "projection": Serializers.default,
        "projection_blend": Serializers.default,
        "texture_mapping": {
            "mapping": Serializers.default,
            "mapping_x": Serializers.default,
            "mapping_y": Serializers.default,
            "mapping_z": Serializers.default,
            "max": Serializers.vector,
            "min": Serializers.vector,
            "rotation": Serializers.vector,
            "scale": Serializers.vector,
            "translation": Serializers.vector,
            "use_max": Serializers.default,
            "use_min": Serializers.default,
            "vector_type": Serializers.default,
        },
    }
    
    basic_types = (bool, int, float, str)
    
    @classmethod
    def curves_props_copy(cls, node, attr_name="mapping"):
        curves = getattr(node, attr_name, None)
        if curves is None: return None
        if not isinstance(curves, bpy.types.CurveMapping): return None
        return Serializers.serialize(curves, cls.curves_copy_attrs)
    
    @classmethod
    def curves_props_paste(cls, node, data, attr_name="mapping"):
        if data is None: return
        curves = getattr(node, attr_name, None)
        if curves is None: return
        if not isinstance(curves, bpy.types.CurveMapping): return
        Serializers.deserialize(curves, cls.curves_copy_attrs, data)
    
    @classmethod
    def ramp_props_copy(cls, node, attr_name="color_ramp"):
        ramp = getattr(node, attr_name, None)
        if ramp is None: return None
        if not isinstance(ramp, bpy.types.ColorRamp): return None
        return Serializers.serialize(ramp, cls.ramp_copy_attrs)
    
    @classmethod
    def ramp_props_paste(cls, node, data, attr_name="color_ramp"):
        if data is None: return
        ramp = getattr(node, attr_name, None)
        if ramp is None: return
        if not isinstance(ramp, bpy.types.ColorRamp): return
        Serializers.deserialize(ramp, cls.ramp_copy_attrs, data)
    
    @classmethod
    def img_props_copy(cls, node):
        if not node.image: return {}
        data = Serializers.serialize(node, cls.img_copy_attrs)
        data["name"] = node.image.name
        return data
    
    @classmethod
    def img_props_paste(cls, node, data):
        if data is None: return
        # Empty/unassigned image should be pasted too
        node.image = bpy.data.images.get(data.get("name", ""))
        if node.image: node.mute = False
        Serializers.deserialize(node, cls.img_copy_attrs, data)
    
    @classmethod
    def sockets_data_copy(cls, node, socket_category, socket_ids):
        sockets = getattr(node, socket_category, None)
        if not sockets: return {}
        
        # Inputs/outputs aren't indexed by identifier
        sockets = {socket.identifier: socket for socket in sockets}
        
        data = {}
        
        for name, label in socket_ids.items():
            socket = sockets.get(name)
            if not socket: continue
            value = getattr(socket, "default_value", None)
            if value is None: continue
            
            if not isinstance(value, cls.basic_types): value = tuple(value)
            data[label] = value
        
        return data
    
    @classmethod
    def sockets_data_paste(cls, node, socket_category, socket_ids, data):
        sockets = getattr(node, socket_category, None)
        if not sockets: return
        
        if not data: return
        
        # Inputs/outputs aren't indexed by identifier
        sockets = {socket.identifier: socket for socket in sockets}
        
        sockets = {label: sockets.get(identifier) for identifier, label in socket_ids.items()}
        
        for label, value in data.items():
            socket = sockets.get(label)
            if not socket: continue
            if not hasattr(socket, "default_value"): continue
            
            try:
                socket.default_value = value
            except Exception:
                pass
    
    @classmethod
    def mat_props_copy(cls, material):
        clipboard = {}
        
        if not (BlRna.is_valid(material) and material.node_tree): return clipboard
        
        parser = TextureFilenameParser(settings.loading_options.categories)
        
        for node in material.node_tree.nodes:
            node_info = cls._mat_props_node_info(node, parser)
            if node_info is None: continue
            
            if node.bl_idname == "ShaderNodeTexImage":
                node_data = clipboard.setdefault(node_info["key"], {})
                
                node_data["image"] = cls.img_props_copy(node)
            else:
                node_data = clipboard.setdefault(node_info["key"], {})
                
                inputs = cls.sockets_data_copy(node, "inputs", node_info["input_ids"])
                node_data.setdefault("inputs", {}).update(inputs)
                
                outputs = cls.sockets_data_copy(node, "outputs", node_info["output_ids"])
                node_data.setdefault("outputs", {}).update(outputs)
                
                if not node_data.get("color_ramp"):
                    node_data["color_ramp"] = cls.ramp_props_copy(node)
                
                if not node_data.get("mapping"):
                    node_data["mapping"] = cls.curves_props_copy(node)
        
        return clipboard
    
    @classmethod
    def mat_props_paste(cls, material, clipboard, all_images=False):
        if not clipboard: return
        
        if not (BlRna.is_valid(material) and material.node_tree): return
        
        parser = TextureFilenameParser(settings.loading_options.categories)
        
        for node in material.node_tree.nodes:
            node_info = cls._mat_props_node_info(node, parser, all_images)
            if node_info is None: continue
            
            if node.bl_idname == "ShaderNodeTexImage":
                node_data = clipboard.get(node_info["key"])
                if not node_data: continue
                
                cls.img_props_paste(node, node_data.get("image"))
            else:
                node_data = clipboard.get(node_info["key"])
                if not node_data: continue
                
                cls.sockets_data_paste(node, "inputs", node_info["input_ids"], node_data.get("inputs"))
                cls.sockets_data_paste(node, "outputs", node_info["output_ids"], node_data.get("outputs"))
                
                cls.ramp_props_paste(node, node_data.get("color_ramp"))
                
                cls.curves_props_paste(node, node_data.get("mapping"))
    
    @classmethod
    def _mat_props_node_info(cls, node, parser, all_images=True):
        info = cls.get(node)
        
        if node.bl_idname == "ShaderNodeTexImage":
            if not info:
                if not all_images: return None
                
                if MaterialLoader.height_to_normal in node.name.lower():
                    key = MaterialLoader.height_to_normal
                else:
                    category = node.label.strip().lower() or "color"
                    category_ids = parser.find(category)
                    if category_ids: category = parser.categories[category_ids[0]]
                    key = "image:" + category
            else:
                key = (info["group"], info["id"])
            
            return dict(key=key)
        else:
            if not info: return None
            
            key = (info["group"], info["id"])
            
            input_ids = cls.get_socket_ids(node.inputs, info)
            output_ids = cls.get_socket_ids(node.outputs, info)
            
            return dict(key=key, input_ids=input_ids, output_ids=output_ids)
    
    @classmethod
    def image_to_clipboard(cls, image, alpha=1.0, image_user=None):
        clipboard = {}
        
        if not BlRna.is_valid(image): return clipboard
        
        parser = TextureFilenameParser(settings.loading_options.categories)
        
        category = "color"
        category_ids = parser.find(category)
        if category_ids: category = parser.categories[category_ids[0]]
        key = "image:" + category
        
        image_user_data = {}
        if image_user:
            image_user_data = Serializers.serialize(image_user, cls.img_copy_attrs["image_user"])
        
        clipboard[key] = {"image": {"name": image.name, "image_user": image_user_data}}
        
        clipboard[("", "Alpha")] = {"inputs": {"Alpha": alpha}}
        
        return clipboard
