# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import itertools
import os

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import gpu

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "utils_gl": "cgl",
    "utils_ui": "NestedLayout, BlUI",
    "utils_addon": "AddonManager",
})

from . import icon_geometry
from . import common
from . import material_utils
from . import decal_utils

from .icon_geometry import get_palette_icon_image
from .common import Jobs, CommonUtils
from .material_utils import MaterialUtils
from .decal_utils import DecalUtils, Decal

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

bpy_Material = bpy.types.Material
bpy_Object = bpy.types.Object

class PaletteItem:
    __slots__ = ["template", "name", "type", "preview", "icon_value",
        "selected", "marked", "users", "enabled", "size", "is_file"]
    
    @property
    def is_dir(self):
        return self.is_file == False
    
    def __init__(self, template=None, preview=None, name="", type="",
                 selected=False, marked=False, users=0, enabled=True, size=None, is_file=None):
        if template:
            self.template = template
            self.name = template.name
            self.type = template.rna_type.name
        else:
            self.template = None
            self.name = name
            self.type = type
        
        if preview:
            self.preview = preview
            self.icon_value = preview.icon_id
        else:
            self.preview = None
            self.icon_value = 0
        
        self.selected = selected
        self.marked = marked
        self.users = users
        self.enabled = enabled
        self.size = size
        self.is_file = is_file
    
    def instantiate(self, context=None):
        if self.is_dir: return None
        
        if self.is_file:
            result = settings.files.open(self, push_undo=False)
            if result != 'LOAD': return None
        else:
            result = bpy.ops.xdecal.instantiate(name=self.name, type=self.type)
            if 'FINISHED' not in result: return None
        
        if not context: context = bpy.context
        return DecalUtils.filter_decals(context.selected_objects, result=[])

class Palette:
    _decals_items = None
    _decals_items_filtered = {}
    _decals_items_selected = set()
    _files_items = None
    _images = {}
    
    _placeholder_pixels = None
    
    _dir_geom = None
    _dir_up_geom = None
    
    @classmethod
    def _find_template_items(cls, context):
        templates = DecalUtils.find_all_templates()
        if not templates: return
        
        if cls._decals_items is not None:
            cls._decals_items_selected = {item.name for item in cls._decals_items if item.selected}
        
        selected_names = cls._decals_items_selected
        
        selected_templates = set()
        for obj in context.selected_objects:
            decal = Decal(obj)
            if decal: selected_templates.add(decal.template)
        
        for template, users in templates.items():
            preview = template.preview
            if isinstance(template, bpy_Material):
                size = MaterialUtils.estimate_image_size(template)
            elif isinstance(template, bpy_Object):
                size = tuple(preview.image_size) if preview else None
            else:
                continue
            
            yield PaletteItem(template=template, preview=preview,
                marked=(template in selected_templates), users=users,
                selected=(template.name in selected_names), size=size)
    
    @classmethod
    def _find_file_items(cls):
        # Postpone icon initialization to avoid "GPU API is not available in background mode" error
        
        if cls._dir_geom is None:
            cls._dir_geom = get_palette_icon_image("browser_folder")
        
        if cls._dir_up_geom is None:
            cls._dir_up_geom = get_palette_icon_image("browser_parent")
        
        for item in settings.files.items:
            if item.is_dir:
                item.preview = (cls._dir_up_geom if item.name == ".." else cls._dir_geom)
                item.size = (1, 1)
            yield item
    
    @classmethod
    def _item_image(cls, item):
        return cls._images.get(item.icon_value) or (item.preview, None, None)
    
    @classmethod
    def _update_items(cls, context):
        # file items are already sorted by (dir/file, name.lower())
        cls._files_items = list(cls._find_file_items())
        
        cls._decals_items = list(cls._find_template_items(context))
        cls._decals_items.sort(key=(lambda item: item.name))
        cls._decals_items_filtered = {}
        
        all_items = itertools.chain(cls._files_items, cls._decals_items)
        
        cls._images = {item.icon_value: cls._item_image(item)
            for item in all_items if item.icon_value != 0}
    
    @classmethod
    def _get_filter_key(cls, palette_type):
        name_filter = settings.decals.name_filter.lower()
        filter_implicit = settings.decals.filter_implicit_get(palette_type)
        filter_explicit = settings.decals.filter_explicit_get(palette_type)
        filter_scene = settings.decals.filter_scene_get(palette_type)
        filter_objects = settings.decals.filter_objects_get(palette_type)
        filter_materials = settings.decals.filter_materials_get(palette_type)
        return (name_filter, filter_implicit, filter_explicit, filter_scene, filter_objects, filter_materials)
    
    @classmethod
    def _update_filtered_items(cls, palette_type, filter_key=None):
        if not filter_key: filter_key = cls._get_filter_key(palette_type)
        items_filtered = list(cls.filtered(cls._decals_items, filter_key))
        cls._decals_items_filtered[palette_type] = (filter_key, items_filtered)
    
    @classmethod
    def filtered(cls, items, filter_key):
        if isinstance(filter_key, str): filter_key = cls._get_filter_key(filter_key)
        name_filter, filter_implicit, filter_explicit, filter_scene, filter_objects, filter_materials = filter_key
        
        scene_templates = (DecalUtils.find_scene_templates() if filter_scene else None)
        
        for item in cls._decals_items:
            if name_filter and (name_filter not in item.name.lower()): continue
            
            check_implicit = filter_implicit and (item.users > 0)
            check_explicit = filter_explicit and CommonUtils.is_template(item.template)
            if not (check_implicit or check_explicit): continue
            
            if filter_scene and (item.template not in scene_templates): continue
            
            if filter_materials and isinstance(item.template, bpy_Material):
                yield item
            elif filter_objects and isinstance(item.template, bpy_Object):
                yield item
    
    @classmethod
    def invalidate(cls, reset_selection=False, reset_images=True):
        if reset_selection:
            cls._decals_items_selected = set()
        elif cls._decals_items is not None:
            cls._decals_items_selected = {item.name for item in cls._decals_items if item.selected}
        
        cls._decals_items = None
        cls._decals_items_filtered = {}
        cls._files_items = None
        
        if reset_images: cls._images = {}
    
    _which_map = {
        "decals": (False, None),
        "decals-ui": (False, "ui"),
        "decals-3d": (False, "3d"),
        "files": (True, None),
        "files-ui": (True, "ui"),
        "files-3d": (True, "3d"),
    }
    
    @classmethod
    def _parse_which(cls, which):
        value = cls._which_map.get(which, None)
        if value is not None: return value
        return (settings.files_show_get(which), which)
    
    @classmethod
    def get_items(cls, which, context=None, force_update=False):
        force_update |= (cls._decals_items is None)
        force_update |= (cls._files_items is None)
        if force_update: cls._update_items(context or bpy.context)
        
        is_files, palette_type = cls._parse_which(which)
        return (cls._files_items if is_files else cls._decals_items)
    
    @classmethod
    def filter_items(cls, which, context=None, force_update=False):
        is_files, palette_type = cls._parse_which(which)
        items = cls.get_items(which, context, force_update)
        
        if is_files: return items # file items are already filtered
        
        filtered_key = cls._get_filter_key(palette_type)
        filtered_kv = cls._decals_items_filtered.get(palette_type, None)
        
        if (not filtered_kv) or (filtered_key != filtered_kv[0]):
            cls._update_filtered_items(palette_type, filtered_key)
            filtered_kv = cls._decals_items_filtered[palette_type]
        
        return filtered_kv[1]
    
    @classmethod
    def get_image(cls, id, force_update=False):
        "Returns ((width, height), pixels)"
        
        if isinstance(id, PaletteItem): id = id.icon_value
        
        images = cls._images
        
        image = images.get(id)
        if not image: return ((0, 0), None)
        
        if not cls._placeholder_pixels:
            cls._placeholder_pixels = gpu.types.Buffer('INT', 1, [0x7F7F7F7F])
        
        preview, size, pixels = image
        
        if isinstance(preview, tuple): return preview
        if isinstance(preview, dict): return ((1, 1), preview) # geometry
        
        if not preview:
            if size or pixels: images[id] = (preview, None, None)
            return ((1, 1), cls._placeholder_pixels)
        
        preview_size = tuple(preview.image_size)
        
        if (not pixels) or (preview_size != size) or force_update:
            size = preview_size
            if (not size) or (size[0] <= 0) or (size[1] <= 0):
                return ((1, 1), cls._placeholder_pixels)
            pixels = preview.image_pixels[:]
            pixels = gpu.types.Buffer('INT', size[0] * size[1], pixels)
            images[id] = (preview, tuple(size), pixels)
        
        return (size, pixels)
    
    @classmethod
    def draw_context_menu_decals(cls, layout, item_name, item_type, palette_type="ui"):
        BlUI.set_ui_metadata(layout, {"item_name":item_name, "item_type":item_type, "palette_type": palette_type})
        
        def draw_op(idname):
            op = layout.operator(idname)
            op.name = item_name
            op.type = item_type
        
        layout.menu(TemplateVisibilityMenu.bl_idname)
        
        draw_op("xdecal.instantiate")
        draw_op("xdecal.mat_props_copy_palette")
        draw_op("xdecal.mat_props_paste_palette")
        draw_op("xdecal.select_instances")
        draw_op("xdecal.select_template")
        draw_op("xdecal.update_icon")
        draw_op("xdecal.update_instances")
        draw_op("xdecal.replace_decals")
        draw_op("xdecal.replace_materials")
        
        layout.menu(TemplateRemoveMenu.bl_idname)
    
    @classmethod
    def draw_context_menu_files(cls, layout, item_name, palette_type="ui"):
        BlUI.set_ui_metadata(layout, {"palette_type": palette_type})
        
        def draw_op(idname):
            op = layout.operator(idname)
            op.name = item_name
        
        dir_path, files = settings.files.get_valid_names_selected(item_name)
        
        row = layout.row()
        row.enabled = bool(dir_path and files)
        row.operator_context = 'EXEC_REGION_WIN'
        op = row.operator("xdecal.load", text="Instantiate")
        op.directory = dir_path
        op.push_undo = True
        for name in files:
            files_item = op.files.add()
            files_item.name = name
        
        # Select Related can operate on multiple selections
        layout.operator("xdecal.select_related_files")
        
        layout.operator("xdecal.replace_decals_with_files")
        layout.operator("xdecal.replace_images_with_files")
    
    @staticmethod
    def multiprop(cls, name_base, prop_data):
        "Define a multi-palette property"
        
        annotations = getattr(cls, "__annotations__", None)
        if annotations is None: cls.__annotations__ = annotations = {}
        
        name_ui = name_base+"_ui"
        name_3d = name_base+"_3d"
        name_default = name_ui
        
        annotations[name_ui] = prop_data
        annotations[name_3d] = prop_data
        
        def get_args(self, palette_type):
            name = (f"{name_base}_{palette_type}" if settings.split_palette_options else name_default)
            return self, name
        
        def get_value(self, palette_type):
            if not settings.split_palette_options: return getattr(self, name_default)
            if palette_type == "any": return getattr(self, name_ui) or getattr(self, name_3d)
            if palette_type == "all": return getattr(self, name_ui) and getattr(self, name_3d)
            return getattr(self, f"{name_base}_{palette_type}")
        
        def set_value(self, palette_type, value):
            if palette_type in ("any", "all"):
                setattr(self, name_ui, value)
                setattr(self, name_3d, value)
            else:
                name = (f"{name_base}_{palette_type}" if settings.split_palette_options else name_default)
                setattr(self, name, value)
        
        setattr(cls, name_base+"_args", get_args)
        setattr(cls, name_base+"_get", get_value)
        setattr(cls, name_base+"_set", set_value)

@addon.Menu(idname="SCENE_MT_xdecal_template_visibility", label="Enable/Disable", description="Enable/disable decals")
class TemplateVisibilityMenu:
    def draw(self, context):
        layout = self.layout
        
        metadata = BlUI.get_ui_metadata(context)
        item_name = metadata.get("item_name", "")
        item_type = metadata.get("item_type", "")
        
        def draw_op(prefix, mode, text, icon):
            idname = f"xdecal.{prefix}_decals"
            op = layout.operator(idname, text=text, icon=icon)
            op.name = item_name
            op.type = item_type
            op.mode = mode
            op.undo_label = "Decals: "+text
        
        draw_op("enable", 'SELECT', "Enable Selection", 'RESTRICT_SELECT_OFF')
        draw_op("enable", 'VIEW_LAYER', "Show in View Layer", 'HIDE_OFF')
        draw_op("enable", 'VIEWPORT', "Enable in Viewport", 'RESTRICT_VIEW_OFF')
        draw_op("enable", 'RENDER', "Enable in Rendering", 'RESTRICT_RENDER_OFF')
        
        layout.separator()
        
        draw_op("disable", 'SELECT', "Disable Selection", 'RESTRICT_SELECT_ON')
        draw_op("disable", 'VIEW_LAYER', "Hide in View Layer", 'HIDE_ON')
        draw_op("disable", 'VIEWPORT', "Disable in Viewport", 'RESTRICT_VIEW_ON')
        draw_op("disable", 'RENDER', "Disable in Rendering", 'RESTRICT_RENDER_ON')

@addon.Menu(idname="SCENE_MT_xdecal_template_remove", label="Remove", description="Remove/delete decals")
class TemplateRemoveMenu:
    def draw(self, context):
        layout = self.layout
        
        metadata = BlUI.get_ui_metadata(context)
        item_name = metadata.get("item_name", "")
        item_type = metadata.get("item_type", "")
        
        def draw_op(idname):
            op = layout.operator(idname)
            op.name = item_name
            op.type = item_type
        
        draw_op("xdecal.restore_to_palette")
        draw_op("xdecal.remove_from_palette")
        draw_op("xdecal.delete_decals")
        draw_op("xdecal.delete_template")
