# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import time

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import numpy as np

import gpu

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "prop",
    "bounds": "Bounds",
    "utils_arrangement": "Arranger",
    "utils_gl": "cgl",
    "utils_ui": "BlUI",
    "utils_addon": "AddonManager",
})

from . import icon_geometry
from . import common
from . import decal_gizmos
from . import palette
from . import palette_options
from . import palette_ui

from .icon_geometry import get_palette_icon_image
from .common import Jobs, CommonUtils, CommonConfig
from .material_utils import MaterialLibrary
from .decal_gizmos import decal_gizmo_state, DecalTransformBase, DecalPositionGizmo
from .palette import Palette
from .palette_options import DirectoryPreset
from .palette_ui import PaletteUI

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

@addon.PropertyGroup
class PaletteGizmoOptions:
    limit_sides: {'LEFT', 'RIGHT', 'BOTTOM', 'TOP'} | prop("Limit by non-floating regions", items=[
        ('LEFT', "Left", "Left"),
        ('BOTTOM', "Bottom", "Bottom"),
        ('RIGHT', "Right", "Right"),
        ('TOP', "Top", "Top"),
    ])
    limit_float: 'LEFT' | prop("Limit by floating regions", items=[
        ('NONE', "None", "None"),
        ('LEFT', "Left", "Left"),
        ('BOTTOM', "Bottom", "Bottom"),
        ('RIGHT', "Right", "Right"),
        ('TOP', "Top", "Top"),
    ])
    
    # xmin, ymin, xmax, ymax
    anchors: (0.0, 0.0, 1.0, 0.0) | prop("Anchors", min=0.0, max=1.0, options={'TEXTEDIT_UPDATE'})
    offsets: (16.0, 16.0, -16.0, 128.0) | prop("Offsets", options={'TEXTEDIT_UPDATE'})
    
    auto_hide: False | prop("Auto-hide", "Minimize the palette when mouse is outside",
        update=(lambda self, context: BlUI.tag_redraw()))

@addon.Gizmo(idname = "VIEW3D_GT_xdecal_palette")
class DecalPaletteGizmo:
    shader_color = None
    shader_image = None
    
    active_rv3d = None
    
    minimized_rect = Bounds.zero(2)
    panel_rect = Bounds.zero(2)
    interior_rect = Bounds.zero(2)
    scroll_area_rect = Bounds.zero(2)
    scroll_handle_rect = Bounds.zero(2)
    cells_rect = Bounds.zero(2)
    edge_xmin_rect = Bounds.zero(2)
    edge_ymin_rect = Bounds.zero(2)
    edge_xmax_rect = Bounds.zero(2)
    edge_ymax_rect = Bounds.zero(2)
    toolbar_rect = Bounds.zero(2)
    toolbar_bkg_rect = Bounds.zero(2)
    panel_bkg_rect = Bounds.zero(2)
    visible_cells = []
    widgets = {}
    
    cell_size = 0.0
    cell_step = 0.0
    scroll_max = 0.0
    active_cell_index = -1
    real_cell_size = 0.0
    active_widget_type = None
    pressed_widget_type = None
    edges_hover = {}
    
    sticky_auto_visible = True
    was_in_panel = False
    
    items = None
    grid_info = None
    
    mouse_position = None
    
    action = None
    
    icon_textures = {}
    
    @classmethod
    def reset(cls):
        cls.active_rv3d = None
        
        cls.scroll_max = 0.0
        cls.active_cell_index = -1
        cls.active_widget_type = None
        cls.pressed_widget_type = None
        cls.edges_hover = {}
        
        cls.items = None
        cls.grid_info = None
        
        cls.minimized_rect = Bounds.zero(2)
        cls.panel_rect = Bounds.zero(2)
        cls.interior_rect = Bounds.zero(2)
        cls.scroll_area_rect = Bounds.zero(2)
        cls.scroll_handle_rect = Bounds.zero(2)
        cls.cells_rect = Bounds.zero(2)
        cls.edge_xmin_rect = Bounds.zero(2)
        cls.edge_ymin_rect = Bounds.zero(2)
        cls.edge_xmax_rect = Bounds.zero(2)
        cls.edge_ymax_rect = Bounds.zero(2)
        cls.toolbar_rect = Bounds.zero(2)
        cls.toolbar_bkg_rect = Bounds.zero(2)
        cls.panel_bkg_rect = Bounds.zero(2)
        cls.visible_cells = []
        cls.widgets = {}
        
        cls.mouse_position = None
        
        cls.clear_icon_textures()
    
    @classmethod
    def tag_redraw(cls, target=None):
        BlUI.tag_redraw(target)
    
    @classmethod
    def bind_icon_texture(cls, shader, name, icon_id, image, force_update=False, mipmap=False):
        tex = cls.icon_textures.get(icon_id, None)
        
        if (tex is None) or force_update:
            image_size, pixels = image
            width, height = image_size
            
            if tex is None:
                tex = cgl.rgba8_texture(image_size, pixels)
                cls.icon_textures[icon_id] = tex
            
            # TODO: Set wrap mode to CLAMP
            # TODO: Set min/mag filter to LINEAR (or the corresponding mipmap version)
            # TODO: Generate mipmap if requested
            # if mipmap:
            #     for mip_level, buf in cls.generate_mipmap(width, height, pixels):
            #         pass # TODO: assign mipmap level
        
        shader.bind()
        shader.uniform_sampler(name, tex)
        
        return tex
    
    @classmethod
    def generate_mipmap(cls, width, height, pixels):
        data = np.array(np.reshape(pixels, (height, width, 1)).view("uint8"), dtype="uint16")
        mip_w, mip_h = width, height
        
        mip_level = 0
        while max(mip_w, mip_h) > 1:
            mip_level += 1
            sub_w = max(mip_w >> 1, 1)
            sub_h = max(mip_h >> 1, 1)
            mip_w = min(mip_w, sub_w*2)
            mip_h = min(mip_h, sub_h*2)
            mip_x = sub_w > 1
            mip_y = sub_h > 1
            mip_factor = int(mip_x) + int(mip_y)
            
            mip = data[:mip_h:2, :mip_w:2, :]
            if mip_x: mip += data[:mip_h:2, 1:mip_w:2, :]
            if mip_y: mip += data[1:mip_h:2, :mip_w:2, :]
            if mip_x and mip_y: mip += data[1:mip_h:2, 1:mip_w:2, :]
            mip += mip_factor
            mip >>= mip_factor
            
            mip = mip.astype("uint8")
            data[:sub_h, :sub_w, :] = mip # copy the result back, to use in the next iteration
            mip = mip.view("int32")
            mip.shape = mip.size # flatten (Note: len() is just the first dimension)
            
            buf = gpu.types.Buffer('INT', mip.size, mip)
            yield mip_level, buf
            
            mip_w, mip_h = sub_w, sub_h
    
    @classmethod
    def remove_icon_texture(cls, icon_id):
        tex = cls.icon_textures.get(icon_id, None)
        if tex is None: return
        # gpu.types.GPUTexture is garbage-collected
        del cls.icon_textures[icon_id]
    
    @classmethod
    def clear_icon_textures(cls, unused_only=False):
        icon_ids = set(cls.icon_textures.keys())
        
        if unused_only:
            for cell in cls.visible_cells:
                icon_ids.discard(cell[0].icon_value)
            
            for widget in cls.widgets.values():
                widget_icon, widget_rect = widget
                if not widget_icon: continue
                icon_ids.discard(widget_icon)
        
        for icon_id in icon_ids:
            # gpu.types.GPUTexture is garbage-collected
            del cls.icon_textures[icon_id]
    
    @classmethod
    def get_active_widget(cls):
        return cls.widgets.get(cls.active_widget_type, None)
    
    @classmethod
    def get_pressed_widget(cls):
        return cls.widgets.get(cls.pressed_widget_type, None)
    
    @classmethod
    def get_active_cell(cls):
        index = cls.active_cell_index
        cells = cls.visible_cells
        return (cells[index] if (index >= 0) and (index < len(cells)) else None)
    
    @classmethod
    def is_in_rect(cls, point, rect):
        if not (point and rect): return None
        return rect.contains(point, end_max=False)
    
    @classmethod
    def is_active(cls, context):
        rv3d = context.region_data
        return bool(cls.active_rv3d) and (cls.active_rv3d == rv3d)
    
    @classmethod
    def is_visible(cls):
        return cls.panel_rect.size.min() > 0
    
    @classmethod
    def scroll_factors(cls, total, window, scroll, scroll_min=0.0):
        if not ((total > 0.0) and (window > 0.0)): return (1.0, 1.0, 0.0)
        if not (window < total): return (1.0, 1.0, 0.0)
        visible = window/total
        size_factor = min(max(window/total, scroll_min/window), 1.0)
        if size_factor == 1.0: return (visible, size_factor, 0.0)
        pos_factor = min(max(scroll, 0.0) / (total * (1.0 - size_factor)), 1.0)
        return visible, size_factor, pos_factor
    
    @classmethod
    def scroll_get(cls, clamp=True):
        ui_scale = BlUI.ui_scale()
        cell_step = (BlUI.ICON_SIZE * settings.icon_scale_get("3d") + settings.palette_padding) * ui_scale
        options = getattr(settings, "decals" if settings.files_show_get("3d") else "files")
        value = options.gizmo_scroll * cell_step
        if clamp: value = min(max(value, 0.0), cls.scroll_max)
        return value
    
    @classmethod
    def scroll_set(cls, value, clamp=True):
        ui_scale = BlUI.ui_scale()
        cell_step = (BlUI.ICON_SIZE * settings.icon_scale_get("3d") + settings.palette_padding) * ui_scale
        if clamp: value = min(max(value, 0.0), cls.scroll_max)
        options = getattr(settings, "decals" if settings.files_show_get("3d") else "files")
        options.gizmo_scroll = value / cell_step
    
    @classmethod
    def update_layout(cls, context, mouse_position=None):
        area = context.area
        region = context.region
        rv3d = context.region_data
        
        ui_scale = BlUI.ui_scale(context)
        
        cell_padding = settings.palette_padding * ui_scale
        scroll_size = settings.palette_scroll_size * ui_scale
        scroll_min = scroll_size
        
        cls.cell_size = BlUI.ICON_SIZE * settings.icon_scale_get("3d") * ui_scale
        cls.cell_step = cls.cell_size + cell_padding
        
        toolbar_size = BlUI.ICON_SIZE * settings.toolbar_scale * ui_scale
        
        anchor_options = settings.palette
        limit_float = anchor_options.limit_float
        limit_sides = anchor_options.limit_sides
        side_anchors = Vector(anchor_options.anchors)
        side_offsets = Vector(anchor_options.offsets) * ui_scale
        
        if mouse_position: cls.mouse_position = Vector(mouse_position)
        
        region_convert = {
            'FLOAT':limit_float,
            'LEFT':('LEFT' if 'LEFT' in limit_sides else 'NONE'),
            'BOTTOM':('BOTTOM' if 'BOTTOM' in limit_sides else 'NONE'),
            'RIGHT':('RIGHT' if 'RIGHT' in limit_sides else 'NONE'),
            'TOP':('TOP' if 'TOP' in limit_sides else 'NONE'),
        }
        safe_rect = BlUI.calc_region_rect(area, region, False, convert=region_convert)
        # Note: region rect is in window coordinates, but draw() and test_select()
        # expect the (0, 0) to be at the region's lower left corner
        safe_rect.move((-region.x, -region.y))
        
        panel_min = safe_rect.normalized_to_point(side_anchors[:2]) + side_offsets[:2]
        panel_max = safe_rect.normalized_to_point(side_anchors[2:]) + side_offsets[2:]
        
        cls.panel_rect = Bounds.MinMax(panel_min, panel_max)
        
        # minimized_size = np.minimum(cls.panel_rect.size, toolbar_size)
        # minimized_offset = (cls.panel_rect.size - minimized_size) * np.mean([side_anchors[:2], side_anchors[2:]], axis=0)
        # cls.minimized_rect = Bounds.MinSize(cls.panel_rect.min + minimized_offset, minimized_size)
        
        #cls.minimized_rect = Bounds.MinSize((safe_rect.center[0] - toolbar_size*0.5, 0), (toolbar_size, toolbar_size))
        cls.minimized_rect = Bounds.MinSize((region.width*0.5 - toolbar_size*0.5, 0), (toolbar_size, toolbar_size))
        
        cls.edge_xmin_rect = cls.panel_rect.split_min("x", cell_padding, clamp=True)[0]
        cls.edge_ymin_rect = cls.panel_rect.split_min("y", cell_padding, clamp=True)[0]
        cls.edge_xmax_rect = cls.panel_rect.split_max("x", cell_padding, clamp=True)[1]
        cls.edge_ymax_rect = cls.panel_rect.split_max("y", cell_padding, clamp=True)[1]
        
        cls.interior_rect = cls.panel_rect.shrunk(cell_padding)
        interior_rect = cls.interior_rect.copy()
        
        cls.scroll_area_rect, interior_rect = interior_rect.split_min("y", scroll_size, clamp=True)
        interior_rect = interior_rect.split_min("y", cell_padding, clamp=True)[1]
        
        interior_rect, cls.toolbar_rect = interior_rect.split_max("y", toolbar_size, clamp=True)
        interior_rect = interior_rect.split_max("y", cell_padding, clamp=True)[0]
        
        cls.panel_bkg_rect, cls.toolbar_bkg_rect = cls.panel_rect.split_max("y", toolbar_size + cell_padding*2, clamp=True)
        
        cls.update_widgets(cell_padding)
        
        cls.update_cells(interior_rect, cell_padding, scroll_min)
        
        cls.detect_active_widget(mouse_position, rv3d, region)
        
        in_panel = cls.is_in_rect(mouse_position, cls.panel_rect)
        in_corner = cls.is_in_rect(mouse_position, cls.minimized_rect)
        
        cls.update_sticky_visibility(in_panel, in_corner)
        
        if settings.palette.auto_hide:
            if settings.auto_hide_mode == 'STICKY':
                return (in_panel if cls.sticky_auto_visible else in_corner)
            else:
                return in_panel or in_corner
        
        return in_panel
    
    @classmethod
    def update_sticky_visibility(cls, in_panel, in_corner):
        if (not settings.palette.auto_hide) or (settings.auto_hide_mode != 'STICKY'):
            cls.sticky_auto_visible = True
            cls.was_in_panel = False
            return
        
        if (in_panel is None) or (in_corner is None):
            return
        
        sticky_auto_visible = cls.sticky_auto_visible
        
        if cls.sticky_auto_visible:
            if cls.was_in_panel:
                cls.sticky_auto_visible &= in_panel
        else:
            cls.sticky_auto_visible |= in_corner
        
        cls.was_in_panel = in_panel
        
        if cls.sticky_auto_visible != sticky_auto_visible:
            BlUI.tag_redraw()
    
    @classmethod
    def detect_active_widget(cls, mouse_position, rv3d, region):
        if not mouse_position: return
        
        prev_edges_hover = cls.edges_hover
        cls.edges_hover = {
            "xmin":cls.edge_xmin_rect.contains(mouse_position),
            "ymin":cls.edge_ymin_rect.contains(mouse_position),
            "xmax":cls.edge_xmax_rect.contains(mouse_position),
            "ymax":cls.edge_ymax_rect.contains(mouse_position),
        }
        edges_changed = (cls.edges_hover != prev_edges_hover)
        
        prev_active_widget_type = cls.active_widget_type
        cls.active_widget_type = None
        
        prev_active_cell_index = cls.active_cell_index
        cls.active_cell_index = -1
        active_icon_id = 0
        
        if (not cls.active_rv3d) or (cls.active_rv3d != rv3d):
            cls.active_rv3d = rv3d
            cls.tag_redraw() # we need to update other regions as well
        
        if cls.is_in_rect(mouse_position, cls.toolbar_rect):
            for widget_type, widget in cls.widgets.items():
                widget_icon, widget_rect = widget
                if cls.is_in_rect(mouse_position, widget_rect):
                    cls.active_widget_type = widget_type
                    break
        elif cls.is_in_rect(mouse_position, cls.cells_rect):
            for i, cell in enumerate(cls.visible_cells):
                item, rect = cell
                if cls.is_in_rect(mouse_position, rect):
                    cls.active_cell_index = i
                    active_icon_id = item.icon_value
                    break
        
        cell_changed = (cls.active_cell_index != prev_active_cell_index)
        widget_changed = (cls.active_widget_type != prev_active_widget_type)
        
        if cell_changed or widget_changed or edges_changed: cls.tag_redraw(region)
    
    widget_tooltips = {
        "close": "Close the palette",
        "pin": "Toggle auto-hide",
        "anchors": "Palette anchoring",
        "options": "Show all options",
        "files": "Switch between palette and file browser modes",
        "inherit": "Inherit properties",
        "template_material": "The material on which new decals will be based",
        "directory_preset": "Directory Presets",
        "directory_refresh": "Refresh the file list",
        "directory": "Directory path",
        "filter_implicit": "Filter implicit templates",
        "filter_explicit": "Filter explicit templates",
        "filter_scene": "Filter scene templates",
        "add_objects": "Add selected objects to decal palette (if already in palette, update the preview)",
        "filter_objects": "Filter object-based decals",
        "add_materials": "Add selected objects' materials to decal palette",
        "filter_materials": "Filter material-based decals",
    }
    
    @classmethod
    def update_widgets(cls, cell_padding):
        cls.widgets.clear()
        
        toolbar_rect = cls.toolbar_rect
        icon_size = toolbar_rect.size[1]
        
        toolbar_rect, widget_rect = toolbar_rect.split_max("x", icon_size)
        if toolbar_rect.size[0] <= 0: return
        cls.widgets["close"] = ("close", widget_rect)
        
        toolbar_rect, widget_rect = toolbar_rect.split_max("x", icon_size)
        if toolbar_rect.size[0] <= 0: return
        icon = ("unpinned" if settings.palette.auto_hide else "pinned")
        cls.widgets["pin"] = (icon, widget_rect)
        
        toolbar_rect, widget_rect = toolbar_rect.split_max("x", icon_size)
        if toolbar_rect.size[0] <= 0: return
        cls.widgets["anchors"] = ("expand", widget_rect)
        
        widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
        if toolbar_rect.size[0] <= 0: return
        cls.widgets["options"] = ("wrench", widget_rect)
        
        widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
        if toolbar_rect.size[0] <= 0: return
        cls.widgets["files"] = ("folder", widget_rect)
        
        widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
        if toolbar_rect.size[0] <= 0: return
        cls.widgets["inherit"] = ("duplicate", widget_rect)
        
        if settings.files_show_get("3d"):
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size*6)
            widget_rect = widget_rect.split_min("x", cell_padding)[1].split_max("x", cell_padding)[0]
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["template_material"] = (None, widget_rect)
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["directory_preset"] = ("preset", widget_rect)
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["directory_refresh"] = ("refresh", widget_rect)
            
            toolbar_rect = toolbar_rect.split_min("x", cell_padding)[1].split_max("x", cell_padding)[0]
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["directory"] = (None, toolbar_rect)
        else:
            toolbar_rect = toolbar_rect.split_min("x", cell_padding)[1]
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["filter_implicit"] = ("ghost_disabled", widget_rect)
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["filter_explicit"] = ("ghost_enabled", widget_rect)
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["filter_scene"] = ("scene", widget_rect)
            
            toolbar_rect = toolbar_rect.split_min("x", cell_padding)[1]
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["add_materials"] = ("plus", widget_rect)
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["filter_materials"] = ("material", widget_rect)
            
            toolbar_rect = toolbar_rect.split_min("x", cell_padding)[1]
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["add_objects"] = ("plus", widget_rect)
            
            widget_rect, toolbar_rect = toolbar_rect.split_min("x", icon_size)
            if toolbar_rect.size[0] <= 0: return
            cls.widgets["filter_objects"] = ("object", widget_rect)
    
    @classmethod
    def update_cells(cls, cells_rect, cell_padding, scroll_min):
        cell_size = cls.cell_size
        cell_step = cls.cell_step
        
        scroll_size = cls.scroll_area_rect.size[1]
        toolbar_size = cls.toolbar_rect.size[1]
        
        items = Palette.filter_items("3d")
        
        cls.items = items
        cls.grid_info = None
        
        if (cells_rect.size.min() < 1) or (len(items) == 0):
            cls.scroll_handle_rect = cells_rect.resized((0, scroll_size), type="MinSize")
            cls.cells_rect = cells_rect.resized((0, 0), type="MinSize")
            cls.scroll_max = 0.0
            cls.real_cell_size = cell_size
            cls.visible_cells.clear()
            return
        
        scroll = max(cls.scroll_get(clamp=False), 0.0)
        
        cls.cells_rect = cells_rect
        
        cell_bbox = Bounds.MinSize(cls.cells_rect.min - (scroll, 0), (cell_step, cell_step))
        
        indices = Arranger.grid_indices(cls.cells_rect, cell_bbox,
            modes=['OVERLAP', 'SIZE_FLOOR'], min_count=[None, 1])
        
        items_size = Arranger.grid_size(indices, (cell_size, cell_size), padding=cell_padding,
            flat=(0, len(items)), sorting=[(1, -1), (0, 1)])
        
        _pos, _size = cls.scroll_area_rect.get("min", "size", convert=Vector)
        visible_factor, size_factor, pos_factor = cls.scroll_factors(items_size[0], _size[0], scroll, scroll_min)
        _pos[0] += pos_factor * _size[0] * (1.0 - size_factor)
        _size[0] *= size_factor
        cls.scroll_handle_rect = Bounds.MinSize(_pos, _size)
        
        cls.scroll_max = items_size[0] * (1.0 - visible_factor)
        
        if scroll > cls.scroll_max:
            scroll = cls.scroll_max
            
            cell_bbox = Bounds.MinSize(cls.cells_rect.min - (scroll, 0), (cell_step, cell_step))
            
            indices = Arranger.grid_indices(cls.cells_rect, cell_bbox,
                modes=['OVERLAP', 'SIZE_FLOOR'], min_count=[None, 1])
        
        cls.grid_info = Arranger.grid(cls.cells_rect, cell_bbox, indices,
            alignments=['ABS', 'CENTER'], flat=(0, len(items)), sorting=[(1, -1), (0, 1)])
        
        cell_size = min(cell_size, cls.cells_rect.size[1])
        cls.real_cell_size = cell_size
        
        cls.visible_cells.clear()
        for i, rect in cls.grid_info["iterator"]():
            cell = (items[i], rect.resized((cell_size, cell_size), type="CenterSize"))
            cls.visible_cells.append(cell)
    
    def setup(self):
        self.shader_color = gpu.shader.from_builtin('SMOOTH_COLOR')
        #self.shader_image = gpu.shader.from_builtin('IMAGE')
        
        self.shader_image = gpu.types.GPUShader('''
        uniform mat4 ModelViewProjectionMatrix;
        in vec2 texCoord;
        in vec3 pos;
        in vec4 color;
        out vec2 texCoord_interp;
        out vec4 color_interp;
        void main() {
            gl_Position = ModelViewProjectionMatrix * vec4(pos.xy, 0.0f, 1.0f);
            gl_Position.z = 1.0;
            texCoord_interp = texCoord;
            color_interp = color;
        }
        ''', '''
        uniform sampler2D image;
        in vec2 texCoord_interp;
        in vec4 color_interp;
        out vec4 fragColor;
        void main() {
            fragColor = texture(image, texCoord_interp) * color_interp;
            fragColor = blender_srgb_to_framebuffer_space(fragColor);
        }
        ''')
    
    def test_select(self, context, mouse_position):
        return (0 if self.update_layout(context, mouse_position) else -1)
    
    @classmethod
    def draw_rect(cls, shader, rect, color=None, uv=False, wire=False):
        if rect.size.min() <= 0: return
        
        pos, size = Vector(rect.min), Vector(rect.size)
        x0, y0 = pos
        x1, y1 = pos + size
        
        attrs = {"pos": [(x0, y0, 0), (x0, y1, 0), (x1, y1, 0), (x1, y0, 0)]}
        if uv: attrs["texCoord"] = [(0, 0), (0, 1), (1, 1), (1, 0)]
        if color: attrs["color"] = [color, color, color, color]
        
        primitive_type = ('LINE_LOOP' if wire else 'TRI_FAN')
        batch = cgl.batch(shader, primitive_type, **attrs)
        batch.draw(shader)
    
    @classmethod
    def draw_geometry(cls, shader, rect, geometry, color=None):
        if rect.size.min() <= 0: return
        
        pos, size = Vector(rect.min), Vector(rect.size)
        x0, y0 = pos
        x1, y1 = pos + size
        
        vmul = (lambda a, b: Vector((a[0]*b[0], a[1]*b[1])))
        
        positions = geometry["pos"]
        attrs = {"pos": [(pos + vmul(v, size)).to_3d() for v in positions]}
        
        if color:
            colors = geometry.get("color")
            if not colors: colors = (1, 1, 1, 1)
            
            tint = (lambda c: (c[0]*color[0], c[1]*color[1], c[2]*color[2], c[3]*color[3]))
            
            if isinstance(colors[0], (int, float)):
                color = tint(colors)
                attrs["color"] = [color for i in range(len(positions))]
            else:
                attrs["color"] = [tint(c) for c in colors]
        
        batch = cgl.batch(shader, 'TRIS', **attrs)
        batch.draw(shader)
    
    def draw(self, context):
        if not self.is_active(context): return
        
        self.update_layout(context) # in case the region is resized
        if not self.is_visible(): return
        
        colors = settings.colors
        
        drag_base = getattr(self, "drag_base", None)
        if drag_base:
            with gpu.matrix.push_pop_projection():
                perspective_matrix = context.region_data.perspective_matrix
                gpu.matrix.load_projection_matrix(perspective_matrix)
                DecalPositionGizmo.draw_shape(drag_base, selection=False, color=colors["decal_active"])
            return
        
        # TODO: adjust this code when antialiasing control
        # gets implemented in the gpu module
        
        # If we don't turn off GL_POLYGON_SMOOTH, blended triangles
        # will be rendered with gaps on GPUs that support smooth polygons
        # polygon_smooth = cgl.POLYGON_SMOOTH
        # cgl.POLYGON_SMOOTH = False
        
        prev_blend = cgl.blend
        cgl.blend = 'ALPHA'
        
        # At some point, I made an attempt to reduce the slowdown for other operators
        # (e.g. Mouselook Navigation) by drawing the panel into a separate texture,
        # but I couldn't figure out how to make it work (or if it would even help).
        # In the end, I removed all the corresponding logic.
        self.draw_main(context)
        
        cgl.blend = prev_blend
        # cgl.POLYGON_SMOOTH = polygon_smooth
    
    def draw_main(self, context):
        colors = settings.colors
        theme = colors.get_theme() # update theme reference
        
        # scroll_roundness = getattr(theme.user_interface.wcol_scroll, "roundness", 1.0)
        
        self.clear_icon_textures(unused_only=True)
        
        if settings.palette.auto_hide:
            in_panel = self.is_in_rect(self.mouse_position, self.panel_rect)
            in_corner = self.is_in_rect(self.mouse_position, self.minimized_rect)
            
            self.update_sticky_visibility(in_panel, in_corner)
            
            if settings.auto_hide_mode == 'STICKY':
                show_mini_icon = not self.sticky_auto_visible
                in_panel = self.sticky_auto_visible
                in_corner = False
            else:
                show_mini_icon = not in_panel
            
            if show_mini_icon:
                self.draw_rect(self.shader_color, self.minimized_rect, color=colors["palette_panel"])
                icon_id = "image"
                image = get_palette_icon_image(icon_id)
                self.draw_item_icon(icon_id, image, self.minimized_rect, mipmap=True)
            
            if not (in_panel or in_corner): return
        
        self.draw_rect(self.shader_color, self.panel_bkg_rect, color=colors["palette_panel"])
        
        self.draw_rect(self.shader_color, self.scroll_area_rect, color=colors["scroll_area"])
        self.draw_rect(self.shader_color, self.scroll_area_rect, color=colors["scroll_outline"], wire=True)
        
        self.draw_rect(self.shader_color, self.scroll_handle_rect, color=colors["scroll_handle"])
        self.draw_rect(self.shader_color, self.scroll_handle_rect, color=colors["scroll_outline"], wire=True)
        
        self.draw_toolbar(context)
        
        self.draw_items(context)
    
    def draw_toolbar(self, context):
        colors = settings.colors
        
        toolbar_bkg_color = colors["palette_panel"]
        toolbar_bkg_color[3] *= settings.toolbar_background_alpha
        self.draw_rect(self.shader_color, self.toolbar_bkg_rect, color=toolbar_bkg_color)
        
        widget_color = colors["toolbar_widget"]
        widget_color_on = colors["toolbar_toggle"]
        widget_pressed = colors["toolbar_active"]
        widget_highlight = Vector(widget_pressed)
        widget_highlight[3] *= 0.5
        widget_alert = (1.0, 0.0, 0.0, 0.5)
        
        #cls.interior_rect = cls.panel_rect
        for edge_key, edge_value in self.edges_hover.items():
            if not edge_value: continue
            rect = getattr(self, f"edge_{edge_key}_rect")
            self.draw_rect(self.shader_color, rect, color=widget_highlight)
        
        if not self.widgets: return # e.g. if not initialized
        
        ui_scale = BlUI.ui_scale(context)
        text_size = settings.palette_text_size * ui_scale
        text_size = min(text_size, self.toolbar_rect.size[1])
        cgl.text.size(text_size)
        
        text_offset = 5.0 * settings.toolbar_scale
        
        active_widget = self.get_active_widget()
        pressed_widget = self.get_pressed_widget()
        
        def draw_widget_toggle(widget_name):
            if widget_name not in self.widgets: return
            widget_icon, widget_rect = self.widgets[widget_name]
            self.draw_rect(self.shader_color, widget_rect, color=widget_color_on)
        
        def draw_widget_text(widget_name, text):
            if widget_name not in self.widgets: return
            widget_icon, widget_rect = self.widgets[widget_name]
            cgl.scissor = (*widget_rect.min, *widget_rect.size)
            cgl.text.draw(text, (widget_rect.min[0] + text_offset, widget_rect.center[1]), origin=(0, 0.5))
            #self.draw_rect(self.shader_color, widget_rect, color=widget_color, wire=True)
        
        # DRAW TOGGLE STATES
        if settings.use_inherit:
            draw_widget_toggle("inherit")
        
        if settings.files_show_get("3d"):
            draw_widget_toggle("files")
        else:
            if settings.decals.filter_implicit_get("3d"):
                draw_widget_toggle("filter_implicit")
            
            if settings.decals.filter_explicit_get("3d"):
                draw_widget_toggle("filter_explicit")
            
            if settings.decals.filter_scene_get("3d"):
                draw_widget_toggle("filter_scene")
            
            if settings.decals.filter_objects_get("3d"):
                draw_widget_toggle("filter_objects")
            
            if settings.decals.filter_materials_get("3d"):
                draw_widget_toggle("filter_materials")
        
        # DRAW HIGHLIGHTS
        if pressed_widget:
            self.draw_rect(self.shader_color, pressed_widget[1], color=widget_pressed)
        elif active_widget:
            self.draw_rect(self.shader_color, active_widget[1], color=widget_highlight)
            
            tooltip = self.widget_tooltips.get(self.active_widget_type, None)
            if tooltip:
                cgl.text.disable('SHADOW')
                cgl.text.color(*widget_color)
                
                cell_padding = settings.palette_padding * ui_scale
                tooltip_batched = cgl.text.compile(tooltip, width=self.toolbar_bkg_rect.size[0])
                tooltip_x_min = active_widget[1].min[0]
                tooltip_x_max = tooltip_x_min + tooltip_batched.size[0]
                delta_min = max(self.toolbar_bkg_rect.min[0] - tooltip_x_min, 0)
                delta_max = min(self.toolbar_bkg_rect.max[0] - tooltip_x_max, 0)
                tooltip_x = tooltip_x_min + delta_min + delta_max
                tooltip_y = self.toolbar_bkg_rect.max[1] + cell_padding*2
                tooltip_rect = Bounds.MinSize((tooltip_x, tooltip_y), tooltip_batched.size)
                self.draw_rect(self.shader_color, tooltip_rect.grown(cell_padding), color=toolbar_bkg_color)
                tooltip_batched.draw(tooltip_rect.min)
        
        if settings.files_show_get("3d"):
            active_path = settings.loading_options.template_material_path
            
            # Draw the alert box with the regular scissor test configuration
            if not MaterialLibrary.availability.get(active_path, True):
                widget_icon, widget_rect = self.widgets["template_material"]
                self.draw_rect(self.shader_color, widget_rect, color=widget_alert)
            
            # DRAW TEXT
            cgl.text.disable('SHADOW')
            cgl.text.color(*widget_color)
            scissor_box = cgl.scissor
            
            draw_widget_text("directory", settings.files.directory)
            
            library_path, material_name = CommonUtils.split(active_path)
            draw_widget_text("template_material", material_name)
            
            cgl.scissor = scissor_box
        
        # DRAW ICONS
        for widget in self.widgets.values():
            widget_icon, widget_rect = widget
            if not widget_icon: continue
            image = get_palette_icon_image(widget_icon)
            self.draw_item_icon(widget_icon, image, widget_rect, mipmap=True)
    
    def draw_items(self, context):
        colors = settings.colors
        
        # Essentially, each time Blender draws an icon, it creates a texture, copies pixel data, draws a quad, frees the texture
        # Perhaps this is done like that to avoid resource leaks or consuming too much memory?
        # Since Blender only updates UI when something changes, this usually isn't too costly.
        
        ui_scale = BlUI.ui_scale(context)
        cgl.text.size(settings.palette_text_size * ui_scale)
        
        show_all_names = settings.palette_show_all_names
        
        label_text_color_active = colors["label_text"]
        label_shadow_color_active = colors["label_shadow"]
        
        label_text_color = Vector(label_text_color_active)
        label_text_color[3] *= 0.85
        label_shadow_color = Vector(label_shadow_color_active)
        #label_shadow_color[3] *= 0.65
        
        item_selection_color = colors["item_selection"]
        
        if settings.palette_trigger_size > 0:
            show_preview = (self.real_cell_size < settings.palette_trigger_size * ui_scale)
        else:
            show_preview = (self.real_cell_size < settings.palette_preview_size * ui_scale)
        
        self.shader_image.bind() # needed to set uniforms
        self.shader_image.uniform_int("image", 0)
        
        scissor_box = cgl.scissor
        
        cgl.scissor = (*self.cells_rect.min, *self.cells_rect.size)
        
        for i, cell in enumerate(self.visible_cells):
            item, rect = cell
            is_active_cell = (self.active_cell_index == i)
            
            if is_active_cell:
                text_shadow = label_shadow_color_active
                text_color = label_text_color_active
            else:
                text_shadow = label_shadow_color
                text_color = label_text_color
            
            show_name = (is_active_cell or show_all_names or item.is_dir) and not show_preview
            
            self.draw_item(item, rect, item_selection_color, text_color, text_shadow,
                show_selection=item.selected, show_name=show_name, force_update=is_active_cell)
        
        cgl.scissor = scissor_box
        
        if show_preview:
            self.draw_preview(context, item_selection_color,
                label_text_color_active, label_shadow_color_active)
    
    def draw_preview(self, context, selection_color, text_color, text_shadow):
        active_cell = self.get_active_cell()
        if (not active_cell) or self.action: return
        
        colors = settings.colors
        
        area = context.area
        region = context.region
        rv3d = context.region_data
        
        ui_scale = BlUI.ui_scale(context)
        
        # Note: region rect is in window coordinates, but draw() and test_select()
        # expect the (0, 0) to be at the region's lower left corner
        safe_rect = BlUI.calc_region_rect(area, region, False)
        safe_rect.move((-region.x, -region.y))
        
        palette_preview_size = settings.palette_preview_size * ui_scale
        
        if settings.palette_trigger_size > 0:
            show_above = True
        else:
            show_above = (self.real_cell_size < palette_preview_size * 0.5)
        
        if show_above:
            cell_padding = settings.palette_padding * ui_scale
        else:
            cell_padding = 1.0 * ui_scale
        
        preview_total_size = palette_preview_size + cell_padding*2
        
        item, rect = active_cell
        
        preview_rect = Bounds.CenterSize(rect.center, (preview_total_size, preview_total_size))
        
        if show_above:
            preview_rect.min = (preview_rect.min[0], self.panel_rect.max[1] + cell_padding)
        
        delta_min = max(safe_rect.min[0] + cell_padding - preview_rect.min[0], 0)
        delta_max = min(safe_rect.max[0] - cell_padding - preview_rect.max[0], 0)
        preview_rect.move(((delta_min+delta_max), 0))
        
        preview_bkg_color = colors["palette_panel"]
        preview_bkg_color[3] *= settings.palette_preview_alpha
        self.draw_rect(self.shader_color, preview_rect, color=preview_bkg_color)
        
        rect = preview_rect.resized(palette_preview_size, type="CenterSize")
        
        self.draw_item(item, rect, selection_color, text_color, text_shadow,
            show_selection=item.selected, show_name=True, force_update=False)
    
    def draw_item(self, item, rect, selection_color, text_color, text_shadow, show_selection, show_name, force_update):
        fitted_rect = Bounds.CenterSize((), item.size or (1.0, 1.0)).fitted(rect)
        
        image = Palette.get_image(item, force_update=force_update)
        self.draw_item_icon(item.icon_value, image, fitted_rect, force_update=force_update)
        
        if show_selection:
            self.draw_rect(self.shader_color, fitted_rect, color=selection_color, wire=True)
            selection_color = Vector(selection_color)
            selection_color.xyz *= 0.5
            self.draw_rect(self.shader_color, fitted_rect.shrunk(1), color=selection_color, wire=True)
        
        if show_name:
            item_label = item.name
            if item.users > 0: item_label = f"[{item.users} users]\n{item_label}"
            self.draw_item_name(item_label, rect.min, rect.size, text_shadow, text_color)
    
    def draw_item_icon(self, icon_id, image, fitted_rect, color=(1,1,1,1), force_update=False, mipmap=False):
        if not image: return
        image_size, pixels = image
        if not (image_size and pixels): return
        
        if isinstance(pixels, gpu.types.Buffer):
            self.bind_icon_texture(self.shader_image, "image", icon_id, image, force_update=force_update, mipmap=mipmap)
            self.draw_rect(self.shader_image, fitted_rect, color=color, uv=True)
        else:
            self.draw_geometry(self.shader_color, fitted_rect, pixels, color=color)
    
    def draw_item_name(self, name, pos, size, text_shadow, text_color):
        x, y = pos[0], pos[1]
        width = size[0]
        
        cgl.text.enable('SHADOW')
        cgl.text.color(*text_shadow)
        cgl.text.shadow(5, *text_shadow)
        cgl.text.shadow_offset(0, 0)
        cgl.text.draw(name, (x-1, y), width=width)
        cgl.text.draw(name, (x+1, y), width=width)
        cgl.text.draw(name, (x, y-1), width=width)
        cgl.text.draw(name, (x, y+1), width=width)
        cgl.text.disable('SHADOW')
        
        cgl.text.color(*text_color)
        cgl.text.draw(name, pos, width=width)
    
    def invoke(self, context, event):
        cls = self.__class__
        
        # Normally, we want most events to go through the palette (e.g. Ctrl+Z).
        # But the events we actually handle should be blocked from propagating further.
        self.use_event_handle_all = False
        
        if not self.is_active(context): return {'PASS_THROUGH'}
        
        self.update_layout(context)
        if not self.is_visible(): return {'PASS_THROUGH'}
        
        mouse_position = Vector((event.mouse_region_x, event.mouse_region_y))
        cls.mouse_position = mouse_position
        self.mouse_start = mouse_position
        self.mouse_last = mouse_position
        self.mouse_near_start = True
        
        active_cell = self.get_active_cell()
        active_item, active_rect = active_cell or (None, None)
        
        active_widget = self.get_active_widget()
        widget_icon, widget_rect = active_widget or (None, None)
        
        self.action = None # reset action by default, override if we actually need modal processing
        
        if event.type == 'LEFTMOUSE':
            if cls.is_in_rect(mouse_position, cls.scroll_area_rect):
                if not cls.is_in_rect(mouse_position, cls.scroll_handle_rect):
                    pos, size = cls.scroll_area_rect.get("min", "size", convert=Vector)
                    cls.scroll_set(cls.scroll_max * (mouse_position.x - pos.x) / size.x)
                self.begin_scroll(context, 'SCROLL_BAR')
                self.use_event_handle_all = True # block further propagation
                return {'RUNNING_MODAL'}
            elif cls.is_in_rect(mouse_position, widget_rect):
                self.begin_widget(context)
                self.use_event_handle_all = True # block further propagation
                return {'RUNNING_MODAL'}
            elif cls.is_in_rect(mouse_position, active_rect):
                if event.value == 'DOUBLE_CLICK':
                    if settings.files_show_get("3d"):
                        settings.files.dclick_action(event, active_item)
                    else:
                        settings.decals.dclick_action(event, active_item)
                    cls.tag_redraw(context.region)
                    self.use_event_handle_all = True # block further propagation
                    # Consume next event to prevent item deselection after Ctrl+Double-click
                    return {'RUNNING_MODAL' if event.ctrl else 'FINISHED'}
                elif event.value in ('PRESS', 'CLICK'):
                    if settings.files_show_get("3d"):
                        if event.ctrl:
                            settings.files.select(active_item.name, 'TOGGLE')
                        elif event.shift:
                            settings.files.select(active_item.name, 'MULTI')
                        else:
                            settings.files.select(active_item.name, 'ACTION')
                            self.begin_item_drag(context, active_item, cursor='CROSSHAIR')
                    else:
                        if event.ctrl:
                            settings.decals.select(active_item.name, 'TOGGLE')
                        elif event.shift:
                            settings.decals.select(active_item.name, 'MULTI')
                        else:
                            settings.decals.select(active_item.name, 'ACTION')
                            self.begin_item_drag(context, active_item, cursor='CROSSHAIR')
                    cls.tag_redraw()
                    self.use_event_handle_all = True # block further propagation
                    return {'RUNNING_MODAL'}
            elif cls.is_in_rect(mouse_position, cls.panel_rect):
                if cls.is_in_rect(mouse_position, cls.interior_rect):
                    self.edge_mask = (1, 1, 1, 1)
                else:
                    self.edge_mask = (
                        int(cls.is_in_rect(mouse_position, cls.edge_xmin_rect)),
                        int(cls.is_in_rect(mouse_position, cls.edge_ymin_rect)),
                        int(cls.is_in_rect(mouse_position, cls.edge_xmax_rect)),
                        int(cls.is_in_rect(mouse_position, cls.edge_ymax_rect)),
                    )
                self.action = 'EDGE_DRAG'
                self.value_start = Vector(settings.palette.offsets)
                cls.active_cell_index = -1
                mask_x = bool(self.edge_mask[0] or self.edge_mask[2])
                mask_y = bool(self.edge_mask[1] or self.edge_mask[3])
                mask_xy = mask_x and mask_y
                cursor_type = ('SCROLL_XY' if mask_xy else ('MOVE_X' if mask_x else 'MOVE_Y'))
                context.window.cursor_modal_set(cursor_type)
                self.use_event_handle_all = True # block further propagation
                return {'RUNNING_MODAL'}
        elif event.type == 'RIGHTMOUSE':
            if settings.files_show_get("3d"):
                settings.files.select(active_item, 'ACTION')
            else:
                settings.decals.select(active_item, 'ACTION')
            bpy.ops.wm.call_panel(name="VIEW3D_PT_xdecal_palette_gizmo_context_menu")
            cls.tag_redraw()
            self.use_event_handle_all = True # block further propagation
            return {'FINISHED'}
        elif event.type == 'MIDDLEMOUSE':
            if event.ctrl:
                self.action = 'SCALE_DRAG'
                self.value_start = settings.icon_scale_get("3d")
                self.zoom_info = cls.get_zoom_info(mouse_position)
                context.window.cursor_modal_set('SCROLL_Y')
            else:
                self.begin_scroll(context, 'SCROLL_DRAG', cursor='HAND')
            self.use_event_handle_all = True # block further propagation
            return {'RUNNING_MODAL'}
        elif event.type == 'WHEELUPMOUSE':
            cls.active_cell_index = -1
            if event.ctrl:
                zoom_info = cls.get_zoom_info(mouse_position)
                step = CommonConfig.palette_zoom_step
                settings.icon_scale_set("3d", settings.icon_scale_get("3d") + step)
                cls.update_layout(context)
                cls.apply_zoom_info(zoom_info)
                cls.tag_redraw()
            else:
                cls.scroll_set(cls.scroll_get() - cls.cell_step)
                cls.tag_redraw(context.region)
            self.use_event_handle_all = True # block further propagation
            return {'FINISHED'}
        elif event.type == 'WHEELDOWNMOUSE':
            cls.active_cell_index = -1
            if event.ctrl:
                zoom_info = cls.get_zoom_info(mouse_position)
                step = CommonConfig.palette_zoom_step
                settings.icon_scale_set("3d", settings.icon_scale_get("3d") - step)
                cls.update_layout(context)
                cls.apply_zoom_info(zoom_info)
                cls.tag_redraw()
            else:
                cls.scroll_set(cls.scroll_get() + cls.cell_step)
                cls.tag_redraw(context.region)
            self.use_event_handle_all = True # block further propagation
            return {'FINISHED'}
        
        return {'PASS_THROUGH'}
    
    def exit(self, context, cancel):
        cls = self.__class__
        
        context.window.cursor_modal_restore()
        
        if self.action:
            if self.action == 'WIDGET':
                self.exit_widget(cancel)
            elif self.action.startswith("SCROLL"):
                self.exit_scroll(cancel)
            elif self.action == 'ITEM_DRAG':
                self.exit_item_drag(context, cancel)
            
            self.action = None
    
    def modal(self, context, event, tweak):
        cls = self.__class__
        
        mouse_position = Vector((event.mouse_region_x, event.mouse_region_y))
        delta = mouse_position - self.mouse_start
        cls.mouse_position = mouse_position
        self.mouse_last = mouse_position
        self.mouse_near_start &= ((self.mouse_last - self.mouse_start).magnitude < 1.5)
        
        if self.action:
            if self.action == 'WIDGET':
                self.update_widget(mouse_position)
            elif self.action.startswith("SCROLL"):
                self.update_scroll(delta)
            elif self.action == 'SCALE_DRAG':
                step = CommonConfig.palette_zoom_step
                settings.icon_scale_set("3d", int((self.value_start + delta.y / 50.0) / step) * step)
                cls.update_layout(context)
                cls.apply_zoom_info(self.zoom_info)
                cls.tag_redraw()
            elif self.action == 'EDGE_DRAG':
                ui_scale = BlUI.ui_scale(context)
                scaled_delta = Vector(self.edge_mask) / ui_scale
                scaled_delta[0] *= delta.x
                scaled_delta[1] *= delta.y
                scaled_delta[2] *= delta.x
                scaled_delta[3] *= delta.y
                settings.palette.offsets = self.value_start + scaled_delta
            elif self.action == 'ITEM_DRAG':
                self.update_item_drag(context, event, mouse_position)
            cls.tag_redraw(context.region)
            return {'RUNNING_MODAL'}
        
        return {'CANCELLED'}
    
    @classmethod
    def get_zoom_info(cls, mouse_position):
        index = -1
        cell_pos = None
        
        if cls.grid_info:
            ndim_to_bounds = cls.grid_info["ndim_to_bounds"]
            point_to_ndim = cls.grid_info["point_to_ndim"]
            to_flat = cls.grid_info["to_flat"]
            
            ndindex = point_to_ndim(mouse_position)
            index = to_flat(ndindex)
            rect = ndim_to_bounds(ndindex)
            cell_pos = rect.point_to_normalized(mouse_position)
        
        return {"abs_pos":mouse_position, "cell_pos":cell_pos, "index":index}
    
    @classmethod
    def apply_zoom_info(cls, zoom_info):
        if not cls.grid_info: return
        
        ndim_to_bounds = cls.grid_info["ndim_to_bounds"]
        to_ndim = cls.grid_info["to_ndim"]
        
        ndindex = to_ndim(zoom_info["index"])
        rect = ndim_to_bounds(ndindex)
        
        abs_pos = zoom_info["abs_pos"]
        cell_pos = rect.normalized_to_point(zoom_info["cell_pos"])
        
        delta = abs_pos[0] - cell_pos[0]
        cls.scroll_set(cls.scroll_get() - delta)
    
    def begin_widget(self, context):
        cls = self.__class__
        self.action = 'WIDGET'
        self.mouse_over_widget = True
        cls.pressed_widget_type = cls.active_widget_type
        cls.tag_redraw(context.region)
    
    def update_widget(self, mouse_position):
        cls = self.__class__
        pressed_widget = self.get_pressed_widget()
        self.mouse_over_widget = cls.is_in_rect(mouse_position, pressed_widget[1])
    
    def exit_widget(self, cancel):
        cls = self.__class__
        
        widget_type = cls.pressed_widget_type
        cls.pressed_widget_type = None
        
        if (not cancel) and self.mouse_over_widget:
            BlUI.tag_redraw()
            
            if widget_type == "close":
                bpy.ops.xdecal.toggle_palette_gizmo()
            elif widget_type == "pin":
                settings.palette.auto_hide = not settings.palette.auto_hide
            elif widget_type == "anchors":
                bpy.ops.wm.call_panel(name="VIEW3D_PT_xdecal_palette_gizmo_anchors")
            elif widget_type == "files":
                settings.files_show_set("3d", not settings.files_show_get("3d"))
            elif widget_type == "options":
                bpy.ops.wm.call_panel(name="VIEW3D_PT_xdecal_palette_gizmo_context_menu")
            elif widget_type == "inherit":
                settings.use_inherit = not settings.use_inherit
            elif widget_type == "template_material":
                bpy.ops.wm.call_menu(name='SCENE_MT_xdecal_template_material')
            elif widget_type == "directory":
                bpy.ops.xdecal.load('INVOKE_DEFAULT')
            elif widget_type == "directory_refresh":
                settings.files.refresh(reset_scroll=False, reset_icons=True)
            elif widget_type == "directory_preset":
                DirectoryPreset.show_popup()
            elif widget_type == "filter_implicit":
                settings.decals.filter_implicit_set("3d", not settings.decals.filter_implicit_get("3d"))
            elif widget_type == "filter_explicit":
                settings.decals.filter_explicit_set("3d", not settings.decals.filter_explicit_get("3d"))
            elif widget_type == "filter_scene":
                settings.decals.filter_scene_set("3d", not settings.decals.filter_scene_get("3d"))
            elif widget_type == "add_objects":
                bpy.ops.xdecal.add_objects_to_palette()
            elif widget_type == "filter_objects":
                settings.decals.filter_objects_set("3d", not settings.decals.filter_objects_get("3d"))
            elif widget_type == "add_materials":
                bpy.ops.xdecal.add_materials_to_palette()
            elif widget_type == "filter_materials":
                settings.decals.filter_materials_set("3d", not settings.decals.filter_materials_get("3d"))
    
    def begin_scroll(self, context, action, cursor='DEFAULT'):
        cls = self.__class__
        self.action = action
        self.value_start = cls.scroll_get()
        cls.active_cell_index = -1
        context.window.cursor_modal_set(cursor)
        cls.tag_redraw(context.region)
    
    def update_scroll(self, delta):
        cls = self.__class__
        if self.action == 'SCROLL_BAR':
            area_rect = cls.scroll_area_rect
            handle_rect = cls.scroll_handle_rect
            scale = area_rect.size[0] / max(handle_rect.size[0], 1)
        else:
            scale = -1
        cls.scroll_set(self.value_start + delta.x * scale)
    
    def exit_scroll(self, cancel):
        cls = self.__class__
        if cancel: cls.scroll_set(self.value_start, clamp=False)
    
    def begin_item_drag(self, context, item, cursor='DEFAULT'):
        cls = self.__class__
        self.action = 'ITEM_DRAG'
        self.drag_instances = None
        self.drag_base = None
        self.drag_item = item
        context.window.cursor_modal_set(cursor)
    
    def update_item_drag(self, context, event, mouse_position):
        cls = self.__class__
        
        if self.drag_item.is_dir: return
        
        if self.drag_base:
            result = self.drag_base.modal(context, event)
            if 'CANCELLED' in result:
                self.drag_base = None
                self.drag_instances = []
                context.window.cursor_modal_set('STOP')
            return
        
        if (self.drag_instances is not None) and (len(self.drag_instances) == 0): return
        
        if cls.is_in_rect(mouse_position, cls.panel_rect): return
        
        self.drag_instances = self.drag_item.instantiate()
        if not self.drag_instances:
            context.window.cursor_modal_set('STOP')
            return
        
        decal_gizmo_state.target = self.drag_instances[0]
        
        self.drag_base = DecalTransformBase()
        self.drag_base.invoke(context, event)
        self.drag_base.origin_delta = Vector((0, 0))
    
    def exit_item_drag(self, context, cancel):
        cls = self.__class__
        
        if cancel:
            if self.drag_instances:
                cls.active_cell_index = -1 # e.g. to hide preview
                # Undo reverts to the *previous* checkpoint,
                # so here we need to push a dummy undo entry
                bpy.ops.ed.undo_push(message="Instantiate decal")
                bpy.ops.ed.undo()
        elif self.drag_base:
            cls.active_cell_index = -1 # e.g. to hide preview
            
            self.drag_base.exit(context, cancel)
        else:
            active_cell = self.get_active_cell()
            active_item, active_rect = active_cell or (None, None)
            
            if self.mouse_near_start:
                settings.files.select(active_item, 'SINGLE')
            
            if self.drag_item.is_dir:
                if cls.is_in_rect(self.mouse_last, active_rect):
                    settings.files.open(self.drag_item)
        
        self.drag_base = None

@addon.GizmoGroup(idname = "OBJECT_GGT_xdecal_palette", label = f"{addon.name} Palette",
    space_type = 'VIEW_3D', region_type = 'WINDOW', options = {'PERSISTENT'})
class DecalPaletteGizmoGroup:
    @classmethod
    def poll(cls, context):
        return ('VIEW3D' in settings.show_palette) and (context.mode == 'OBJECT')
    
    @classmethod
    def setup_keymap(cls, keyconfig):
        # By default, gizmos only have a left mouse press/tweak defined in their keymap
        km = keyconfig.keymaps.find("xdecal:palette", space_type='EMPTY', region_type='WINDOW')
        if km: keyconfig.keymaps.remove(km)
        km = keyconfig.keymaps.new("xdecal:palette", space_type='EMPTY', region_type='WINDOW')
        # LEFTMOUSE is also necessary if gizmo.use_event_handle_all == True
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFTMOUSE', 'PRESS', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'MIDDLEMOUSE', 'PRESS', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'RIGHTMOUSE', 'PRESS', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFTMOUSE', 'DOUBLE_CLICK', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELUPMOUSE', 'PRESS', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELDOWNMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELINMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'WHEELOUTMOUSE', 'PRESS', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'MOUSEMOVE', 'ANY', any=True)
        # kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'INBETWEEN_MOUSEMOVE', 'ANY', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFT_CTRL', 'ANY', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFT_ALT', 'ANY', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'LEFT_SHIFT', 'ANY', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'RIGHT_ALT', 'ANY', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'RIGHT_CTRL', 'ANY', any=True)
        kmi = km.keymap_items.new('gizmogroup.gizmo_tweak', 'RIGHT_SHIFT', 'ANY', any=True)
        return km
    
    @addon.on_unregister
    def clear_keymap():
        wm = bpy.context.window_manager
        for kc in wm.keyconfigs:
            km = kc.keymaps.find("xdecal:palette", space_type='EMPTY', region_type='WINDOW')
            if km: kc.keymaps.remove(km)
    
    def setup(self, context):
        priority = 1000
        
        gizmo = self.gizmos.new(DecalPaletteGizmo.bl_idname)
        gizmo.use_draw_modal = True
        gizmo.select_bias = priority
        
        # Enable by default so that the first right-click would get blocked
        gizmo.use_event_handle_all = True

@addon.Panel(idname="VIEW3D_PT_xdecal_palette_gizmo_context_menu", label="Decals", space_type='CONSOLE', region_type='WINDOW')
class PaletteGizmoContextMenuPanel:
    def draw(self, context):
        layout = self.layout
        
        PaletteUI.draw_popup(layout, context, draw_items=False, palette_type="3d")
        
        active_cell = DecalPaletteGizmo.get_active_cell()
        if not active_cell: return
        
        item = active_cell[0]
        
        layout.separator()
        box = layout.box()
        box.label(text=item.name, icon='IMAGE_PLANE')
        col = box.column(align=True)
        
        if not settings.files_show_get("3d"):
            Palette.draw_context_menu_decals(col, item.name, item.type, palette_type="3d")
        else:
            Palette.draw_context_menu_files(col, item.name, palette_type="3d")

@addon.Panel(idname="VIEW3D_PT_xdecal_palette_gizmo_anchors", label="Anchors", space_type='CONSOLE', region_type='WINDOW')
class PaletteGizmoAnchorsPanel:
    side_icons = {'NONE':'X', 'LEFT':'TRIA_RIGHT', 'BOTTOM':'TRIA_UP', 'RIGHT':'TRIA_LEFT', 'TOP':'TRIA_DOWN'}
    side_order = ['LEFT', 'BOTTOM', 'RIGHT', 'TOP']
    
    def draw(self, context):
        layout = self.layout
        
        anchor_options = settings.palette
        
        row = layout.row(align=True)
        
        col = row.column(align=True)
        col.label(icon=self.side_icons[anchor_options.limit_float])
        for side in self.side_order:
            if side not in anchor_options.limit_sides: side = 'NONE'
            col.label(icon=self.side_icons[side])
        
        col = row.column(align=True)
        col.prop_menu_enum(anchor_options, "limit_float", text="Float")
        for side in self.side_order:
            col.prop_enum(anchor_options, "limit_sides", side)
        
        col = row.column(align=True)
        col.prop(anchor_options, "anchors")
        
        col = row.column(align=True)
        col.prop(anchor_options, "offsets")
