# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import os
import fnmatch
import re

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "prop, bpy_file_filters",
    "utils_blender": "BpyPath, TextureFilenameParser, BlUtil",
    "utils_ui": "BlUI",
    "utils_addon": "AddonManager",
})

from . import common
from . import palette
from pathlib import Path

from .common import Jobs, CommonUtils
from .palette import Palette, PaletteItem
from .material_utils import MaterialLoader
from .decal_utils import DecalUtils, Decal

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

def any_decals_selected():
    return any(Decal.exists(obj) for obj in bpy.context.selected_objects)

@addon.Preset("directory.preset", sorting='NONE', popup='PANEL', title="", options={'ADD'})
class DirectoryPreset:
    def apply(self, context):
        settings.files.directory = self.id
    
    class OpEditMixin:
        def execute(self, context):
            DirectoryPreset.presets.add(DirectoryPreset(settings.files.directory))
            return {'FINISHED'} 

class SelectableListBase:
    def index(self, name):
        if not isinstance(name, str): name = name.name
        for i, item in enumerate(self.items):
            if item.name == name: return i
        return -1
    
    def find(self, name, type=None):
        for item in self.items:
            if type and (item.type != type): continue
            if item.name == name: return item
    
    @property
    def selected_names(self):
        return [item.name for item in self.items if item.selected]
    
    def select(self, item, mode):
        # TOGGLE, MULTI, SINGLE, ACTION
        if isinstance(item, str): item = self.find(item)
        
        cls = self.__class__
        
        if isinstance(mode, bool):
            if not item: return
            item.selected = (item.is_file != False) and mode
            cls.active_index = self.index(item.name)
        elif mode == 'TOGGLE':
            if not item: return
            item.selected = (item.is_file != False) and (not item.selected)
            cls.active_index = self.index(item.name)
        elif mode == 'MULTI':
            if not item: return
            if (cls.active_index < 0) or (cls.active_index >= len(self.items)):
                self.select(item, 'SINGLE')
            else:
                index = self.index(item.name)
                i_min = min(cls.active_index, index)
                i_max = max(cls.active_index, index)
                for i, item2 in enumerate(self.items):
                    item2.selected = (item2.is_file != False) and (i >= i_min) and (i <= i_max)
        elif mode == 'SINGLE':
            if not item:
                for item2 in self.items:
                    item2.selected = False
                cls.active_index = -1
            else:
                for item2 in self.items:
                    item2.selected = (item2.is_file != False) and (item2 == item)
                cls.active_index = self.index(item.name)
        elif mode == 'ACTION':
            if (not item) or (not item.selected):
                self.select(item, 'SINGLE')
            else:
                cls.active_index = self.index(item.name)
    
    def dclick_action(self, event, item):
        if not item: return 'NONE'
        
        if event.alt:
            action_type = settings.palette_alt_dclick_action
        elif event.ctrl:
            action_type = settings.palette_ctrl_dclick_action
            self.select(item, True)
        else:
            action_type = settings.palette_dclick_action
        
        action = getattr(self, f"dclick_{action_type.lower()}", None)
        if not action: return 'NONE'
        
        result = action(item)
        if result is False: return 'NONE'
        
        BlUI.tag_redraw()
        
        return action_type

@addon.PropertyGroup
class DecalsPaletteOptions(SelectableListBase):
    gizmo_scroll: 0.0 | prop("Palette scroll", "Palette scroll", min=0.0)
    page_scroll: 0.0 | prop("Page scroll", "Page scroll", min=0.0, max=1.0, subtype='PERCENTAGE', options={'TEXTEDIT_UPDATE'})
    name_filter: "" | prop("Filter decals", "Decal name filter")
    
    items = property(lambda self: Palette.get_items("decals"))
    active_index = -1
    
    def dclick_select(self, item):
        bpy.ops.xdecal.select_instances(name=item.name, type=item.type, push_undo=True)
    
    def dclick_add_replace(self, item):
        if any_decals_selected():
            bpy.ops.xdecal.replace_decals(name=item.name, type=item.type, push_undo=True)
        else:
            bpy.ops.xdecal.instantiate(name=item.name, type=item.type, push_undo=True)
    
    def dclick_add_modify(self, item):
        if any_decals_selected():
            if item.type != "Material": return False
            
            material = bpy.data.materials.get(item.name)
            if not material: return False
            
            bpy.ops.xdecal.replace_materials_in_objects(prototype_name=item.name, push_undo=True)
        else:
            bpy.ops.xdecal.instantiate(name=item.name, type=item.type, push_undo=True)

Palette.multiprop(DecalsPaletteOptions, "filter_objects", True | prop("Filter object-based decals", ""))
Palette.multiprop(DecalsPaletteOptions, "filter_materials", True | prop("Filter material-based decals", ""))
Palette.multiprop(DecalsPaletteOptions, "filter_implicit", True | prop("Filter implicit templates", "Show implicit templates in the palette"))
Palette.multiprop(DecalsPaletteOptions, "filter_explicit", True | prop("Filter explicit templates", "Show explicit templates in the palette"))
Palette.multiprop(DecalsPaletteOptions, "filter_scene", True | prop("Filter scene templates", "Show templates from the current scene only"))

@addon.PropertyGroup
class FilesPaletteOptions(SelectableListBase):
    gizmo_scroll: 0.0 | prop("Palette scroll", "Palette scroll", min=0.0)
    page_scroll: 0.0 | prop("Page scroll", "Page scroll", min=0.0, max=1.0, subtype='PERCENTAGE', options={'TEXTEDIT_UPDATE'})
    name_filter: "" | prop("Filter files", "Filename filter", update=(lambda self, context: self.refresh(reset_icons=False)))
    
    directory: "//" | prop("", "Directory", update=(lambda self, context: self.refresh(reset_scroll=True, reset_icons=True)))
    
    items = []
    active_index = -1
    
    file_icons = {}
    file_icons.update((ext, 'FILE_BLEND') for ext in bpy_file_filters["blender"])
    file_icons.update((ext, 'FILE_IMAGE') for ext in bpy_file_filters["image"])
    file_icons.update((ext, 'FILE_MOVIE') for ext in bpy_file_filters["movie"])
    
    icons_enum: 'NONE' | prop(items=[(icon, icon, icon, icon, i) for i, icon in
        enumerate(bpy.types.UILayout.bl_rna.functions["prop"].parameters["icon"].enum_items.keys())
    ])
    
    def dclick_select(self, item):
        self.select_related()
    
    def dclick_add_replace(self, item):
        if any_decals_selected():
            self.replace(item.name, push_undo=True)
        else:
            dir_path, files = self.get_valid_names_selected(item.name)
            return self.load(files, push_undo=True)
    
    def dclick_add_modify(self, item):
        if any_decals_selected():
            selected_names = self.selected_names
            if not selected_names: return False
            
            self.replace_images(selected_names, push_undo=True)
        else:
            dir_path, files = self.get_valid_names_selected(item.name)
            return self.load(files, push_undo=True)
    
    def get_icon_value(self, icon):
        return bpy.types.UILayout.enum_item_icon(self, "icons_enum", icon)
    
    @property
    def has_parent(self):
        directory = self.directory
        if directory.startswith("//") and not BpyPath.abspath("//"): return False
        abs_dir = os.path.normpath(BpyPath.abspath(directory))
        parent_dir = os.path.dirname(abs_dir)
        return parent_dir and (len(parent_dir) < len(abs_dir))
    
    def select_related(self):
        parser = TextureFilenameParser(settings.loading_options.categories)
        
        related_map = {}
        
        for item in self.items:
            if not item.is_file: continue
            
            shared_key, category_id = parser.parse(item.name)
            
            related_info = related_map.get(shared_key)
            if not related_info:
                related_map[shared_key] = {"selected":item.selected, "items":[item]}
            else:
                related_info["selected"] |= item.selected
                related_info["items"].append(item)
        
        for related_info in related_map.values():
            if not related_info["selected"]: continue
            for item in related_info["items"]:
                item.selected = True
    
    def open(self, item, push_undo=True):
        if isinstance(item, str): item = self.find(item)
        if not item: return 'NONE'
        
        if item.is_file:
            self.select(item, 'ACTION')
            success = self.load(self.selected_names, push_undo=push_undo)
            return ('LOAD' if success else 'NONE')
        else:
            success = self.go(item.name, relative=True)
            return ('GO' if success else 'NONE')
    
    def go(self, path, relative=False):
        # Assigning to directory triggers icon reloading (intentionally),
        # but here we don't actually want it if the directory hasn't changed
        
        if relative:
            if (not path) or (path == "."): return False
            
            directory = self.directory
            if directory.startswith("//") and not BpyPath.abspath("//"): return False
            
            # os.path.join("C:", "name") returns "C:name" for some reason
            if not BpyPath.normslash(directory).endswith("/"): directory += "/"
            
            if path == "..":
                # if path_mode == 'DONT_CHANGE', we don't want
                # to just add ".." each time we go to subfolder
                parts = re.split("([\\\\/])", directory)
                # Since we ensure a slash at the end, split()
                # will return [..., "something", ("/" or "\\"), ""]
                if (len(parts) > 2) and (parts[-3] not in "\\/"):
                    directory, path = "", "".join(parts[:-3])
            
            if directory:
                if not directory.endswith("//"):
                    # For some reason, if directory ends with "/",
                    # (but not "//") os.path.join() adds "\\" on windows
                    if directory.endswith("/"): directory = directory[:-1]
                path = BpyPath.join(directory, path)
        
        path_mode = settings.loading_options.real_path_mode
        if not BpyPath.abspath("//"): path_mode = 'ABSOLUTE'
        
        if path_mode != 'DONT_CHANGE':
            path = os.path.normpath(BpyPath.abspath(path))
            if path_mode == 'RELATIVE':
                path = BpyPath.relpath(path)
                if path == "//.": path = "//"
            path = BpyPath.normslash(path)
        else:
            path_abs = os.path.normpath(BpyPath.abspath(path))
            directory_abs = os.path.normpath(BpyPath.abspath(self.directory))
            if path_abs == directory_abs: return True
        
        if self.directory != path:
            self.directory = path
        
        return True
    
    def get_valid_names_selected(self, item_name=None):
        selected_names = self.selected_names
        if item_name and (item_name not in selected_names): selected_names = [item_name]
        return self.get_valid_names(selected_names)
    
    def get_valid_names(self, names):
        dir_path = BpyPath.abspath(self.directory)
        if not os.path.isdir(dir_path): return "", []
        
        if isinstance(names, str): names = (names,)
        
        files = []
        
        for name in names:
            path = os.path.join(dir_path, name)
            if not os.path.isfile(path): continue
            
            ext = os.path.splitext(name)[1].lower()
            icon = self.file_icons.get(ext)
            #if icon not in ('FILE_IMAGE', 'FILE_MOVIE'): continue
            if icon not in ('FILE_IMAGE', 'FILE_MOVIE', 'FILE_BLEND'): continue
            
            files.append(name)
        
        return dir_path, files
    
    def load(self, names, push_undo=True):
        if not names: return False
        
        dir_path, files = self.get_valid_names(names)
        if not files: return False
        
        files = [dict(name=name) for name in files]
        result = bpy.ops.xdecal.load(files=files, directory=dir_path, push_undo=push_undo)
        return ('FINISHED' in result)
    
    def replace(self, name, push_undo=True):
        if not isinstance(name, str): raise ValueError("name must be a string")
        
        context = bpy.context
        
        # Store the current selection, since self.load() adds new objects (and selects them)
        selected_objs = list(context.selected_objects)
        active_obj = context.object
        
        decals = []
        for obj in selected_objs:
            if obj.library: continue
            decal = Decal(obj)
            if decal: decals.append(decal)
        
        if not decals: return False
        
        load_success = self.load(name, push_undo=False)
        if not load_success: return False
        
        obj = context.object
        if (not obj) or (obj == active_obj): return False
        
        decal = Decal(obj)
        if not decal: return False
        
        template = decal.template
        if not template: return False
        
        mesh = obj.data
        bpy.data.objects.remove(obj)
        bpy.data.meshes.remove(mesh)
        
        BlUtil.Object.select_activate(selected_objs, 'SOLO', active=None)
        BlUtil.Object.active_set(active_obj)
        
        for decal in decals:
            decal.template = template
            decal.update()
        
        if push_undo: bpy.ops.ed.undo_push(message="Replace Decals")
        return True
    
    def replace_images(self, names, push_undo=True):
        dir_path, files = self.get_valid_names(names)
        if not files: return False
        
        paths = [os.path.join(dir_path, name) for name in files]
        DecalUtils.replace_images(bpy.context.selected_objects, paths)
        
        if push_undo: bpy.ops.ed.undo_push(message="Replace Decal Images")
        return True
    
    def refresh(self, reset_scroll=False, reset_icons=True):
        cls = self.__class__
        
        if reset_scroll:
            self.page_scroll = 0
        
        cls.active_index = -1
        self.items.clear()
        if not settings.files_show_get("any"): return
        
        dir_path = BpyPath.abspath(self.directory)
        if not os.path.isdir(dir_path): return
        
        pcoll = previews["files"]
        
        if reset_icons:
            pcoll.clear()
            Jobs.icon_loading.clear()
        
        self._add_file_entry(pcoll, dir_path, "..", False, self.has_parent)
        
        for is_file, modification_time, name_low, name in self._get_file_entries(dir_path):
            self._add_file_entry(pcoll, dir_path, name, is_file)
    
    def _get_file_entries(self, dir_path):
        names = os.listdir(dir_path)
        
        if ("*" in self.name_filter) or ("?" in self.name_filter):
            # Note: fnmatch can match wildcards (*, ?) and character ranges ([...], [!...])
            names = fnmatch.filter(names, self.name_filter) # case-insensitive
        elif self.name_filter:
            filter_low = self.name_filter.lower()
            names = [name for name in names if filter_low in name.lower()]
        
        sort_by_time = (settings.palette_sorting == 'TIME')
        
        file_entries = []
        for name in names:
            entry_path = os.path.join(dir_path, name)
            is_file = os.path.isfile(entry_path)
            modification_time = (os.path.getmtime(entry_path) if sort_by_time else 0)
            file_entries.append((is_file, modification_time, name.lower(), name))
        file_entries.sort()
        
        return file_entries
    
    def _add_file_entry(self, pcoll, dir_path, name, is_file, enabled=True):
        icon = ('FILE_FOLDER' if name != ".." else 'FILE_PARENT')
        if is_file:
            ext = os.path.splitext(name)[1].lower()
            icon = self.file_icons.get(ext)
            if not icon: return
        
        item = PaletteItem(name=name, enabled=enabled, is_file=is_file)
        self.items.append(item)
        
        if is_file:
            filepath = os.path.join(dir_path, item.name)
            
            def job():
                preview = pcoll.get(item.name)
                
                if not preview: preview = pcoll.load(item.name, filepath, icon[len("FILE_"):])
                
                item.preview = preview
                item.icon_value = preview.icon_id
                item.size = tuple(preview.image_size)
            
            Jobs.icon_loading.append(job)
        else:
            item.icon_value = self.get_icon_value(icon)
