# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# =========================================================================== #

import math
import time
import os

import bpy
from mathutils import Vector, Matrix, Quaternion, Euler, Color

import importlib
import importlib.util

exec(("" if importlib.util.find_spec("dairin0d") else "from . ")+"import dairin0d")

dairin0d.load(globals(), {
    "bpy_inspect": "BpyOp, prop",
    "utils_blender": "BpyPath",
    "utils_ui": "NestedLayout, BlUI",
    "utils_addon": "AddonManager",
})

from . import common
from . import palette
from . import palette_options

from .common import CommonConfig
from .palette import Palette
from .palette_options import DirectoryPreset

addon = AddonManager()
settings = addon.settings
previews = addon.previews

#============================================================================#

@addon.Operator(idname="xdecal.filename_click", label="Open/select", options={'INTERNAL'})
class FilenameClickOperator:
    "Click: select, Double click: open (if folder) or custom action (if file)"
    
    name: "" | prop()
    mode: "" | prop()
    
    click_time = 0.0
    
    def execute(self, context):
        mode = getattr(self, "mode", 'OPEN')
        
        if mode == 'OPEN':
            settings.files.open(self.name)
        elif mode == 'REPLACE':
            settings.files.replace(self.name)
        elif mode == 'SELECT_ACTION':
            settings.files.select(self.name, 'ACTION')
        elif mode == 'SELECT_TOGGLE':
            settings.files.select(self.name, 'TOGGLE')
        elif mode == 'SELECT_MULTI':
            settings.files.select(self.name, 'MULTI')
        
        BlUI.tag_redraw()
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        cls = self.__class__
        
        click_time = time.perf_counter()
        click_time_delta = click_time - cls.click_time
        cls.click_time = click_time
        
        double_click_time = context.preferences.inputs.mouse_double_click_time/1000.0
        
        # Blender 2.80 does not support double-click event on buttons
        if click_time_delta < double_click_time:
            item = settings.files.find(self.name)
            if not item: return {'CANCELLED'}
            
            if item.is_file:
                settings.files.dclick_action(event, item)
                return {'FINISHED'}
            else:
                self.mode = 'OPEN'
        else:
            if event.shift:
                self.mode = 'SELECT_MULTI'
            elif event.ctrl:
                self.mode = 'SELECT_TOGGLE'
            else:
                self.mode = 'SELECT_ACTION'
        
        return self.execute(context)

@addon.Operator(idname="xdecal.palette_click", label="Click", options={'INTERNAL'})
class PaletteClickOperator:
    "Click: select, Double click: custom action (defined in preferences)"
    
    name: "" | prop()
    mode: "" | prop()
    type: "" | prop()
    
    click_time = 0.0
    
    def execute(self, context):
        mode = getattr(self, "mode", 'SELECT_INSTANCES')
        
        if mode == 'REPLACE':
            bpy.ops.xdecal.replace_decals(name=self.name, type=self.type)
        elif mode == 'SELECT_INSTANCES':
            bpy.ops.xdecal.select_instances(name=self.name, type=self.type)
        elif mode == 'INSTANTIATE':
            bpy.ops.xdecal.instantiate(name=self.name, type=self.type)
        elif mode == 'SELECT_ACTION':
            settings.decals.select(self.name, 'ACTION')
        elif mode == 'SELECT_TOGGLE':
            settings.decals.select(self.name, 'TOGGLE')
        elif mode == 'SELECT_MULTI':
            settings.decals.select(self.name, 'MULTI')
        
        BlUI.tag_redraw()
        
        return {'FINISHED'}
    
    def invoke(self, context, event):
        cls = self.__class__
        
        click_time = time.perf_counter()
        click_time_delta = click_time - cls.click_time
        cls.click_time = click_time
        
        double_click_time = context.preferences.inputs.mouse_double_click_time/1000.0
        
        # Blender 2.80 does not support double-click event on buttons
        if click_time_delta < double_click_time:
            item = settings.decals.find(self.name, self.type)
            if not item: return {'CANCELLED'}
            
            settings.decals.dclick_action(event, item)
            return {'FINISHED'}
        else:
            if event.shift:
                self.mode = 'SELECT_MULTI'
            elif event.ctrl:
                self.mode = 'SELECT_TOGGLE'
            else:
                self.mode = 'SELECT_ACTION'
        
        return self.execute(context)

@addon.context_menu('APPEND')
def palette_context_menu_draw(self, context):
    btn_op = getattr(context, "button_operator", None)
    if btn_op is None: return
    
    layout = self.layout
    op_idname = BpyOp.convert(btn_op.rna_type.identifier, py=True)
    
    if op_idname == PaletteClickOperator.bl_idname:
        layout.separator()
        layout.label(text="Decal Template Operations:", icon='IMAGE_PLANE')
        Palette.draw_context_menu_decals(layout, btn_op.name, btn_op.type)
    elif op_idname == FilenameClickOperator.bl_idname:
        layout.separator()
        layout.label(text="File Operations:", icon='IMAGE_PLANE')
        Palette.draw_context_menu_files(layout, btn_op.name)

class PaletteUI:
    decals_label = "Decal Palette"
    files_label = "Decal Import"
    
    @classmethod
    def draw_header(cls, layout, context, palette_type="ui"):
        row = layout.row(align=True)
        row.prop(*settings.files_show_args(palette_type), text="", icon='FILEBROWSER')
        row.prop(*settings("show_settings"), text="", icon='TOOL_SETTINGS')
        row.prop_enum(*settings("show_palette"), 'PANEL', text="")
        row.prop_enum(*settings("show_palette"), 'VIEW3D', text="")
    
    @classmethod
    def draw_popup(cls, layout, context, width=None, draw_items=True, palette_type="ui"):
        layout = NestedLayout(layout)
        
        with layout.row():
            cls.draw_header(layout, context, palette_type=palette_type)
            
            if settings.files_show_get(palette_type):
                layout.label(text=cls.files_label)
            else:
                layout.label(text=cls.decals_label)
        
        cls.draw_panel(layout, context, width=width, draw_items=draw_items, palette_type=palette_type)
    
    @classmethod
    def draw_panel(cls, layout, context, width=None, draw_items=True, palette_type="ui"):
        layout = NestedLayout(layout)
        
        BlUI.set_ui_metadata(layout, {"palette_type": palette_type})
        
        draw_items = draw_items and ('PANEL' in settings.show_palette)
        
        if not settings.files_show_get(palette_type):
            options = settings.decals
            
            if settings.show_settings:
                cls.draw_decal_settings(layout, palette_type)
            
            cls.draw_display_options(layout, options, palette_type)
            
            if not draw_items: return
            items = Palette.filter_items(f"decals-{palette_type}", context=context)
            
            cls.draw_page_scroll(layout, options, items)
            
            cls.draw_items(layout[-1], items, width, options.page_scroll, "xdecal.palette_click")
        else:
            options = settings.files
            
            if settings.show_settings:
                cls.draw_files_settings(layout)
            
            cls.draw_display_options(layout, options, palette_type)
            
            if not draw_items: return
            items = options.items
            
            cls.draw_page_scroll(layout, options, items)
            
            cls.draw_items(layout[-1], items, width, options.page_scroll, "xdecal.filename_click")
    
    @classmethod
    def draw_decal_settings(cls, layout, palette_type="ui"):
        with layout.row():
            with layout.row(align=True):
                icon = ('PLAY' if settings.default_dynamic else 'PAUSE')
                layout.prop(*settings("default_dynamic"), text="", icon=icon)
                layout.prop(*settings("default_intersect"), text="", icon='MOD_BOOLEAN')
            with layout.row(align=True):
                layout.prop(*settings("default_scale"), text="Scale")
                layout.prop(*settings("default_rotation"), text="", icon_only=True)
        
        settings.draw_inherit_options(layout)
        
        with layout.row():
            with layout.row(align=True):
                layout.prop(*settings.decals.filter_implicit_args(palette_type), toggle=True, text="", icon='GHOST_DISABLED')
                layout.prop(*settings.decals.filter_explicit_args(palette_type), toggle=True, text="", icon='GHOST_ENABLED')
                layout.prop(*settings.decals.filter_scene_args(palette_type), toggle=True, text="", icon='SCENE_DATA')
            
            with layout.row(align=True):
                layout.operator("xdecal.add_materials_to_palette", text="", icon='ADD')
                layout.prop(*settings.decals.filter_materials_args(palette_type), toggle=True, text="", icon='MATERIAL_DATA')
            
            with layout.row(align=True):
                layout.operator("xdecal.add_objects_to_palette", text="", icon='ADD')
                layout.prop(*settings.decals.filter_objects_args(palette_type), toggle=True, text="", icon='OBJECT_DATA')
    
    @classmethod
    def draw_files_settings(cls, layout):
        with layout.row(align=True):
            layout.prop_enum_filtered(settings.loading_options, "draw_category")
        settings.loading_options.draw(layout, settings.loading_options.draw_category)
        
        settings.draw_inherit_options(layout)
        
        dir_exists = os.path.isdir(BpyPath.abspath(settings.files.directory))
        with layout.row(align=True):
            DirectoryPreset.draw_popup(layout, text="", icon='PRESET')
            with layout.row(align=True)(alert = not dir_exists):
                layout.prop(settings.files, "directory", text="")
            layout.operator("xdecal.load", text="", icon='FILE_FOLDER')
    
    @classmethod
    def draw_display_options(cls, layout, options, palette_type="ui"):
        with layout.row():
            with layout.row(align=True):
                # layout.label(icon='FILTER')
                layout.prop_menu_enum(*settings("palette_sorting"), text="", icon='FILTER')
                layout.prop(options, "name_filter", text="")
            with layout.row(align=True)(ui_units_x=7):
                layout.label(icon='FULLSCREEN_ENTER')
                #layout.prop(*settings.icon_scale_args(palette_type), slider=True)
                layout.prop(*settings.icon_scale_args(palette_type), text="")
    
    @classmethod
    def draw_page_scroll(cls, layout, options, items):
        if cls.get_lines_per_page(items) > 0:
            layout.prop(options, "page_scroll", slider=True)
    
    @classmethod
    def get_lines_per_page(cls, items):
        # Attempting to display too many items will simply crash Blender
        if (settings.lines_per_page > 0) or (len(items) < 500):
            return settings.lines_per_page
        return 30
    
    @classmethod
    def draw_items(cls, layout, items, width, page_scroll, op_idname):
        if not items: return
        
        scale = settings.icon_scale_get("ui")
        
        def draw_item(col, item, show_icon):
            text = (f"{item.name} [{item.users}]" if item.users > 0 else item.name)
            icon_value = (item.icon_value if show_icon else 0)
            op = col.operator(op_idname, text=text, depress=item.selected, translate=False, icon_value=icon_value)
            op.name = item.name
            if hasattr(op, "type"): op.type = item.type
        
        if scale <= 1:
            def callback(col, item):
                col.active = item.enabled
                draw_item(col, item, True)
            cls.draw_icon_list(layout, items, callback, page_scroll)
        else:
            def callback(col, item, scale):
                col.active = item.enabled
                col.template_icon(item.icon_value, scale=scale)
                draw_item(col, item, False)
            cls.draw_icon_grid(layout, items, callback, page_scroll, scale, width)
    
    @classmethod
    def draw_icon_list(cls, layout, items, callback, page_scroll):
        lines_per_page = cls.get_lines_per_page(items)
        
        count = len(items)
        i_start, i_end = 0, count
        if lines_per_page > 0:
            max_index = max(count - lines_per_page, 0)
            i_start = min(int((max_index+1) * page_scroll), max_index)
            i_end = min(i_start + lines_per_page, count)
        
        col_main = layout.column(align=True)
        for i in range(i_start, i_end):
            col = col_main.column(align=True)
            callback(col, items[i])
    
    @classmethod
    def _calc_grid_params(cls, count, scale, width):
        ui_scale = BlUI.ui_scale()
        
        if width is None:
            region = bpy.context.region
            width = region.width - (BlUI.PANEL_PADDING + BlUI.POPUP_PADDING) * ui_scale
            region_scale = BlUI.region_scale()
        else:
            width *= ui_scale
            region_scale = 1.0
        
        # Sometimes (width / actual_icon_size) can be NaN, for some reason
        actual_icon_size = max(1, BlUI.ICON_SIZE * scale * ui_scale * region_scale)
        
        num_cols = max(1, math.floor(width / actual_icon_size))
        num_rows = math.ceil(count / num_cols)
        
        return num_cols, num_rows, scale
    
    @classmethod
    def _calc_grid_range(cls, count, num_rows, num_cols, scale, lines_per_page, page_scroll):
        if lines_per_page <= 0: return 0, count
        
        rows_per_page = max(int(lines_per_page / (scale + 1)), 1)
        max_row = max(num_rows - rows_per_page, 0)
        row_start = min(int((max_row+1) * page_scroll), max_row)
        row_end = min(row_start + rows_per_page, num_rows)
        i_start = min(max(row_start * num_cols, 0), count-1)
        i_end = min(max(row_end * num_cols, 0), count)
        
        return i_start, i_end
    
    @classmethod
    def draw_icon_grid(cls, layout, items, callback, page_scroll, scale, width=None):
        # It seems that Blender UI slows down when icons have big size
        # (or possibly because layouted UI has large height), but it
        # doesn't seem to depend on the actual number of items per row.
        
        lines_per_page = cls.get_lines_per_page(items)
        count = len(items)
        num_cols, num_rows, scale = cls._calc_grid_params(count, scale, width)
        i_start, i_end = cls._calc_grid_range(count, num_rows, num_cols, scale, lines_per_page, page_scroll)
        
        col_main = layout.column(align=True)
        
        if num_cols == 1:
            for i in range(i_start, i_end):
                col = col_main.column(align=True)
                callback(col, items[i], scale)
        else:
            i = i_start
            while i < i_end:
                i_stop = i + num_cols
                
                if i_stop > count:
                    row = col_main.split(factor=(count-i)/(i_stop-i), align=True).row(align=True)
                    i_stop = count
                else:
                    row = col_main.row(align=True)
                
                while i < i_stop:
                    col = row.column(align=True)
                    callback(col, items[i], scale)
                    i += 1

@addon.Panel(idname="VIEW3D_PT_xdecal", label=" ",
    space_type='VIEW_3D', region_type='UI', category=CommonConfig.sidebar_category)
class DecalsPanel:
    @classmethod
    def poll(cls, context):
        return settings.show_in_sidebar
    
    def draw_header(self, context):
        layout = self.layout
        layout.enabled = (context.mode == 'OBJECT')
        
        PaletteUI.draw_header(layout, context)
        
        if settings.files_show_get("ui"):
            layout.label(text=PaletteUI.files_label)
        else:
            layout.label(text=PaletteUI.decals_label)
    
    def draw(self, context):
        layout = self.layout
        layout.enabled = (context.mode == 'OBJECT')
        PaletteUI.draw_panel(layout, context)

@addon.Operator(idname="xdecal.popup_palette", label="Decals", description="Decals")
class DecalsPopupPaletteView3D:
    def execute(self, context):
        return {'FINISHED'}
    
    def invoke(self, context, event):
        if not settings.popup_palette_use: return {'CANCELLED'}
        if context.mode != 'OBJECT': return {'CANCELLED'}
        
        wm = context.window_manager
        width = min(settings.popup_dialog_width, context.window.width)
        return wm.invoke_popup(self, width=int(width))
    
    def draw(self, context):
        width = min(settings.popup_dialog_width, context.window.width)
        PaletteUI.draw_popup(self.layout, context, width=width)

@addon.Operator(idname="xdecal.toggle_palette_gizmo")
class TogglePaletteGizmo:
    def execute(self, context):
        if 'VIEW3D' in settings.show_palette:
            settings.show_palette = settings.show_palette - {'VIEW3D'}
        else:
            settings.show_palette = settings.show_palette | {'VIEW3D'}
        BlUI.tag_redraw()
        return {'FINISHED'}

@addon.Panel(idname="VIEW3D_PT_xdecal_header_popover", label="Decals", space_type='CONSOLE', region_type='WINDOW')
class DecalsHeaderPopoverPanelView3D:
    def draw(self, context):
        PaletteUI.draw_popup(self.layout, context, width=200)

#@addon.ui_draw("VIEW3D_HT_header", 'APPEND') # right-side
@addon.ui_draw("VIEW3D_MT_editor_menus", 'APPEND') # left-side
def draw_view3d_header(self, context):
    layout = self.layout
    
    if not settings.show_in_header: return
    
    if context.mode != 'OBJECT': return
    
    layout.popover("VIEW3D_PT_xdecal_header_popover", text="", icon='IMAGE_PLANE')
